
/* C++ includes */
#include <stream.h>
#include <String.h>


/* C includes */

/* Local includes */
#include "vovlib.h"
#include "trace.h"
#include "place.h"
#include "voverror.h"
#include "dialog.h"
#include "path.h"
#include "annotation.h"
#include "assist.h"
#include "asshelp.h"

extern String vov_prettyTime( int );

String latex_bf( String& s )
{
    return "{\\bf " + s + "}";
}

String latex_tt( String& s )
{
    return "{\\tt " + s + " }";
}
String latex_verb( String& s )
{
    return "\\verb+" + s + "+";
}

String& latex( String& s )
{
    s.gsub( "_" , "\\_" );
    s.gsub( "%%" , "\\%%" );
    s.gsub( "&" , "\\&" );
    return s;
}

void latex_begin()
{
    cout << "\\documentstyle[times]{article}\n";
    cout << "\\begin{document}\n";
}

void latex_end()
{
    cout << "\\end{document}\n";
}

VovStatus VT_printLatex( VovNode& node, void* data )
{
    // Print something only if the node has notes or if it is a primary input.
    int input_count = node.numberOfInputs();
    if ( node.countNotes() > 0 || input_count == 0  ) {
	cout << "\n\n\\noindent\n";
	if ( node.nodeType() == VOV_PLACE ) {
	    VovPlace place( node );
	    String ps = vov_getPlaceTypeString( place.getPlaceType() );	// props.
	    String ns = place.getCommonName();				// Name
	    if  ( !input_count ) {
		ps += ", primary input";		// Properties.
	    }
	    cout << latex_verb( ns );
	    cout << "\n\nProperties of this place: " << ps << "." ;
	} else {
	    VovTransition tran( node );
	    String cs = tran.getCommandLine();
	    String s;
	    int linelength = 60;
	    while ( cs.length() > linelength ) {
		s += cs.at( 0, linelength ) + "\n";
		cs = cs.after( linelength - 1 );
	    }
	    s += cs;
	    cout << "\\begin{verbatim}\n";
	    cout << s ;
	    cout << "\\end{verbatim}\n";
	    cout << "\n\nProperties of this transition:  duration " << vov_prettyTime( tran.duration() ) << "\n" ;
	}
	VA_hardcopyNotes( node );
    }
    return VOV_OK;
}


void VA_hardcopyNotes( VovObject& obj )
{
    static VovAnnotation* notes = 0;
    int count = obj.countNotes();
    if ( count ) {
	if ( notes == 0 ) {
	    notes = new VovAnnotation[40];
	}
	count = obj.getNotes(  notes, count );

	cout << "\\begin{itemize}\n";
	for ( int i = 0 ; i < count ; i++ ) {
	    cout << "\\item\n";
	    cout << latex( notes[i].string() );
	}
	cout << "\n\\end{itemize}\n";
    }
}


VovStatus VA_hardcopy( String& traceName, String& setName )
{
    vovMessage( "HARDCOPY: " + traceName + "," + setName );
    VovTrace trace;
    if ( trace.open( traceName, "<no host>", "r", VOV_TRACE_TYPE_LOCAL ) != VOV_OK ) {
	vovFatalError( traceName + " cannot open" );
    }
    VovNodeSet set( setName );
    
    if ( trace.find( set ) == VOV_NOT_FOUND ) {
	vovError( "set not found: " + setName );
	return VOV_ERROR;
    } else {
	latex_begin();

	cout << "\\noindent\n";
	cout << "{\\large {\\bf TRACE } " << trace.getProjectName() << "}\n";
	
	set.cover( VT_printLatex, 0 );
    
	latex_end();
	return VOV_OK;
    }
}

