/* C++ includes */
#include <stream.h>
#include <String.h>


/* C includes */
extern "C"
{
#include "utility.h"		// Octtools
}

/* Local includes */
#include "vovlib.h"
#include "trace.h"
#include "annotation.h"
#include "user.h"
#include "voverror.h"
#include "dialog.h"
#include "path.h"
#include "ddsplus.h"
#include "assist.h"
#include "asshelp.h"

//
// Global variables.
//
extern int VAAutomatic;
extern int VAdebug;

//
// Externally defined function.
//
extern void vovEditSet( VovTrace& trace, VovNodeSet& set );
extern void vov_assist_click_me_ccb( ddsHandle );


String VA_newPlaceName( VovPlace& examplePlace, VovVarSet& varSet )
{
    String name = examplePlace.getName();
    for ( Pix i = varSet.first(); i != 0 ; varSet.next( i ) ) {
	VovVar& var = varSet(i);
	name = var.to_new( name );
    }
    if ( VAdebug ) vovMessage( examplePlace.getName() + " --> " + name );
    return name ;
}

void VA_updatePlaceName( VovPlace& examplePlace, VovVarSet& varSet )
{
    for ( Pix i = varSet.first(); i != 0 ; varSet.next( i ) ) {
	VovVar& var = varSet(i);
	examplePlace.mvlib( var.oldValue(), var.newValue() );
    } 
}

void VA_startNewSearch( VovTrace& exampleTrace )
{
    VovNodeSet set(";;;NODES;;;");
    VOV_ASSERT( exampleTrace.find( set ) );
    VovContentsGenerator genNode( &set, VOV_NODE_OCTMASK );
    VovNode node;
    if ( octGenerate( &genNode, &node ) != OCT_OK ) {
	vovFatalError( "No nodes in node set" );
    }
    node.newVisit();
}

#ifdef OBSOLETE
VovStatus VA_prepareTrace( 
			  VovTrace& mainTrace, 		// The main design trace.
			  VovTrace& exampleTrace,	// The trace to be used as example.
			  VovNodeSet& set,		// The set of input places.
			  VovVarSet& varSet		// Set for variable substitution.
			 ) 
{
    int count = 1;		// To generate a unique name for the new set.
    String setname;		// 
    VovNodeSet newSet;		// Set in the main trace.
    while ( 1 ) {		// Until we have a nice new set to work with.
	setname = shortTraceName( exampleTrace.getProjectName() ) + form( ".%d", count );
	VovNodeSet tmpset( setname );
	if ( mainTrace.find( tmpset ) == VOV_OK ) {
	    count++;		// to generate the new name.
	} else {		// OK, set does not exist yet, create it.
	    VOV_ASSERT( mainTrace.create( tmpset ) ); 
	    newSet.id() = tmpset.id();
	    VOV_ASSERT( mainTrace.getById( newSet ) );
	    break;
	}
    }
    
    VA_startNewSearch( exampleTrace );
    //
    // Find all transitions in the set in the example trace.
    // Transfer each transition to the main trace.
    //
    VovGenerator genTran;
    VOV_ASSERT( exampleTrace.initGenContents( set, VOV_NODE_OCTMASK, genTran ) ); 
    VovNode exampleNode;
    while ( exampleTrace.generate( genTran, exampleNode ) == VOV_OK ) {
	VOV_ASSERT( exampleTrace.getById( exampleNode ) );

	// Do only transitions: places will follow as inputs and outputs of transitions.
	if ( exampleNode.nodeType() == VOV_TRANSITION ) {
	    VovTransition exampleTran;
	    exampleTran.id() = exampleNode.id();
	    VOV_ASSERT( exampleTrace.getById( exampleTran ) );
	    VovStatus s  = VA_setupInputs( mainTrace, exampleTrace, exampleTran, newSet, varSet );
	    if ( s == VOV_ERROR || s == VOV_STOP ) return s;
	}
    }
    return VOV_OK;
}
#endif


String VA_formatNotes( VovTrace& exampleTrace, VovNode& exampleNode )
{
    String s;
    int n = exampleTrace.numberOfAnnotations( exampleNode );
    if ( n ) {
	VovAnnotation noteArray[n];
	if ( exampleTrace.getAnnotations( exampleNode, noteArray, n ) != n ) {
	    vovError( "wrong number of annotations" );
	} else {
	    for ( int i = 0; i < n ; i++ ) {
		s += form( "[%d]:: ", i+1 );
		s += noteArray[i].string() + "\n";
	    }
	}
    } else {
	s = "\n\nNo notes available.\n\n";
    }
    return s;
}

VovStatus VA_copyNotes( VovTrace& mainTrace, VovTrace& exampleTrace, VovNode& newNode, VovNode& exampleNode )
{
    int n = exampleTrace.numberOfAnnotations( exampleNode );
    if ( n ) {
	VovAnnotation noteArray[n];
	if ( exampleTrace.getAnnotations( exampleNode, noteArray, n ) != n ) {
	    vovError( "wrong number of annotations" );
	} else {
	    for ( int i = 0; i < n ; i++ ) {
		VOV_ASSERT( mainTrace.addAnnotation( newNode, noteArray[i] ) );
	    }
	}
    }
    return VOV_OK;
}



// ----------------------------------------------
// ----------------------------------------------
// ----------------------------------------------
// ----------------------------------------------
// ----------------------------------------------

typedef struct {
    VovTrace* tracePtr;
    VovNode* nodePtr;
    VovDialogEditText* oldnameDlgPtr;
    VovDialogEditText* newnameDlgPtr;
    String   newname;
    String   oldname;
    String   ctrl;
} VovLocalEditNodeData;

void vov_assist_copy_node_ccb( ddsHandle item )
{
    VovDialogControl button( item );
    VovLocalEditNodeData& data = *(VovLocalEditNodeData*)button.getData();
    
    if ( data.newnameDlgPtr ) {
	VovDialogEditText& newnameDlg = *data.newnameDlgPtr;
	data.newname = newnameDlg.getTextItem();
    }
    
    if ( data.oldnameDlgPtr ) {
	VovDialogEditText& oldnameDlg = *data.oldnameDlgPtr;
	if ( data.oldname != oldnameDlg.getTextItem() ) {
	    vovMessage( "oldname changes ignored" );
	}
    }
    data.ctrl = button.name();
    dds_destroy( item );
}

void vov_assist_edit_node_ccb( ddsHandle item )
{
    VovDialogControl ctrl( item );
    VovLocalEditNodeData& data = *(VovLocalEditNodeData*)ctrl.getData();
    VovTrace& trace = *data.tracePtr;
    VovNode& node = *data.nodePtr;
    // vovMessage( "About to edit: " + node.format() );
    trace.edit( node , 1 );	// Just view this node.
}


void VA_doCopy( VovPlace& examplePlace, VovPlace& newPlace, VovVarSet& varSet )
{
    if ( newPlace.getName() == examplePlace.getName() ) {
	return ;		// No need to copy a place onto itself.
    }

    if ( newPlace.copy( examplePlace ) != VOV_OK ) {
	vovError( newPlace.format() + "\n\tcould not be copied." );
	return;
    }

    // for ( Pix i = varSet.first(); i != 0 ; varSet.next( i ) ) {
    // VovVar& var = varSet(i);
    // data.newname = var.to_new( data.newname );
    // 
    // } 
}


VovStatus VA_copyOnePlace( 
			  VovTrace& mainTrace, 		// The main design trace.
			  VovTrace& exampleTrace,	// The trace to be used as example.
			  VovPlace& examplePlace,	// The place to be copied.
			  VovPlace& mainPlace,		// The place which is created/copied.
			  VovNodeSet& set,
			  VovVarSet& varSet
			  ) 
{
    VovLocalEditNodeData data;
    data.tracePtr = &exampleTrace;
    data.nodePtr = (VovNode*)&examplePlace;
    data.oldname = examplePlace.getName();
    data.ctrl = 0;

    data.newname = VA_newPlaceName( examplePlace, varSet );

    if ( VAAutomatic ) {
	data.ctrl = "Copy";
	vovMessage( "PRIM. INPUT: " + data.newname ); // Feedback for user.
    } else {
	VovDialog top_dialog( "VOV ASSIST: COPY PRIMARY INPUT", DDS_VERTICAL, 3 );
	VovDialogEditText exampleNameDlg( top_dialog, "Name in example", data.oldname,  1, 80 );
	VovDialogEditText newNameDlg( top_dialog,  "New name", data.newname, 1, 80 );
	VovDialogEditText notesDlg( top_dialog,  "Notes", VA_formatNotes( exampleTrace, examplePlace), 10, 80 );
	
	top_dialog.addControl( "CLICK ME", vov_assist_click_me_ccb, 0 );
	top_dialog.addControl( "Copy" , vov_assist_copy_node_ccb, (ddsData)&data);
	top_dialog.addControl( "Skip", vov_assist_copy_node_ccb, (ddsData)&data  );
	top_dialog.addControl( "Automatic", vov_assist_copy_node_ccb, (ddsData)&data );
	top_dialog.addControl( "View place", vov_assist_edit_node_ccb, (ddsData)&data );
	top_dialog.addHelp( VHA_COPY_PLACE );

	data.oldnameDlgPtr = &exampleNameDlg;
	data.newnameDlgPtr = &newNameDlg;
	top_dialog.postModed();
    }
    
    VovPlace newPlace( examplePlace.getPlaceType(), data.newname );
    if ( data.ctrl == "Copy" || data.ctrl == "Automatic" ) {
	if ( VAdebug ) vovMessage( "Copying place " + newPlace.format() );
	VA_doCopy( examplePlace, newPlace, varSet );
	VOV_ASSERT( mainTrace.findOrCreate( newPlace ) );
	VOV_ASSERT( mainTrace.setOperation( set, newPlace, VOV_SET_ADD ) );
	if ( data.ctrl == "Automatic" ) VAAutomatic = 1;
	mainPlace.id() = newPlace.id();
	return VOV_OK;
    } else if ( data.ctrl == "Skip" ) {
	VOV_ASSERT( mainTrace.findOrCreate( newPlace ) );
	VOV_ASSERT( mainTrace.setOperation( set, newPlace, VOV_SET_ADD ) );
	return VOV_NOT_OK;
    } else {
	return VOV_STOP;
    }
}


VovStatus VA_setupInputs( 
			 VovTrace& mainTrace, 		// The main design trace.
			 VovTrace& exampleTrace,	// The trace to be used as example.
			 VovTransition& exampleTran,	// The transition to be copied.
			 VovNodeSet& set,		// 
			 VovVarSet& varSet
			 ) 
{
    // Avoid doing the same transition too many times.
    if ( exampleTran.alreadyVisited() == VOV_OK ) {
	vovMessage( exampleTran.getCommandLine() + " already visited" );
	return VOV_OK;
    }

    VovLocalEditNodeData data;
    data.tracePtr = &exampleTrace;
    data.nodePtr = (VovNode*)&exampleTran;
    data.newname = 0;
    data.oldname = 0;
    data.oldnameDlgPtr = 0;
    data.newnameDlgPtr = 0;
    data.ctrl = 0;
    
    VovVar var_cwd( "CWD", exampleTran.getWd(), path_canonicalize( utcc_getwd() ) );
    varSet.add( var_cwd );
    
    for ( Pix i = varSet.first(); i != 0 ; varSet.next( i ) ) {
	VovVar& var = varSet(i);
	if ( VAdebug ) vovMessage( var.format() );
	exampleTran.mvlib( var.oldValue(), var.newValue() );
    } 
    
    
    // If the transition exists already, skip it.
    if ( mainTrace.find( exampleTran ) == VOV_OK ) {
	vovMessage( exampleTran.getCommandLine() + " already in the trace: skipping it." );
	return VOV_OK;		// 
    }
    
    
    if ( VAAutomatic ) {
	vovMessage( "TRANS: " +  exampleTran.getCommandLine() );	// Give feedback on what is going on.
	data.ctrl = "Copy";					// Always copy in automatic.
    } else {
	VovDialog top_dialog( "VOV_ASSISTANT: COPYING A TRANSITION", DDS_VERTICAL, 10 );
	VovDialogEditText cwdDlg( top_dialog, "Cwd", exampleTran.getWd(), 1, 80 );
	VovDialogEditText cmdDlg( top_dialog,  "Transition", exampleTran.getCommandLine(), 5, 80 );
	VovDialogEditText notesDlg( top_dialog, "Notes", VA_formatNotes( exampleTrace, exampleTran ), 8, 80 );
	
	top_dialog.addControl( "CLICK ME", vov_assist_click_me_ccb, 0 );
	top_dialog.addControl( "Copy", vov_assist_copy_node_ccb, (ddsData)&data );
	top_dialog.addControl( "Skip", vov_assist_copy_node_ccb, (ddsData)&data );
	top_dialog.addControl( "Automatic", vov_assist_copy_node_ccb, (ddsData)&data );
	top_dialog.addControl( "View", vov_assist_edit_node_ccb, (ddsData)&data );
	top_dialog.addHelp( VHA_COPY_TRANSITION );
	top_dialog.postModed();
    } 
    
    VovTransition newTran( exampleTran );
    if ( data.ctrl == "Copy" || data.ctrl == "Automatic") {
	VOV_ASSERT( mainTrace.create( newTran ) );
	VOV_ASSERT( mainTrace.setOperation( set, newTran, VOV_SET_ADD ) );
	VovUser user( utcc_getenv("USER") );
	VOV_ASSERT( mainTrace.findOrCreate( user ) );
	VOV_ASSERT( mainTrace.activeDesigner( user , newTran ) );
	VOV_ASSERT( VA_copyNotes( mainTrace, exampleTrace, newTran, exampleTran ) );
	if ( data.ctrl == "Automatic" ) VAAutomatic = 1;
    } else if ( data.ctrl == "Skip" ) {
	return VOV_NOT_OK;	// Skip this transition.
    } else {
	return VOV_STOP;	// Probably clicked on Cancel. 
    }
    
    
    // Now do all the inputs of the transition.
    VovGenerator genPlace;
    VOV_ASSERT( exampleTrace.initGenContainers( exampleTran, VOV_PLACE_OCTMASK, genPlace ) ); 
    
    VovPlace examplePlace;
    while ( exampleTrace.generate( genPlace, examplePlace ) == VOV_OK ) {
	VOV_ASSERT( exampleTrace.getById( examplePlace ) );
	String newPlaceName = VA_newPlaceName( examplePlace, varSet );
	
	VovPlace newPlace( examplePlace.getPlaceType(), newPlaceName );

	if (VAdebug) vovMessage(form("New place is %s (%s)",(char*)newPlace.format(),(char*)newPlaceName));
	if ( mainTrace.find( newPlace ) == VOV_NOT_FOUND ) {
	    if (VAdebug) vovMessage( form( "Not found %s\n\t(%s)", (char*)newPlace.format(),(char*)newPlaceName ) );
	    VovGenerator genTran;
	    VOV_ASSERT( exampleTrace.initGenContainers( examplePlace, VOV_TRANSITION_OCTMASK, genTran ) );
	    VovTransition parentTran;
	    if ( exampleTrace.generate( genTran, parentTran ) == VOV_OK ) {
		VOV_ASSERT( exampleTrace.getById( parentTran ) );
		VovStatus s = VA_setupInputs( mainTrace, exampleTrace, parentTran, set,  varSet );
		if ( s != VOV_OK ) {
		    return s;
		}
		if (  mainTrace.find( newPlace ) != VOV_OK ) {
		    vovError( newPlace.format() + "\n\tis not in trace as I expected." );
		}
	    } else {
		// The examplePlace is a primary input. See if it is the case that
		// We further help the user by copying the example place to a new place
		// in the current working directory.
		VovPlaceType type = examplePlace.getPlaceType();
		if ( type != VOV_UNIX_FILE &&  type != VOV_OCT_FACET ) {
		    if ( VAdebug ) vovMessage( "Skipping " + examplePlace.format() );
		    continue;
		}
		VovStatus s = VA_copyOnePlace( mainTrace, exampleTrace, examplePlace, newPlace, set, varSet );
		if ( s != VOV_OK ) return s;
	    }
	}
	mainTrace.declareInput( newTran, newPlace, 0 );
	VA_updatePlaceName( examplePlace, varSet );
	VA_copyNotes( mainTrace, exampleTrace, newPlace, examplePlace );
    }
    
    {
	// Now operate on the outputs of the transition.
	VovGenerator genOutputs;
	VOV_ASSERT( exampleTrace.initGenContents( exampleTran, VOV_PLACE_OCTMASK, genOutputs ) ); 
	
	VovPlace examplePlace;
	while ( exampleTrace.generate( genOutputs, examplePlace ) == VOV_OK ) {
	    VOV_ASSERT( exampleTrace.getById( examplePlace ) );

	    String newName = VA_newPlaceName( examplePlace, varSet );
	    VovPlace newPlace( examplePlace.getPlaceType(), newName );
	    
	    VOV_ASSERT( mainTrace.findOrCreate( newPlace ) );
	    VOV_ASSERT( mainTrace.setOperation( set, newPlace, VOV_SET_ADD ) );
	    
	    VA_copyNotes( mainTrace, exampleTrace, newPlace, examplePlace );
	    String errmsg;
	    VOV_ASSERT( mainTrace.declareOutput( newTran, newPlace, 1, 0,  errmsg ) );
	    VA_updatePlaceName( examplePlace, varSet );
	}
    }
    VOV_ASSERT( mainTrace.changeStatus( newTran, VOV_PLAN_NODE, VOV_CHANGE_STATUS_SELF_REC ) );
    return VOV_OK;
}





// -----------------------------------------------------------------------------
// -----------------------------------------------------------------------------
// -----------------------------------------------------------------------------
// -----------------------------------------------------------------------------
// -----------------------------------------------------------------------------
// -----------------------------------------------------------------------------
// -----------------------------------------------------------------------------
// -----------------------------------------------------------------------------

typedef struct {
    VovTrace* exampleTracePtr;
    VovTrace* mainTracePtr;
    VovVarSet* varSetPtr;
    VovNodeSet* newSetPtr;
} VA_struct;

VovStatus VA_visitFunction( VovNode& node, void* d )
{
    VA_struct& data = *(VA_struct*)d;
    VovTrace& mainTrace = *data.mainTracePtr;
    VovTrace& exampleTrace = *data.exampleTracePtr;
    VovVarSet& varSet = *data.varSetPtr;
    VovNodeSet& newSet = *data.newSetPtr;

    if ( node.nodeType() == VOV_PLACE ) {
	// vovMessage( "Cover: visiting and skipping place." );
    } else {
	VovTransition exampleTran;
	exampleTran.id() = node.id();
	VOV_ASSERT( exampleTrace.getById( exampleTran ) );
	VovStatus s  = VA_setupInputs( mainTrace, exampleTrace, exampleTran, newSet, varSet );
	if ( s == VOV_ERROR || s == VOV_STOP ) return s;
    }
    return VOV_OK;
}



VovNodeSet VA_createNewSet( VovTrace& trace, String& setRootName )
{
    int count = 1;						// To generate a unique name for the new set.
    while ( 1 ) {						// Until we have a nice new set to work with.
	String setName = setRootName +  form( ".%d", count );	// 
	VovNodeSet set( setName );				// 
	if ( trace.find( set ) == VOV_OK ) {			// 
	    count++;						// to generate the new name.
	} else {						// OK, set does not exist yet, create it.
	    VOV_ASSERT( trace.create( set ) );			// 
	    return set;						// 
	}
    }
}


VovNodeSet VA_fillExampleSet( VovTrace& trace, VovNodeSet& set )
{
    VovNodeSet full( "FULL SET FOR ASSIST" );
    VOV_ASSERT( trace.create( full ) );
    VovSelectRule selectAllNodes; 		// By default, selects all.
    VOV_ASSERT( trace.getIntermediateNodes( set, set, selectAllNodes, full ) );
    
    int fl = full.length( VOV_NODE_OCTMASK );
    int sl = set.length( VOV_NODE_OCTMASK );
    if ( fl < sl ) {
	vovError( "should never happen: full set smaller than initial set." );
    }
    if ( VAdebug ) vovMessage( form( "Set lengths: original %d, full %d\n", sl, fl ) ); // Debug only.

    // vovMessage( form( "id %d", full.id() ));
    return full;
}

VovStatus VA_prepareTrace( 
			  VovTrace& mainTrace, 		// The main design trace.
			  VovTrace& exampleTrace,	// The trace to be used as example.
			  VovNodeSet& set,		// The set of input places.
			  VovVarSet& varSet		// Set for variable substitution.
			 ) 
{
    //
    // Set in the main trace to collect the new transitions and places.
    //
    VovNodeSet newSet = VA_createNewSet( mainTrace, exampleTrace.getProjectName() ); 
    
    //
    // Make sure that the example tracae has no holes.
    // 
    VovNodeSet fullSet = VA_fillExampleSet( exampleTrace, set );
    if ( VAdebug ) vovMessage( form( "id %d\n", fullSet.id()  ) + fullSet.format() );

    VA_struct data;
    data.exampleTracePtr = &exampleTrace;
    data.mainTracePtr = &mainTrace;
    data.varSetPtr = &varSet;
    data.newSetPtr = &newSet;
    

    if ( VAdebug ) vovMessage( "Covering " + fullSet.format() );
    fullSet.cover( VA_visitFunction, &data );

    return VOV_OK;
}








