/* C++ includes */
#include <stream.h>
#include <String.h>


/* C includes */
extern "C"
{
#include "utility.h"		// Octtools
}

/* Local includes */
#include "vovlib.h"
#include "user.h"
#include "voverror.h"
#include "assist.h"
#include "asshelp.h"

extern int VAdebug;

VovVar::VovVar()
{
    
}
VovVar::VovVar( String& n, String& cv, String& nv )
{
    set( n, cv, nv );
}

void VovVar::set( String& n, String& cv, String& nv )
{
    name= n;
    cvalue = cv;
    nvalue = nv;
}

String VovVar::format()
{
    String s = getName() + ": " + cvalue + " --> " + nvalue;
    return s;
}

String VovVar::to_new( String& s )
{
    String ss = s;
    int c = 0;
    c += ss.gsub( getName(), nvalue );
    c += ss.gsub( cvalue, nvalue );
    // if ( c ) vovMessage( "ToNEW: " + s + "   " + ss  );
    return ss;
}

String VovVar::to_var( String& s )
{
    int c = 0;
    String ss = s;
    c += ss.gsub( nvalue, getName() );
    c += ss.gsub( cvalue, getName() );
    // if ( c ) vovMessage( "ToVAR: " + s + "   " + ss  );
    return ss;
}


VovVarSet::VovVarSet()
{
    size = 0;
    count = 0;
}

void VovVarSet::resize( int ns )
{
    if ( ns <= size ) return;
    ns = ns >? size * 2 ;


    VovVar* tmparray = new VovVar[ns];
    if ( count ) {
	for( int i = 0; i < count; i++ ) tmparray[i] = array[i];
	delete array;
    }
    size = ns;
    array = tmparray;
}

void VovVarSet::add( VovVar& var )
// Add the variable in the set, making sure that the order is correct.
// CORRECT order: if a var contains another var, the longest var should come first.
{
    resize( count + 1 );
    
    for ( int i = 0 ; i < count ; i++ ) {
	if ( var.cvalue.contains( array[i].cvalue ) ) {
	    if ( VAdebug ) vovMessage( var.cvalue + " Contains " + array[i].cvalue );
	    for ( int j = count; j > i ; j-- ) {
		array[j] = array[j-1];
	    }
	    break;
	}
    }
    array[i] = var;
    count++;
}

Pix VovVarSet::first() { return (Pix)1; }
void VovVarSet::next( Pix& i ) 
{
    if ( (int)i < count ) {
	i++;			// get next index.
    } else {
	i = 0;			// Set Pix to 0; end of list.
    }
}

VovVar& VovVarSet::operator()( Pix& i )
{
    return array[(int)i - 1];
}
