#include "vovinternal.h"
#include "utcc.h"
#include "annotation.h"
#include "date.h"
#include "voverror.h"
#include "oh.h"

VovAnnotation::VovAnnotation()
    :
    ( "NOTE", OCT_STRING ),
    signature("SIGNATURE", OCT_STRING ),
    date( "DATE", OCT_INTEGER )
{
    // Remember that this constructors can be called before other constructors 
    // e.g. Strings, so not all the classes are initialized here.
}

VovAnnotation::VovAnnotation( String& n )
    :
    ( "NOTE", OCT_STRING ),
    signature("SIGNATURE", OCT_STRING ),
    date( "DATE", OCT_INTEGER )
{
    contents.prop.value.string = utcc_strsav_and_track( (char*)n );
    signature = utcc_getenv("USER");
    date = getDate();
}

VovAnnotation::VovAnnotation( String& n , String& t )
    :
    ( "NOTE", OCT_STRING ),
    signature("SIGNATURE", OCT_STRING ),
    date( "DATE", OCT_INTEGER )
{
    vovMessage("This routine should not be used" );
    contents.prop.value.string = (char*)t;
    signature = utcc_getenv("USER");
    date = getDate();
}

VovAnnotation::~VovAnnotation()
{
}

#ifndef VOVCLIENT
VovStatus VovAnnotation::create()
{
    OH_ASSERT( octCreate( this, &signature) );
    OH_ASSERT( octCreate( this, &date ) );
    return VOV_OK;
}
#endif

#ifndef VOVCLIENT
VovStatus VovAnnotation::get()
{
    OH_ASSERT( octGetByName( this, &signature) );
    OH_ASSERT( octGetByName( this, &date ) );
    return VOV_OK;
}
#endif

#ifndef VOVCLIENT
VovStatus VovAnnotation::modify()
{
    OH_ASSERT( octModify( this ) );
    OH_ASSERT( octModify( &signature) );
    OH_ASSERT( octModify( &date ) );
    return VOV_OK;
}
#endif

#ifndef VOVCLIENT
VovStatus VovAnnotation::del()
{
    OH_ASSERT( octDelete( &signature) );
    OH_ASSERT( octDelete( &date ) );
    OH_ASSERT( octDelete( this ) );
    return VOV_OK;
}
#endif

String VovAnnotation::format() 
{
    Date d = date.integer();
    String s = d.format() + "  " + signature.string() + "\n";
    s +=  string();
    return s;
}




char NOTES_BAG_NAME[] = "<^>NOTES<^>";

#ifndef VOVCLIENT
void VovObject::addNote( VovAnnotation& note )
{
    VovBag bag(NOTES_BAG_NAME);
    OH_ASSERT( octGetOrCreate( this, &bag ) );
    OH_ASSERT( octCreate( &bag, &note ) );
    VOV_ASSERT( note.create() );
}
#endif

#ifndef VOVCLIENT
int  VovObject::countNotes( )
{
    int count = 0;
    VovBag bag(NOTES_BAG_NAME);
    if ( octGetByName( this, &bag ) == VOV_OK ) {
	VovContentsGenerator genNote( &bag, VOV_ANNOTATION_OCTMASK );
	VovObject note;
	while ( octGenerate( &genNote, &note ) == OCT_OK ) count++;
    }
    return count;
}
#endif

#ifndef VOVCLIENT
int  VovObject::getNotes( VovAnnotation* noteArray, int length )
{
    int count = 0;
    VovBag bag(NOTES_BAG_NAME);
    if ( octGetByName( this, &bag ) == VOV_OK ) {
	VovContentsGenerator genNote( &bag, VOV_ANNOTATION_OCTMASK );
	while ( count < length && octGenerate( &genNote, &noteArray[count] ) == OCT_OK ) {
	    VOV_ASSERT( noteArray[count].get() );
	    count++;
	}
    }
    return count;
}
#endif

#ifndef VOVCLIENT
void VovObject::copyNotes( VovObject& obj ) 
{
    static VovAnnotation* noteArray = 0;
    if ( noteArray == 0 ) {
	vovMessage( "Allocating array for copyNotes() " );
	noteArray = new VovAnnotation[VOV_MAX_ANNOTATIONS];
    }
    int count = obj.countNotes();
    if ( count ) {
	if ( obj.getNotes( noteArray, count ) != count ) {
	    vovBug( "Annotations corrupted" );
	}
	for ( int i = 0 ; i < count ; i++ ) addNote( noteArray[i] );
    }
}
#endif


#ifndef VOVCLIENT
void VovObject::deleteNotes()
{
    VovBag bag(NOTES_BAG_NAME);
    if ( octGetByName( this, &bag ) == VOV_OK ) {
	VovContentsGenerator genNote( &bag, VOV_ANNOTATION_OCTMASK );
	VovAnnotation note;
	while ( octGenerate( &genNote, &note ) == OCT_OK ) {
	    VOV_ASSERT( note.get() );
	    VOV_ASSERT( note.del() );
	}
    }
}
#endif

#ifndef VOVCLIENT
String VovObject::formatNotes()
{
    String s;
    VovBag bag(NOTES_BAG_NAME);
    if ( octGetByName( this, &bag ) == OCT_OK ) {
	VovContentsGenerator genNote( &bag, OCT_PROP_MASK );
	VovAnnotation note;
	while ( octGenerate( &genNote, &note ) == OCT_OK ) {
	    VOV_ASSERT( note.get() );
	    s += note.format();
	    s += "\n";
	}
    }
    return s;
}
#endif
