#include "arc.h"
#include "oh.h"

extern void vov_setGeoColor( VovObject* , char* );
extern char* vov_nodeColorName( VovNodeStatus );

VovArc::VovArc()
    :
    ( OCT_PATH )
{
}

VovArc::~VovArc()
{
}

VovArc::VovArc( VovNode& inNode, VovNode& outNode, VovNodeStatus status )
    :
    ( OCT_PATH )
{
    VovObject inTerm, outTerm;
    ohGetByTermName( &outNode,  &inTerm, "in" ); 		
    ohGetByTermName( &inNode, &outTerm, "out" );

    VovObject facet;				// The facet where all this happens.
    VovObject layer;				// ... giving the color to the path.
    const int PATH_WIDTH = 60;			// width of the path.
    octGetFacet( &inNode, &facet );		// Get the facet (VOID function!)
    ohGetByLayerName( &facet, &layer,		// 
		     vov_nodeColorName(status));// Get the layer.
    ohCreatePath( &layer, this, PATH_WIDTH );	// Create the path.
    OH_ASSERT( octAttach( this, &inTerm) );	// Do the two attachments to terminals.
    OH_ASSERT( octAttach( this, &outTerm) );	// 
    adjust();					// Adjust the position of the path.
}

void VovArc::changeStatus( VovNodeStatus newstatus )
{
    vov_setGeoColor( this, vov_nodeColorName( newstatus ) );
}

/*
** Geometrically adjust the position of the arc.
*/
void VovArc::adjust()
{
    VovObject inTerm;
    VovObject outTerm;
    VovContentsGenerator genTerm( this, OCT_TERM_MASK );
    OH_ASSERT( octGenerate( &genTerm, &inTerm ));
    OH_ASSERT( octGenerate( &genTerm, &outTerm ));
    
    octBox inBB, outBB;
    ohTerminalBB(&inTerm,&inBB);
    ohTerminalBB(&outTerm,&outBB); 				// and its bounding box

    octPoint points[2];						// extremes of the arc
    points[0].x = (inBB.lowerLeft.x + inBB.upperRight.x) / 2; 	// center of the terminal
    points[0].y = (inBB.lowerLeft.y + inBB.upperRight.y) / 2;	// 
    points[1].x = (outBB.lowerLeft.x + outBB.upperRight.x) / 2;	// Center of the out term
    points[1].y = (outBB.lowerLeft.y + outBB.upperRight.y) / 2;	// 
    
 
    octPutPoints( this, 2, points);	  			// 

}

void VovArc::other( VovNode& n , VovNode& other )
{
    VovObject term;
    VovContentsGenerator genTerm( this, OCT_TERM_MASK );
    while (  octGenerate( &genTerm, &term ) == OCT_OK  ) {
	other.id() = term.contents.term.instanceId;
	if ( other.id() == n.id() ) continue;
	other.get();
	break;
    }
}

void VovArc::del() 
{
    OH_ASSERT( octDelete( this ) );
}


