/* C++ includes */
#include <stream.h>
#include <sys/time.h>

/* C includes */
extern "C"
{
#include <errno.h>
#include <sys/types.h>
#include <sys/socket.h> 
#include <netinet/in.h>

// PATCH!! This should be "fcntl.h", but it does not work now (Jan23.90). 
// #include <fcntl.h>
// #include "//usr/include/sys/file.h"
#ifdef vax
u_short htons(  );
u_long ntohl();
u_long htonl();
#endif
}

/* Local includes */
#include "buf.h"
#include "utcc.h"		// Utils  


#define SWAP_MINOR 0x1		/* Swap bytes */
#define SWAP_MAJOR 0x2		/* Swap byte-pairs */

#define BUF_DEBUG 0


Buf::Buf()
{
    size = 0;
    cnt = 0;
    debug = BUF_DEBUG;
}


Buf::Buf( Buf& b )
{
    size = b.size;
    cnt = b.cnt;
    debug = b.debug;
}

Buf::~Buf()
{
    if ( cnt ) {
	cerr.form("\t~Buf(): Buffer non empty: %d characters left\n" , cnt );
    }
    if (size) {
	if (debug) cerr << "\tBuf:Destroying buffer\n";
	delete buf;
    }
}


#ifdef OLDCODE
/*
* Set the swap mask so that the Bzocket can correctly decode the
* integers coming in. The argument test is the test number coming in.
* It should be the same as 'expected'.
*/
int Buf::setSwapMask( int test )
{
    int expected  = SWAP_TEST_VECTOR; 

    
    swapMask = 0;		/* Initialize. */
    if (test != expected) {
	cerr.form( "SET SWAP MASK: probe: %x expected: %x\n", test, expected );
	if (sizeof(int) != 4 ) {
	    utcc_error("Sizeof(int) is not 4. Cannot do the byte swapping" );
	    return 0;		/* Return and go on. */
	}
	int maskArray[3] = {SWAP_MINOR, SWAP_MAJOR, SWAP_MAJOR | SWAP_MINOR};
	for ( int i = 0 ; i < 3 ; i++ ) {
	    swapMask = maskArray[i];
	    if ( swapBytes( test ) == expected ) {
		break;		// Found correct swap mask.
	    }
	}
    }
    if ( swapBytes( test ) != expected ) {
	utcc_fatalerror("Could not set up swapmask correctly");
    }
    return swapMask;
}


/* Union used to make the swapping. */
/* Assume that sizeof(int) is 4. */
union swapUnion {
    char c[4];
    int  n;
};

int Buf::swapBytes( int n )
{
    union swapUnion u;
    u.n = n;
    if ( swapMask & SWAP_MINOR ) {
	char c;
	c = u.c[0] ; u.c[0] = u.c[1] ; u.c[1] = c;
	c = u.c[2] ; u.c[2] = u.c[3] ; u.c[3] = c;
    } 
    if ( swapMask & SWAP_MAJOR ) {
	char c;
	c = u.c[0] ; u.c[0] = u.c[2] ; u.c[2] = c;
	c = u.c[1] ; u.c[1] = u.c[3] ; u.c[3] = c;
    }
    if (debug) {
	cerr.form("\tBuf:swapped %d into %d\n", n, u.n );
    }
    return u.n;
}
#endif /* OLDCODE */

void Buf::print()
{
    cerr.form("Size %6d:",size);
    if (size) {
	cerr.form("Buf %6x, rptr %6x, wptr %6x, pptr %6x, Count %6d", buf, rptr, wptr, pptr, cnt);
    }
    cerr << "\n";
}

void Buf::clear()
{
    cnt = 0;
    rptr = buf;
    wptr = buf;
    pptr = buf;
}

void Buf::cleanup()
{
    if ( rptr != buf ) {		// If the read pointer is not at the beginning of buf.
	if ( cnt ) {			// If there is data, move it in the buffer.
	    bcopy (rptr, buf, cnt);	// Copy the existing data.
	}				// 
	int d = rptr - buf;		// Compute delta (by how much data has been moved).
	rptr = buf;			// Reset read pointer.
	wptr -= d;			// Correct write pointer and ...
	pptr -= d;			// ... packet pointer
    }
}

void Buf::freeMemory()
{
    if (size) {
	delete buf;
    }
    size = 0;
    cnt = 0;
}



int Buf::resize(int ns)
{
    if ( ns < DEFBUFSIZE ) {
	ns = DEFBUFSIZE;
    }
    if (size == 0) {
	buf = new char[ns];
	size = ns;
	wptr = rptr =buf;
	pptr = buf;
	cnt = 0;
	return BUF_OK;
    } 
    if (ns > size) {
	/* must increase the buffer to accomodate what's coming */
	if (ns <= MAXBUFSIZE) {
	    int pptr_offset = pptr - rptr;
	    int wptr_offset = wptr - rptr;
	    if ( wptr_offset != cnt ) utcc_fatalerror("BUF: cnt is wrong");

	    char* tmp = new char[ns];
	    bcopy (rptr, tmp, cnt);
	    delete buf;

	    buf = rptr = tmp;
	    pptr = rptr + pptr_offset;
	    wptr = rptr + wptr_offset;
	    size = ns;
	    return BUF_OK;
	} else {
	    utcc_error ( "Buf: The new size is too big ");
	    return BUF_ERROR;
	}
    }
    return BUF_OK;
}

/*
* Get a pointer to a string of n characters from the buffer.
*/
char* Buf::get(int n)
{
    if ( n > cnt ) {
	return 0;		// Not enough characters.
    }
    char* ret = rptr;
    rptr += n;
    cnt -= n;
    return ret;
}

int Buf::getInt(  )
{
    int* p = (int*)get( sizeof( int ) );
    if ( p == 0 ) {
	utcc_error( "failed Buf::getInt()" );
	return 0;
    }
    return ntohl( *p );
}

int Buf::length()
{
    return cnt;
}

char* Buf::getPacket( int* length )
{
    int n = pptr - rptr;	// The length of the rest of the packet.
    if ( n == 0 ) {
	return 0;		// No packet to get.
    }
    char* ret = rptr;		// The return pointer.
    rptr += n;			// Move read pointer.
    cnt -= n;			// Update counter.
    *length = n;		// Length is also returned.
    if ( debug ) {
	cerr.form( "\tBuf:read packet %d long", *length );
    }
    return ret;
}

/*
*  Back up n charachters in the buffer.
*/
int Buf::unget(int n)
{
    if (n > ( rptr - buf )) {
	utcc_error("Buf: Cannot unget so many characters");
	return BUF_ERROR;
    }
    rptr -= n;
    cnt += n;
    return BUF_OK;
}

/*
*  Put an integer at the end of the buffer.
*/
int  Buf::put( int n )
{
    int n_n = htonl( n );
    return  put( (char*)&n_n, sizeof(int) );
}

/*
* Put a string of bytes starting from p at the end of buffer.
*/
int  Buf::put(char* p, int n)
{
    int nl = cnt + n;			// New length of the packet.
    if ( resize(nl) == BUF_ERROR) {	// Make sure the buffer is big enough.
	return BUF_ERROR;		// Resize failed.
    }					// 
    bcopy(p,wptr,n);			// Copy data.
    wptr += n;				// Move pointer.
    cnt = nl;				// New byte count.
    return BUF_OK;			// OK.
}

/*
* Read as many bytes as you can from socket fd.
*/
int Buf::readSocket(int fd)
{
    cleanup();			// Throw away parts already read.
    
    //
    // Then read as much as we can.
    //
    int bufferleft;
    const int MINIMUM_BUFFER_LEFT = 100; // If there is not enough space in the buffer
					 // resize it.
    while  ( ( bufferleft = buf + size - wptr) <= MINIMUM_BUFFER_LEFT) {
        // utcc_message("Buf:running out of space in buffer: resizing it");
	// print();
	if ( resize( 2 * size ) == BUF_ERROR ) {
	    utcc_fatalerror("Buf:Resizing failed");
	} 
	// print();
	// utcc_message("Buf:buffer successfully resized");
    }
    int numread = 0;
    if (( numread = ::read (fd, wptr, bufferleft )) <= 0) {
	// Read may fail because the socket is now dead. This is a normal occurrence.
	if ( numread == -1 && errno != EBADF ) {
	    utcc_printerrno("Buf:readSocket");
	}
	return BUF_ERROR;
    }
    cnt += numread;
    wptr += numread;
    if ( debug ) {
	cerr.form("\tBuf:read %d bytes from %d\n", numread, fd);
    }
    return BUF_OK;
}



/* Write data to client.
 * We might have to wait, if the client isn't keeping up with us.  We wait for
 * a short time, then close the connection.  This isn't a wonderful solution,
 * but it rarely seems to be a problem right now, and buffering output for
 * asynchronous delivery sounds complicated and expensive.
 */

/*
* Send all data in the buffer
* but first send the total length of the message.
*/
int Buf::sendSocket(int fd)
{
    int n_cnt = htonl( cnt );
    if (write( fd, (char*)&n_cnt, sizeof(int) ) != sizeof(int)) {
	utcc_message("Buf: Could not send buffer");
	return BUF_ERROR;
    }
    
    int total = cnt;
    int count = cnt;
    register int n;
    int max_attempts = 1000;	// Do not stay here for ever.

    while ( max_attempts-- ) {
	if ((n = write (fd, rptr, count)) == cnt ) {
	    clear();					// Buffer is sent.
	    return BUF_OK;				// OK.
	}
	if (n > 0) {
	    rptr += n;
	    cnt -= n;
	    if (cnt < count) {
		count = cnt;
	    }
	} else if (errno == EMSGSIZE) {
	    count >>= 1;			// Decrease message size.
	} else if (errno != EWOULDBLOCK) {	// 
	    utcc_printerrno( "writing to socket" );
	    return BUF_ERROR;			// 
	}					// 
	const int OUTTIME = 10;			// Wait 10 seconds to see if recipient
	// utcc_message( form("Buf:Wait %d s for fd %d", OUTTIME, fd ));	// can keep up with us.
	fd_set mask;
	FD_ZERO( &mask );
	FD_SET( fd, &mask );
	struct timeval outtime = {OUTTIME, 0};	// 
	n = select (fd+1, (int *) NULL, &mask,	// 
		    (int *) NULL, &outtime);	// 
	if (n != 1) {				// 
	    utcc_printerrno("sending buffer");	// Socket must be dead.
	    return BUF_ERROR;			// 
	}
    }
    utcc_fatalerror( "BUF: Could not send message" );
    /* Not reached */
}

/*
* Is there a packet ready in the buffer ?
*/
int Buf::packetReady()
{
    if (pptr > rptr) {		// Previous packet not finished yet.
	return BUF_OK;		// Yes, ready to read on.
    }
    if (pptr < rptr) {		// Paranoic check.
	utcc_error("Buf: Reading across packet boundary");
	print();
	/* continue anyway, this might not be an error */
    }

    char* lp = get(sizeof(int)); // Get the length of the packet.
    if (lp == 0) {		 // 
	return BUF_ERROR;	 // Not even enough to read the length.
    } 
    int length = ntohl( *(int*)lp );
    if (length <= 0) {		// Really paranoid.
	utcc_fatalerror("Buf:Message length cannot be zero or negative");
    }
    if (cnt >= length) {	// 
	pptr = rptr + length;	// Yes, the whole packet has been read.
	return BUF_OK;
    }
    unget( sizeof(int) );	// Put back the packet length.
    return BUF_ERROR;		// Not ready yet.
}



