// Some communication utilities.

/* C++ includes */
#include <stream.h>
#include <builtin.h>
#include <String.h>
/* C++ includes */
#include "vovinternal.h"
#include "bzocket.h"
#include "VOVprotocol.h"
#include "voverror.h"
#include "utcc.h"

// A client waits for a reply.
VovStatus VC_get_reply(Bzocket& zoc)
{
    while  ( zoc.ready() != BUF_OK ) {
	if ( ! zoc.active() || zoc.read() == BUF_ERROR) {
	    vovError("lost connection");
	    return VOV_ERROR;
	}
    }
    return VOV_OK;
}


// A client gets some text.
VovStatus VC_get_text(Bzocket& zoc, String& text)
{
    const int MAX_TEXT_LENGTH = 10000; 	// Must have some bounds, anyway.
    int text_truncated = 0;
    while (1) {
	if  ( VC_get_reply( zoc ) == VOV_ERROR) {
	    return VOV_ERROR;
	}
	int more = zoc.unpackInt(); 	// More text to come ?
	if ( more == VOV_Error ) {
	    return VOV_ERROR;
	}
	if ( !more ) break;		// No, then exit loop.
	String Msg = zoc.unpackString();
	if ( text.length() < MAX_TEXT_LENGTH ) { // 
	    text += Msg;		// Add text to string.
	} else {
	    text_truncated += Msg.length();
	}
    }
    if ( text_truncated ) {
	text += ((String)"\nWarning: " + dec( text_truncated ) + " characters were truncated." );
    }
    return VOV_OK;
}


// A client waits for a perticular code.
VovStatus VC_wait_for_code(Bzocket& zoc, VovCode expectedCode )
{
    while  ( zoc.ready() != BUF_OK ) {
	if ( ! zoc.active() || zoc.read() == BUF_ERROR) {
	    vovError("lost connection");
	    return VOV_ERROR;
	}
    }

    int code = zoc.unpackInt( );
    if (code == VOV_Error) {
	vovMessage( zoc.unpackString() );
	return VOV_ERROR;
    } else if ( code == expectedCode ) {
	return VOV_OK;
    } else {
	vovError( String("Wrong code:") + dec( code ) + " instead of " + dec( expectedCode ) );
	return VOV_ERROR;
    }
}

// A client waits for the code VOV_Ok.
VovStatus VC_wait_for_ok_reply(Bzocket& zoc)
{
    return VC_wait_for_code( zoc, VOV_Ok );
}



// A client cleans up the Bzocket.
VovStatus VC_empty( Bzocket& zoc )
{
    if ( zoc.ready() == BUF_OK ) {
	vovError("Some unread message left in bzocket" );
	zoc.discardPacket();
	return VOV_ERROR;
    }
    return VOV_OK;
}




// Utilities used by the server.


/*
 * Send an Error back to the client.
 */
void VS_oops (Bzocket& zock, const String& error_string)
{
    zock.discardPacket();	// Clear up anything left in the buffer.

    zock.pack( VOV_Error );
    zock.pack( error_string );
    zock.flush();
}

// Send the message VOV_Ok.
void VS_ok (Bzocket& zock)
{
    zock.pack( VOV_Ok );
    zock.flush();
}

// Server sends text.
void VS_sendText( Bzocket& zock, String& s )
{
    if ( s != 0 ) {
	zock.pack( 1 );
	zock.pack( s );
    } else {
	zock.pack( 0 );
    }
}

