#include "trace.h"
#include "annotation.h"
#include "place.h"
#include "transition.h"
#include "event.h"
#include "retrace.h"
#include "voverror.h"
#include "vovlib.h"
#include "ddsplus.h"
#include "path.h"
#include "dialog.h"

typedef struct {
    VovDialogHandles* handlesPtr;
    VovSuperObject* objPtr;
} VovEditData;

VovStatus VovAnnotation::edit()
{
    VovDialog top_dialog( "EDIT ANNOTATION", DDS_VERTICAL, 3 );
    top_dialog.addControl( "Ok" );


    VovDialogHandles noteHandles( 5 );
    setupDialog( top_dialog, noteHandles );

    
    top_dialog.postModed();

    return VOV_NO_CHANGE;
}

VovStatus VovSelectRule::edit()
{
    VovDialog top_dialog( "VOV: Select Rule", DDS_VERTICAL, 4 );

    VovDialogHandles handles(5);
    setupDialog( top_dialog, handles );

    top_dialog.addControl("Ok");

    top_dialog.postModed();
    if ( top_dialog.getControl() == "Ok" ) {
	return  readDialog( handles );
    } else {
	return VOV_NO_CHANGE;
    }
}

VovStatus VovTransition::edit()
{
#ifdef SKIPIT
    VovStatus rtn = VOV_OK;
    // vovMessage( "EDITING:\n" + format() );

    VovDialog top_dialog( 2 );
    (void)setupDialog( top_dialog );

    top_dialog.setTitle( "EDIT TRANSITION" );
    
    top_dialog.addControl( "Ok" );
    top_dialog.addControl( "Fire" );
    top_dialog.addControl( "Fire (trace only)" );
    
    
    top_dialog.postModed();

    if ( top_dialog.getControl() == "Ok" ) {
	rtn = readDialog( 0 ) ;
    } else if ( top_dialog.getControl() == "Fire" ) {
	util_unsetenv( "VOV_TRACE_ONLY" );
	fire( 0 );
    } else if ( top_dialog.getControl() == "Fire (trace only)" ) {
	util_setenv( "VOV_TRACE_ONLY", "TRUE" );
	fire( 0 );
    } else {
	rtn = VOV_ERROR;
    }
    return rtn;
#else 
    vovMessage( "Function commented out" );
#endif
}

VovStatus VovPlace::edit()
{
#ifdef SKIPIT
    VovStatus rtn = VOV_OK;
    VovDialog top_dialog(3);
    top_dialog.setTitle( "EDIT A PLACE" );
    top_dialog.addHelp( "Edit the place or change its status" );
    
    (void)setupDialog( top_dialog );
    
    top_dialog.addControl( "Ok" );
    
    top_dialog.postModed();

    if ( top_dialog.getControl() == "Ok" ) {
	int offset = 0;
	rtn = readDialog( offset );
    } else {
	rtn = VOV_ERROR;
    }
#else 
    vovMessage( "Function commented out" );
#endif
}

VovStatus VovNodeSet::edit()
{
#ifdef SKIPIT
    const int MAX_ITEMS = 20;
    int countItems = 0;
    VovDialog top_dialog(1);

    VovDialog& dialog = top_dialog.getNewDialog( DM_WHICH_ONE );
    
    dialog.setTitle( VovBag::format() );
    VovContentsGenerator genNode( this, VOV_NODE_OCTMASK );
    VovNode node;
    String label;
    while ( octGenerate( &genNode, &node ) == OCT_OK ) {
	node.get();
	if ( node.nodeType() == VOV_PLACE ) {
	    VovPlace place = node;
	    label = place.getName();
	} else {
	    VovTransition tran = node;
	    label = tran.getCommandLine();
	}
	dialog.addItem(label, node.id() );
	if ( countItems++ > MAX_ITEMS ) break;
    }
    dialog.setItem( label );

    if ( countItems == 0 ) {
	vovMessage("The set is empty" );
	return VOV_OK;
    }

    top_dialog.addControl( "Edit" );
    top_dialog.addControl( "Quit" );
    top_dialog.addHelp( "You can edit all nodes in this set" );
    
    while (1) {
	top_dialog.postModed();
	String ctrl =  top_dialog.getControl();
	if ( ctrl == "Quit" ) return VOV_OK;

	node.id() = dialog.getWhichOne();
	node.get();
	if ( node.nodeType() == VOV_PLACE) {
	    VovPlace place = node;
	    if ( place.edit() == VOV_OK ) place.modify();
	} else {
	    VovTransition tran = node;
	    if ( tran.edit() == VOV_OK ) tran.modify();
	}
    }
#else 
    vovMessage( "Function commented out" );
#endif
}


// Navigation through the trace.
// 

typedef struct {
    VovTrace* tracePtr;
    VovNode*  nodePtr;
    int   direction;
    octId newId;
    VovDialogList* listDlgPtr;
    VovDialogShowText* titleDlgPtr;
} VovLocalNavigateData;

void vov_navigate_update( VovLocalNavigateData& data )
{
    VovTrace& trace = *data.tracePtr;
    VovDialogList& listDlg = *data.listDlgPtr;
    VovNode& n = *data.nodePtr;

    listDlg.clear();
    int count  = 0;
    VovGenerator genElts;
    if ( data.direction == VOV_INPUT_MASK ) {
	VOV_ASSERT( trace.initGenContainers( n, VOV_NODE_OCTMASK, genElts ) );
    } else {
	VOV_ASSERT( trace.initGenContents( n, VOV_NODE_OCTMASK, genElts ) );
    }
    VovNode node;
    while ( trace.generate( genElts, node ) == VOV_OK ) {
	VOV_ASSERT( trace.getById( node ) );
	count++;
	listDlg.addItem( node.format(), node.id(), 0 );
    }
    if ( count == 0 ) {
	listDlg.addItem( "<--->", 0, 0 ); // Dummy, to prevent widget 0 size.
    }
    listDlg.endOfList( 0, 0 );


    char* dirString = ( data.direction == VOV_INPUT_MASK ) ? "INPUT" : "OUTPUT";
    
    VovDialogShowText& titleDlg = *data.titleDlgPtr;
    titleDlg.modifyText( form( "%d %s%s", count, dirString, (count != 1) ? "S":"" ) );
    
}

void vov_navigate_direction_ccb( ddsHandle item )
{
    VovDialogControl ctrl( item );
    VovLocalNavigateData& data = *(VovLocalNavigateData*)ctrl.getData();

    if ( ctrl.name() == "In" ) {
	data.direction = VOV_INPUT_MASK;
	ctrl.modifyName( "Out" );
    } else {
	data.direction = VOV_OUTPUT_MASK;
	ctrl.modifyName( "In" );
    }

    vov_navigate_update( data );
}

void vov_navigate_go_ccb( ddsHandle item )
{
    VovDialogControl ctrl( item );
    VovLocalNavigateData& data = *(VovLocalNavigateData*)ctrl.getData();
    VovDialogList& listDlg = *data.listDlgPtr;

    if (  listDlg.isSelected( listDlg.selectedIndex() ) ) {
	octId id = (int)listDlg.getWhichOne();
	if ( id != oct_null_id ) {
	    data.newId = id;	// Store the new id.
	    dds_destroy( item );
	    return;
	}
    } 
    vovConfirm( "Go where?", "You must select a new node where to go.", "Continue", 0 );

}

void vov_navigate_disconnect_ccb( ddsHandle item )
{
    VovDialogControl ctrl( item );
    VovLocalNavigateData& data = *(VovLocalNavigateData*)ctrl.getData();
    VovDialogList& listDlg = *data.listDlgPtr;

    int newId = (int)listDlg.getWhichOne();

    if ( (! listDlg.isSelected( listDlg.selectedIndex() )) || newId == oct_null_id ) {
	vovConfirm( "Disconnect what?", "You must select a new node to disconnect.", "Continue", 0 );
    } else {
	VovTrace& trace = *data.tracePtr;
	VovNode& node = *data.nodePtr;
	VovNode nn;
	nn.id() = newId;
	VOV_ASSERT( trace.getById( nn ) );
	VOV_ASSERT( trace.disconnect( node, nn ) );
	
	listDlg.modifyItem( "<--- disconnected --->", 0 , listDlg.selectedIndex() );
    }
}

VovStatus vovNavigate( VovTrace& trace, VovNode& n, int direction )
    // Return VOV_CHANGE, VOV_NO_CHANGE
{
    
    
    VovDialog top_dialog( "NAVIGATION DIALOG", DDS_VERTICAL, 6 );
    VovDialogShowText note(top_dialog, "", n.format(), 4, 60 );
    VovDialogShowText titleDlg( top_dialog, "", "" , 1, 60 );
    VovDialogList  dialog( top_dialog, DDS_TRUE );
    
    
    VovLocalNavigateData data;

    
    top_dialog.addControl( "Go" , vov_navigate_go_ccb, (ddsData)&data );
    
    if ( direction == VOV_OUTPUT_MASK ) top_dialog.addControl( "In", vov_navigate_direction_ccb, (ddsData)&data );
    if ( direction == VOV_INPUT_MASK )  top_dialog.addControl( "Out", vov_navigate_direction_ccb, (ddsData)&data );
    
    top_dialog.addControl( "Disconnect", vov_navigate_disconnect_ccb, (ddsData)&data );
    
    
    top_dialog.addHelp( "\
This dialog allows you to navigate through the trace following the\n\
input/output links.\n\
\n\
If you are now looking at the inputs of the current node, you can click\n\
`Out' to look at the outputs instead, and a similar operation is valid\n\
if you are now looking at the outputs.\n\
\n\
Remember that the trace is a bipartite graph, with two types of nodes:\n\
PLACES and TRANSITIONS. Thus, inputs and outputs of places are transitions,\n\
inputs and outputs of transitions are places.\n\n\
Each node here is identified by a short LABEL and by its status\n\
(e.g. VALID, NOT VALID). The meaning of status is explained by another dialog,\n\
the one used to change the status of a node.\n\
\n\
Select a node and then `Go' to it. If you decide you do not want to go anywhere,\n\
click on `Cancel'.\n\n\
FOR EXPERT USERS ONLY: Use `Disconnect' to eliminate a node from the list\n\
of inputs/outputs of the current node. This may be helpful \n\
to eliminate some erroneous or obsolete dependencies.\n\
" );
    
    data.direction = direction;
    data.tracePtr = &trace;
    data.nodePtr = &n;
    data.listDlgPtr = &dialog;
    data.titleDlgPtr = &titleDlg;
    data.newId = n.id();

    vov_navigate_update( data );
    top_dialog.postModed();
	
    if ( n.id() != data.newId ) {
	n.id() = data.newId;
	VOV_ASSERT( trace.getById( n ) );
	return VOV_CHANGE;
    } else {
	return VOV_NO_CHANGE;
    }
}


// Set editing.
// Special function to edit sets of nodes.

typedef struct {
    VovTrace* tracePtr;
    VovNodeSet* setPtr;
    VovDialogList* listDlgPtr;
} VovLocalEditSetData;

void vov_edit_set_all_ccb( ddsHandle item )
{
    VovDialogControl ctrl( item );
    VovLocalEditSetData& data = *(VovLocalEditSetData*)ctrl.getData();
    VovDialogList& list = *data.listDlgPtr;
    if ( ctrl.name() == "All" ) {
	list.doAll( DDS_TRUE );
    } else {
	list.doAll( DDS_FALSE );
    }
}

void vov_edit_set_op_ccb( ddsHandle item )
{
    VovDialogControl button( item );
    String ctrl = button.name();
    
    VovLocalEditSetData& data = *(VovLocalEditSetData*)button.getData();
    VovTrace& trace = *data.tracePtr;
    VovDialogList& elementsDlg = *data.listDlgPtr;
    VovNodeSet& set = *data.setPtr;

    elementsDlg.suspendRefresh();
    for ( int i = 0 ; i < elementsDlg.length() ; i++ ) {
	if ( ! elementsDlg.isSelected( i ) ) continue;
	VovNode node;
	node.id() =  (int)elementsDlg.getItemData( i );
	if ( trace.getById( node ) != VOV_OK ) continue;

	if ( ctrl == "View" ) {
	    if ( trace.edit( node, 0 ) == VOV_ERROR ) { 
		int remaining = 0;
		for ( int j = i+1; j < elementsDlg.length(); j++ ){
		    if ( elementsDlg.isSelected( j ) ) remaining++;
		}
		String q = form("%d more nodes scheduled to be edited." , remaining );
		if ( remaining == 0 || vovConfirm( "EDIT SET INTERRUPTED", q, "Continue Editing", "Abort" ) != "Continue") {
		    break;
		}
	    }
	} else if ( ctrl == "Trash" ) {
	    if ( node.nodeType() == VOV_PLACE ) {
		VovPlace p;
		p.id() = node.id();
		trace.getById( p );
		VOV_ASSERT( trace.freeResources( p ) );
	    }
	} else if ( ctrl == "Forget" ) {
	    VOV_ASSERT( trace.destroy( node ) );
	    elementsDlg.modifyItem( "<-- forgotten -->", 0 , i );
	} else if ( ctrl == "Detach" ) {
	    trace.setOperation( set, node, VOV_SET_DEL );
	    elementsDlg.modifyItem( "<-- detached -->", 0 , i );
	} else {
	    vovMessage( "What is this button: " + ctrl );
	    break;
	}
    }
    elementsDlg.refresh();
    // end for loop.
}

void vovEditSet( VovTrace& trace, VovNodeSet& set )
{
    octId lockId;
    int forceLock = 0;
    Date now;
    String desc = utcc_getenv("USER") + " " + now.format();
    String msg;
    if ( trace.lock( set, desc, forceLock, lockId, msg ) != VOV_OK ) {
	String q = "\n\n\tLOCK CONFLICT!! LOCK CONFLICT!! LOCK CONFLICT!!\n\n" +
	  set.format() + "\nis already locked.\n\n" + msg;
	String o1 = "Force the lock";
	String o2 = "Abort editing";
	if ( userDecision( q, o1, o2 ) != 1 ) {
	    return;
	}
	forceLock = 1;
	desc += " (FORCING)";
	VOV_ASSERT( trace.lock( set, desc, forceLock, lockId, msg ) );
    }

    VovGenerator genElts;
    VOV_ASSERT( trace.initGenContents( set, VOV_NODE_OCTMASK, genElts ) );
    
    VovLocalEditSetData data;
    VovNode node;
    
    int count_elts = 0;
    VovDialog top_dialog( "EDIT SET DIALOG", DDS_VERTICAL, 5 );
    VovDialogEditText titleDlg( top_dialog, "NAME",  set.getName(), 1 , 40  );
    VovDialogHandles handles( 10 );
    set.getRule().setupDialog( top_dialog, handles );

    VovDialog elementDlg( top_dialog, "ELEMENT LIST", DDS_VERTICAL, 2 );
    VovDialogList listDlg( elementDlg, DDS_FALSE );
    elementDlg.addControl( "All", vov_edit_set_all_ccb, (ddsData)&data );
    elementDlg.addControl( "None", vov_edit_set_all_ccb, (ddsData)&data );
	
    //
    // Get some nodes, up to MAX_ELTS_PER_DIALOG for each pane.
    // 
    while ( trace.generate( genElts, node ) == VOV_OK ) {
	VOV_ASSERT( trace.getById( node ) );
	count_elts++;
	String uniquelabel = node.format();
	listDlg.addItem( uniquelabel, node.id(), 0 );
    }
    listDlg.endOfList( 0, 0 );
	
    if ( count_elts == 0 ) {
	vovConfirm( "EDIT SET", form( "Set `%s':\nSet is empty.\n", (char*)set.getName()), "Continue", 0 );
	return;
    }

    data.tracePtr = &trace;
    data.setPtr = &set;
    data.listDlgPtr = &listDlg;

    top_dialog.addControl( "Detach", vov_edit_set_op_ccb, (ddsData)&data );
    top_dialog.addControl( "View", vov_edit_set_op_ccb, (ddsData)&data );
    top_dialog.addControl( "Forget",vov_edit_set_op_ccb, (ddsData)&data );
    top_dialog.addControl( "Trash", vov_edit_set_op_ccb, (ddsData)&data );
	
    top_dialog.addHelp( "\
You are editing a set of nodes in the trace.\n\
Select the nodes on which you want to operate,\n\
then choose the operation:\n\n\
`Forget': eliminate node from the trace\n\
`Trash':  physically eliminated the node. If the node is\n\
\ta place, it will be removed from disk.\n\
`Cancel': abort the editing of this set.\n\
" );
	
	
    top_dialog.postModed();	// Keep handles and set in scope.


    if ( trace.unlock( set, lockId ) != VOV_OK ) {
	vovError( "Unlocking failed" );
    }
    return;
}




// --------------------------------------------------
// --------------------------------------------------
// --------------------------------------------------
// Event queue editing.
// --------------------------------------------------
// --------------------------------------------------



typedef struct {
    int index;
    VovDialogEditText* eventTextPtr;
    VovDialogNumber*   eventNumberPtr;
    VovDialogList*     subjectMaskDlgPtr;
    VovDialogList*     eventMaskDlgPtr;
    VovEventQueue*     queuePtr;
    int sm;			// Subject mask.
    int em;			// Event mask.
} VovEditSetData;

void vov_edit_queue_update( VovEventQueue& queue, int index, VovDialogEditText& eventDlg, VovDialogNumber& indexDlg )
{
    String eventDescription;		
	
    if ( index >= queue.length() ) {
	eventDescription = "\n\n\tEnd of event list";
	index = queue.length();
    } else if ( index < 0 ) {
	eventDescription = "\n\n\tBeginning of event list";
	index = -1;
    } else {
	VovEvent& evt = queue.get( index );
	eventDescription = evt.format();
	eventDescription.gsub( path_canonicalize( utcc_getwd() ), String(".") );
    }
    eventDlg.modifyText( eventDescription );
    indexDlg.modifyValue( index );
}

void vov_edit_queue_index_ccb( ddsHandle item )
{
    if ( dds_type( item ) == DDS_NUMBER ) {
	VovDialogNumber number( item );
	VovEditSetData& data = *(VovEditSetData*)number.getData();
	data.index =  (int)number.getValue();
	vov_edit_queue_update( *data.queuePtr, data.index, *data.eventTextPtr, *data.eventNumberPtr );
    } else {
	VovDialogControl ctrl( item );
	VovEditSetData& data = *(VovEditSetData*)ctrl.getData();
	VovEventQueue& queue = *data.queuePtr;
	if ( ctrl.name() == "Next" ) {
	    data.index = queue.newIndexForEventQueueEdit( data.index, data.sm, data.em, +1 );
	} else {
	    data.index = queue.newIndexForEventQueueEdit(  data.index, data.sm, data.em, -1 );
	}
	vov_edit_queue_update( *data.queuePtr, data.index, *data.eventTextPtr, *data.eventNumberPtr );
    }
}


void vov_edit_queue_unreported_ccb( ddsHandle item )
{
    VovDialogControl ctrl( item );
    VovEditSetData& data = *(VovEditSetData*)ctrl.getData();
    VovEventQueue& queue = *data.queuePtr;
    int newevents = VOVsynchNotify( queue.getTrace(), queue, 400, 1 );
    if ( newevents == 0 ) {
	vovShortMessage( "Information", "No more events yet." );
    } else {
	vovShortMessage( "Information", form( "%d events added", newevents ) );
    }
    
    VovDialogNumber& numberDlg = *data.eventNumberPtr;
    numberDlg.modifyBounds( 0, queue.length() - 1 );
    vov_edit_queue_update( *data.queuePtr, data.index, *data.eventTextPtr, *data.eventNumberPtr );
}

void vov_edit_queue_event_mask_ccb( ddsHandle item, int toggled )
{
    VovDialogList list( item );
    VovEditSetData& data = *(VovEditSetData*)list.getData();
    data.em = list.reduceOR();	// Could be made more efficient
    // vovMessage( form( "em = %d", data.em ) );
}

void vov_edit_queue_subject_mask_ccb( ddsHandle item , int toggled )
{
    VovDialogList list( item );
    VovEditSetData& data = *(VovEditSetData*)list.getData();
    data.sm = list.reduceOR();
    // vovMessage( form( "sm = %d", data.sm ) );
}

void vov_edit_queue_edit_ccb( ddsHandle item )
{
    VovDialogControl ctrl( item );

    int viewOnlyFlag = ( ctrl.name() == "View" );

    VovEditSetData& data = *(VovEditSetData*)ctrl.getData();
    VovEventQueue& queue = *data.queuePtr;
    int index = data.index;
    
    VovEvent& evt = queue.get( index );
    if ( evt.getId() == 0 ) {
	vovConfirm("Warning", "The object you want to edit is not retrievable at this time.","Continue",0 );
	return;
    } 
    
    VovTrace& trace = queue.getTrace();
    int mask = evt.getSubjectMask();
    
    if ( mask ==  VOV_TRANSITION_MASK ) {
	VovTransition trans;
	trans.id() = evt.getId();
	if ( trace.getById( trans ) != VOV_OK ) {
	    vovShortMessage( "Warning", "This transition no longer exists.");
	} else {
	    trace.edit( trans, viewOnlyFlag );
	}
    } else if ( mask ==  VOV_PLACE_MASK ) {
	VovPlace place;
	place.id() = evt.getId();
	if ( trace.getById( place ) != VOV_OK ) {
	    vovShortMessage("Warning", "This place no longer exists.");
	} else {
	    trace.edit( place, viewOnlyFlag );
	}
    } else if ( mask ==  VOV_NODE_SET_MASK ) {
	VovNodeSet set;
	set.id() = evt.getId();
	if ( trace.getById( set ) != VOV_OK ) {
	    vovShortMessage( "Warning", "This set no longer exists." );
	} else {
	    vovEditSet( trace, set );
	}
    } else {
	vovConfirm("Sorry", "This function is not yet implemented.\n", "Continue", 0);
    }
}


void vov_edit_queue_toggle_all_ccb( ddsHandle item )
{
    static ddsBoolean flag = DDS_FALSE;
    flag = ( flag == DDS_FALSE ) ? DDS_TRUE : DDS_FALSE;
    
    VovDialogControl ctrl( item );
    VovEditSetData& data =  *(VovEditSetData*)ctrl.getData();

    VovDialogList& subjectMaskDlg = *data.subjectMaskDlgPtr;
    VovDialogList& eventMaskDlg = *data.eventMaskDlgPtr;

    
    eventMaskDlg.doAll( flag );
    subjectMaskDlg.doAll( flag );
    
    data.em = eventMaskDlg.reduceOR();
    data.sm = subjectMaskDlg.reduceOR();
}

VovStatus VovEventQueue::edit()
{
    if ( length() == 0 ) {
	vovShortMessage( "EVENT QUEUE",  "No events in queue" );
	return VOV_OK;
    }
    
    VovEditSetData data;

    VovTrace& trace = *eqtrace;

    String title = "EVENT QUEUE FOR TRACE " + trace.getProjectName();
    VovDialog top_dialog( title, DDS_VERTICAL, 3 );
    int max = ( length() > 2 ) ? length()  - 1 : 2 ;
    VovDialogNumber indexDlg( top_dialog, "Index", 0, max, 1, 0, "%.0f", vov_edit_queue_index_ccb , (ddsData)&data );

    VovDialog filters( top_dialog, "FILTERS", DDS_HORIZONTAL, 2 );
    VovDialogList subjectMaskDlg( filters, DDS_FALSE );

    subjectMaskDlg.addItem( "TRANSITIONS", VOV_TRANSITION_MASK, 1 );
    subjectMaskDlg.addItem( "PLACES", VOV_PLACE_MASK ,0 );
    subjectMaskDlg.addItem( "RETRACE", VOV_RETRACE_MASK, 0 );
    subjectMaskDlg.addItem( "SLAVES", VOV_SLAVE_MASK, 0 );
    subjectMaskDlg.addItem( "SETS", VOV_NODE_SET_MASK, 0 );
    subjectMaskDlg.endOfList( vov_edit_queue_subject_mask_ccb, (ddsData)&data );
    
    VovDialogList eventMaskDlg( filters , DDS_FALSE);
    eventMaskDlg.addItem( "START", VE_START, 0 );
    eventMaskDlg.addItem( "END",   VE_END, 	1  );
    eventMaskDlg.addItem( "FORGET", VE_FORGET, 0  );
    eventMaskDlg.addItem( "CHANGE", VE_CHANGE, 0 );
    eventMaskDlg.addItem( "ERROR",  VE_ERROR,  1 );
    eventMaskDlg.endOfList( vov_edit_queue_event_mask_ccb, (ddsData)&data );
    
    filters.addControl( "Toggle all", vov_edit_queue_toggle_all_ccb, (ddsData)&data );
   
    VovDialogEditText ed( top_dialog, "Event", "<no desc>", 10, 70 );
    
    top_dialog.addControl( "Previous", vov_edit_queue_index_ccb, (ddsData)&data );
    top_dialog.addControl( "Next", vov_edit_queue_index_ccb, (ddsData)&data );
    top_dialog.addControl( "View", vov_edit_queue_edit_ccb, (ddsData)&data );
    top_dialog.addControl( "Edit", vov_edit_queue_edit_ccb, (ddsData)&data );
    top_dialog.addControl( "Unreported", vov_edit_queue_unreported_ccb, (ddsData)&data );
    
    
    String helpStr = "\
You are looking at a list of events happened to the trace.\n\
You can browse the list with the `Previous' and the `Next' buttons,\n\
or with the horizontal scrollbar.\n\
\n\
You can select the type of events you want either by subject or\n\
by event type (see `Filter subject' and `Filter event' pane).\n\n\
`Unreported': get more events.\n\
`Edit': get more information about an event; you will be presented with\n\
\tthe object that was the subject of the event.\n\
";
    top_dialog.addHelp( helpStr );
    

    data.index = 0;
    data.sm = subjectMaskDlg.reduceOR();
    data.em = eventMaskDlg.reduceOR();
    
    data.subjectMaskDlgPtr = &subjectMaskDlg;
    data.eventMaskDlgPtr = &eventMaskDlg;
    data.eventTextPtr = &ed;
    data.eventNumberPtr = &indexDlg;
    data.queuePtr = this;

    vov_edit_queue_update( *this, data.index, ed, indexDlg );

    top_dialog.postModed( );	// Must be moded to keep queue in scope.

}


// ---------------------------------------------------
// ---------------------------------------------------
// Editing of annotations.
// ---------------------------------------------------
// ---------------------------------------------------
typedef struct {
    VovTrace* tracePtr;
    VovSuperObject* objPtr;
    VovAnnotation* notePtr;
    VovAnnotation* array;

    VovDialogEditText* noteDlgPtr;
    VovDialogNumber* indexDlgPtr;
    VovDialogHandles* handlesPtr;
    
    int index;
    int count;
} VovLocalNotesData;

void vov_notes_update( VovLocalNotesData& data )
{
    int index = data.index;
    VovDialogEditText& noteDlg = *data.noteDlgPtr;
    VovDialogNumber& indexDlg = *data.indexDlgPtr;
    
    String note_desc;
    if ( index >= data.count ) {
	note_desc = "\n\n\tEnd of note list.";
	index = data.count;
    } else if ( index < 0 ) {
	note_desc = "\n\n\tBeginning of note list.";
	index = -1;
    } else {
	note_desc = data.array[index].format();
    }
    noteDlg.modifyText( note_desc );
    indexDlg.modifyValue( data.index );
}

void vov_notes_index_ccb( ddsHandle item )
{
    if ( dds_type( item ) == DDS_NUMBER ) {
	VovDialogNumber number( item );
	VovLocalNotesData& data = *(VovLocalNotesData*)number.getData();
	data.index =  (int)number.getValue();
	vov_notes_update( data );
    } else {
	VovDialogControl ctrl( item );
	VovLocalNotesData& data = *(VovLocalNotesData*)ctrl.getData();
	if ( ctrl.name() == "Next" ) {
	    data.index++;
	} else {
	    data.index--;
	}
	vov_notes_update( data );
    }
}


void vov_notes_add_ccb( ddsHandle item )
{
    VovDialogControl ctrl( item );
    VovLocalNotesData& data = *(VovLocalNotesData*)ctrl.getData();

    if ( data.count == VOV_MAX_ANNOTATIONS ) {
	vovError("Lazy programmer:cannot add this node. Dismiss and try again" ); // 
	return;
    }
	
    VovAnnotation& note = *data.notePtr;
    VovDialogHandles& handles = *data.handlesPtr;
    VovTrace& trace = *data.tracePtr;
    VovSuperObject& obj = *data.objPtr;
    VovAnnotation* array = data.array;

    if ( note.readDialog( handles ) != VOV_CHANGE ) {
	vovConfirm("ATTENTION","\
In order to add a new note, you must\n\
make some changes to the editable note.\n\
", "Continue", 0 );
	return;
    }

    if ( trace.addAnnotation( obj, note ) == VOV_OK ) {
	array[data.count] = note;
	data.index = data.count++;
    }
    vov_notes_update( data );
}

void vov_notes_forget_ccb( ddsHandle item )
{
    VovDialogControl ctrl( item );
    VovLocalNotesData& data = *(VovLocalNotesData*)ctrl.getData();

    VovAnnotation& note = *data.notePtr;
    VovDialogHandles& handles = *data.handlesPtr;
    VovTrace& trace = *data.tracePtr;
    VovSuperObject& obj = *data.objPtr;
    VovAnnotation* array = data.array;


    if ( data.index >= 0 && data.index < data.count && array[data.index].id() != 0 ) {
	trace.destroy( array[data.index] );
	VovAnnotation newnote("<No longer existing>" );
	array[data.index] = newnote;
	array[data.index].id() = 0; // Cannot forget it again.
    }
    vov_notes_update( data );
}

void vovEditNotes( VovTrace& trace, VovSuperObject& obj )
{
    // I think this automatic allocation of array causes a problem on ds3100, upon destruction.
    // VovAnnotation array[arraySize];
    // Replace with this  dynamic allocation.
    static VovAnnotation* array = 0;
    static int arraySize = VOV_MAX_ANNOTATIONS;		// Very large number of annotations.
    if ( array == 0 ) {
	array = new VovAnnotation[arraySize];
    }


    VovLocalNotesData data;
    data.index = 0;
    data.count = trace.numberOfAnnotations( obj );

    if ( data.count > arraySize ) {
	vovError( form( "More than %d notes: cannot handle that (bug on ds3100 g++)", arraySize ) );
	data.count = arraySize;
    }

    if ( data.count ) {
	int n = trace.getAnnotations( obj, array, arraySize );
	if ( n != data.count ) vovError( form( "could not get the right number of notes %d instead of %d", n, data.count ) );
    }
    
    VovAnnotation note("");	// A new note to be added if needed
    VovDialog top_dialog( "EDIT ANNOTATIONS", DDS_VERTICAL, 20 );

    VovDialogHandles handles( 4 );
    
    note.setupDialog( top_dialog, handles );
    
    VovDialogShowText descDlg( top_dialog, "", obj.format(), 4, 70 );
    VovDialogNumber   indexDlg( top_dialog, "INDEX", 0, VOV_MAX_ANNOTATIONS - 1, 1, 0 , "%.0f", vov_notes_index_ccb, (ddsData)&data );
    VovDialogEditText noteDlg( top_dialog, "Note", "<no desc>", 10, 70 );
    
    
    top_dialog.addControl( "Previous", vov_notes_index_ccb, (ddsData)&data  );
    top_dialog.addControl( "Next", vov_notes_index_ccb, (ddsData)&data  );
    top_dialog.addControl( "Add" , vov_notes_add_ccb, (ddsData)&data  );
    top_dialog.addControl( "Forget", vov_notes_forget_ccb, (ddsData)&data );
    
    String helpStr = "\
You are looking at the annotations which have been attached to this object.\n\n\
You can browse the list of annotations using the `Previous',\n\
the `Next' buttons, or the scroll bar.\n\
\n\
If you want to add a new note, use to top part of the menu to compose the note\n\
then click `Add'.\n\n\
Click on `Forget' to eliminate an existing annotation.\n\
";

    data.tracePtr = &trace;
    data.objPtr = &obj;
    data.notePtr = &note;
    data.array = array;
    data.indexDlgPtr = &indexDlg;
    data.noteDlgPtr = &noteDlg;
    data.handlesPtr = &handles;


    top_dialog.addHelp( helpStr );
    vov_notes_update( data );
    top_dialog.postModed( );
}



// -----------------------------------------------------
// -----------------------------------------------------
// -----------------------------------------------------
// -----------------------------------------------------
// -----------------------------------------------------

typedef struct {
    VovNode* nodePtr;
    VovTrace* tracePtr;
    int flags;
    VovNodeStatus status;
} VovEditStatusData;

void vov_flags_status_ccb( ddsHandle item, int previous )
{
    VovDialogList dialog( item );
    VovEditStatusData& data = *(VovEditStatusData*)dialog.getData();

    // String sel = dialog.getItemData();
    // vovMessage( "Flags should be " + sel ); // Must change status now.
    
    data.flags = (int)dialog.getItemData();
}

void vov_status_status_ccb( ddsHandle item, int previous )
{
    VovDialogList dialog( item );
    VovEditStatusData& data = *(VovEditStatusData*)dialog.getData();
    data.status = (VovNodeStatus)dialog.getWhichOne();

    // vovMessage( "Status is " + vov_getStatusString( data.status ) );

}

void vov_ok_status_ccb( ddsHandle item )
{
    VovDialogControl ctrl( item );
    VovEditStatusData& data = *(VovEditStatusData*)ctrl.getData();
    VovNode& node = *data.nodePtr;
    VovTrace& trace = *data.tracePtr;
    VovNodeStatus newstatus = data.status;

    int flags = data.flags;
    
    if ( flags == 0 ) {
	switch ( newstatus ) {
	case VOV_PLAN_NODE:
	case VOV_DEAD_NODE:
	    flags = VOV_CHANGE_STATUS_SELF_REC;
	    break;
	case VOV_TRACE_NODE:
	case VOV_NONEXISTING_NODE:
	case VOV_TRACING_NODE:
	case VOV_RETRACING_NODE:
	default:
	    flags = VOV_CHANGE_STATUS_SELF;
	}
    }

    if ( newstatus != node.getStatus() ) {
	VOV_ASSERT( trace.changeStatus( node, newstatus, flags ) );
    }
    dds_destroy( item );
}

void vov_touch_status_ccb( ddsHandle item )
{
    VovDialogControl ctrl( item );
    VovEditStatusData& data = *(VovEditStatusData*)ctrl.getData();
    VovNode& node = *data.nodePtr;
    VovTrace& trace = *data.tracePtr;

    VovNodeStatus cur_status = node.getStatus();
    VovNodeStatus new_status;
    switch ( node.getStatus() ) {
    case VOV_PLAN_NODE:
	VOV_ASSERT( trace.changeStatus( node, VOV_TRACE_NODE, VOV_CHANGE_STATUS_SELF ) );
	VOV_ASSERT( trace.changeStatus( node, VOV_PLAN_NODE, VOV_CHANGE_STATUS_SELF_REC ) );
	break;
    case VOV_TRACE_NODE:
	VOV_ASSERT( trace.changeStatus( node, VOV_PLAN_NODE, VOV_CHANGE_STATUS_SELF_REC ) );
	VOV_ASSERT( trace.changeStatus( node, VOV_TRACE_NODE, VOV_CHANGE_STATUS_SELF ) );
	break;
    default:
	vovConfirm("Wrong node status for a touch","I can only touch nodes which are either VALID or NOT VALID",
		   "Continue", 0);
    }
    dds_destroy( item );
}

void vovEditNodeStatus( VovTrace& trace,  VovNode& node )
{
    VovEditStatusData data;
    data.tracePtr = &trace;
    data.nodePtr = &node;
    data.flags = 0;
    data.status = node.getStatus();

    VovDialog top_dialog( "EDIT NODE STATUS", DDS_VERTICAL, 4  );
    
    VovDialog flagsDlg( top_dialog, "", DDS_HORIZONTAL, 2 );
    
    VovNodeStatus s = node.getStatus();
    VovDialogList statusDlg( flagsDlg, DDS_TRUE );
    statusDlg.addItem( vov_getStatusString( VOV_TRACE_NODE ), VOV_TRACE_NODE, s == VOV_TRACE_NODE );
    statusDlg.addItem( vov_getStatusString( VOV_PLAN_NODE ), VOV_PLAN_NODE, s == VOV_PLAN_NODE );
    statusDlg.addItem( vov_getStatusString( VOV_DEAD_NODE ), VOV_DEAD_NODE, s == VOV_DEAD_NODE );
    if ( s != VOV_TRACE_NODE && s != VOV_PLAN_NODE && s != VOV_DEAD_NODE ) {
	statusDlg.addItem( vov_getStatusString( node.getStatus() ), node.getStatus(), 1 );
    }
    statusDlg.endOfList( vov_status_status_ccb, (ddsData)&data  );
    
    VovDialogList flagDialog( flagsDlg , DDS_TRUE );
    
    flagDialog.addItem( "AUTOMATIC", 0 , 1);
    flagDialog.addItem( "ONLY SELF", VOV_CHANGE_STATUS_SELF, 0 );
    flagDialog.addItem( "ONLY DEPENDENT", VOV_CHANGE_STATUS_RECURSIVE, 0 );
    flagDialog.addItem( "SELF AND DEPENDENT", VOV_CHANGE_STATUS_SELF_REC, 0 );
    flagDialog.endOfList( vov_flags_status_ccb, (ddsData)&data );
    
    VovDialogShowText title( top_dialog, "NODE", node.getName(), 2, 80 );
    
    
    top_dialog.addControl("Ok", vov_ok_status_ccb, (ddsData)&data );
    top_dialog.addControl("Touch" , vov_touch_status_ccb, (ddsData)&data );
    top_dialog.addHelp("\
This dialog allows you to change the status of a node in the trace.\n\
\n\
Choose a new status for this node.\n\
VALID     means that the node is good, that it does\n\
          not need retracing.\n\
NOT VALID means that the node needs to be retraced.\n\
DEAD      means that the server ignores the node:\n\
          no checking, no retracing.\n\
\n\
`Ok' select a new status and click `Ok'.\n\
`Touch' can be applied only to nodes whose status is\n\
  either VALID or NOT VALID. Touching means to change \n\
  the status from VALID to NOT VALID or viceversa and \n\
  to immediately change it back to the previous value.\n\
  It is a good way to get VOV to invalidate all nodes below \n\
  the one being edited.\n
\n\
You should generally let VOV decide automatically how to propagate\n\
the status from one node to all its dependents. \n\
ONLY SELF: changes the status of the node;\n\
ONLY DEPENDENT: changes the status of all dependent nodes;\n\
SELF AND DEPENDENT: changes the status of the node and of all\n\
     its dependents.\n\
AUTOMATIC: let VOV decide which nodes should be affected.\n\
");
    
    top_dialog.postModed();
    
    return ;  /* VOID */
}


// -----------------------------------------------------
// -----------------------------------------------------
// -----------------------------------------------------
// -----------------------------------------------------
typedef struct {
    VovTrace* tracePtr;
    VovSuperObject* objPtr;
    VovDialogHandles* handlesPtr;
} VovTraceEditData;

void vov_navigate_ccb( ddsData item )
{
    // vovMessage( "In vov_navigate_ccb" );
    VovDialogControl ctrl( item );
    VovTraceEditData& data = *(VovTraceEditData*)ctrl.getData();
    VovTrace& trace = *data.tracePtr;
    VovSuperObject& obj = *data.objPtr;

    int direction = ( ctrl.name() == "In" ) ? VOV_INPUT_MASK : VOV_OUTPUT_MASK ;

    VovNode node;
    node.id() = obj.contents.object->id();
    if ( trace.getById( node  ) != VOV_OK ) {
	vovConfirm( "ERROR" , form( "Obsolete id: %d", node.id() ), "Continue", 0 );	      		
	return;
    }
    VovStatus s = vovNavigate( trace, node, direction );
    if ( s == VOV_CHANGE ) {
	trace.edit( node, 1 );
	dds_destroy( item );
    } 
}

void vov_notes_ccb( ddsHandle item )
{
    VovDialogControl ctrl( item );
    VovTraceEditData& data = *(VovTraceEditData*)ctrl.getData();
    VovTrace& trace = *data.tracePtr;
    VovSuperObject& obj = *data.objPtr;

    vovEditNotes( trace, obj );
} 

void vov_change_status_ccb( ddsHandle item )
{
    VovDialogControl ctrl( item );
    VovTraceEditData& data = *(VovTraceEditData*)ctrl.getData();
    VovTrace& trace = *data.tracePtr;
    VovSuperObject& obj = *data.objPtr;

    vovEditNodeStatus( trace, *obj.contents.node );
}

void vov_edit_ccb( ddsHandle item )
{
    VovDialogControl ctrl( item );
    VovTraceEditData& data = *(VovTraceEditData*)ctrl.getData();
    VovTrace& trace = *data.tracePtr;
    VovSuperObject& obj = *data.objPtr;

    dds_destroy( item );

    trace.edit( obj );
}

void vov_apply_changes_ccb( ddsHandle item )
{
    VovDialogControl ctrl( item );
    VovTraceEditData& data = *(VovTraceEditData*)ctrl.getData();
    VovTrace& trace = *data.tracePtr;
    VovSuperObject& obj = *data.objPtr;
    VovDialogHandles& handles = *data.handlesPtr;

    if ( obj.readDialog( handles ) == VOV_CHANGE ) {
	VOV_ASSERT( trace.modify( obj ) );
    }
}

void vov_retrace_ccb( ddsHandle item )
{
    VovDialogControl ctrl( item );
    VovTraceEditData& data = *(VovTraceEditData*)ctrl.getData();

    VovSuperObject& obj = *data.objPtr;
    VovTrace& trace = *data.tracePtr;
    String msg;		
    if ( obj.type == VOV_TRANSITION ) {
	VovTransition& t = *obj.contents.transition;
	VovRetrace retracetransition( "to " + t.getCommandLine() );
	VOV_ASSERT( trace.create( retracetransition ) );
	trace.evolve( retracetransition, t, VOV_BACKWARD, VOV_RETRACE_SLOW, msg );
    } else {
	VovPlace& p = *obj.contents.place;
	VovRetrace retraceplace( "to " + p.format() );
	VOV_ASSERT( trace.create( retraceplace ) );
	trace.evolve( retraceplace, p, VOV_BACKWARD, VOV_RETRACE_SLOW, msg );
    }
    vovShortMessage( "RETRACE" , (char*)msg );	      
}


VovStatus VovTrace::edit( VovSuperObject& obj, int viewonly = 0 )
{
    // Nodes cannot be edited as such.
    if ( obj.type == VOV_NODE ) {
	VovNode& node = *obj.contents.node;
	if ( node.nodeType() == VOV_PLACE ) {
	    VovPlace p;
	    p.id() = node.id();
	    getById( p );
	    return edit( p, viewonly );
	} else {
	    VovTransition t;
	    t.id() = node.id();
	    getById( t );
	    return edit( t, viewonly );
	}
    }

    Bzocket& zoc = getBzocket();
    VovStatus rtn = VOV_OK;
    octId lockId;		// Number of lock, in case object needs to be locked.
    
    if ( ! viewonly ) {
	int forceLock = 0;
	Date now;
	String desc = utcc_getenv("USER") + " " + now.format();
	String msg;
	if( lock( *obj.contents.object, desc, forceLock, lockId, msg ) != VOV_OK ) {
	    String q = "\n\n\tLOCK CONFLICT!! LOCK CONFLICT!! LOCK CONFLICT!!\n\n" +
	      obj.format() + "\nis already locked.\n\n" + msg;
	    String o1 = "Force the lock";
	    String o2 = "Abort editing";
	    if ( userDecision( q, o1, o2 ) != 1 ) {
		return VOV_NO_CHANGE;
	    }
	    forceLock = 1;
	    desc += " (FORCING)";
	    VOV_ASSERT( lock( *obj.contents.object, desc, forceLock, lockId, msg ) );
	}
    }
    
    String title = form( "%s DIALOG", viewonly ? "VIEW":"EDIT" );
    VovDialog top_dialog( title, DDS_VERTICAL, 15 );
    top_dialog.addHelp( "You are editing an object in the trace.\n\n" );
    
    
    VovDialogHandles handles( 10 ); // Dialog handles for readDialog/setupDialog.
    
    VovTraceEditData data;
    data.tracePtr = this;
    data.objPtr = &obj;
    data.handlesPtr = &handles;
    
    obj.setupDialog( top_dialog, handles, viewonly );
    
    // No notes can be attached to other notes.
    if ( obj.type != VOV_ANNOTATION ) {
	int notes = numberOfAnnotations( obj );
	if ( notes ) {
	    top_dialog.addControl( "See notes", vov_notes_ccb, (ddsData)&data );
	    top_dialog.addHelp( "`See notes': this object has been annotated. Click here to look at the notes.\n\n" );
	}
	top_dialog.addControl( "Add notes", vov_notes_ccb, (ddsData)&data );
	top_dialog.addHelp( "`Add notes': add a new note to this object.\n\n" );
    }
    
    if ( obj.type == VOV_PLACE || obj.type == VOV_TRANSITION ) {
	top_dialog.addControl( "Change status", vov_change_status_ccb, (ddsData)&data );
	top_dialog.addControl( "In", vov_navigate_ccb, (ddsData)&data );
	top_dialog.addControl( "Out", vov_navigate_ccb, (ddsData)&data );
	top_dialog.addHelp("\
`Change status': this object is a node and it has a status, which\n\tcan be changed with this button.\n\n\
`In',`Out': these buttons allow you to navigate through the trace.\n\n\
");
	top_dialog.addControl( "Retrace", vov_retrace_ccb, (ddsData)&data );
	top_dialog.addHelp( "`Retrace': schedule the transition for retracing.\n\n" );
	
    }
    
    if ( viewonly ) {
	top_dialog.addControl( "Edit", vov_edit_ccb, (ddsData)&data );
    } else {
	top_dialog.addControl( "Apply changes", vov_apply_changes_ccb, (ddsData)&data );
    }
    
    top_dialog.postModed();	// It must be moded to maintain handles valid.
    
    
    
    if ( ! viewonly ) {
	if ( unlock( *obj.contents.object, lockId ) != VOV_OK ) {
	    vovError( "Unlocking failed: somebody must have overwritten the lock." );
	}
    }
    
    return rtn;
    // End of ::edit()
}



























