/* C++ includes */
#include <stream.h>
#include <String.h>

/* C includes */
#include "port.h"
#include "errtrap.h"
/* Local includes */

#include "voverror.h"


#define VOV_MAIL_ADDRESS "octtools@ic.Berkeley.EDU"

void vov_dflt_output_func( const String& s )
{
    cerr << s ;
}

static void (*output_function)( const String& s ) = vov_dflt_output_func;

VovErrorOutputFunc vovSetErrorOutputFunction( VovErrorOutputFunc func )
{
    VovErrorOutputFunc rtn = output_function;
    output_function = func;
    return rtn;
}

void vovFatalError( String& s)
{
    errRaise("vov",-1,"Fatal Error: %s", (char*)s);
}

void vovError( String& s )
{
    (*output_function)( "VOV error: " + s + "\n" );
}

void vovMessage( String& s )
{
    (*output_function)( "VOV message: " + s + "\n" );
}

void vovBug( String& s )
{
    String errmsg = "VOV internal bug discovered: " +  s + "\n";
    errmsg += form("\tPlease mail complaints to %s\n",VOV_MAIL_ADDRESS);
    (*output_function)( errmsg );
    errRaise( "vov", -2, "BUG: %s", (char*)s );
}


class VovErrorRecord {
    String desc;
    String cause;
    String fix;
public:
    VovErrorRecord();
    ~VovErrorRecord();
    
    String format();
};


VovErrorRecord::VovErrorRecord()
{
    desc = 0;
    cause = 0;
    fix = 0;
}

VovErrorRecord::~VovErrorRecord()
{
}


String VovErrorRecord::format()
{
    String s = desc;
    if ( cause != 0 ) {
	s += "\nCAUSE: " + cause;
    }
    if ( fix != 0 ) {
	s += "\nPOSSIBLE SOLUTION: " + fix;
    }
    return s;
}
