#include "trace.h"
#include "place.h"
#include "transition.h"
#include "VOVprotocol.h"
#include "voverror.h"

#include "comm.h"

struct history_data {
    String* string;		// The string collecting the data
    int   min_level;		// 
    int   max_level;		// 
    int   flags;
    int   direction;
};


VovStatus his_visit(VovNode& node, struct history_data* data)
{
    const int MAX_TEXT_LENGTH = 10000; 	// Must have some bounds, anyway.
    
    String newText;
    int nodevisited = 0;
 
    if ( data->flags & VOV_NODE_MASK ) {
	newText = "\n" +  node.format();
	nodevisited = 1;
    } else if ( node.nodeType() == VOV_PLACE  && ( data->flags & VOV_PLACE_MASK )  ) {
	VovPlace p( node );
	newText +=  ("\n" + p.getStatusString() + " " + p.format());
	nodevisited = 1;
    } else if ( node.nodeType() == VOV_TRANSITION && ( data->flags & VOV_TRANSITION_MASK ) ) {
	VovTransition t( node );
	newText +=( "\n" + t.getStatusString() + " " + t.getCommandLine()) ;
	nodevisited = 1;
    }
    if ( nodevisited && ( data->flags & VOV_ANNOTATION_MASK ) ) {
	String notes = node.formatNotes();
	if ( notes != 0 ) {
	    notes.gsub( "\n", "\n# " );
	    newText += "\n# " + notes;
	}
    }
    
    // Make sure the history string does not get too long.
    static int truncated = 0;
    if ( data->string->length() + newText.length() < MAX_TEXT_LENGTH ) {
	*(data->string) += newText;
	truncated = 0;
	return VOV_OK;
    } else {
	if ( ! truncated ) {
	    *(data->string) += "\n\nWARNING:History is too long: truncating.";
	    truncated = 1;
	}
	return VOV_ERROR;
    }
}

VovStatus  VT_history(VovNode& node, void* d )
{
    struct history_data* data = (struct history_data*)d;
    if ( node.getLevel() <= data->max_level && node.getLevel() >= data->min_level) {
	return  his_visit( node, data );
    } else {
	return VOV_NOT_OK;
    }
}


//
// Prepare a textual description of the history of a place
// 

VovStatus VovTrace::history(VovNode& node, 		// The place 
			    int flags,			// 
			    int direction,		// 
			    int level, 			// 
			    String& historyString	// RETURNED string.
			    )
{
    VovStatus rtn = VOV_OK;
    Bzocket& zoc = getBzocket();

    if ( isClient() ) {
	// place.canonicalize();
	zoc.pack( VOV_GetHistory );
	zoc.pack( level );
	zoc.pack( direction );
	zoc.pack( flags );
	zoc.pack( node.id() );
	zoc.flush();
    } else {
	if ( isServer(VOV_SERVER_ENTRY) ) {
	    level = zoc.unpackInt(); 	
	    direction = zoc.unpackInt();
	    flags = zoc.unpackInt();	
	    node.id() = zoc.unpackInt();
	    if ( node.get() != VOV_OK ) rtn = VOV_ERROR;
	}
	
	if ( rtn == VOV_OK ) {
	    if ( node.nodeType() == VOV_PLACE ) {
		VovPlace place( node );
		historyString += place.getName() + "\n" ;
		historyString += ( place.getStatusString() + "  " + place.getTimeStamp().format() + "\n");
		historyString += place.formatMeasures();
	    } else {
		VovTransition trans( node );
		historyString += trans.format() + "\n";
	    }
	    String notes = node.formatNotes();
	    if ( notes !=  0 ) {
		notes.gsub( "\n", "\n# ");
		historyString += "# " + notes ;
	    }
	    
	    struct history_data data;
	    data.string = &historyString;
	    data.min_level = node.getLevel() - level;
	    data.max_level = node.getLevel() + level;
	    data.flags = flags;
	    data.direction = direction;
	    if ( node.bfs( direction, VT_history, 0, 0, 0, (void*)&data )!=VOV_OK) {
		vovMessage("History not completed");
	    }
	}
	if ( isServer(VOV_SERVER_EXIT) ) {
	    zoc.pack( historyString );
	    zoc.pack( rtn );
	    zoc.flush();
	}
    }
    if ( isClient() ) {
	VC_get_reply( zoc );
	historyString = zoc.unpackString();
	rtn = zoc.unpackInt();
    }
    return rtn;
}

