// File kill.cc
//
#include "transition.h"
#include "utcc.h"
#include "voverror.h"
#include "buf.h"
#include "utility.h"

extern "C" {
#include "//usr/include/sys/wait.h"
#include <sys/time.h>
#include <sys/resource.h>
int wait( );
int wait3();
typedef union wait WaitStatus;
}

extern void vov_get_redirect( Buf& buf );
extern String vov_end_redirect( Buf& buf );

VovStatus VovTransition::waitCompletion( int* ret, String& errmsg )
    // Wait for the transition to complete.
    // Returns return status of completed transition.
{
    VovStatus rtn = VOV_OK;	// Value returned by this function.
    WaitStatus status;		// Return status from system call wait();
    struct rusage rusage;	// 
    int w;			// Returned by wait().
    
    *ret = 127;			// Default error condition.
    
    Buf pipeBuf;		// A buffer to read from the pipe.
    do {
	w = ::wait3( (int*)&status, 0, &rusage ); 			// 
	vov_get_redirect( pipeBuf );
	if (w == -1) {							// Check for no children.
	    vovFatalError("waiting for transition to complete" );	// 
	}								// 
	if ( w != pid.integer() ) {					// 
	    vovError( form( "Strange child died: pid %d != %d",		// 
			     w,pid.integer()));				// Continue anyway.
	}								// 
    } while ( w != pid.integer() ) ;					// 
    
    //
    // Read more characters from the pipe, then close the pipe.
    // Close filedescriptors associated with the pipe.
    // Remember that fildes[1] is already closed.
    //
    errmsg = vov_end_redirect( pipeBuf );
    
    if ( WIFSIGNALED( status ) ) {
	errmsg +=  form("Transition received signal %d.\n",status.w_termsig);
	rtn = VOV_ERROR;
    }
    if ( WIFSTOPPED( status ) ) {
	errmsg += form( "Transition stopped  by signal %d.\n", status.w_stopsig);
	rtn = VOV_ERROR;
    }
    if ( status.w_coredump ) {
	errmsg += "Transition dumped its core";
	rtn = VOV_ERROR;
    }

    if ( WIFEXITED( status ) ) {
	*ret = status.w_retcode;
	if ( *ret == 127 ) {
	    errmsg += "Transition could not be fired.\n";
	    rtn =  VOV_ERROR;
	}
    }
    if ( *ret != 0 ) {
	// Print a message only if status is not the normal 0
	vovMessage( form("Status is %d", *ret ) );
    }
    return rtn;
}

VovStatus VovTransition::kill( int signal )
    // Kill a transition that is running.
    // Three cases:
    // 1 TRANSITION active on local host: just use kill();
    // 2 TRANSITION started by slave on different host: send signal to slave.
    // 3 TRANSITION active on remote host, with known PID: use rsh to send signal.
{
    if ( pid.integer() == 0 ) {
	vovError( "Illegal pid = 0 in kill" );
	return VOV_ERROR;
    }

    if ( getStatus() == VOV_RETRACING_NODE ) {
	if ( retracingInfo.integer() ) {
	    VovTransition rt = retracingInfo.integer();
	    rt.kill( SIGKILL );
	}
    }


    if ( host.string() != utcc_gethostname() ) {
	// Need a kill on a different host.
	String cmd = form( "rsh %s kill -%d %d \"&\"", (char*)host.string(), signal, pid.integer() );
	vovMessage( "Killing command: " + cmd );
	int ss = util_csystem( (char*)cmd ) ; // System status
	if (ss == 127) {
	    vovError( "Cannot exec the kill" );
	    return VOV_ERROR;
	}
    } else {
	vovMessage( form( "Direct kill -%d %d", signal, pid.integer() ) );
	if ( ::kill( pid.integer(), signal ) == -1 ) {
	    utcc_printerrno( "Cannot kill transition" );
	    return VOV_ERROR;
	}
    }
    return VOV_OK;
}


// End of kill.cc ---------------

