#include "node.h"
#include "arc.h"
#include "placematrix.h"
#include "travfunc.h"
#include "voverror.h"
#include "invalidate.h"

extern void vov_setGeoColor( VovObject* , char* );
extern char* vov_nodeColorName( VovNodeStatus );


VovNode::VovNode( )
    :
    (OCT_INSTANCE),
    label( OCT_LABEL ),
    level("LEVEL", OCT_INTEGER ),
    status( "STATUS", OCT_INTEGER ),
    searchId("SEARCH_ID", OCT_INTEGER ),
    node_type("NODETYPE", OCT_INTEGER )
{
    objectId = 0;
}

#ifndef VOVCLIENT
VovNode::VovNode( octId i )
    :
    (OCT_INSTANCE),
    label( OCT_LABEL ),
    level("LEVEL", OCT_INTEGER ),
    status( "STATUS", OCT_INTEGER ),
    searchId("SEARCH_ID", OCT_INTEGER ),
    node_type("NODETYPE", OCT_INTEGER )
{
    objectId = i;
    if ( i ) {
	if (  getById( VOV_NODE_OCTMASK ) != OCT_OK ||  get() != VOV_OK ) {
	    vovError( form("VovNode:: get by id (%d) failed.", i ) );
	    objectId = 0;	// 
	}
    }
}
#endif

VovNode::VovNode( char* cell , VovType nt )
    :
    (OCT_INSTANCE),
    label( OCT_LABEL ),
    level("LEVEL", OCT_INTEGER ),
    status( "STATUS", OCT_INTEGER ),
    searchId("SEARCH_ID", OCT_INTEGER ),
    node_type("NODETYPE", OCT_INTEGER )
{

    status = VOV_PLAN_NODE;
    node_type = nt;
    searchId = 0;

    label.contents.label.label = "<no label>";
    // Place and create the label
    label.contents.label.textHeight = LABEL_TEXT_HEIGHT ;
    label.contents.label.vertJust = OCT_JUST_CENTER;
    label.contents.label.horizJust = OCT_JUST_CENTER;
    label.contents.label.lineJust = OCT_JUST_CENTER;
    label.contents.label.region.lowerLeft.x = 0;
    label.contents.label.region.lowerLeft.y = 0;
    label.contents.label.region.upperRight.x = 0;
    label.contents.label.region.upperRight.y = 0;

    contents.instance.master = cell;
    contents.instance.view = "plan";			// The view is now fixed!!
    contents.instance.facet = "contents";
    contents.instance.version = OCT_CURRENT_VERSION;
    contents.instance.transform.transformType = OCT_NO_TRANSFORM;
    contents.instance.transform.translation.x = 0;
    contents.instance.transform.translation.y = 0;

    level = -1;			// Illegal value, to force the setlevel to act.
}

VovNode::VovNode( VovNode& n ) 
    :
    ( n ),
    label( OCT_LABEL ),
    level("LEVEL", OCT_INTEGER ),
    status( "STATUS", OCT_INTEGER ),
    searchId("SEARCH_ID", OCT_INTEGER ),
    node_type("NODETYPE", OCT_INTEGER )
{
    VovObject::operator=( n );
    contents.instance = n.contents.instance;
    label = n.label;
    level = n.level;
    status = n.status;
    node_type = n.node_type;
    searchId = n.searchId;
}

VovNode::~VovNode( )
{
}

void VovNode::operator=( VovNode& n) 
{
    VovObject::operator=( n );
    contents.instance = n.contents.instance;
    label = n.label;
    level = n.level;
    status = n.status;
    node_type = n.node_type;
    searchId = n.searchId;
}

void VovNode::setName( String& n )
{
    contents.instance.name = utcc_strsav_and_track( (char*)n );
    label.contents.label.label = contents.instance.name;
}

String VovNode::getName()
{
    return contents.instance.name;
}

#ifndef VOVCLIENT
VovStatus VovNode::create( )
{
    VovObject box( OCT_BOX );		// Used to give a color to the node.

    OH_ASSERT( octCreate( this, &level ) );
    OH_ASSERT( octCreate( this, &status ) );
    OH_ASSERT( octCreate( this, &node_type ) );
    OH_ASSERT( octCreate( this, &searchId ) );

    

    OH_ASSERT( octBB( this, &box.contents.box ) );
    OH_ASSERT( octCreate( this, &box ) );
    vov_setGeoColor( &box, vov_nodeColorName( status.integer() ) );


    OH_ASSERT( octCreate( this, &label ) );
    vov_setGeoColor( &label, LABEL_COLOR ); 

    level = -1; setLevel( 1 );	// Force replacement.
    return VOV_OK;
}
#endif

#ifndef VOVCLIENT
VovStatus VovNode::get( )
{
    VovStatus rtn = VOV_ERROR;
    if ( getById( VOV_NODE_OCTMASK ) == OCT_OK ) {
	if ( octGenFirstContent( this, OCT_LABEL_MASK, &label ) != OCT_OK ) {
	    vovError("Missing label for " + getName() );
	    OH_ASSERT( octCreate( this, &label ) );
	}
	if (octGetByName( this, &level ) == OCT_OK 
	    && octGetByName( this, &status ) == OCT_OK
	    && octGetByName( this, &node_type ) == OCT_OK
	    && octGetOrCreate( this, &searchId ) == OCT_OK
	    )
	  {
	      rtn = VOV_OK;
	  }

    }
    if ( rtn == VOV_ERROR ) {
	vovError( "Corrupt node:" + getName() );
	id() = 0;
    }
    return rtn;
}
#endif

#ifndef VOVCLIENT
VovStatus VovNode::modify( )
{
    OH_ASSERT( octModify( this ) );
    OH_ASSERT( octModify( &label ) );
    OH_ASSERT( octModify( &level ) );
    OH_ASSERT( octModify( &status ) );
    OH_ASSERT( octModify( &node_type ) );
    OH_ASSERT( octModify( &searchId ) );
    return VOV_OK;
}
#endif

#ifndef VOVCLIENT
VovStatus VovNode::del( )
{

    extern octPoint octToMatrix( octPoint point );

    disconnect( VOV_IOS_MASK );

    VovPlacementMatrix* matrix = getOrAddPlacementMatrix(0);
    octPoint oldCoord = octToMatrix( contents.instance.transform.translation );
    if (matrix->del( id(), oldCoord.y, oldCoord.x ) == VOV_ERROR) {
	vovError("doPlacement: error deleting from matrix");
    }
    VovObject box(OCT_BOX);
    OH_ASSERT( octGenFirstContent( this, OCT_BOX_MASK, &box ) );

    OH_ASSERT( octDelete( &box ) );
    OH_ASSERT( octDelete( &label ) );
    OH_ASSERT( octDelete( &level ) );
    OH_ASSERT( octDelete( &status ) );
    OH_ASSERT( octDelete( &node_type ) );
    OH_ASSERT( octDelete( &searchId ) );
    OH_ASSERT( octDelete( this ) );
    return VOV_OK;
}
#endif


#ifndef VOVCLIENT
int VovNode::numberOfInputs()
{
    int count = 0;
    VovContainersGenerator genNode( this, VOV_NODE_OCTMASK );
    VovNode node = 0;
    while ( octGenerate( &genNode, &node ) == OCT_OK ) count++;
    return count;
}
#endif


#ifndef VOVCLIENT
int VovNode::numberOfOutputs()
{
    return ohCountContents( this, VOV_NODE_OCTMASK );
}
#endif

#ifndef VOVCLIENT
VovStatus VovNode::hasInputs()
{
    VovObject obj;
    if ( octGenFirstContainer( this, VOV_NODE_OCTMASK, &obj ) == OCT_OK )  {
	return VOV_OK;
    } else {
	return VOV_NOT_OK;
    }
}
#endif

#ifndef VOVCLIENT
VovStatus VovNode::hasOutputs()
{
    VovObject obj;
    if ( octGenFirstContent( this, VOV_NODE_OCTMASK, &obj ) == OCT_OK )  {
	return VOV_OK;
    } else {
	return VOV_NOT_OK;
    }
}
#endif



String vov_getStatusString( VovNodeStatus status )
{
    switch ( status ) {
    case VOV_PLAN_NODE:
	return "NOT VALID";
    case VOV_TRACE_NODE:
	return "VALID";
    case VOV_TRACING_NODE:
	return "EXECUTING";
    case VOV_RETRACING_NODE:
	return "RETRACING";
    case VOV_DEAD_NODE:
	return "DEAD";
    case VOV_NONEXISTING_NODE:
	return "MISSING";
    case VOV_FIRST_STATUS_NODE:
    case VOV_LAST_STATUS_NODE:
    case VOV_UNSPECIFIED_NODE:
	return "WEIRD";
    default:
	return "BOGUS";
    }
}

String vov_getPlaceTypeString( VovPlaceType type )
{
    switch ( type ) {
    case VOV_UNIX_FILE:
	return "FILE";
    case VOV_OCT_FACET:
	return "FACET";
    case VOV_PROGRAM_OPTION:
	return "OPTION";
    case VOV_EXECUTABLE:
	return "EXEC";
    case VOV_EXIT_STATUS:
	return "EXIT";
    case VOV_BOOLEAN:
	return "BOOLEAN";
    case VOV_MEASURE:
	return "MEASURE";
    case VOV_FIRST_PLACE_TYPE:
    case VOV_LAST_PLACE_TYPE:
	return "*** WEIRD SENTINEL VALUE ***";
    default:
	return "ILLEGAL PLACE TYPE";
    }
}

String VovNode::getStatusString( ) 
{
    return vov_getStatusString( getStatus() ) ;
}

#ifndef VOVCLIENT
void  VovNode::setStatus( VovNodeStatus s )
{
    if ( getStatus() == s ) {
	return;			// Nothing to do.
    }
    status.modify( s );		// Modify the status property.
    
    // Change the color of the instance. Originally this was done by 
    // changing the instance view with ohReplaceInstance. This has the
    // disadvantage that the octId of the node is changed and that caused some
    // octGenerate to generate this node more than once. 
    // The alternative way is to just change the color of a box attached to the instance.

    VovObject box;
    if (octGenFirstContent( this, OCT_BOX_MASK , &box ) != OCT_OK ) {
	vovError("No box attached to this node. Cannot change its color" );
    } else {
	vov_setGeoColor( &box,  vov_nodeColorName( status.integer() ) );
    }

    // Change color of output arcs.
    VovObject term;
    ohGetByTermName( this, &term, "out" );
    VovContainersGenerator genPath( &term, OCT_PATH_MASK );
    VovArc arc;
    while ( octGenerate( &genPath, &arc ) == OCT_OK ) {
	arc.changeStatus( getStatus() );
    }
    
    return;
}
#endif

// This routine is a bit confusing.
// It should be rewritten with better semantics.
#ifndef VOVCLIENT
VovStatus VovNode::changeStatus( VovNodeStatus s, int flags, String& msg)
{
    VovStatus rtn = VOV_OK;
    VovNodeStatus oldstatus = getStatus();
    if ( oldstatus == s ) {
	return VOV_OK;			// No need to change status.
    }					// 
    
    if ( flags & VOV_CHANGE_STATUS_RECURSIVE ) {
	VovInvalidateInfo info( *this, s );
	info.revive = ( oldstatus == VOV_DEAD_NODE );
	if ( bfs( VOV_FORWARD, VT_invalidate, 0, 0, 0, (void*)&info ) != VOV_OK ) {
	    msg += "Failed recursive status change.";
	    rtn =  VOV_ERROR;
	} else {
	    msg += "\n"+ info.format();
	}
    }
    if ( flags & VOV_CHANGE_STATUS_SELF ) {
	setStatus( s );			// 
    }  else {
	if ( getStatus() != oldstatus ) setStatus( oldstatus );
    }

    return rtn;
}
#endif

#ifndef VOVCLIENT
void VovNode::setLevel( int l )
// The level l==-1 is a special level, illegal,
// The node is not placed at that level. Useful
// to assign a temporary illegal level to a node.
{
    if ( level.integer() != l ) {
	level.modify( l );
	if ( l != -1 ) 	doPlacement();
    }
}
#endif

#ifndef VOVCLIENT
void VovNode::levellize( int newlevel ) 
{

    if ( level.integer() < newlevel ) {
	setLevel( newlevel );
	int lev = -9999;			// Smaller than any legal level 
	bfs( VOV_FORWARD, VT_levelbefore, 0, VT_levelvisit_out, 0, &lev );

	// Some inputs might have to be relevelled also,
	// in order to bring them closer to this node.
	VovContainersGenerator genIn( this, VOV_NODE_OCTMASK );
	VovNode node = 0;
	while ( octGenerate( &genIn, &node ) == OCT_OK ) {
	    node.get();
	    node.levellize( node.getLevel() ); // Relevellize at same level.
	}
    }
    
    // Now move the node as close as possible to its outputs,
    // but only if the node has no inputs
    if ( hasInputs() == VOV_NOT_OK  && hasOutputs() == VOV_OK ) {
	int minLevel = 100000;	//  Impossibly high level.
	foreachOutput( VT_levelmin_output_level, (void*)&minLevel );
	setLevel( minLevel - 1) ;
    }
}
#endif



//
// Connectivity routines.
//

#ifndef VOVCLIENT
VovStatus VovNode::connectOutput(VovNode& n)
    // Returns: VOV_OK, VOV_CYCLE.
{
    extern VovStatus vov_cycle( VovNode& start, VovNode& end );
    
    if ( vov_cycle( n, *this ) == VOV_ERROR) {
	return VOV_CYCLE;
    }
    
    if ( octIsAttached ( this, &n ) == OCT_OK ) {
	vovError( form( "nodes %d and %d already attached", id(), n.id() ) );
    } else {
	OH_ASSERT( octAttach( this, &n ) );
	VovArc arc( *this, n, getStatus() );
	n.levellize( level.integer()+1 );
    }
    
    return VOV_OK;
}
#endif

#ifndef VOVCLIENT
VovStatus VovNode::disconnectOutput(VovNode& n)
{
    VovObject term;
    OH_ASSERT( ohGetByTermName( this, &term, "out" ) );
    VovContainersGenerator genPath( &term, VOV_ARC_OCTMASK );
    VovArc arc;
    VovNode other = 0;
    while ( octGenerate( &genPath, &arc ) == OCT_OK ) {
	 arc.other( *this, other );
	 if ( other.id() == n.id() ) { 
	    arc.del();
	    break;
	}
    }
    
    OH_ASSERT( octDetach( this, &n ) );
    return VOV_OK;
}
#endif


#ifndef VOVCLIENT
VovStatus VovNode::connectInput(VovNode& n)
{
    return n.connectOutput( *this );
}
#endif

#ifndef VOVCLIENT
VovStatus VovNode::disconnectInput(VovNode& n)
{
    return n.disconnectOutput( *this );
}
#endif


#ifndef VOVCLIENT
void VovNode::disconnect( int mask = VOV_IOS_MASK )
{
    if ( mask & VOV_INPUT_MASK ) {
	VovContainersGenerator genInputs( this, VOV_NODE_OCTMASK );
	VovNode node = 0;
	while ( octGenerate( &genInputs, &node ) == OCT_OK ) {
	    OH_ASSERT( octDetach( &node, this ) );
	}
	VovObject term;
	OH_ASSERT( ohGetByTermName( this, &term, "in" ) );
	VovContainersGenerator genPath( &term, VOV_ARC_OCTMASK );
	VovArc arc;
	while ( octGenerate( &genPath, &arc ) == OCT_OK ) {
	    arc.del();
	}

    }
    if ( mask & VOV_OUTPUT_MASK ) {
	VovContentsGenerator genInputs( this, VOV_NODE_OCTMASK );
	VovNode node = 0;
	while ( octGenerate( &genInputs, &node ) == OCT_OK ) {
	    OH_ASSERT( octDetach( this, &node ) );
	}
	VovObject term;
	OH_ASSERT( ohGetByTermName( this, &term, "out" ) );
	VovContainersGenerator genPath( &term, VOV_ARC_OCTMASK );
	VovArc arc;
	while ( octGenerate( &genPath, &arc ) == OCT_OK ) {
	    arc.del();
	}
    }
}
#endif




// Test whether a node is a direct input or output
// of another node.
#ifndef VOVCLIENT
VovStatus VovNode::isInput( VovNode& node ) 
{
    return  (octIsAttached( &node , this )==OCT_OK) ? VOV_OK : VOV_NOT_OK;
}
#endif

#ifndef VOVCLIENT
VovStatus VovNode::isOutput( VovNode& node ) 
{
    return  (octIsAttached( this , &node  )==OCT_OK) ? VOV_OK : VOV_NOT_OK;
}
#endif


#ifndef VOVCLIENT
void VovNode::updateLabel()
{
    VovObject box(OCT_BOX);
    OH_ASSERT( octGenFirstContent( this, OCT_BOX_MASK , &box ) );

    int boxhalfwidth = box.contents.box.upperRight.x - box.contents.box.lowerLeft.x;
    boxhalfwidth >>= 1;

    label.contents.label.region.lowerLeft.x = box.contents.box.lowerLeft.x - boxhalfwidth;
    label.contents.label.region.upperRight.x = box.contents.box.upperRight.x + boxhalfwidth;
    
    extern octPoint  octToMatrix( octPoint );
    octPoint p = octToMatrix( contents.instance.transform.translation );
    int col = p.x;

    if ( col & 0x1 ) {
	label.contents.label.region.lowerLeft.y = box.contents.box.upperRight.y;
	label.contents.label.region.upperRight.y = label.contents.label.region.lowerLeft.y + LABEL_REGION_HEIGHT ;
    } else {
	label.contents.label.region.upperRight.y = box.contents.box.lowerLeft.y;
	label.contents.label.region.lowerLeft.y = label.contents.label.region.upperRight.y - LABEL_REGION_HEIGHT ;
    }
    OH_ASSERT( octModify( &label ) );

    // Temporary test of consistency for the label;
    String l = label.contents.label.label;
    String n = contents.instance.name;
    if ( l != n ) {
	vovError( "Inconsistent label name: " + l );
	vovMessage( "Correct label name to: " + n );
	setName( getName() );
    }
    // VOID;
}
#endif


#ifndef VOVCLIENT
void VovNode::move( octPoint point )
{
    // Move instance.
    contents.instance.transform.translation = point; 	
    OH_ASSERT( octModify( this ) );

    // Move the colored box in the right place.
    VovObject box;
    OH_ASSERT( octGenFirstContent( this, OCT_BOX_MASK , &box ) );
    OH_ASSERT( octBB( this, &box.contents.box ) );
    OH_ASSERT( octModify( &box ) );

    // Adjust all arcs attached to node.
    VovObject term;
    VovContentsGenerator genTerm( this, OCT_TERM_MASK );
    while ( octGenerate( &genTerm, &term ) == OCT_OK ) {
	VovContainersGenerator genPath( &term, OCT_PATH_MASK );
	VovArc arc;
	while ( octGenerate( &genPath, &arc ) == OCT_OK ) {
	    arc.adjust();
	}
    }
    // Update the label.
    updateLabel();
}
#endif


String VovNode::format()
{
    String s = form( "[%s] %-9s ", ( nodeType() == VOV_TRANSITION ) ? "T" : "P", (char*)getStatusString() );
    s += getName();
    s += form( " <%d>", id() );
    return s;
}


