#include "node.h"

#include "voverror.h"
#include "date.h"

#define DEFAULT_SET_OWNER "VOV_SERVER"

VovNodeSet::VovNodeSet( )
    :
    ( "<Set with no name>" ),
    owner( "OWNER", OCT_STRING ),
    created( "CREATED", OCT_INTEGER ),
    selectRule()
{
}


VovNodeSet::VovNodeSet( String& name )
    :
    ( utcc_strsav_and_track( (char*)name ) ),
    owner( "OWNER", OCT_STRING ),
    created( "CREATED", OCT_INTEGER ),
    selectRule( 0, 0, 0, 0 )
{
    owner = DEFAULT_SET_OWNER;
    created = getDate();
}

VovNodeSet::VovNodeSet( String& name, VovSelectRule& rule  )
    :
    ( utcc_strsav_and_track( (char*)name ) ),
    owner( "OWNER", OCT_STRING ),
    created( "CREATED", OCT_INTEGER ),
    selectRule( rule )
{
    owner = DEFAULT_SET_OWNER;
    created = getDate();
}


VovNodeSet::VovNodeSet( VovNodeSet& set )
    :
    ( set ),
    owner( "OWNER", OCT_STRING ),
    created( "CREATED", OCT_INTEGER ),
    selectRule( set.selectRule )
{
    owner = set.owner;		// 
    created = set.created;
    selectRule = set.selectRule;
}

void VovNodeSet::operator=( VovNodeSet& set )
{
    VovObject::operator=( set );
    contents.bag = set.contents.bag;
    owner = set.owner;		// 
    created = set.created;
    selectRule = set.selectRule;
}

VovNodeSet::~VovNodeSet()
{
}

#ifndef VOVCLIENT
VovStatus VovNodeSet::create()
{
    OH_ASSERT( octCreate( this, &owner ) );
    OH_ASSERT( octCreate( this, &created ) );
    OH_ASSERT( octCreate( this, &selectRule ) );
    selectRule.create();
    return VOV_OK;
}
#endif

#ifndef VOVCLIENT
VovStatus VovNodeSet::get()
{
    if ( octGetByName( this, &owner ) != OCT_OK ) {
	owner = DEFAULT_SET_OWNER;
	vovError("Set is missing owner" );
	OH_ASSERT( octCreate( this, &owner ) );
    }
    if ( octGetByName( this, &created ) != OCT_OK ) {
	created = getDate();
	vovError( "set is missing creation date" );
	OH_ASSERT( octCreate( this, &created ) );
    }
    if (  octGetByName( this, &selectRule )  != OCT_OK )  {
	OH_ASSERT( octCreate( this, &selectRule ) );
	vovError( "Set is missing selectrule" );
	selectRule.create();
    }
    selectRule.get();
    return VOV_OK;
}
#endif



#ifndef VOVCLIENT
VovStatus VovNodeSet::modify()
{
    OH_ASSERT( octModify( &owner ) );
    OH_ASSERT( octModify( &created ) );
    OH_ASSERT( octModify( &selectRule ) );
    selectRule.modify();
    return VOV_OK;
}
#endif

#ifndef VOVCLIENT
VovStatus VovNodeSet::del()
{
    OH_ASSERT( octDelete( &owner ) );
    OH_ASSERT( octDelete( &created ) );
    selectRule.del();
    OH_ASSERT( octDelete( &selectRule ) );
    OH_ASSERT( octDelete( this ) );
    return VOV_OK;
}
#endif


#ifndef VOVCLIENT
int VovNodeSet::select( VovNodeSet& set )
{
    int count = 0;
    VovContentsGenerator genNodes( &set, VOV_NODE_OCTMASK );
    VovNode node;
    // selectRule.debug = 1;
    while ( octGenerate( &genNodes, &node ) == OCT_OK ) {
	node.get();
	if ( selectRule.select( node ) ) {
	    add( node );
	    count++;
	}
    }
    // selectRule.debug = 0;
    return count;
}
#endif

#ifndef VOVCLIENT
int VovNodeSet::subselect( VovSelectRule& rule )
{
    int count = 0;
    VovContentsGenerator genNodes( this, VOV_NODE_OCTMASK );
    VovNode node = 0;
    while ( octGenerate( &genNodes, &node ) == OCT_OK ) {
	node.get();
	if ( ! rule.select( node ) ) {
	    del( node );
	} else {
	    count++;
	}
    }
    return count;
}
#endif

String VovNodeSet::format()
{
    String s = form( "Set: %s", contents.bag.name );
    if ( owner.string() != 0 ) s += ("\nowned by " + owner.string() );
    s += ("\nselection rule: " + selectRule.format() + "\n" );
    return  s;
}




#ifndef VOVCLIENT
VovStatus VovNodeSet::visit( VovNode& node, VovStatusFunc visitFunc, void* data ) 
{
    VovContainersGenerator genInput( &node, VOV_NODE_OCTMASK );
    VovNode in;
    while ( octGenerate( &genInput, &in ) == OCT_OK ) {
	VOV_ASSERT( in.get () );
	if ( in.alreadyVisited() ) continue; 	// If input already visited, skip it.
	if ( contains( in ) ) {			// If input node is in trace ...
	    visit( in, visitFunc, data );	// Visit the input first.
	}
    }
    
    (*visitFunc)( node, data );
    node.markVisited();
    return VOV_OK;
}
#endif 

#ifndef VOVCLIENT
VovStatus VovNodeSet::cover( VovStatusFunc visitFunc, void* data ) 
{
    VovNode node;
    node.newVisit();		// Start a new traversal on the trace.
    VovContentsGenerator genNode( this, VOV_NODE_OCTMASK );
    while ( octGenerate( &genNode, &node ) == OCT_OK )  {
	VOV_ASSERT( node.get() );
	if ( node.alreadyVisited() ) continue;
	visit( node, visitFunc, data );
    }
    return VOV_OK;
}
#endif
