#include <stream.h>
#include "base.h"
#include "utcc.h"
#include "bzocket.h"
#include "voverror.h"

// Construct an oct object. 
// Make sure that all important fields are initialized.
VovObject::VovObject()
{
    type = 0;
    objectId = OCT_NULL_ID;
}

VovObject::VovObject( int t )
{
    type = t;
    switch ( type ) {
    case OCT_INSTANCE:
	contents.instance.name = "";
	contents.instance.master = "";
	contents.instance.view = "";
	contents.instance.facet = "";
	contents.instance.version = "";
	break;
    case OCT_BAG:
	contents.bag.name = "";
	break;
    case OCT_LABEL:
	contents.label.label = "";
	break;
    case OCT_NET:
	contents.net.name = "";
	break;
    default:
	;
    }
}


VovObject::VovObject( VovObject& o ) 
{
    operator = ( o );
}

void VovObject::operator=( VovObject& o )
{
    objectId = o.objectId;
    type = o.type ;
    switch ( type ) {
    case OCT_INSTANCE:
	contents.instance.name = utcc_strsav_and_track(  o.contents.instance.name );
	contents.instance.master = utcc_strsav_and_track(  o.contents.instance.master );
	contents.instance.view = utcc_strsav_and_track(  o.contents.instance.view );
	contents.instance.facet = utcc_strsav_and_track(  o.contents.instance.facet );
	contents.instance.version = utcc_strsav_and_track(  o.contents.instance.version );
	contents.instance.transform.transformType = o.contents.instance.transform.transformType;
	contents.instance.transform.translation.x = o.contents.instance.transform.translation.x;
	contents.instance.transform.translation.y = o.contents.instance.transform.translation.y;
	break;
    case OCT_BAG:
	contents.bag.name = utcc_strsav_and_track(  o.contents.bag.name );
	break;
    case OCT_PROP:
	contents.prop.name = utcc_strsav_and_track(  o.contents.prop.name );
	contents.prop.type = o.contents.prop.type;
	switch ( contents.prop.type ) {
	case OCT_INTEGER:
	    contents.prop.value.integer = o.contents.prop.value.integer;
	    break;
	case OCT_STRING:
	    contents.prop.value.string = utcc_strsav_and_track(  o.contents.prop.value.string );
	    break;
	case OCT_REAL:
	    contents.prop.value.real = o.contents.prop.value.real;
	    break;
	default:
	    vovError("Wrong prop type in ::unpack()" );
	}
	break;
    case OCT_LABEL:
	contents.label.label = utcc_strsav_and_track(  o.contents.label.label );
	contents.label.region.lowerLeft.x = o.contents.label.region.lowerLeft.x;
	contents.label.region.lowerLeft.y = o.contents.label.region.lowerLeft.y;
	contents.label.region.upperRight.x = o.contents.label.region.upperRight.x;
	contents.label.region.upperRight.y = o.contents.label.region.upperRight.y;
	contents.label.textHeight = o.contents.label.textHeight;
	contents.label.vertJust = o.contents.label.vertJust;
	contents.label.horizJust = o.contents.label.horizJust;
	contents.label.lineJust = o.contents.label.lineJust;
	break;
    case OCT_NET:
	contents.net.name  = utcc_strsav_and_track( o.contents.net.name );
	break;
    default:
	vovMessage( "Cannot unpack this object: " + format() );
    }
}


VovObject::~VovObject(  )
{
}


void VovObject::pack( Bzocket& z )
{
    z.pack( objectId );
    z.pack( type );
    switch ( type ) {
    case OCT_INSTANCE:
	z.pack( contents.instance.name );
	z.pack( contents.instance.master );
	z.pack( contents.instance.view );
	z.pack( contents.instance.facet );
	z.pack( contents.instance.version );
	z.pack( contents.instance.transform.transformType );
	z.pack( contents.instance.transform.translation.x );
	z.pack( contents.instance.transform.translation.y );
	break;
    case OCT_BAG:
	z.pack( contents.bag.name );
	break;
    case OCT_PROP:
	z.pack( contents.prop.name );
	z.pack( contents.prop.type );
	switch ( contents.prop.type ) {
	case OCT_INTEGER:
	    z.pack( contents.prop.value.integer );
	    break;
	case OCT_STRING:
	    z.pack( contents.prop.value.string );
	    break;
	case OCT_REAL:
	    z.pack( form( "%e", contents.prop.value.real ) ); // Send a real as a string.
	    break;
	default:
	    vovError("Wrong prop type in ::pack()" );
	}
	break;
    case OCT_LABEL:
	z.pack( contents.label.label );
	z.pack( contents.label.region.lowerLeft.x );
	z.pack( contents.label.region.lowerLeft.y );
	z.pack( contents.label.region.upperRight.x );
	z.pack( contents.label.region.upperRight.y );
	z.pack( contents.label.textHeight );
	z.pack( contents.label.vertJust );
	z.pack( contents.label.horizJust );
	z.pack( contents.label.lineJust );
	break;
    default:
	vovMessage( "Cannot pack this object: " + format() );
    }
}

void VovObject::unpack( Bzocket& z )
{
    objectId = z.unpackInt();
    type = z.unpackInt();
    switch ( type ) {
    case OCT_INSTANCE:
	contents.instance.name = utcc_strsav_and_track( z.unpackStr() );
	contents.instance.master = utcc_strsav_and_track( z.unpackStr() );
	contents.instance.view = utcc_strsav_and_track( z.unpackStr() );
	contents.instance.facet = utcc_strsav_and_track( z.unpackStr() );
	contents.instance.version = utcc_strsav_and_track( z.unpackStr() );
	contents.instance.transform.transformType = z.unpackInt();
	contents.instance.transform.translation.x = z.unpackInt();
	contents.instance.transform.translation.y = z.unpackInt();
	break;
    case OCT_BAG:
	contents.bag.name = utcc_strsav_and_track( z.unpackStr() );
	break;
    case OCT_PROP:
	contents.prop.name = utcc_strsav_and_track( z.unpackStr() );
	contents.prop.type = z.unpackInt();
	switch ( contents.prop.type ) {
	case OCT_INTEGER:
	    contents.prop.value.integer = z.unpackInt();
	    break;
	case OCT_STRING:
	    contents.prop.value.string = utcc_strsav_and_track( z.unpackStr() );
	    break;
	case OCT_REAL:
	    char* s = z.unpackStr();
	    contents.prop.value.real = atof( s );
	    break;
	default:
	    vovError("Wrong prop type in ::unpack()" );
	}
	break;
    case OCT_LABEL:
	contents.label.label = 	utcc_strsav_and_track( z.unpackStr() );
	contents.label.region.lowerLeft.x = z.unpackInt();
	contents.label.region.lowerLeft.y = z.unpackInt();
	contents.label.region.upperRight.x =z.unpackInt();
	contents.label.region.upperRight.y =z.unpackInt();
	contents.label.textHeight = z.unpackInt();
	contents.label.vertJust = z.unpackInt();
	contents.label.horizJust =z.unpackInt();
	contents.label.lineJust = z.unpackInt();
	break;
    default:
	vovMessage( "Cannot unpack this object: " + format() );
    }
}

#ifndef VOVCLIENT
octStatus VovObject::getById( int typemask = OCT_ALL_MASK )
{
    octStatus rtn = octGetById( this );
    if ( rtn == OCT_OK ) {
	if ( ! ((1 << type ) & typemask) ) {
	    utcc_error("Get by id: good id, wrong type" );
	    rtn = OCT_ERROR;
	}
    }
    return rtn;
}
#endif


VovStatus VovObject::lock(  String& desc, int force, octId& lockId, String&  msg )
{
    VovProp lock( "LOCK", OCT_STRING );			// The lock is a STRING property.
    octStatus s = octGetByName( this, &lock );		// Does it exist already?
    if ( force || s == OCT_NOT_FOUND ) {		// If no, or if force is on, 
	lock = desc;					// Change value of prop.
	OH_ASSERT( octCreateOrModify( this, &lock ) );	// 
	lockId = lock.id();				// Return lockid, used to do unlocking.
	return VOV_OK;					// Locking succeded.
    } else if ( s == OCT_OK ) {				// The lock exists already.
	msg = "LOCKED: " + lock.string();		// Return descriptive message.
	return VOV_NOT_OK;				// Locking failed.
    } 
}

VovStatus VovObject::unlock( octId lockId )
{
    VovProp lock( "LOCK", OCT_STRING );			// The lock is a STRING property.
    octStatus s = octGetByName( this, &lock );		// Does it exist already?
    if ( s == OCT_NOT_FOUND ) {				// No, not locked.
	return VOV_OK;					// Good anyway, the object is unlocked.
    }							// 
    if ( lock.id() == lockId ) {			// The lock exists, do we have the right Id ?
	OH_ASSERT( octDelete( &lock ) );		// Yes, delete the lock.
	return VOV_OK;					// Place is now unlocked.
    }
    return VOV_NOT_OK;
}

String VovObject::getName()
{
    switch ( contents.prop.type ) {
    case OCT_BAG:
	return contents.bag.name;
    case OCT_PROP:
	return contents.prop.name;
    case OCT_INSTANCE:
	return contents.instance.name;
    case OCT_NET:
	return contents.net.name;
    default:
	utcc_error( format() + " has no name" ); 
	return "****";
    }
}

/* 
 * Print the object on stdout
 */
String VovObject::format()
{
    String s = form( "{[%d]}%s",objectId, ohFormatName( this ) );
    return s;
}


VovProp::VovProp( VovProp& p )
    : ( p.type )
{
    utcc_message( "new routine: I did not expect this to be used" );
}

VovProp::~VovProp()
{
}

String VovProp::format()
{
    String s = contents.prop.name;
    s += "=";
    switch ( contents.prop.type ) {
    case OCT_INTEGER:
	s += dec(contents.prop.value.integer);
	break;
    case OCT_STRING:
	s += form( "\"%s\"", contents.prop.value.string );
	break;
    case OCT_REAL:
	s += form("%lf",contents.prop.value.real);
	break;
    default:
	utcc_error("Wrong type for a property");
    }
    return s;
}


String VovProp::string()
{
    return contents.prop.value.string;
}

int VovProp::integer()
{
    return contents.prop.value.integer;
}

double VovProp::real()
{
    return contents.prop.value.real;
}

#ifndef VOVCLIENT
void  VovProp::modify( String& s )
{
    operator = ( s );
    OH_ASSERT( octModify( this ) );
}
#endif
#ifndef VOVCLIENT
void  VovProp::modify( int s )
{
    operator = ( s );
    OH_ASSERT( octModify( this ) );
}
#endif
#ifndef VOVCLIENT
void  VovProp::modify( double s )
{
    operator = ( s );
    OH_ASSERT( octModify( this ) );
}
#endif


VovBag::VovBag( char* name ) 
    : 
    (OCT_BAG)
{
    contents.bag.name = name;
}

VovBag::VovBag( VovBag& b ) 
    : 
    (b)
{
}

VovBag::~VovBag()
{
    // Just call the destructor for VovObject.
}

// Add elements of b to this bag.
// UNION operator
void VovBag::operator |= ( VovBag& b )
{
    VovContentsGenerator genObj( &b, OCT_ALL_MASK );
    VovObject obj;
    while ( octGenerate( &genObj, &obj ) == OCT_OK ) {
	OH_ASSERT(octAttachOnce( this, &obj ));
    }
}

// Intersection operator.
void VovBag::operator &= ( VovBag& b )
{
    VovContentsGenerator genObj( this, OCT_ALL_MASK );
    VovObject obj;
    while ( octGenerate( &genObj, &obj ) == OCT_OK ) {
	if ( octIsAttached( &b, &obj ) != OCT_OK ) {
	    OH_ASSERT( octDetach( this, &obj ) );
	}
    }
}

// COMPLEMENT operator
void VovBag::operator ^= ( VovBag& b )
{
    VovBag complement("xxx");
    VovObject facet(OCT_FACET);
    octGetFacet( this, &facet );
    octCreate( &facet, &complement );

    VovContentsGenerator genObj( &b, OCT_ALL_MASK );
    VovObject obj;
    while ( octGenerate( &genObj, &obj ) == OCT_OK ) {
	if ( octIsAttached( this, &obj ) != OCT_OK ) {
	    OH_ASSERT(octAttach( &complement, &obj ));
	}
    }
    clear( OCT_ALL_MASK );
    operator |= ( complement );
    OH_ASSERT(octDelete( &complement ));
}



String VovBag::format()
{
    String s = ( String(contents.bag.name) + ": " + dec( ohCountContents( this, OCT_ALL_MASK ) ) );
    return s;
}

int VovBag::length(int mask )
{
    if ( type != OCT_BAG ) {
	cout << "BUG: bag is not a bag!!\n ";
	return -1;
    }
    return ohCountContents( this, mask );
}

void  VovBag::add( VovObject& o )
{
    OH_ASSERT(octAttachOnce( this, &o ));
}

void  VovBag::del( VovObject& o )
{
    if (octIsAttached( this, &o ) == OCT_OK ) {
	OH_ASSERT(octDetach( this, &o ));
    }
}

int VovBag::contains( VovObject& o ) 
{
    return ( octIsAttached( this, &o ) == OCT_OK ) ;
}

void VovBag::clear( int mask )
{
    VovContentsGenerator genObj( this, mask );
    VovObject obj;		// Any obj type is OK.
    while ( octGenerate( &genObj, &obj ) == OCT_OK ) {
	OH_ASSERT(octDetach( this, &obj ));
    }
}


