#include "place.h"
#include "transition.h"
#include "user.h"
// #include "measure.h"

#include "voverror.h"

#define ABSOLUTE_PATH_TO_OCT_VOV_CELLS "$OCTTOOLS/lib/vov/OCTgraphics"


static char* initCellName( char* cell )
{
    char* n = new char[128];
    strcpy( n ,  ABSOLUTE_PATH_TO_OCT_VOV_CELLS);
    strcat( n , cell );
    return n;		// 
}

char* VovTransition::getCellName()
{
    static char* n = 0;
    if ( ! n ) {
	n = initCellName("/transition" );
    }
    return n;
}


char* VovPlace::getCellName( VovPlaceType type )
{
    static char* facet = 0;
    static char* exec = 0;
    static char* option = 0;
    static char* exit = 0;
    static char* file = 0;
    static char* bool = 0;
    static char* meas = 0;	// Measure
    if ( ! facet ) {
	facet = initCellName( "/place_oct" );
	exec  = initCellName( "/place_exec" );
	option= initCellName( "/place_option" );
	exit  = initCellName( "/place_status" );
	file  = initCellName( "/place" );
	bool  = initCellName( "/place_boole" );
	meas  = initCellName( "/place_measure" );
    }
    switch ( type ) {
    case VOV_OCT_FACET:
	return facet;
    case VOV_EXECUTABLE:
	return exec;
    case VOV_PROGRAM_OPTION:
	return option;
    case VOV_EXIT_STATUS:
	return exit;
    case VOV_BOOLEAN:
	return bool;
    case VOV_MEASURE:
	return meas;
    case VOV_UNIX_FILE:
	return file;
    default:
	return file;
    }
}

char* vov_nodeColorName( VovNodeStatus status )
{
    switch (  status ) {
    case VOV_NONEXISTING_NODE:
	return "PDIF";
    case VOV_TRACE_NODE:
	return "MET2";
    case VOV_PLAN_NODE:
	return "MET1";
    case VOV_TRACING_NODE:
	return "POLY";
    case VOV_RETRACING_NODE:
	return "NWEL";
    case VOV_DEAD_NODE:
	return "GLAS";
    default:
    case VOV_UNSPECIFIED_NODE:
	vovError("nodeColorName: Illegal node type");
	return "GLAS";		// NOT REACHED.
    }
}

void vov_setGeoColor( VovObject* geo, char* color )
{
    VovObject facet;
    octGetFacet( geo, &facet );	// This is a VOID function!!
    VovObject oldLayer;
    VovObject newLayer;
    if ( octGenFirstContainer( geo, OCT_LAYER_MASK, &oldLayer ) == OCT_OK ) {
	OH_ASSERT(octDetach( &oldLayer, geo ));
    }
    ohGetOrCreateLayer( &facet, &newLayer, color );
    OH_ASSERT(octAttach( &newLayer, geo ));
}



char* VovNode::getViewName()
{
    vovMessage( "This function should not be used" );
    switch ( status.integer() ) {
    case VOV_TRACE_NODE:
	return "vov";
    case VOV_PLAN_NODE:
	return "plan";
    case VOV_TRACING_NODE:
	return "fired";
    default:
	vovError("getViewName:: Illegal node type");
	return "plan";		// NOT REACHED.
    }
}


#ifdef OBSOLETE
char* VovGoal::getCellName()
{
    static char* n = 0;
    if ( ! n ) {
	n = initCellName( "/goal" );
    }
    return n;
}
#endif

char* VovUser::getCellName()
{
    static char* n = 0;
    if ( ! n ) {
	n = initCellName( "/user" );
    }
    return n;
}






