#include "port_internal.h"
#include "path.h"
#include "utcc.h"

#include "utility.h" 		/* The octtools utility. */

extern "C"
{
#include <sys/types.h>
#include <sys/stat.h>
int stat();
int lstat();
typedef struct stat stat_struct;
}

#define PATH_NET_PREFIX "/net/"
#define PATH_MAX_COMPONENTS  40
static String path_components[PATH_MAX_COMPONENTS];

extern String vov_stripWord( String& );

void path_read_equivalence_file()
{
    char* file = util_file_expand( "~/.vovequivalence" );
    if ( path_test( file ) ) {
	String value = utcc_popen( file );
	String var = 0;
	String path = 0;
	while ( value.length() ) {
	    if ( var == 0 ) {
		var = vov_stripWord( value );
	    } else {
		path = vov_stripWord( value );
		if ( path != 0  ) {
		    path_register_logical( (char*)var, (char*)path );
		    path_register_logical( (char*)var, (char*)path_expand( path ) );
		    var = 0;
		}
	    }
	}
    }
}

void path_init()
{
    static int init = 0;
    if ( !init ) {
 	init = 1;
	path_read_equivalence_file();
	path_register_logical( "OCTTOOLS", getenv( "OCTTOOLS" ) );
	path_register_logical( "OCTTOOLS", path_expand( getenv( "OCTTOOLS" )) );

#ifdef OLD_CODE
	// This code uses the intermediate variable OCTROOT.
	// This is no longer needed since we have the nicer 
	// mechanism provided by ~/.vovequivalence file.

	String ot = utcc_getenv( "OCTTOOLS" );
	String tail = ot.after( "/", -1 );
	String machine = utcc_machine();
	if ( tail == machine || tail == "common" ) {
	    String root =  ot.before( "/", -1 );
	    util_setenv( "OCTROOT", (char*)root );
	    path_register_logical( "OCTROOT", getenv( "OCTROOT" ) );
	    path_register_logical( "OCTTOOLS", "${OCTROOT}/common" );
	    path_register_logical( "OCTTOOLS", "${OCTROOT}/"+machine );
	    path_register_logical( "OCTTOOLS", path_expand( getenv( "OCTTOOLS" )) );
	}
#endif
    }
}


int path_test( const String& path )
{
    stat_struct  buf;
    if ( ::stat( (char*)path, &buf ) == -1) {
	return 0;				// Does not exist.
    } else {					// 
	return 1;				// Yes, path is valid.
    }
}


String path_hostNetPrefix()
{
    static char* prefix_static = 0;
    if ( ! prefix_static )  {
	String prefix = PATH_NET_PREFIX + utcc_gethostname();
	prefix_static = utcc_strsav( (char*)prefix );
    }
    return (String)prefix_static;
}

int path_hasNetPrefix( const String& path )
{
    Regex netPrefix = (String)PATH_NET_PREFIX + ".*";
    return path.matches( netPrefix );
}

String path_extractNetPrefix(   String& path )
{
    Regex rx = String(PATH_NET_PREFIX) + "[A-Za-z]+/";
    String prefix = path.at( rx );
    return prefix;
}

String path_stripNetPrefix(  String& path )
{
    if ( path_hasNetPrefix( path ) ) {
	String p = path.after( path_extractNetPrefix( path ) );
	return p;
    } else {
	utcc_error( "Cannot strip net prefix");
	return path;
    }
}

String path_localFull( const String& path )
    //
    // Get the fullpath, but keep it compressed 
    // with ~ and logical names.
    // 
{
    path_init();
    String newpath = path;
    if ( newpath.length() ) {
	if ( newpath[0] == '$' ) {
	    newpath = util_logical_expand( (char*)newpath );
	}
	if ( newpath[0] == '/' || newpath[0] == '~' ) {
	    return path;
	} else {
	    newpath = utcc_getwd() + "/" + path;
	}
    }
    return newpath;
}


String path_networkFull( const String& path )
{
    utcc_message( "Obsolete network full" );
    String newpath = path_localFull( path );

    if ( ! path_hasNetPrefix( newpath ) ) {
	newpath = path_hostNetPrefix() + newpath;
    }

    return newpath;
}

// Eliminate ".." and "." from a path
// /a/b/c/../d/e  --> /a/b/d/e
// /a/b/c/./d/e  -->  /a/b/c/d/e
String path_removedotdot( const String& path )
{
    String fullpath = path;
    if ( fullpath[0] != '/' ) {
	fullpath = path_localFull( path );
    }
    int    nextcomp[PATH_MAX_COMPONENTS];
    int    preccomp[PATH_MAX_COMPONENTS];
    Regex slash = "/";
    int count = split( fullpath, path_components, PATH_MAX_COMPONENTS, slash);
    
    // for (int j = 0; j<count; j++ ) cout << j << "  " << path_components[j] << "\n";

    int last = count-1;		// Index of the last component
    int first = 1;		// The first component is always empty when path begins with a /
    for (int i = first; i < last ; i++ ) {
	nextcomp[i] = i+1;
	preccomp[i] = i-1;
    }
    nextcomp[last] = -1;	// Illegal index
    preccomp[first] = -1;	// Illegal index
    String dot = ".";
    String dotdot = "..";
    if ( path_components[first] == dotdot ) {
	cout << path_components[first] << " matches ..\n";
	utcc_fatalerror( "bad full path" );
    }
    for ( i = first; i < last ; i++ ) {
	if ( path_components[i] == dotdot ) {
	    if ( preccomp[i] < 0 ) {
		utcc_error( "Illegal path" );
		return path;
	    }
	    int prec = preccomp[preccomp[i]];
	    if ( prec < 0 ) {
		first = nextcomp[i];
		preccomp[first] = -1;
	    } else {
		int next = nextcomp[i];
		nextcomp[prec] = next;
		if (next >= 0) preccomp[next] = prec;
	    }
	} else if ( path_components[i] ==  dot ) {
	    int prec = preccomp[i];
	    if (prec < 0 ) {
		first = nextcomp[i];
		preccomp[first] = -1;
	    } else {
		int next = nextcomp[i];
		nextcomp[prec] = next;
		if (next >= 0) preccomp[next] = prec;
	    }
	}
    }
    // Now build the resulting canonical path
    String ret;
    int index = first;
    while ( index >= 0 ) {
	ret += ("/" + path_components[index]);
	index = nextcomp[index];
    }
    return ret;
}

//
// Returns 0 if path is a file or a symbolic link and 1 otherwise.
// Returns -1 if an error occurs.
int path_isDirectory( const String& path ) 
{
    stat_struct buf;
    if ( lstat( (char*)path, &buf ) == -1) return -1;
    return  ( buf.st_mode & S_IFDIR ) ? 1 : 0;
}

int path_isSymbolicLink( const String& path ) 
    // RETURN  1 if path is a symbolic link and 
    //         0 otherwise.
    //        -1 if an error occurs.
{
    stat_struct buf;
    if ( lstat( (char*)path, &buf ) == -1) return -1;
    return  ( (buf.st_mode & S_IFMT) == S_IFLNK ) ? 1 : 0;
}
 
String path_removeSymbolicLinks( const String& path ) 
    // Break workpath into components.
    // Analyze a piece at a time until you find a symlink. 
    // Replace the symlink with its content and repeat. 
    // Stop when there are no symlinks in workpath.
    // RETURN workpath. 
{
    int    nextcomp[PATH_MAX_COMPONENTS];
    int    preccomp[PATH_MAX_COMPONENTS];
    Regex slash = "/";
    
    String workpath = path;
    int symlink_found = 0;
    do {
	int count = split( workpath, path_components, PATH_MAX_COMPONENTS, slash);
	// for (int j = 0; j<count; j++ ) cout << j << "  " << path_components[j] << "\n";
	
	int last = count-1;		// Index of the last component
	// Look for an empty component starting form the end.
	// Empty components are the same as a root.
	// Remember that there might be no empty components.
	int emptycomp = last;
	while ( path_components[emptycomp].length() != 0) {
	    if ( --emptycomp < 0 ) break;
	}
	int first =  emptycomp + 1;
	
	String symlink;
	symlink_found = 0;
	for ( int i = first; i <= last ; i++ ) {
	    if ( i == 0 ) {
		symlink = path_components[i]; 		// the path does not start from root
	    } else {
		symlink += ( "/" + path_components[i] );
	    }
	    int sl = path_isSymbolicLink( symlink ) ;
	    if ( sl == -1 ) {
		return path;	// Exit immediately on error
	    }
	    if ( sl ) {
		char buf[1024];
		int cc = readlink( (char*)symlink, buf, 1024 );
		if ( cc == -1 ) {
		    utcc_printerrno( (char*)symlink );
		    utcc_error("path_removeSymlk");
		}
		buf[cc] = '\0';		// terminate the string
		// cerr << "Replacing " << symlink << " with " << buf << "\n";
		if ( buf[0] == '/' ) {
		    path_components[i] = &buf[1];   // Skip the slash
		    first = i;
		} else {
		    path_components[i] = buf;
		}
		symlink_found = 1;
		break;
	    }
	}
	// Now build the resulting canonical path
	workpath = "";
	for (int index = first; index <= last; index++ ) {
	    if ( index == 0 ) {
		workpath = path_components[index];
	    } else {
		workpath += ("/" + path_components[index]);
	    }
	}
    } while ( symlink_found );
    return workpath;
}

// Get the name of the host which owns the file system
// in which the file is kept.
String path_host( const String& path )
{
    cerr << "In path_host: " << path << "\n";
    String host;
    String fullpath = path_networkFull( path );
    Regex netPrefix = (String)PATH_NET_PREFIX + ".*";
    if ( fullpath.matches( netPrefix ))  {
	 host = fullpath.after( PATH_NET_PREFIX );
	 host = host.before( "/" );
    }
    return host;
}

String path_lastComponent( String& path )
{
    int pos = path.index( "/" , -1 ); // Rightmost occurrence of '/' in path.
    if ( pos == -1 ) {
	return path;
    } else {
	return path.after( pos );
    }
}



// path_canonicalize()
// Main routine to make all paths good for all hosts in the local network.
// 
#define LOGICAL_ARRAY_SIZE 32
static char* logicalNameArray[LOGICAL_ARRAY_SIZE];
static char* logicalValueArray[LOGICAL_ARRAY_SIZE];

static int logicalCount = 0;
void path_register_logical( String& name, String& value )
{
    if ( logicalCount >= LOGICAL_ARRAY_SIZE ) {
	utcc_message( form("logical name array full (%d entries)", LOGICAL_ARRAY_SIZE ));
	return;
    }
    if ( value != 0 ) {
	char* actualName  = utcc_strsav( form("${%s}",(char*)name) );
	for ( int i = 0 ; i < logicalCount ; i++ ) {
	    if (value==String(logicalValueArray[i]) && String(actualName) == String(logicalNameArray[i]) ) {
		// utcc_message( "value/name already in array" );
		return;
	    }
	}
	// utcc_message( form( "Equivalent %s %s", actualName, (char*)value ) );
	logicalNameArray[logicalCount] = actualName;
	logicalValueArray[logicalCount] = utcc_strsav( (char*)value );
	logicalCount++;
    }
}

String path_logical_compress( const String& path )
{
    String s = path;
    int count = 0;		// Count succesful substitutions.
    for ( int i = 0 ; i < logicalCount; i++ ) {
	String o = logicalValueArray[i];
	String n =  logicalNameArray[i];
	if ( common_prefix( s, o ) == o ) {
	    s = n + s.after( o );
	}
    }
    if ( count ) utcc_message( path + " becomes " + s );
    return s;
}


String path_canonicalize( const String& path )
{
    path_init();

    String name = path;
    // name = util_logical_expand( (char*)name );  // Expand env variables.
    name = path_localFull( name ); 		// Get the full path first.
    name = path_expand( name );
    name = path_logical_compress( name ); 	// 
    return name;
}

String path_expand( const String& path )
    //
    // Expand file name, eliminating logical components and ~.
    // Eliminate symbolic links and all occurrences of ".."
    //
{
    path_init();

    String name ;
    name = util_logical_expand( (char*)path );  // Expand env variables.
    name = path_localFull( name ); 		// Get the full path  first.
    name = path_removeSymbolicLinks( name );	// 
    name = path_removedotdot( name );		// Remove ../'s and ./'s.
    
    return name;
}
