#include "place.h"
#include "placelist.h"
#include "voverror.h"

VovPlaceList::VovPlaceList ( int s )
{
    const int MIN_SIZE = 8;
    if ( s < MIN_SIZE ) {
	vovMessage( form("Place list size should be at least %d", MIN_SIZE) );
	size = MIN_SIZE;
    } else {
	size = s;
    }
    array = new VovPlaceListElt[size];
    n = 0;
}

VovPlaceList::~VovPlaceList()
{
    if ( size ) {
	clear();
	delete array;
    }
}

void VovPlaceList::clear()
{
#ifndef mips    
    for ( int i = 0; i < n ; i++ ) {
	delete array[i].placePtr;
    }
#endif
    n = 0;
}

void VovPlaceList::add( VovPlace& place, int ioType )
{
    if ( n >= size ) {
	int ns = size * 2;
	VovPlaceListElt* tmp = new VovPlaceListElt[ns];
	for ( int i = 0 ; i < n ; i++ ) {
	    tmp[i].placePtr = array[i].placePtr;
	    tmp[i].ioType = array[i].ioType;
	    tmp[i].status = array[i].status;
	}
	delete array;
	array = tmp;
	size = ns;		// This was missing!!
    }

    // place.canonicalize();    // This should not be necessary.
    array[n].placePtr = new VovPlace(place);
    array[n].ioType = ioType;
    array[n].status = 0;
    n++;
}

void VovPlaceList::add( VovPlaceType t , String& name, int ioType )
{
    VovPlace place( t, name );
    add( place, ioType );
}

void VovPlaceList::del( int index )
{
    checkBounds( index );
    delete array[index].placePtr;
    for ( int i = index ; i < length() - 1; i++ ) {
	array[i] = array[i+1];
    }
    n--;
}


int VovPlaceList::length()
{
    return n;
}

void VovPlaceList::checkBounds( int i )
{
    if ( i < 0 || i >= n ) {
	vovFatalError( form( "VovPlaceList:: index %d out of bound [0,%d]", i, n-1 ) );
    }
}

VovPlace& VovPlaceList::getPlace( int i )
{
    checkBounds( i );
    return *array[i].placePtr;
}
int& VovPlaceList::getType( int i )
{
    checkBounds( i );
    return array[i].ioType;
}
VovStatus& VovPlaceList::getStatus( int i )
{
    checkBounds( i );
    return array[i].status;
}

