/* LINTLIBRARY */
#include "port_internal.h"
#include "utcc.h" 

class PointersArray {
    char** array;
    int size;
    int count;
    void resize( int i );
public:
    PointersArray() { size = 0; count = 0; }
    ~PointersArray();
    
    void add( char* p );
    void del( char* p );
    int  length() { return count; }
    int  freeall();
};

PointersArray::~PointersArray()
{
    if ( size ) delete array;
}

void PointersArray::resize( int index )
{
    if ( size == 0 ) {
	array = new char*[256];
	size = 256;
	count = 0;
    } 

    if ( index >= size ) {
	do { size *= 2; } while ( index >= size );
	char** tmp = new char*[size];
	for (int i = 0; i < count; i++ ) tmp[i] = array[i];
	delete array;
	array = tmp;
    }
}

void PointersArray::add( char* p )
{
    if ( count >= size ) {	// Of course it should never be greater.
	// Array is full, must grow it.
	resize( count + 1 );
    }
    array[count++] = p;
}

void PointersArray::del( char* p )
{
    // A very inefficient way to find the pointer. But this should only be
    // needed for debugging.
    for ( int i = 0; i< count ; i++ ) {
	if ( array[i] == p ) {
	    delete p;
	    array[i] = array[--count];
	    return;
	}
    }
    utcc_error( form( "%s: Not an string allocated with utcc_strsav", (char*)p ) );
}

int PointersArray::freeall()
{
    int n = 0;
    for ( int i = 0 ; i < count ; i++ ) {
	// utcc_message( form( "Deleting string: %s", (char*)array[i] ) );
	n += strlen( (char*)array[i] );
	delete array[i];
    }
    if ( n ) {
	static int totalsaved = 0;
	totalsaved += n;
	// utcc_message( form( "Reclaimed %d strings, %d bytes (%d total)." , count, n, totalsaved ) );
    }
    count = 0;
    return n;
}

static PointersArray allocated_strings;

/*
 *  util_strsav -- save a copy of a string
 */
// Strings allocated with this function can be freed using
// utcc_strfreeall()
//
char*
utcc_strsav_and_track( const char* s  )
{
    char* ns = utcc_strsav( s );
    allocated_strings.add( ns );
    return ns;
}


char*
utcc_strsav( const char* s  )
{
    char* ns = new char[ strlen(s) + 1 ];
    (void)strcpy( ns , s );
    return ns;
}

void utcc_strfree( char* p ) 
{
    allocated_strings.del( p );
}

//
// Free all string allocated with utcc_strsav_and_track;
//
void utcc_strfreeall(  ) 
{
    allocated_strings.freeall( );
}
