#include "vovinternal.h"
#include "user.h"
#include "place.h"
#include "transition.h"
#include "annotation.h"
#include "retrace.h"
// #include "measure.h"		/*  */
#include "bzocket.h"
#include "voverror.h"

VovSuperObject::VovSuperObject()
{
    type = VOV_UNDEFINED;
    contents.object = 0 ;
    allocated = 0;
}

VovSuperObject::VovSuperObject( VovSuperObject& o )
{
    vovMessage( "VSO(VSO&)" );
    type = o.type;
    contents = o.contents;
    allocated = 0;
}

VovSuperObject::VovSuperObject( VovPlace& x)
{
    type = VOV_PLACE; contents.place = &x;
    allocated = 0;
}

VovSuperObject::VovSuperObject( VovTransition& x)
{
    type = VOV_TRANSITION; contents.transition = &x;
    allocated = 0;
}

VovSuperObject::VovSuperObject( VovNode& x)
{
    type = VOV_NODE; contents.node = &x;
    allocated = 0;
}

VovSuperObject::VovSuperObject( VovNodeSet& x)
{
    type = VOV_NODE_SET; contents.set = &x;
    allocated = 0;
}

#ifdef OBSOLETE
VovSuperObject::VovSuperObject( VovGoal& x )
{
    type = VOV_GOAL; contents.goal = &x ;
    allocated = 0;
}
#endif

VovSuperObject::VovSuperObject( VovUser& x )
{
    type = VOV_USER; contents.user = &x ;
    allocated = 0;
}

VovSuperObject::VovSuperObject( VovAnnotation& x )
{
    type = VOV_ANNOTATION; contents.note = &x ;
    allocated = 0;
}

VovSuperObject::VovSuperObject( VovRetrace& x )
{
    type = VOV_RETRACE; contents.retrace = &x ;
    allocated = 0;
}


 
void VovSuperObject::pack( Bzocket& zoc )
{
    zoc.pack( type );
    switch (type ) {
    case VOV_UNDEFINED:
	vovError("Cannot pack UNDEFINED superObject");
	break;
    case VOV_PLACE:
	contents.place->pack( zoc );
	break;
    case VOV_TRANSITION:
	contents.transition->pack( zoc );
	break;
    case VOV_NODE:
	contents.node->pack( zoc );
	break;
    case VOV_NODE_SET:
	contents.set->pack( zoc );
	break;
    case VOV_USER:
	contents.user->pack( zoc );
	break;
    case VOV_ANNOTATION:
	contents.note->pack( zoc );
	break;
    case VOV_RETRACE:
	contents.retrace->pack( zoc );
	break;
    default:
	vovFatalError("VovSuperObject::pack(): unknown type");
    }
}


void VovSuperObject::unpack( Bzocket& zoc )
{
    type = zoc.unpackInt();

    if ( contents.object == 0) {
	allocated = 1;		// We need to allocate memory for this object.
    }
    switch (type ) {
    case VOV_PLACE:
	if ( contents.place == 0 ) {
	    contents.place = new VovPlace;
	}
	contents.place->unpack( zoc );
	break;
    case VOV_TRANSITION:
	if ( contents.transition == 0 ) {
	    contents.transition = new VovTransition;
	}
	contents.transition->unpack( zoc );
	break;
    case VOV_NODE:
	if ( contents.node == 0 ) {
	    contents.node = new VovNode;
	}
	contents.node->unpack( zoc );
	break;
    case VOV_USER:
	if ( contents.user == 0 ) {
	    contents.user = new VovUser("no user");
	}
	contents.user->unpack( zoc );
	break;
    case VOV_ANNOTATION:
	if ( contents.note == 0 ) {
	    contents.note = new VovAnnotation;
	}
	contents.note->unpack( zoc );
	break;
    case VOV_NODE_SET:
	if ( contents.set == 0 ) {
	    contents.set = new VovNodeSet;
	}
	contents.set->unpack( zoc );
	break;
    case VOV_RETRACE:
	if ( contents.retrace == 0 ) {
	    contents.retrace = new VovRetrace;
	}
	contents.retrace->unpack( zoc );
	break;
    default:
	vovFatalError("VovSuperObject::unpack() Unknown type");
    }
}



octId VovSuperObject::getId()
{
    if ( contents.object ) {
	return contents.object->id();
    } else {
	return oct_null_id;
    }
}

#ifndef VOVCLIENT
VovStatus VovSuperObject::get()
{
    switch (type ) {
    case VOV_NODE:
	return contents.node->get();
    case VOV_TRANSITION:
	return contents.transition->get();
    case VOV_PLACE:
	return contents.place->get();
    case VOV_NODE_SET:
	return contents.set->get();
    case VOV_USER:
	return contents.user->get();
    case VOV_ANNOTATION:
	return contents.note->get();
    case VOV_RETRACE:
	return contents.retrace->get();
    default:
	vovError("illegal superobject type");
    }
    return VOV_ERROR;
}
#endif

String VovSuperObject::format()
{
    switch ( type ) {
    case VOV_TRANSITION:
	return contents.transition->format();
    case VOV_PLACE:
	return contents.place->format();
    case VOV_ANNOTATION:
	return contents.note->format();
    case VOV_NODE_SET:
	return contents.set->format();
    case VOV_NODE:
	return contents.node->format();
    case VOV_USER:
	return contents.user->format();
    case VOV_RETRACE:
	return contents.retrace->format();
	break;
    default:
	vovError("illegal superobject type");
    }
    return 0;
}

#ifndef VOVCLIENT
void VovSuperObject::setupDialog( VovDialog& top_dialog, VovDialogHandles& handles, int viewonly )
{
    switch ( type ) {
    case VOV_TRANSITION:
	contents.transition->setupDialog( top_dialog, handles, viewonly  );
	return;
    case VOV_PLACE:
	contents.place->setupDialog( top_dialog, handles, viewonly );
	return;
    case VOV_ANNOTATION:
	contents.note->setupDialog( top_dialog, handles );
	return;
    case VOV_NODE:
    case VOV_NODE_SET:
    case VOV_USER:
    case VOV_RETRACE:
	vovError( "cannot setupDialog for NODES, GOALS, SETS, USERS, RETRACE" );
	break;
    default:
	vovError("illegal superobject type");
    }
}
#endif

#ifndef VOVCLIENT
VovStatus VovSuperObject::readDialog( VovDialogHandles& handles )
{
    switch ( type ) {
    case VOV_TRANSITION:
	return contents.transition->readDialog( handles );
    case VOV_PLACE:
	return contents.place->readDialog(  handles );
    case VOV_ANNOTATION:
	return contents.note->readDialog( handles );
    case VOV_NODE:
    case VOV_NODE_SET:
    case VOV_USER:
    case VOV_RETRACE:
	vovError( "cannot readDialog for these objects" );
	break;
    default:
	vovError("illegal superobject type");
    }
    return VOV_ERROR;
    
}
#endif

VovSuperObject::~VovSuperObject()
{
    if ( allocated ) {
	int debug = 0;
	if ( debug ) {
	    vovMessage( "Freeing superobject" + format() );
	}
	switch ( type ) {
	case VOV_PLACE:
	    delete contents.place;
	    break;
	case VOV_TRANSITION:
	    delete contents.transition;
	    break;
	case VOV_NODE:
	    delete contents.node;
	    break;
	case VOV_USER:
	    delete contents.user;
	    break;
	case VOV_ANNOTATION:
	    delete contents.note;
	    break;
	case VOV_NODE_SET:
	    delete contents.set;
	    break;
	case VOV_RETRACE:
	    delete contents.retrace;
	    break;
	default:
	    vovFatalError("VovSuperObject::destructor: Unknown type");
	}
    }
}
