#include "trace.h"
#include "travfunc.h"
#include "user.h"
#include "VOVprotocol.h"
#include "place.h"
#include "transition.h"
#include "annotation.h"
#include "retrace.h"
#include "placelist.h"

// #include "measure.h"		/*  */

#include "voverror.h"
#include "clients.h"
#include "comm.h"



static int debug = 0;

char* vovCodeString( VovCode code )
{
    char* s;
    switch ( code ) {
    case VOV_NoCode: 
	s = "NoCode"; break;
    case VOV_Error: 
	s = "Error"; break;
    case VOV_Ok: 
	s = "Ok"; break;
    case VOV_StartUp: 
	s = "StartUp"; break;
    case VOV_Close: 
	s = "Close"; break;
    case VOV_KillServer: 
	s = "KillServer"; break;
    case VOV_Create: 
	s = "Create"; break;
    case VOV_GetById: 
	s = "GetById"; break;
    case VOV_GetIntermediateNodes: 
	s = "GetIntermediateNodes"; break;
    case VOV_List: 
	s = "List"; break;
    case VOV_Find: 
	s = "Find"; break;
    case VOV_Modify: 
	s = "Modify"; break;
    case VOV_Delete: 
	s = "Delete"; break;
    case VOV_FreeResources: 
	s = "FreeResources"; break;
    case VOV_Edit: 
	s = "Edit"; break;
    case VOV_Lock: 
	s = "Lock"; break;
    case VOV_Unlock: 
	s = "Unlock"; break;
    case VOV_ChangeStatus: 
	s = "ChangeStatus"; break;
    case VOV_1SetOperation: 
	s = "1SetOperation"; break;
    case VOV_2SetOperation: 
	s = "2SetOperation"; break;
    case VOV_SetRecompute: 
	s = "SetRecompute"; break;
    case VOV_InitGenerator: 
	s = "InitGenerator"; break;
    case VOV_Generate: 
	s = "Generate"; break;
    case VOV_FreeGenerator: 
	s = "FreeGenerator"; break;
    case VOV_SaveOntoDisk: 
	s = "SaveOntoDisk"; break;
    case VOV_CheckPlaces: 
	s = "CheckPlaces"; break;
    case VOV_StopAllTransitions: 
	s = "StopAllTransitions"; break;
    case VOV_SynchClocks: 
	s = "SynchClocks"; break;
    case VOV_Debug: 
	s = "Debug"; break;
    case VOV_Sync: 
	s = "Sync"; break;
    case VOV_Acknowledge: 
	s = "Acknowledge"; break;
    case VOV_EchoLog: 
	s = "EchoLog"; break;
    case VOV_PlaceTrace: 
	s = "PlaceTrace"; break;
    case VOV_StartTransition: 
	s = "StartTransition"; break;
    case VOV_EndTransition: 
	s = "EndTransition"; break;
    case VOV_ForgetTransition: 
	s = "ForgetTransition"; break;
    case VOV_ForgetSet: 
	s = "ForgetSet"; break;
    case VOV_TestConnectivity:
	s = "TestConnectivity"; break;
    case VOV_Disconnect: 
	s = "Disconnect"; break;
    case VOV_AddInput: 
	s = "AddInput"; break;
    case VOV_AddOutput: 
	s = "AddOutput"; break;
    case VOV_AddIOs: 
	s = "AddIOs"; break;
    case VOV_AddAnnotation: 
	s = "AddAnnotation"; break;
    case VOV_CountAnnotations: 
	s = "CountAnnotations"; break;
    case VOV_GetAnnotations: 
	s = "GetAnnotations"; break;
    case VOV_ToggleStatus: 
	s = "ToggleStatus"; break;
    case VOV_Get: 
	s = "Get"; break;
    case VOV_GetPlaceById: 
	s = "GetPlaceById"; break;
    case VOV_GetHistory: 
	s = "GetHistory"; break;
    case VOV_GetInfo: 
	s = "GetInfo"; break;
    case VOV_Format: 
	s = "Format"; break;
    case VOV_UserFormat: 
	s = "UserFormat"; break;
    case VOV_MvLib: 
	s = "MvLib"; break;
    case VOV_ActiveDesigner: 
	s = "ActiveDesigner"; break;
    case VOV_EvolveAll: 
	s = "EvolveAll"; break;
    case VOV_EvolveSet: 
	s = "EvolveSet"; break;
    case VOV_EvolveNode: 
	s = "EvolveNode"; break;
    case VOV_EvolveStop: 
	s = "EvolveStop"; break;
    case VOV_GetExecutionScript: 
	s = "GetExecutionScript"; break;
    case VOV_GetScript: 
	s = "GetScript"; break;
    case VOV_SubSelect: 
	s = "SubSelect"; break;
    case VOV_FreezeConfiguration: 
	s = "FreezeConfiguration"; break;
    case VOV_Snapshot: 
	s = "Snapshot"; break;
    case VOV_SlaveReady: 
	s = "SlaveReady"; break;
    case VOV_Exec: 
	s = "Exec"; break;
    case VOV_CannotExec: 
	s = "CannotExec"; break;
    case VOV_GetHostLoad: 
	s = "GetHostLoad"; break;
    case VOV_Notify: 
	s = "Notify"; break;
    case VOV_SynchNotify: 
	s = "SynchNotify"; break;
    case VOV_AsynchNotify: 
	s = "AsynchNotify"; break;
    case VOV_RequestPage: 
	s = "RequestPage"; break;
    case VOV_StartPage: 
	s = "StartPage"; break;
    case VOV_EndPage: 
	s = "EndPage"; break;
    case VOV_Measurement: 
	s = "Measurement"; break;
    case VOV_LastCode: 
	s = "LastCode"; break;
    default:
	return  form( "%d", code );
	break;
    }
    return s;
}



int VovTrace::isClient()
{ 
    return type == VOV_TRACE_TYPE_CLIENT;
}

int VovTrace::isServer( int entry_flag )
{
    int srv = ( type == VOV_TRACE_TYPE_SERVER); 
    if ( entry_flag ) {
	return  ( srv && serverEntryCount++ == 0 );
    } else {
	return  ( srv && --serverEntryCount == 0 );
    }
}

int VovTrace::isLocal()
{ 
    return type == VOV_TRACE_TYPE_LOCAL;
}


VovStatus VovTrace::serverDispatch( Client& zoc, VovCode code, octId& id )
{
    VovStatus rtn = VOV_OK;
#ifndef VOVCLIENT
    VovTransition  theTransition;
    VovPlace       thePlace;
    String         theString;
    
    setBzocket( zoc );
    serverEntryCount = 0;	// Exec the next trace function as a "server".
    
    if ( debug ) cerr << "Doing code " <<  vovCodeString( code );

    switch ( code ) {
    case VOV_Ok:
	{
	    String msg;
	    OK( msg );
	}
	break;
    case VOV_SaveOntoDisk:
	saveOntoDisk(0,0);
	break;
    case VOV_CheckPlaces:
	checkPlaces( 0 );
	break;
    case VOV_StopAllTransitions:
	stopAllTransitions();
	break;
    case VOV_PlaceTrace:
	doPlacement();
	break;
    case VOV_Create:
	{
	    VovSuperObject theObj;
	    create( theObj );
	    id = theObj.getId();
	}
	break;
    case VOV_Measurement:
	{
	    VovPlace place;
	    VovPlace measurePlace;
	    measure( place, measurePlace );
	}
	break;
    case VOV_ActiveDesigner:
	{
	    VovUser user("no user");
	    activeDesigner( user, theTransition);
	}
	break;
    case VOV_Modify:
	{
	    VovSuperObject theObj;
	    modify( theObj );
	}
	break;
    case VOV_List:
	{ 
	    String rx, l;
	    list( 0, rx, l );
	}
	break;
    case VOV_Find:
	{
	    VovSuperObject theObj;
	    find( theObj );
	}
	break;
    case VOV_Delete:
	{
	    VovSuperObject theObj;
	    destroy( theObj ) ;
	}
	break;
    case VOV_FreeResources:
	freeResources( thePlace );
	break;
    case VOV_AddAnnotation:
	{
	    VovAnnotation note;
	    VovSuperObject theObj;
	    addAnnotation( theObj, note );
	}
	break;			// 
    case VOV_CountAnnotations:
	{
	    VovSuperObject theObj;
	    numberOfAnnotations( theObj );
	}
	break;
    case VOV_GetAnnotations:
	{
	    static VovAnnotation* noteArray = 0;
	    if ( noteArray == 0 ) {
		vovMessage( "Allocating noteArray in traceDispatch" );
		noteArray = new VovAnnotation[VOV_MAX_ANNOTATIONS]; 
	    }
	    VovSuperObject theObj;
	    getAnnotations( theObj, noteArray, VOV_MAX_ANNOTATIONS );
	}
	break;
    case VOV_Edit:
	{
	    VovSuperObject theObj;
	    edit( theObj );
	}
	break;
    case VOV_Lock:
	{
	    VovObject o(OCT_INSTANCE);
	    String d,m;
	    int f;
	    octId i;
	    lock( o, d, f, i, m );
	}
	break;
    case VOV_Unlock:
	{
	    VovObject o(OCT_INSTANCE);
	    octId i;
	    unlock( o, i );
	}
	break;
    case VOV_StartTransition:
	initiate( theTransition );
	id = theTransition.id();
	break;
    case VOV_EndTransition:
	terminate( theTransition, VOV_OK, 0, theString );
	break;
    case VOV_TestConnectivity:
	{
	    VovNode n;
	    testConnectivity( n );
	}
	break;
    case VOV_Disconnect:
	{
	    VovNode n1,n2;
	    disconnect( n1, n2 );
	}
	break;
    case VOV_AddInput:
	declareInput( theTransition, thePlace, 0 );
	break;
    case VOV_AddOutput:
	{
	    String errmsg;
	    declareOutput( theTransition, thePlace, 0, 0, errmsg );
	}
	break;
    case VOV_AddIOs:
	{
	    VovPlaceList list(16); // Start list with some space, so it won't have to grow right away.
	    declareIOs( theTransition, list );
	}
	break;
    case VOV_GetById:
	{
	    VovSuperObject theObj;
	    getById( theObj  );
	}
	break;
    case VOV_GetIntermediateNodes:
	{
	    VovNodeSet s1, s2, s3;
	    VovSelectRule r;
	    getIntermediateNodes( s1,s2,r,s3 );
	}
	break;
    case VOV_1SetOperation:
	{
	    VovNodeSet set;
	    VovNode node;
	    setOperation( set, node, VOV_SET_ADD );
	} 
	break;
    case VOV_2SetOperation:
	{
	    VovNodeSet set1, set2;
	    setOperation( set1, set2, VOV_SET_UNION );
	} 
	break;
    case VOV_InitGenerator:
	{
	    VovObject obj(OCT_INSTANCE);
	    VovGenerator gen;
	    initGenerator( 0, obj, 0, gen );
	} 
	break;
    case VOV_Generate:
	{
	    VovObject obj(OCT_INSTANCE);
	    VovGenerator gen;
	    generate( gen, obj );
	} 
	break;
    case VOV_FreeGenerator:
	{
	    VovGenerator gen;
	    freeGenerator( gen );
	} 
	break;
    case VOV_ChangeStatus:
	{
	    VovNode node = 0;
	    VovNodeStatus s;
	    changeStatus( node, s, 0);
	}
	break;
    case VOV_GetHistory:
	{
	    int level, direction, flags;
	    String historyString;
	    history( thePlace, flags, direction, level, historyString );
	}
	break;
    case VOV_GetExecutionScript:
	{
	    String script;
	    prepareExecutionScript( script );
	}
	break;
    case VOV_GetScript:
	{
	    String script;
	    VovNodeSet set;
	    prepareScript( set, 0,script);
	}
	break;
    case VOV_EvolveStop:
	stopRetracing( theTransition, theString );
	break;
    case VOV_EvolveAll:
    case VOV_EvolveSet:
    case VOV_EvolveNode:
	{
	    VovNode theNode;
	    VovRetrace retrace;
	    VovNodeSet set;
	    switch( code ) {
	    case VOV_EvolveAll:
		evolve( retrace, VOV_RETRACE_SLOW, theString);
		break;
	    case VOV_EvolveSet:
		evolve( retrace, set, VOV_RETRACE_SLOW, theString );
		break;
	    case VOV_EvolveNode:
		evolve( retrace, theNode, VOV_BACKWARD, VOV_RETRACE_SLOW, theString );
		break;
	    default:
		vovBug( "impossible");
	    }
	}
	break;
    case VOV_SubSelect:
	{
	    VovNodeSet bag;
	    VovSelectRule rule;
	    subselect(bag, rule );
	}
	break;
    case VOV_MvLib:
	{
	    VovNodeSet set;
	    String dummyString1 = "old";
	    String dummyString2 = "new";
	    String msg;
	    mvlib( set, dummyString1, dummyString2, msg );
	}
	break;
    case VOV_UserFormat:
	userFormat( 0 );
	break;
    case VOV_GetInfo:
	getInfo( 0, 0 );
	break;
    case VOV_Format:
	format();
	break;
    default:
	rtn = VOV_NOT_OK;
    }
    serverEntryCount = 1;	// Exec the next trace function as a "local trace".

    if ( debug) cerr << " Done. "  << rtn << "\n";	// Debugging
    
#endif
    return rtn;
}


