#include "trace.h"
#include "travfunc.h"
#include "user.h"
#include "VOVprotocol.h"
#include "place.h"
#include "placelist.h"
#include "transition.h"
#include "annotation.h"
#include "retrace.h"
// #include "measure.h"		/*  */

#include "voverror.h"
#include "comm.h"


extern VovStatus vov_cycle( VovNode&, VovNode&);
extern void vovEditSet( VovTrace& trace, VovNodeSet& set );


// Static arrays: I'd rather have them local, but g++ punts on this.
#ifndef VOVCLIENT
VovStatus VovTrace::copyTrace( VovTrace& trace, VovNodeSet& set, VovNodeSet& newset )
{
    VovStatus rtn = VOV_OK;
    
    // findOrCreate( newset );
    // newset.clear( VOV_NODE_OCTMASK );
    VovContentsGenerator genNodes( &set, VOV_NODE_OCTMASK );
    VovNode node;
    while ( octGenerate( &genNodes, &node ) == OCT_OK  ) {
	VOV_ASSERT( node.get() );
	if ( node.nodeType() == VOV_TRANSITION ) {
	    VovTransition tran = node;
	    VovTransition newtran = tran;
	    vovMessage( "Copying:" + newtran.getCommandLine() );
	    if ( create( newtran ) != VOV_OK ) {
		rtn = VOV_ERROR; 
		break;
	    }
	    // newset.add( newtran );
	    changeStatus( newtran, VOV_PLAN_NODE, VOV_CHANGE_STATUS_SELF_REC );
	    VovContainersGenerator genInputs( &tran, VOV_PLACE_OCTMASK );
	    VovPlace place;
	    while ( octGenerate( &genInputs, &place ) == OCT_OK ) {
		VOV_ASSERT( place.get() );
		findOrCreate( place );
		// newset.add( place );
		vovMessage( "Add Input: " + place.getName() );
		if ( declareInput( newtran, place, 0 ) != VOV_OK ) {
		    vovError("Error adding input" );
		    rtn = VOV_ERROR;
		    break;
		}
	    }

	    if ( rtn == VOV_ERROR ) break;

	    VovContentsGenerator genOutputs( &tran, VOV_PLACE_OCTMASK );
	    while ( octGenerate( &genOutputs, &place ) == OCT_OK ) {
		VOV_ASSERT( place.get() );
		findOrCreate( place );
		// newset.add( place );
		vovMessage( "Add Output: " + place.getName() );
		String errmsg;
		if ( declareOutput( newtran, place, 0, 0, errmsg ) != VOV_OK ) {
		    vovError( errmsg );
		    rtn = VOV_ERROR;
		    break;
		}
	    }
	    if ( rtn == VOV_ERROR ) break;
	}
    }
    
    if ( rtn==VOV_ERROR) {
	vovError( "Copy of trace failed" );
    }
    return rtn;
}
#endif



VovStatus VovTrace::freeResources( VovPlace& place )
{
    if ( type == VOV_TRACE_TYPE_DISABLED ) return VOV_OK;
    Bzocket& zoc = getBzocket();
    VovStatus rtn = VOV_OK;
    
    if ( isClient() ) {
	zoc.pack( VOV_FreeResources );
	place.pack( zoc );
	zoc.flush();
    } else {
#ifndef VOVCLIENT
	if ( isServer(VOV_SERVER_ENTRY) ) {
	    place.unpack( zoc );
	    rtn = place.get();
	}
	
	if ( rtn == VOV_OK ) {
	    if ( place.freeResources() == VOV_OK )  {
		getQueue().enq( VOV_PLACE_MASK, VE_CHANGE, place.format() + "\nResources have been freed", place.id(), 0 );
	    }
	}
	if ( isServer(VOV_SERVER_EXIT) ) {
	    zoc.pack ( rtn );
	    zoc.flush();
	}
#endif
    }
    if ( isClient() ) {
	VC_get_reply( zoc );
	rtn = zoc.unpackInt();
    }
    return rtn;
    // End of ::freeResources()
}



VovStatus VovTrace::addAnnotation( VovSuperObject& obj, VovAnnotation& note ) 
{
    if ( type == VOV_TRACE_TYPE_DISABLED ) return VOV_OK;
    Bzocket& zoc = getBzocket();
    VovStatus rtn = VOV_OK;
    String errmsg;

    if ( isClient() ) {
	zoc.pack( VOV_AddAnnotation );
	obj.pack( zoc );
	note.pack( zoc );
	zoc.flush();
    } else {
#ifndef VOVCLIENT
	if ( isServer(VOV_SERVER_ENTRY) ) {
	    obj.unpack( zoc );
	    note.unpack( zoc );
	}

	if (octGetById( obj.contents.object ) != OCT_OK ) {
	    errmsg = "AddAnnotation:Corrupt id";
	    rtn = VOV_ERROR;
	} else {
	    VovObject& o = *obj.contents.object;
	    o.addNote( note );
	    // vovMessage( obj.format() + note.format() );
	} 

	if ( isServer(VOV_SERVER_EXIT) ) {
	    zoc.pack( rtn );
	    note.pack( zoc );
	    if ( rtn != VOV_OK ) {
		zoc.pack( errmsg );
	    }
	    zoc.flush();
	}
#endif
    }
    if ( isClient() ) {
	VC_get_reply( zoc );
	rtn = zoc.unpackInt();
	note.unpack( zoc );
	if ( rtn != VOV_OK ) {
	    errmsg = zoc.unpackString();
	}
    }
    if ( rtn != VOV_OK ) vovError( errmsg );
    return rtn;
    // End ::annotate()
}

int VovTrace::numberOfAnnotations( VovSuperObject& obj ) 
{
    if ( type == VOV_TRACE_TYPE_DISABLED ) return 0;
    Bzocket& zoc = getBzocket();
    int  rtn = 0;
    String errmsg;

    VovObject o(OCT_BAG);	// Constructor needs argument, but arg is not important here.
    o.id() = obj.getId();

    if ( isClient() ) {
	zoc.pack( VOV_CountAnnotations );
	zoc.pack( o.id() );
	zoc.flush();
    } else {
#ifndef VOVCLIENT
	if ( isServer(VOV_SERVER_ENTRY) ) {
	    o.id() =  zoc.unpackInt();
	}

	if (octGetById( &o ) != OCT_OK ) {
	    errmsg = "NumberOfAnnotations:Corrupt id";
	} else {
	    rtn = o.countNotes( );
	} 

	if ( isServer(VOV_SERVER_EXIT) ) {
	    zoc.pack( rtn );
	    zoc.flush();
	}
#endif
    }
    if ( isClient() ) {
	VC_get_reply( zoc );
	rtn = zoc.unpackInt();
    }
    return rtn;
    // End ::numberofAnnotations()
}

int VovTrace::getAnnotations( VovSuperObject& obj, VovAnnotation*& noteArray , int length ) 
{
    if ( type == VOV_TRACE_TYPE_DISABLED ) return 0;
    Bzocket& zoc = getBzocket();
    int  rtn = 0;
    String errmsg;
    
    if ( length > VOV_MAX_ANNOTATIONS ) {
	vovError( form("Max number of annotations per object is %d (< %d)", VOV_MAX_ANNOTATIONS, length) );
	length = VOV_MAX_ANNOTATIONS;
    }
    VovObject o( OCT_BAG );
    o.id() =  obj.getId();
    if ( isClient() ) {
	zoc.pack( VOV_GetAnnotations );
	zoc.pack( o.id() );
	zoc.pack( length );
	zoc.flush();
    } else {
#ifndef VOVCLIENT
	if ( isServer(VOV_SERVER_ENTRY) ) {
	    o.id() = zoc.unpackInt();
	    length = zoc.unpackInt();
	}
	
	if (octGetById( &o ) != OCT_OK ) {
	    errmsg = "GetAnnotations:Corrupt id";
	    rtn = -1;
	} else {
	    rtn = o.getNotes( noteArray, length );
	} 
	    
	if ( isServer(VOV_SERVER_EXIT) ) {
	    zoc.pack( rtn );
	    for ( int i = 0; i < rtn; i++ ) {
		noteArray[i].pack( zoc );
	    }
	    zoc.flush();
	}
#endif
    }
    if ( isClient() ) {
	VC_get_reply( zoc );
	rtn = zoc.unpackInt();
	for ( int i = 0; i < rtn; i++ ) {
	    noteArray[i].unpack( zoc );
	}
    }
    return rtn;
    // End ::getAnnotations()
}



// ROUTINE: VovTrace::declareInput()
// Check whether there is already a path between rp and rt.
// If so, than the place is already an input (even in transitive sense)
// for the transition.
VovStatus VovTrace::declareInput(
				 VovTransition& tran, 		// The transition.
				 VovPlace& place,		// The input place.
				 int suspendAccounting		// A control flag.
				 )
{
    if ( type == VOV_TRACE_TYPE_DISABLED ) return VOV_OK;
    Bzocket& zoc = getBzocket();
    VovStatus rtn = VOV_OK;
    String errmsg;
    
    if ( isClient() ) {
	zoc.pack( VOV_AddInput );
	zoc.pack( tran.id() );
	zoc.pack( place.id() );
	zoc.pack( suspendAccounting );
	zoc.flush();
    } else {
#ifndef VOVCLIENT
	if ( isServer(VOV_SERVER_ENTRY) ) {
	    tran.id() = zoc.unpackInt();
	    place.id() = zoc.unpackInt();
	    suspendAccounting = zoc.unpackInt();
	    if ( tran.get() != VOV_OK || place.get() != VOV_OK ) {
		errmsg = "Bogus ids in declare input";
		rtn = VOV_ERROR;
	    }
	}
	if ( rtn == VOV_OK ) {
	    VOV_ASSERT( place.correctInputInChain(*this,tran)); 	// Deal with chains.
	    if ( tran.isInput( place ) == VOV_OK ) {		// Skip place if already an input.
		errmsg = "Place is already an input";		// 
	    } else if ( tran.isOutput( place ) == VOV_OK ) {
		errmsg = "Place is already an output";
	    } else {						// 
		rtn = tran.connectInput( place );		// Connect input, but check for cycles.
		if ( rtn == VOV_CYCLE ) {			// 
		    errmsg = place.format() + "->" + tran.format() + " would create cycle.";
		    if ( !suspendAccounting ) {
			accumulateStatistics( String("CONFLICT_CYCLES"), 1 );
		    }
		} else {
		    if ( tran.getStatus() == VOV_TRACING_NODE && place.getStatus() != VOV_TRACE_NODE ) {
			if ( ! suspendAccounting ) {
			    accumulateStatistics( String("CONFLICT_INVALID_INPUT"), 1 );
			}
		    }
		    /* Check for input conflicts. */
		    if ( place.getStatus() != VOV_TRACE_NODE ) {
			rtn = VOV_INPUT_CONFLICT;
		    }
		}
	    }
	}
	if ( isServer(VOV_SERVER_EXIT) ) {
	    zoc.pack( rtn );
	    zoc.pack( place.id() ); // The id might have changed in this routine.
	    zoc.pack( errmsg );
	    zoc.flush();
	}
#endif
    } 
    if ( isClient() ) {
	VC_get_reply( zoc );
	rtn = zoc.unpackInt();
	place.id() = zoc.unpackInt();
	errmsg = zoc.unpackString();
    }
    return rtn;
}

VovStatus VovTrace::declareOutput(
				  VovTransition& tran, 		// The transition.
				  VovPlace& place,		// The output place.
				  int force,			// Force place to be output.
				  int suspendAccounting,	// Do not account conflicts.
				  String& errmsg		// Returned error message.
				  )
    // RETURN: 
    // VOV_OK 
{
    if ( type == VOV_TRACE_TYPE_DISABLED ) return VOV_OK;
    Bzocket& zoc = getBzocket();
    VovStatus rtn = VOV_OK;
    
    if ( isClient() ) {
	zoc.pack( VOV_AddOutput );
	zoc.pack( tran.id() );
	zoc.pack( place.id() );
	zoc.pack( force );
	zoc.pack( suspendAccounting );
	zoc.flush();
    } else {
#ifndef VOVCLIENT
	if ( isServer(VOV_SERVER_ENTRY) ) {
	    tran.id() = zoc.unpackInt();
	    place.id() = zoc.unpackInt();
	    force = zoc.unpackInt();
	    suspendAccounting = zoc.unpackInt();
	    if ( tran.get() != VOV_OK || place.get() != VOV_OK ) {
		errmsg = "Bogus ids in declare output";
		rtn = VOV_ERROR;
	    }
	}
	
	if ( rtn == VOV_OK  ) {
	    place.correctOutputInChain( *this, tran );
	    if ( tran.isOutput( place ) == VOV_OK ) {
		errmsg = "place is already an output";
	    } else {
		// Check for output conflicts.
		if ( place.getStatus() != VOV_RETRACING_NODE && place.hasInputs() == VOV_OK  ) {
		    // Output conflict. Place already output of another transition.
		    // Get transition which already produces this place and delete it.
			
		    VovTransition oldTran;
		    OH_ASSERT( octGenFirstContainer(&place,OCT_INSTANCE_MASK,&oldTran ) );
		    VOV_ASSERT( oldTran.get() );
		    if ( force ) {
			journalLog( "Forgetting transition:\n" + oldTran.format() );
			destroy( oldTran );
		    } else {
			if ( ! suspendAccounting ) {
			    accumulateStatistics( String("CONFLICT_REDECLARING_OUTPUT"), 1 );
			}
			errmsg=place.format()+"\nis already the output of\n"+oldTran.format();
			tran = oldTran;
			rtn =  VOV_ERROR;
		    } 
		}
		if ( rtn == VOV_OK ) {
		    rtn = tran.connectOutput( place );
		    if ( rtn == VOV_CYCLE ) {
			errmsg = tran.format() + "->" + place.format() + "\nwould create cycle.";
			if ( ! suspendAccounting ) {
			    accumulateStatistics( String("CONFLICT_CYCLES"), 1 );
			}
		    } 
		}
	    }
	    if ( rtn == VOV_OK ) {
		VovNodeStatus s = tran.getStatus();
		place.changeStatus( s, VOV_CHANGE_STATUS_SELF, errmsg );
		place.changeStatus( VOV_PLAN_NODE, VOV_CHANGE_STATUS_RECURSIVE, errmsg );
		place.setCurrent(); // The output of the transition is the current place in the chain.
	    }
	}
	
	if ( isServer(VOV_SERVER_EXIT) ) {
	    zoc.pack( rtn );
	    zoc.pack( place.id() ); // The id might have changed in this routine.
	    zoc.pack( errmsg );
	    place.pack( zoc );
	    zoc.flush();
	}
#endif
    }
    if ( isClient() ) {
	VC_get_reply( zoc );
	rtn = zoc.unpackInt();
	place.id() = zoc.unpackInt();
	errmsg = zoc.unpackString();
	place.unpack( zoc );
    }
    
    // if ( rtn != VOV_OK ) vovMessage( errmsg );
    return rtn;
}

//
// ROUTINE: VovTrace::declareIOs();
//
// More efficient routine to declare lots of inputs and outputs.
// Package all inputs and outputs in one big packet.
// Analyze all dependencies. If some place gives problems, we
// have to check it again individually.
// Accounting of conflicts is suspended, because any conflict is
// to be detected again on second check.
//
VovStatus VovTrace::declareIOs(VovTransition& tran, VovPlaceList& list )
{
    if ( type == VOV_TRACE_TYPE_DISABLED ) return VOV_OK;
    Bzocket& zoc = getBzocket();
    VovStatus rtn = VOV_OK;
    String errmsg;
    
    if ( isClient() ) {
	zoc.pack( VOV_AddIOs );
	zoc.pack( tran.id() );
	zoc.pack( list.length() );
	for ( int i = 0 ; i < list.length() ; i++ ) {
	    VovPlace& place = list.getPlace( i );
	    int type = list.getType( i );
	    place.pack( zoc );
	    zoc.pack( type );
	}
	zoc.flush();
    } else {
#ifndef VOVCLIENT
	struct timeval serverTime;
	getDate( serverTime );

	if ( isServer(VOV_SERVER_ENTRY) ) {
	    
	    list.clear();
	    tran.id() = zoc.unpackInt();
	    int length = zoc.unpackInt();
	    for ( int i = 0; i < length ; i++ ) {
		VovPlace place;	// Must be inside loop for unpack() to work.
		place.unpack( zoc );
		int type = zoc.unpackInt();
		VOV_ASSERT( findOrCreate( place ) ); 		// Notice, place is created here!!
		list.add( place, type );
	    }
	    if ( tran.get() != VOV_OK ) {
		vovError( "Transition is busted" );
		rtn = VOV_ERROR;
	    }
	}   
	
	struct timeval measureTime;
	getDate( measureTime );
	// If everything worked while getting the long list of places,
	// run trhough the list and do the checking.
	if ( rtn == VOV_OK ) {
	    for ( int i = 0 ; i < list.length() ; i++ ) {
		VovPlace& place = list.getPlace( i );
		VovStatus localReturnStatus = VOV_OK;
		//
		// If this place is an input ...
		//
		if ( list.getType( i ) & VOV_INPUT_MASK ) {
		    localReturnStatus = declareInput( tran, place, 1 );
		    if ( place.getStatus() != VOV_TRACE_NODE ) {
			localReturnStatus = VOV_NOT_OK;
		    }
		}
		//
		// If all OK and this place is an output ...
		//
		if ( (localReturnStatus == VOV_OK ) && ( list.getType( i ) & VOV_OUTPUT_MASK ) ) {
		    String msg;
		    localReturnStatus = declareOutput( tran, place, 0, 1, msg );
		}
		if ( localReturnStatus != VOV_OK ) rtn = VOV_NOT_OK;
		list.getStatus( i ) = localReturnStatus;
	    }
	}
	double gutsTime = getDateInterval( measureTime );
	
	if ( isServer(VOV_SERVER_EXIT) ) {
	    zoc.pack( rtn );
	    for ( int i = 0 ; i < list.length() ; i++ ) {
		zoc.pack( list.getPlace(i).id() ); // Id's might have changed.
		zoc.pack( list.getStatus( i ) );
	    }
	    zoc.flush();
	}
	
	double totTime = getDateInterval( serverTime );
	if ( list.length() ) {
	    vovMessage( form( "%d places in %.3f s: %.3f s/place", list.length(), totTime,
			     totTime / list.length() ) );
	    vovMessage( form( "Communication overhead %.3f s", totTime - gutsTime ) );
	}

#endif
    } 
    if ( isClient() ) {
	VC_get_reply( zoc );
	rtn = zoc.unpackInt();
	for ( int i = 0 ; i < list.length() ; i++ ) {
	    list.getPlace( i ).id() = zoc.unpackInt();
	    list.getStatus( i ) = zoc.unpackInt();
	}
    }
    return rtn;
} // End of VovTrace::declareIOs();


VovStatus VovTrace::disconnect( VovNode& n1, VovNode& n2 )
{
    if ( type == VOV_TRACE_TYPE_DISABLED ) return VOV_OK;
    Bzocket& zoc = getBzocket();
    VovStatus rtn = VOV_OK;
    
    if ( isClient() ) {
	zoc.pack( VOV_Disconnect );
	zoc.pack( n1.id() );
	zoc.pack( n2.id() );
	zoc.flush();
    } else {
#ifndef VOVCLIENT
	if ( isServer(VOV_SERVER_ENTRY) ) {
	    n1.id() = zoc.unpackInt();
	    n2.id() = zoc.unpackInt();
	    if ( n1.get() != VOV_OK || n2.get() != VOV_OK ) {
		rtn = VOV_ERROR;		// errmsg = "Bogus ids in disconnect";
	    }
	}
	
	if ( rtn == VOV_OK  ) {
	    if ( n1.isOutput( n2 ) == VOV_OK ) {
		n1.disconnectOutput( n2 );
	    } else if ( n1.isInput( n2 ) == VOV_OK ) {
		n1.disconnectInput( n2 );
	    } else {
		rtn = VOV_ERROR;
	    }
	}
	if ( rtn == VOV_OK ) {
	    getQueue().enq( VOV_NODE, VE_DISCONNECT, n1.format() + "\n" + n2.format(), n1.id(), 0 );
	}
	
	if ( isServer(VOV_SERVER_EXIT) ) {
	    zoc.pack( rtn );
	    zoc.flush();
	}
#endif
    }
    if ( isClient() ) {
	VC_get_reply( zoc );
	rtn = zoc.unpackInt();
    }
    
    return rtn;
    // ::disconnect()
}


VovStatus VovTrace::activeDesigner( VovUser& user, VovTransition& trans )
{
    if ( type == VOV_TRACE_TYPE_DISABLED ) return VOV_OK;
    Bzocket& zoc = getBzocket();
    VovStatus rtn = VOV_OK;
    
    if ( isClient() ) {
	zoc.pack( VOV_ActiveDesigner );
	zoc.pack( user.id() );
	zoc.pack( trans.id() );
	zoc.flush();
    } else {
#ifndef VOVCLIENT
	if ( isServer(VOV_SERVER_ENTRY) ) {
	    user.id() = zoc.unpackInt();
	    trans.id() = zoc.unpackInt();
	    if ( user.get() != VOV_OK || trans.get() != VOV_OK ) {
		rtn = VOV_ERROR;
	    }
	}
	if ( rtn == VOV_OK ) {
	    user.connected();				// 
	    if ( trans.isUserInitiated() == VOV_OK ) { 	// Because you do not want the slave to mess with this.
		VOV_ASSERT( user.active( trans ) );
	    } else {
		VovUser server_user( "vov_server" );
	        VOV_ASSERT( findOrCreate( user ) );
		VOV_ASSERT( user.active( trans ) );
	    }	
	}
	if ( isServer(VOV_SERVER_EXIT) ) {
	}
#endif
    }
    return rtn;
}


VovStatus VovTrace::initiate( VovTransition& t )
{
    VovStatus rtn = VOV_OK;
    if ( type == VOV_TRACE_TYPE_DISABLED ) return VOV_OK;
    Bzocket& zoc = getBzocket();
    
    if ( isClient() ) {
	zoc.pack( VOV_StartTransition );
	zoc.pack( t.id() );
	zoc.flush();
    } else {
#ifndef VOVCLIENT
	if ( isServer(VOV_SERVER_ENTRY) ) {
	    t.id() = zoc.unpackInt();
	    rtn =  t.get();
	    if ( rtn != VOV_OK ) {
		vovError( "Wrong id. Cannot initiate transition" );
	    }
	}
	
	if ( rtn == VOV_OK ) {
	    t.initiate( VOV_TRACING_NODE );
	}

	if ( isServer(VOV_SERVER_EXIT) ) {
	    if ( rtn == VOV_OK )  {
		VS_ok( zoc );
	    } else {
		VS_oops( zoc, "Termination faiiled." );
	    }
	}
#endif
    }
    if ( isClient() ) {
	VC_wait_for_ok_reply( zoc );
    }
    return VOV_OK;
}

#ifndef VOVCLIENT
void vov_accumulateTimeAndCount( VovTrace& trace, String& label, int duration )
{
    trace.accumulateStatistics( label + "_TRANS_TIME", duration );
    trace.accumulateStatistics( label + "_TRANS_COUNT", 1 );
}
#endif

#ifndef VOVCLIENT
void vov_accumulateTransitionStats( 
				   VovTrace& trace, 
				   int userInitiated, 
				   int duration, 
				   VovTransitionTerminationCode code )
{
    int success = code == VOV_TRANSITION_TERMINATION_OK;

    String u = userInitiated ? "USER" : "AUTO";
    String s = success ? "_SUCCESS" : "_FAILED";
    
    vov_accumulateTimeAndCount( trace, "TOTAL", duration );
    vov_accumulateTimeAndCount( trace, u , duration );
    vov_accumulateTimeAndCount( trace, u + s , duration );
    
    if ( ! success ) {
	String c = "";
	switch ( code ) {
	case VOV_TRANSITION_TERMINATION_WRONG_STATUS:
	    c = "_WRONG_STATUS"; break;
	case VOV_TRANSITION_TERMINATION_FAILED:
	    c = "_GENERIC" ; break;
	case VOV_TRANSITION_TERMINATION_INVALID_INPUT:
	    c = "_INVALID_INPUT"; break;
	case VOV_TRANSITION_TERMINATION_NO_OUTPUTS:
	    c = "_NO_OUTPUTS"; break;
	case VOV_TRANSITION_TERMINATION_DEFECTIVE_OUTPUT:
	    c = "_DEFECTIVE_OUTPUTS"; break;
	case VOV_TRANSITION_TERMINATION_INCONSISTENT:
	    c = "_INCONSISTENT_TIMESTAMP"; break;
	default:
	    vovError( "Weird termination code" );
	}
	vov_accumulateTimeAndCount( trace, u + s + c, duration );
    }
}
#endif

VovStatus VovTrace::terminate( VovTransition& t, VovStatus termStatus, int exit_status, String& pipedMsg )
{
    VovStatus rtn = VOV_OK;
    if ( type == VOV_TRACE_TYPE_DISABLED ) return VOV_OK;
    Bzocket& zoc = getBzocket();
    
    if ( isClient() ) {
	zoc.pack( VOV_EndTransition );
	zoc.pack( t.id() );
	zoc.pack( termStatus);
	zoc.pack( exit_status );
	zoc.pack( pipedMsg );
	zoc.flush();
    } else {
#ifndef VOVCLIENT
	if ( isServer(VOV_SERVER_ENTRY) ) {
	    t.id() = zoc.unpackInt();
	    rtn =  t.get();
	    if ( rtn != VOV_OK ) {
		vovError( "Wrong id. Cannot terminate transition" );
		// Do something.
	    }
	    termStatus  = zoc.unpackInt();
	    exit_status = zoc.unpackInt();
	    pipedMsg = zoc.unpackString();
	}
	
	if ( rtn == VOV_OK ) {
	    String msg;
	    if ( stopRetracing( t, msg ) != VOV_OK ) {
		vovError( msg );
	    }
	    
	    int userInitiated = ( t.isUserInitiated() == VOV_OK );

	    VovTransitionTerminationCode code;
	    VovStatus s = t.terminate( getQueue(), termStatus, exit_status, pipedMsg, code );

	    int duration      = t.duration();
	    int outputs       = t.numberOfOutputs(); // Every transition should have at least 1 output.
	    if ( s == VOV_ERROR ) {
		vovError("Problems terminating the transition" );
	    } else if ( s == VOV_DELETE_NODE ) {
		t.del();
	    } else  {
		vov_accumulateTransitionStats( *this, userInitiated, duration, code );
	    }

	    if ( s != VOV_DELETE_NODE && outputs == 0 ) {
		vovMessage( "Forget transition due to lack of outputs\n" + t.format() );
		t.del();
	    }  
	}
	if ( isServer(VOV_SERVER_EXIT) ) {
	    zoc.pack( rtn );
	    t.pack( zoc );
	    zoc.flush();
	}
#endif
    }
    if ( isClient() ) {
	VC_get_reply( zoc );
	rtn = zoc.unpackInt();
	t.unpack( zoc );
    }
    return rtn;
}





/*
** Check one place.
*/
#ifndef VOVCLIENT
VovStatus VovTrace::checkPlace(VovPlace& place,int* exist_par, Date* timeStamp_par )
{
    int& existflag = *exist_par;
    Date& curTimestamp = *timeStamp_par;
    Date knownTimestamp = place.getTimeStamp();
    

    int mode;
    place.info( &existflag, &mode, &curTimestamp, 0 );
    
    switch ( place.getStatus() ) {
    case VOV_TRACE_NODE:
	if ( ! existflag ) {
	    String s =  place.getName() + "\n\thas disappeared.";
	    VOV_ASSERT( place.changeStatus( VOV_NONEXISTING_NODE, VOV_CHANGE_STATUS_SELF, s ) );
	    place.setCurrent();
	    getQueue().enq( VOV_PLACE_MASK, VE_CHANGE, s, place.id(), 0 ) ;
	} else if ( curTimestamp != knownTimestamp ) {
	    String s = place.getName() + "\n has a new timestamp:";
	    s += ("\n\twas:" + knownTimestamp.format());
	    s += ("\n\t is:" + curTimestamp.format() );
	    s += "\n";
	    VovChainInfo chain;
	    if ( place.getChain( chain ) == VOV_OK ) {
		VovPlace f;
		chain.getFirst( f );
		f.changeStatus( VOV_PLAN_NODE, VOV_CHANGE_STATUS_RECURSIVE, s );
		f.setCurrent();
		f.info( &existflag, &mode, &curTimestamp, 1 );		      // Update time stamp.
		getQueue().enq( VOV_PLACE_MASK, VE_CHANGE, s , f.id(), 0 ) ;
	    } else {
		place.changeStatus(VOV_PLAN_NODE,VOV_CHANGE_STATUS_RECURSIVE,s);  // Invalidate all dependents,
		place.info( &existflag, &mode, &curTimestamp, 1 );		      // Update time stamp.
		getQueue().enq( VOV_PLACE_MASK, VE_CHANGE, s , place.id(), 0 ) ;
	    }
	}
	break;
    case VOV_PLAN_NODE:
	if ( existflag ) {
	    if ( curTimestamp > knownTimestamp ) {
		String s = place.getName() + "\nwas invalid and has a new timestamp.";
		s += ("\t\nwas:" + knownTimestamp.format());
		s += ("\t\n is:" + curTimestamp.format() );
		s +=  "\t\nI assume the place is now valid.\n";
		place.info( &existflag, &mode, &curTimestamp, 1 );
		VOV_ASSERT( place.changeStatus( VOV_TRACE_NODE, VOV_CHANGE_STATUS_SELF, s) );
		getQueue().enq( VOV_PLACE_MASK, VE_CHANGE, s, place.id(), 0 ) ;
	    }
	} else {
	    String s = place.getName() + "\ndoes not exist.";
	    VOV_ASSERT( place.changeStatus( VOV_NONEXISTING_NODE, VOV_CHANGE_STATUS_SELF, s ) );
	    getQueue().enq( VOV_PLACE_MASK, VE_CHANGE, s, place.id(), 0 );
	}
	break;
    case VOV_NONEXISTING_NODE:
	if ( existflag ) {
	    String s = place.getName() + "\nhas reappeared.";
	    s +=  "\t\nI assume the place is now valid.";
	    place.info( &existflag, &mode, &curTimestamp, 1 );
	    VOV_ASSERT( place.changeStatus( VOV_TRACE_NODE, VOV_CHANGE_STATUS_SELF, s ) );
	    getQueue().enq( VOV_PLACE_MASK, VE_CHANGE, s, place.id(), 0 ) ;
	}
	break;
    case VOV_TRACING_NODE:
    case VOV_RETRACING_NODE:
    case VOV_DEAD_NODE:
	break;
    default:
	vovBug("checkPlace: illegal node status" );
	break;
    }
    return VOV_OK;
}
#endif

VovStatus VovTrace::checkPlaces( int maxWork )
{
    Bzocket& zoc = getBzocket();
    VovStatus rtn = VOV_OK;
    String errmsg = "<no error>";
    
    if ( isClient() ) {
	zoc.pack( VOV_CheckPlaces );
	zoc.pack( maxWork );
	zoc.flush();
    } else {
#ifndef VOVCLIENT
	if ( isServer(VOV_SERVER_ENTRY) ) {
	    maxWork = zoc.unpackInt();
	}
	
	static int lastGlobalCheck = 0;
	static int lastCheck = 0;
	int globalCheckInterval = 3600;		// Do a check every hour.
	int minInterval = 2;			// Do not do too many checks.
	VovPage* pagePtr = &workPage();		// The reference gives me problems
	Date currentDate;			// 
	
	start_timer();		// Measure how long it takes to do the checking.
	
	if ( maxWork != 0 && currentDate.get() - lastCheck < minInterval ) {
	    // vovMessage("Whow!! too many checks!!");
	    goto skipcheck;
	} 
	lastCheck = currentDate.get();
	if ( currentDate.get() - lastGlobalCheck > globalCheckInterval ) {
	    // vovMessage("We are due for a global check of all places" );
	    maxWork = 0 ;
	}
	// vovMessage( form("CheckPlaces: maxWork %d", maxWork) );
	if ( maxWork == 0 ) {
	    // vovMessage( "Checking all the places in the design" );
	    pagePtr->placesToCheck.clear( VOV_PLACE_OCTMASK );
	    VovContentsGenerator genPlace( &pagePtr->places, VOV_PLACE_OCTMASK );
	    VovPlace place = 0;
	    while (	octGenerate( &genPlace, &place ) == OCT_OK ) {
		pagePtr->placesToCheck.add( place );
	    }
	    maxWork = pagePtr->placesToCheck.length( VOV_PLACE_OCTMASK );
	    lastGlobalCheck = getDate();
	}
	
	
	{
	    int l = pagePtr->placesToCheck.length( VOV_PLACE_OCTMASK );
	    if ( l > 0 ) {
		maxWork = maxWork <? l;
		//
		// Get the place from the queue and put it back at the end of the queue.
		//
		int count = 0;				// 
		VovContentsGenerator genPlace(&pagePtr->placesToCheck,VOV_PLACE_OCTMASK);
		VovPlace place = 0;
		int existflag = 0;		// 
		Date curTimeStamp( 0 );	// The current date timestamp for the place.
		while ( octGenerate( &genPlace, &place ) == OCT_OK ) {
		    if ( ++count >= maxWork ) break;
		    VOV_ASSERT( place.get() );
		    
		    if ( checkPlace( place, &existflag, &curTimeStamp ) == VOV_OK ) {
			pagePtr->placesToCheck.del( place );			    
			// Check if the place is "stable". 
			if ( existflag && ( currentDate.get() - curTimeStamp.get() < 3*DATE_DAY )){ 
			    pagePtr->placesToCheck.add( place ); 	// If not stable, schedule place for check.
			}
		    } else {
			vovMessage( "Some error occurred while checking a place." );
		    }
		}
	    }
	}
	
	{
	    double t = return_elapsed_time( 0 );
	    if ( t > 0.5 ) {
		// vovMessage( form("Excessive time to do the checking: %.2f s.", t ) );
	    }
	}
    skipcheck:
	if ( isServer(VOV_SERVER_EXIT) ) {
	    zoc.pack( rtn );
	    zoc.flush();
	}
#endif
    }
    if ( isClient() ) {
	VC_get_reply( zoc );
	rtn = zoc.unpackInt();
    }
    
    return rtn;
}

//
// ROUTINE: getTransitionToFire()
// Get all transitions useful to evolve to the place.
// First get a list of all the transitions which are either ready to fire or
// already firing. If the list is empty, this is an error, because we cannot
// evolve to the place. If the list is not empty look for a transition having
// a long duration and fire that first.
// 
VovStatus VovTrace::getTransitionsToFire( VovPlace& place, VovBag& transSet )
{
#ifndef VOVCLIENT
    if ( place.bfs(VOV_BACKWARD, 
		   VT_usefulTransition, 0 , 0 , 0 , 
		   (void*)&transSet ) == VOV_ERROR) {
        vovError("in getTransitions to Fire" );
	return VOV_ERROR;
    }
    
    if ( transSet.length( VOV_TRANSITION_OCTMASK ) == 0 ) {
	vovMessage( "No active transitions in the path to " + place.format() );
	return VOV_ERROR;
    } 
#endif
    return VOV_OK;
}


VovStatus VovTrace::measure( VovPlace& place, VovPlace& measure )
{
    if ( type == VOV_TRACE_TYPE_DISABLED ) return VOV_OK;
    Bzocket& zoc = getBzocket();
    VovStatus rtn = VOV_OK;
    
    if ( isClient() ) {
	zoc.pack( VOV_Measurement ); // 
	zoc.pack( place.id() );	     // 
	measure.pack( zoc );	     // Ship the whole measurement.
	zoc.flush();
    } else {
#ifndef VOVCLIENT
	if ( isServer( VOV_SERVER_ENTRY ) ) {
	    place.id() = zoc.unpackInt();
	    if ( place.get() != VOV_OK ) {
		rtn = VOV_ERROR;
	    }
	    measure.unpack( zoc );
	}
	if ( rtn == VOV_OK ) {
	    measure.modify( );
	    place.updateMeasure( measure );
	    // OH_ASSERT( octCreate( &measures, &measure ) );
	    // measure.create();
	    // vovMessage( format() + "\n\t" + measure.format() );
	}
	
	if ( isServer( VOV_SERVER_EXIT ) ) {
	    zoc.pack( rtn );
	    zoc.flush();
	}
#endif
    }
    if ( isClient() ) {
	VC_get_reply( zoc );
	rtn = zoc.unpackInt();
    }
    return rtn;
}



VovStatus VovTrace::mvlib( VovNodeSet& bag, String& oldpath, String& newpath, String& msg )
{
    if ( type == VOV_TRACE_TYPE_DISABLED ) return VOV_OK;
    Bzocket& zoc = getBzocket();
    int count = 0;
    int countPlaces = 0;
    int countTransitions = 0 ;
    
    
    if ( oldpath == newpath ) {
	vovError( "Equal args in VovTrace::mvlib()" );
	return VOV_ERROR;
    } 
    
    if (isClient() ) {
	zoc.pack( VOV_MvLib );
	bag.pack( zoc );
	zoc.pack( oldpath );
	zoc.pack( newpath );
	zoc.flush();
    } else {
#ifndef VOVCLIENT
	if ( isServer(VOV_SERVER_ENTRY) ) {
	    bag.unpack( zoc );
	    oldpath = zoc.unpackString();
	    newpath = zoc.unpackString();
	    bag.get();
	}
	
	VovContentsGenerator genPlace( &bag, VOV_NODE_OCTMASK );
	VovNode node;
	
	while ( octGenerate( &genPlace, &node ) == OCT_OK ) {
	    count++;
	    VOV_ASSERT( node.get() );
	    if ( node.nodeType() == VOV_PLACE ) {
		VovPlace place = node;
		if ( place.mvlib( oldpath, newpath ) == VOV_OK ) {
		    countPlaces++;
		}
	    } else {
		VovTransition tran = node;
		if ( tran.mvlib( oldpath, newpath ) == VOV_OK ) {
		    countTransitions++;
		}
	    }
	}
	msg = form("%d nodes examined: %d places, %d transitions modified.",
		   count, countPlaces, countTransitions );
	if ( isServer(VOV_SERVER_EXIT) ) {
	    zoc.pack( msg );
	    zoc.pack( count );
	    zoc.flush();
	}
	
#endif
    }
    if ( isClient() ) {
	VC_get_reply( zoc );
	msg = zoc.unpackString();
	count = zoc.unpackInt();
    }
    return VOV_OK;
}



#ifndef VOVCLIENT
int VovTrace::accumulateStatistics( String& name, int n )
{
    VovPage& page = workPage();
    VovProp prop( (char*)name, OCT_INTEGER );
    VovBag  bag( "PROJECT_STATISTICS" );

    VOV_ASSERT( octGetOrCreate( &page, &bag ) );
    if ( octGetByName( &bag, &prop ) != OCT_OK ) {
	prop = n;
	VOV_ASSERT( octCreate( &bag, &prop ) );
    } else {
	prop = ( prop.integer() + n );
	VOV_ASSERT( octModify( &prop ) );
    }
    // vovMessage( form( "%s %d", ohFormatName( &prop ), prop.integer() ) );
    return prop.integer();
}

double VovTrace::accumulateStatistics( String& name, double n )
{
    VovPage& page = workPage();
    VovProp prop( (char*)name, OCT_REAL );
    VovBag  bag( "PROJECT_STATISTICS" );

    VOV_ASSERT( octGetOrCreate( &page, &bag ) );
    if ( octGetByName( &bag, &prop ) != OCT_OK ) {
	prop = n;
	VOV_ASSERT( octCreate( &bag, &prop ) );
    } else {
	prop = ( prop.real() + n );
	VOV_ASSERT( octModify( &prop ) );
    }
    // vovMessage( form( "%s %f", ohFormatName( &prop ), prop.real() ) );
    return prop.real();
}

String VovTrace::formatStatistics()
{
    String s = "PROJECT STATISTICS\n";
    
    VovPage& page = workPage();
    VovBag  bag( "PROJECT_STATISTICS" );
    VOV_ASSERT( octGetOrCreate( &page, &bag ) );
    
    VovContentsGenerator genProp( &bag, OCT_PROP_MASK );
    VovProp prop( "", OCT_INTEGER ); // Just an initializer.
    while ( octGenerate( &genProp, &prop ) == OCT_OK ) {
	s += "     " + prop.format() + "\n";
    }
    // vovMessage( form( "%s %f", ohFormatName( &prop ), prop.real() ) );
    
    
    
    if ( ohGetByPropName( &bag, &prop, "TOTAL_TRANS_TIME" ) == OCT_OK && prop.integer() > 0 ) {
	double validDuration = 0.0;
	double expectedDuration = 0.0;
	int    executingTrans = 0;
	VovContentsGenerator genTran( &page.transitions, VOV_TRANSITION_OCTMASK );
	VovTransition t;
	while ( octGenerate( &genTran, &t ) == OCT_OK ) {
	    VOV_ASSERT( t.get() );
	    switch ( t.getStatus() ) {
	    case VOV_TRACE_NODE:
		validDuration += t.duration();
		break;
	    case  VOV_PLAN_NODE:
		expectedDuration += t.duration();
		break;
	    case  VOV_RETRACING_NODE:
		executingTrans++;
		break;
	    default:
		break;
	    }
	}
	double 	quality = validDuration / (double)prop.integer() ;
	double  perceivedQuality = ( validDuration + expectedDuration ) / ( prop.integer() + expectedDuration );
	s += form( "Quality = %.3f   Perceived = %.3f   Currently executing %d.\n", quality, perceivedQuality, executingTrans );
    }
    
    return s;
}

#endif

