
#include "trace.h"
#include "transition.h"
#include "date.h"
#include "place.h"
#include "event.h"
#include "user.h"

#include "path.h"
#include "utcc.h"
#include "travfunc.h"
#include "voverror.h"

extern int vov_gsub( VovProp& prop, const String& os, const  String& ns );
extern String vov_normalizeString( const String& line );

VovTransition::VovTransition( )
    :
    user( "USER", OCT_STRING ),
    host( "HOST", OCT_STRING ),
    cwd ( "CWD", OCT_STRING ),
    pid ( "PID", OCT_INTEGER ),
    retracingInfo("RETRACING_INFO", OCT_INTEGER ),
    affinity( "AFFINITY", OCT_STRING ),

    capsule( "CAPSULE" , OCT_STRING ),  
    commandLine( "DECODED_COMMAND_LINE", OCT_STRING ),
    started( "STARTED", OCT_INTEGER ),
    ended( "ENDED", OCT_INTEGER ),
    exitStatus( "EXIT_STATUS" , OCT_INTEGER ),
    okStatus( "OK_STATUS", OCT_STRING )
{
    okStatus = "0";
}

VovTransition::VovTransition( const String& actualline, const String& capsuleline) 
    :
    ( getCellName(), VOV_TRANSITION ),
    user( "USER", OCT_STRING ),
    host( "HOST", OCT_STRING ),
    cwd ( "CWD", OCT_STRING ),
    pid ( "PID", OCT_INTEGER ),
    retracingInfo("RETRACING_INFO", OCT_INTEGER ),
    affinity( "AFFINITY", OCT_STRING ),

    capsule( "CAPSULE" , OCT_STRING ),  
    commandLine( "DECODED_COMMAND_LINE", OCT_STRING ),
    started( "STARTED", OCT_INTEGER ),
    ended( "ENDED", OCT_INTEGER ),
    exitStatus( "EXIT_STATUS" , OCT_INTEGER ),
    okStatus( "OK_STATUS", OCT_STRING )
{
    setCommandLine( actualline );


    capsule = capsuleline;
    

    user = utcc_getenv("USER");
    host = utcc_gethostname();
    cwd  = path_canonicalize( utcc_getwd() );
    pid  = getpid();
    retracingInfo = oct_null_id;
    affinity = "";
    started = 0;
    ended = 0;

    okStatus = "0";
}

VovTransition::VovTransition(VovTransition& t)
    :
    ( t ),			// Construct base class VovNode.
    user( "USER", OCT_STRING ),
    host( "HOST", OCT_STRING ),
    cwd ( "CWD", OCT_STRING ),
    pid ( "PID", OCT_INTEGER ),
    retracingInfo("RETRACING_INFO", OCT_INTEGER ),
    affinity( "AFFINITY", OCT_STRING ),

    capsule( "CAPSULE" , OCT_STRING ),  
    commandLine( "DECODED_COMMAND_LINE", OCT_STRING ),
    started( "STARTED", OCT_INTEGER ),
    ended( "ENDED", OCT_INTEGER ),
    exitStatus( "EXIT_STATUS" , OCT_INTEGER ),
    okStatus( "OK_STATUS", OCT_STRING )
{
    contents.instance = t.contents.instance;
    capsule = t.capsule;
    commandLine = t.commandLine;
    user = t.user;
    host = t.host;
    cwd = t.cwd;
    pid = t.pid;
    retracingInfo = t.retracingInfo;
    affinity = t.affinity;
    started = t.started;
    ended = t.ended;

}

#ifndef VOVCLIENT
VovTransition::VovTransition( octId i )
    :
    ( i ),
    user( "USER", OCT_STRING ),
    host( "HOST", OCT_STRING ),
    cwd ( "CWD", OCT_STRING ),
    pid ( "PID", OCT_INTEGER ),
    retracingInfo("RETRACING_INFO", OCT_INTEGER ),
    affinity( "AFFINITY", OCT_STRING ),

    capsule( "CAPSULE" , OCT_STRING ),  
    commandLine( "DECODED_COMMAND_LINE", OCT_STRING ),
    started( "STARTED", OCT_INTEGER ),
    ended( "ENDED", OCT_INTEGER ),
    exitStatus( "EXIT_STATUS" , OCT_INTEGER ),
    okStatus( "OK_STATUS", OCT_STRING )
{
    if ( id() ) {
	get();
    }
}
#endif

#ifndef VOVCLIENT
VovTransition::VovTransition( VovNode& node  )
    :
    ( node ),
    user( "USER", OCT_STRING ),
    host( "HOST", OCT_STRING ),
    cwd ( "CWD", OCT_STRING ),
    pid ( "PID", OCT_INTEGER ),
    retracingInfo("RETRACING_INFO", OCT_INTEGER ),
    affinity( "AFFINITY", OCT_STRING ),

    capsule( "CAPSULE" , OCT_STRING ),  
    commandLine( "DECODED_COMMAND_LINE", OCT_STRING ),
    started( "STARTED", OCT_INTEGER ),
    ended( "ENDED", OCT_INTEGER ),
    exitStatus( "EXIT_STATUS" , OCT_INTEGER ),
    okStatus( "OK_STATUS", OCT_STRING )
{
    get();
}
#endif


VovTransition::~VovTransition()
{
}


void VovTransition::operator = (VovTransition& t)
{
    VovNode::operator=( t );
    capsule = t.capsule;
    commandLine = t.commandLine;
    user = t.user;
    host = t.host;
    cwd = t.cwd;
    pid = t.pid;
    retracingInfo = t.retracingInfo;
    affinity = t.affinity;
    started = t.started;
    ended = t.ended;
    exitStatus = t.exitStatus;
    okStatus = t.okStatus;
}

#ifndef VOVCLIENT
VovStatus VovTransition::create() 
{
    VovNode::create();
	
    OH_ASSERT( octCreate( this, &capsule ) );
    OH_ASSERT( octCreate( this, &commandLine ) );
    OH_ASSERT( octCreate( this, &user ) );
    OH_ASSERT( octCreate( this, &host ) );
    OH_ASSERT( octCreate( this, &cwd ) );
    OH_ASSERT( octCreate( this, &pid ) );
    OH_ASSERT( octCreate( this, &retracingInfo ) );
    OH_ASSERT( octCreate( this, &affinity ) );
    OH_ASSERT( octCreate( this, &started ) );
    OH_ASSERT( octCreate( this, &ended ) );
    OH_ASSERT( octCreate( this, &exitStatus ) );
    OH_ASSERT( octCreate( this, &okStatus ) );
    
    return VOV_OK;
}
#endif

#ifndef VOVCLIENT
VovStatus VovTransition::get(  ) 
{
    VovStatus rtn = VOV_ERROR;
    if ( VovNode::get() == VOV_OK ) {
	if ( octGetByName( this, &commandLine ) == OCT_OK ) {
	    OH_ASSERT( octGetByName( this, &capsule ) );
	    OH_ASSERT( octGetByName( this, &user ) );
	    OH_ASSERT( octGetByName( this, &host ) );
	    OH_ASSERT( octGetByName( this, &cwd ) );
	    OH_ASSERT( octGetByName( this, &pid ) );
	    OH_ASSERT( octGetByName( this, &retracingInfo ) );
	    OH_ASSERT( octGetByName( this, &affinity ) );
	    OH_ASSERT( octGetByName( this, &started ) );
	    OH_ASSERT( octGetByName( this, &ended ) );
	    OH_ASSERT( octGetByName( this , &exitStatus ) );
	    OH_ASSERT( octGetByName( this , &okStatus ) );
	    rtn = VOV_OK;
	}
    }
    if ( rtn == VOV_ERROR ) {
	id() = 0;
    }
    return rtn;
}
#endif

#ifndef VOVCLIENT
VovStatus VovTransition::modify(  ) 
{
    VovNode::modify();
    OH_ASSERT( octModify( &capsule ) );
    OH_ASSERT( octModify( &commandLine ) );
    OH_ASSERT( octModify( &user ) );
    OH_ASSERT( octModify( &host ) );
    OH_ASSERT( octModify( &cwd ) );
    OH_ASSERT( octModify( &pid ) );
    OH_ASSERT( octModify( &retracingInfo ) );
    OH_ASSERT( octModify( &affinity ) );
    OH_ASSERT( octModify( &started ) );
    OH_ASSERT( octModify( &ended ) );
    OH_ASSERT( octModify( &exitStatus ) );
    OH_ASSERT( octModify( &okStatus ) );
    return VOV_OK;
}
#endif

#ifndef VOVCLIENT
VovStatus VovTransition::del(  ) 
{
    OH_ASSERT( octDelete( &capsule ) );
    OH_ASSERT( octDelete( &commandLine ) );
    OH_ASSERT( octDelete( &user ) );
    OH_ASSERT( octDelete( &host ) );
    OH_ASSERT( octDelete( &cwd ) );
    OH_ASSERT( octDelete( &pid ) );
    OH_ASSERT( octDelete( &retracingInfo ) );
    OH_ASSERT( octDelete( &affinity ) );
    OH_ASSERT( octDelete( &started ) );
    OH_ASSERT( octDelete( &ended ) );
    OH_ASSERT( octDelete( &exitStatus ) );
    OH_ASSERT( octDelete( &okStatus ) );
    VovNode::del();
    return VOV_OK;
}
#endif


#ifndef VOVCLIENT
VovStatus VovTransition::mvlib( String& oldpath, String& newpath )
    // Return: VOV_OK if changed, VOV_NOT_OK if no change.
{
    int changes = 0;
    changes += vov_gsub( cwd, oldpath, newpath );
    changes += vov_gsub( capsule, oldpath, newpath );
    changes += vov_gsub( commandLine, oldpath, newpath );
    
    if ( changes ) {
	setCommandLine( commandLine.string() );
	VOV_ASSERT( modify() );
	return VOV_OK;
    } else {
	return VOV_NOT_OK;
    }
}
#endif

#ifndef VOVCLIENT
VovStatus VovTransition::match( VovTransition& t )
{
    if ( cwd.string() != t.cwd.string() ) {
	return VOV_NOT_OK;
    } else if ( commandLine.string() != t.commandLine.string() ) {
	return VOV_NOT_OK;
    } else {
	return VOV_OK;
    }
}
#endif

#ifdef NOT_USED_FOR_NOW
VovStatus vov_encodeCommandLine(String& cn, String& nickName, String& cl )
{
    if (nickName != 0) {
	// cn could be a full path, in which case we also try to match the
	// last component of the path.

	if ( cl.contains( cn ) ) {
	    cl.gsub( cn, nickName );
	} else {
	    if ( cn.contains('/') ) {
		cn = cn.after( cn.index( '/' , -1 )) ;
		cl.gsub( cn, nickName );
	    }
	}
    }
    return VOV_OK;
}

String VovTransition::encodeCommandLine()
{
    String cl = commandLine.string();
    VovPlace place = 0;

    VovContainersGenerator genInput( this, OCT_INSTANCE_MASK );
    while ( octGenerate( &genInput, &place )  == OCT_OK ) {
	if ( place.get() == VOV_OK ) 
	vov_encodeCommandLine(	place.getCommonName(), "XXX" , cl );
    }

    VovContentsGenerator genOutput( this, OCT_INSTANCE_MASK );
    while ( octGenerate( &genOutput, &place )  == OCT_OK ) {
	if ( place.get() == VOV_ERROR ) continue; // Skip this node. Should never happen.
	vov_encodeCommandLine(	place.getCommonName(), "XXX" , cl );
    }

    vovMessage( "Decoded command line:" + commandLine.string() );
    return cl;
}
#endif

#ifndef VOVCLIENT
VovStatus VovTransition::initiate( VovNodeStatus s ) 
{
    if ( s != VOV_TRACING_NODE && s != VOV_RETRACING_NODE ) {
	vovFatalError("initiating transition: must be either a tracing or a retracing" );
    }

    setStatus( s );

    VovContentsGenerator genOut( this, VOV_PLACE_OCTMASK );
    VovPlace place = 0;
    String msg;
    while ( octGenerate( &genOut, &place ) == OCT_OK ) {
	if ( place.get() == VOV_ERROR ) continue; // Skip this node. Should never happen.
	place.changeStatus( s, VOV_CHANGE_STATUS_SELF, msg );
	place.changeStatus( VOV_PLAN_NODE, VOV_CHANGE_STATUS_RECURSIVE, msg );
	// place.setStatus( s );
	place.setCurrent();
    }

    // retracingInfo.modify( oct_null_id ); 
    started.modify( getDate() ); 	// Set start date of transition.
    ended.modify( 0 );			// Not ended yet.

    return VOV_OK;
}
#endif

void VovTransition::setAffinity( const String& s )
{
    String a = vov_normalizeString( s );
    affinity = a;
}


//
// Some characteristics of transitions are maintained during retracing.
// Keep: affinity, legal exit status, user, sets, annotations.
//
#ifndef VOVCLIENT
VovStatus VovTransition::retracingInvariants( VovTrace& trace, VovTransition& oldTransition )
    //
    // oldTransition is the transition that is already in the trace.
    //
{
    VOV_ASSERT( setRetracingInfo( oldTransition ) );

    affinity.modify( oldTransition.affinity.string() );
    okStatus.modify( oldTransition.okStatus.string() );

    VovUser u = oldTransition.getUser();
    if (  trace.find( u ) == VOV_OK ) {
	u.active( *this );
	setUser( oldTransition.getUser() );
    }

    //
    // Maintain all sets.
    //
    VovContainersGenerator genSet( &oldTransition, VOV_NODE_SET_OCTMASK );
    VovNodeSet set;
    while ( octGenerate( &genSet, &set ) == OCT_OK ) {
	VOV_ASSERT( set.get() );
	set.add( *this );
    }

    //
    // Maintain all notes.
    //
    if ( oldTransition.countNotes() ) {
	copyNotes( oldTransition );
    }
		     
    return VOV_OK;
}
#endif


#ifndef VOVCLIENT
VovStatus VovTransition::setRetracingInfo( VovTransition& t ) 
{
    if ( retracingInfo.integer() != 0 || t.retracingInfo.integer() != 0 ) {
	vovError( "garbage in setRetracingInfo" );
    }
    retracingInfo.modify( t.id() );
    t.retracingInfo.modify( id() );
    return VOV_OK;
}
#endif

#ifndef VOVCLIENT
VovStatus vov_checkExitStatus( String ok, int exit_status  )
{
    extern String vov_stripWord( String& );

    /* Default behavior: 0 is OK all else not OK */
    if ( ok == 0 ) {
	return ( exit_status == 0 ) ? VOV_OK : VOV_ERROR;
    }
    
    // vovMessage( form("Exit status check: %s %d",(char*)ok, exit_status ) );
    
    ok = vov_normalizeString( ok );
    int l = ok.length() + 1;	// To stop infinite loops.
    while ( ok.length() < l ) {
	l = ok.length();
	String c = vov_stripWord( ok );
	// vovMessage( form("String %s l=%d, word = %s, l=%d",(char*)ok,ok.length(),(char*)c,c.length()) );
	if ( atoi( (char*)c ) == exit_status ) {
	    return VOV_OK;
	} else 	if ( c.matches( "[0-9]+-[0-9]+" ) ) {
	    String los = c.before( "-" );
	    String his = c.after( "-" );
	    int lo = atoi( (char*)los );
	    int hi = atoi( (char*)his );
	    if ( exit_status >= lo && exit_status <= hi ) {
		return VOV_OK;
	    }
	} 
    }
    return VOV_ERROR;
}
#endif

#ifndef VOVCLIENT
void VovTransition::checkAllInputs(
				   VovEventQueue& queue,
				   VovTransitionTerminationCode& terminationCode, 
				   String& failureExplanation
				   )
{
    VovPlace place;
    VovContainersGenerator genInput( this, VOV_PLACE_OCTMASK );
    while ( octGenerate( &genInput, &place ) == OCT_OK ) {
	VOV_ASSERT( place.get() );
	if ( place.getStatus() != VOV_TRACE_NODE ) {
	    failureExplanation += ("Input place: " + place.getName() + "\nis not (or no longer) valid.\n");
	    terminationCode = VOV_TRANSITION_TERMINATION_INVALID_INPUT;
	}
    }
}
#endif

#ifndef VOVCLIENT
void VovTransition::checkAllOutputs(
				    VovEventQueue& queue,
				    VovTransitionTerminationCode& terminationCode, 
				    String& failureExplanation
				    )
{
    Date startTime = started.integer();
    
    
    int countOutputs = 0;
    VovContentsGenerator genOutput( this, VOV_PLACE_OCTMASK );
    VovPlace place;
    while ( octGenerate( &genOutput, &place ) == OCT_OK ) {
	if ( place.get() == VOV_ERROR ) continue; // Skip this node. Should never happen.
	countOutputs++;

	switch ( place.getPlaceType() ) {
	case VOV_EXIT_STATUS:
	    place.setStatus( VOV_TRACE_NODE );
	    break;
	default:
	    {
		int exists,size;
		Date timestamp(0);
		int mode;

		Date knownTimestamp = place.getTimeStamp() ;
		place.setCurrent();
		(void) place.info( &exists, &mode, &timestamp, 0);
		
		if ( ! exists ) {
		    failureExplanation += ( place.format() + " does not exist.\n" );
		    terminationCode = VOV_TRANSITION_TERMINATION_DEFECTIVE_OUTPUT;
		} else if ( timestamp < knownTimestamp ) {
		    failureExplanation += ( place.format() + "'s timestamp has decreased!\n" );
		    terminationCode = VOV_TRANSITION_TERMINATION_INCONSISTENT;
		} else {
		    if ( timestamp < startTime ) {
			String msg = place.format() + "\nhas a time stamp\n" +
			  timestamp.format() + " which is less than\n" + startTime.format();
			queue.enq( VOV_PLACE_MASK, VE_ERROR, msg, place.id(), 0 );
		    } else {
			// This event is important because it makes it easier to edit the modified place,
			// rather than navigating from the transition.
			String msg = place.format() + "\n";
			queue.enq(VOV_PLACE_MASK, VE_CHANGE, msg, place.id(), 0 );
		    }
		}
		place.setStatus( terminationCode != VOV_TRANSITION_TERMINATION_OK ?  VOV_PLAN_NODE : VOV_TRACE_NODE );
		(void) place.info( &exists, &mode, &timestamp, 1);
	    }
	}
    }
    if ( countOutputs == 0 ) {
	terminationCode == VOV_TRANSITION_TERMINATION_NO_OUTPUTS;
    }


}
#endif

#ifndef VOVCLIENT
VovStatus VovTransition::terminate( 
				   VovEventQueue& queue, 
				   VovStatus termStatus, 
				   int exit_status, 
				   String& pipedMsg,
				   VovTransitionTerminationCode& terminationCode
				   ) 
    // Returns :
    // VOV_OK: transition succeeded.
    // VOV_NOT_OK : transition failed
    // VOV_ERROR: bogus data.
    // VOV_DELETE_NODE: the transition should be removed.
{
    if ( getStatus() != VOV_TRACING_NODE && getStatus() != VOV_RETRACING_NODE ) {
	// Almost never happens.
	vovError( format() + " is not firing now. Does not have to be terminated");
	return VOV_ERROR;	
    }
    
    // vovMessage( form("terminate %d,%d", id(), retracingInfo.integer() ) );
    if ( retracingInfo.integer() != 0 ) {
	VovTransition otherTrans = retracingInfo.integer();
	if ( getStatus()==VOV_RETRACING_NODE && otherTrans.getStatus()==VOV_TRACING_NODE) {
	    vovError("The TRACING transition must terminate before the RETRACING one.");
	    vovMessage("Forcing termination of TRACING transition");
	    if ( otherTrans.terminate( queue, VOV_OK, 0, pipedMsg, terminationCode ) != VOV_OK ) {
		return VOV_ERROR;
	    }
	}
	
	if ( otherTrans.pid.integer() == 0 ) {
	    vovMessage( "Deleting manually retraced transition" );
	    otherTrans.del();
	    retracingInfo.modify( oct_null_id );
	}

    } else {
	// vovMessage( "terminate tracing of " + getCommandLine() );
    }
    

    ended.modify( getDate() );
    exitStatus.modify( exit_status );
    terminationCode = VOV_TRANSITION_TERMINATION_OK;	

    if ( getStatus() == VOV_RETRACING_NODE && retracingInfo.integer() != 0) {
	VovTransition otherTransition = retracingInfo.integer();
	if ( otherTransition.id() == retracingInfo.integer() ) {
	    // OK, the TRACING transition is good.
	    // Check all outputs of the old transition, then return a code that
	    // forces this transition to be deleted.
	    VovContentsGenerator genOutput( this, VOV_PLACE_OCTMASK );
	    VovPlace place;
	    String msg;
	    while ( octGenerate( &genOutput, &place ) == OCT_OK ) {
		VOV_ASSERT( place.get() );
		if ( place.getStatus() == VOV_RETRACING_NODE ) {
		    place.changeStatus( VOV_PLAN_NODE, VOV_CHANGE_STATUS_SELF_REC, msg );
		}
	    }
	    setStatus( otherTransition.getStatus() ); // Inherit status from other transition.
	    return VOV_DELETE_NODE;		      // A code meaning: must be deleted.
	} 
	vovError( "BOGUS retracing info: terminating RETRACING trans");
	// Continue with the termination of this transition.
	retracingInfo.modify( 0 );
    }
    

    String failureExplanation;	// 

    if ( vov_checkExitStatus( okStatus.string() , exitStatus.integer() ) != VOV_OK ) {
	terminationCode = VOV_TRANSITION_TERMINATION_WRONG_STATUS;
	failureExplanation = form( "Transition exited with error status %d.\n", exitStatus.integer() );
    }
    
    if ( termStatus != VOV_OK ) {
	terminationCode = VOV_TRANSITION_TERMINATION_FAILED;
	failureExplanation = "The transition itself claims to have failed.\n";
    }
    
    manageRetracing( queue );
    
    
    setStatus( VOV_TRACE_NODE ); // Assume the transition is going to be good. 
    // We must do change the status from TRACING, otherwise we might 
    // go into VT_invalidate and kill the transition, which is
    // no longer there.
    
    
    checkAllInputs( queue, terminationCode, failureExplanation );
    checkAllOutputs( queue, terminationCode, failureExplanation );


    if ( terminationCode != VOV_TRANSITION_TERMINATION_OK ) {
	String msg = getCommandLine() + "\nFAILED. REASONS FOR THE FAILURE:\n" + failureExplanation + "\n" + pipedMsg;
	changeStatus( VOV_PLAN_NODE, VOV_CHANGE_STATUS_RECURSIVE, msg );
	queue.enq( VOV_TRANSITION_MASK, VE_ERROR, msg, id(), 0 );
    } else {
	String msg = getCommandLine() + "\n\tSUCCESSFUL.\n";
	changeStatus( VOV_TRACE_NODE, VOV_CHANGE_STATUS_SELF, msg );
	queue.enq( VOV_TRANSITION_MASK, VE_END, msg, id(), 0 );
    }

    
    // Find a better spot for the transition, now that all IO's have been found.
    doPlacement();		
    
    return VOV_OK;
}
#endif

#ifndef VOVCLIENT
VovStatus VovTransition::manageRetracing( VovEventQueue& queue )
{
    //
    // If this is not a retracing transition, just return.
    //
    if ( retracingInfo.integer() == 0 ) return VOV_OK;
    
    if ( getStatus() != VOV_TRACING_NODE ) {
	vovError( "manageRetracing: called with bogus args" );
	return VOV_ERROR;
    }
    
    int    changedIO = 0;	// A flag. Set to 1 if the IO lists of retracing differ
    // from original lists.
    String changedIOdesc;	// A textual description of the changes.
    VovPlace place = 0;		// Used to run through the list of IO's.
    VovTransition retracingTrans = retracingInfo.integer();
    if ( retracingTrans.id() != retracingInfo.integer() ) {
	vovError("Bogus id in manageRetracing()" );
	return VOV_ERROR;
    } 
	
    VovContentsGenerator genOutput( this , VOV_PLACE_OCTMASK );
    while ( octGenerate( &genOutput, &place ) == OCT_OK ) {
	if ( place.get() == VOV_ERROR ) continue; // Skip this node. Should never happen.
	if ( retracingTrans.isOutput( place ) != VOV_OK ) {
	    changedIO = 1;
	    changedIOdesc += ( "NEW OUTPUT: " + place.format() + "\n" );
	} 
    }
    VovContainersGenerator genInput( this , VOV_PLACE_OCTMASK );
    while ( octGenerate( &genInput, &place ) == OCT_OK ) {
	if ( place.get() == VOV_ERROR ) continue; // Skip this node. Should never happen.
	if ( retracingTrans.isInput( place ) != VOV_OK ) {
	    changedIO = 1;
	    changedIOdesc += ( "NEW INPUT: " + place.format() + "\n" );
	} 
    }
    VovContentsGenerator genOutput2( &retracingTrans , VOV_PLACE_OCTMASK );
    while ( octGenerate( &genOutput2, &place ) == OCT_OK ) {
	if ( place.get() == VOV_ERROR ) continue; // Skip this node. Should never happen.
	if ( isOutput( place ) != VOV_OK ) {
	    changedIO = 1;
	    changedIOdesc += ( "OLD OUTPUT: " + place.format() + "\n" );
	} 
    }
    VovContainersGenerator genInput2( &retracingTrans , VOV_PLACE_OCTMASK );
    while ( octGenerate( &genInput2, &place ) == OCT_OK ) {
	if ( place.get() == VOV_ERROR ) continue; // Skip this node. Should never happen.
	if ( isInput( place ) != VOV_OK ) {
	    changedIO = 1;
	    changedIOdesc += ( "OLD INPUT: " + place.format() + "\n" );
	} 
    }
    retracingInfo.modify( 0 );
    
    if ( changedIO ) {
	queue.enq( VOV_TRANSITION_MASK, VE_CHANGE, getCommandLine() + "\n" + changedIOdesc, id(), 0 ) ;
    }
    return VOV_OK;
}
#endif

// 
// A transition is ready to fire if 
// all its inputs are valid ('TRACE')
// and all its outputs aren't.
//
#ifndef VOVCLIENT
VovStatus VovTransition::isReadyToFire( String& msg )
{
    if (getStatus() != VOV_PLAN_NODE ) {
	return VOV_NOT_OK;
    }

    VovNode node = 0;
    VovContainersGenerator genIn( this, VOV_NODE_OCTMASK );
    while (octGenerate( &genIn, &node ) == OCT_OK ) {
	if ( node.get() == VOV_ERROR ) continue; // Skip this node. Should never happen.
	if ( node.getStatus() != VOV_TRACE_NODE ) {
	    msg += "INPUT " + node.getName() + " is " + node.getStatusString() + "\n";
	    return VOV_NOT_OK;
	}
    }
    VovContentsGenerator genOut( this, VOV_NODE_OCTMASK );
    while (octGenerate( &genOut, &node ) == OCT_OK ) {
	if ( node.get() == VOV_ERROR ) continue; // Skip this node. Should never happen.
	if ( node.getStatus() != VOV_PLAN_NODE && node.getStatus() != VOV_NONEXISTING_NODE ) {
	    msg += "OUTPUT " + node.getName() + " is " + node.getStatusString() + "\n";
	    return VOV_NOT_OK;
	}
    }
    return VOV_OK;
}
#endif

#ifndef VOVCLIENT
VovStatus VovTransition::isUserInitiated()
{
    if ( getStatus() == VOV_RETRACING_NODE ) { // Only server can set node to RETRACING.
	return VOV_NOT_OK;		       // This happens in fast retracings.
    }					       // 
    if ( retracingInfo.integer() != 0 ) {      // 
	return VOV_NOT_OK;
    }
    return VOV_OK;
}
#endif

#ifndef VOVCLIENT
VovStatus VovTransition::isSuccessful()
{
    return VOV_OK;
}
#endif

long  VovTransition::duration()
{ 
    int d = ended.integer() - started.integer();
    return (d>=0) ? d : -1 ;
}

String VovTransition::getCommandLine() 
{ 
    return commandLine.string(); 
}

void VovTransition::setCommandLine( String& cmd )
{
    commandLine = vov_normalizeString( cmd );
    // Find a shorter name to be used in the trace label.
    const int SHORT_NAME_LENGTH = 15;
    String shortName = commandLine.string();
    while ( shortName.length() > SHORT_NAME_LENGTH ) {
	String shorter = shortName.before( RXwhite, -1 );
	if ( shorter.length() == shortName.length() || shorter.length() < SHORT_NAME_LENGTH ) {
	    break;
	} 
	shortName = shorter;
    }
    VovNode::setName( shortName );	// Name the node.
}

String VovTransition::format( )
{
    extern String vov_prettyTime( int );
    Date start = started.integer();
    Date end   = ended.integer();

    String s = "cd " + cwd.string();
    s += form("\n%s", (char*)getCommandLine() );
    s += form("\n# Start: %s\n#   End: %s\n# Duration: %s. Exit status: %d  (%s@%s) %s",
	      (char*)start.format(), (char*)end.format(), (char*)vov_prettyTime( duration()), exitStatus.integer(),
	      (char*)user.string(), (char*)host.string(), (char*)getStatusString() );
    return s;
}


