#include "defs.h"
#define VOVCLIENT
#include "vov.h"


extern "C"
{
#include "options.h"
}

optionStruct optionList[] = {
    { "A",	0, 		"Measure area" },
    { "N",	0, 		"Measure net length" },
    { "I",	0, 		"Measure number of instances" },
    { "b",	0,		"compute bounding boxes of the nets" },
    { "l",	"integer",	"number of longest nets that will be displayed" },
    { OPT_RARG,	"cell:view",	"cell and view to get statistics on" },
    { 0,	0,		0 }
};

#define MEASURE_AREA       0x01
#define MEASURE_NET_LENGTH 0x02
#define MEASURE_INSTANCES  0x04


main(int argc, char **argv )
{
    extern int optind;
    extern char *optarg;
    int		bb = 0;		// flag to show bounding boxes.
    int		num_longest_nets = 6;

    int measures = 0;		// measurements requested.

    ChipArea area;
    ChipInstances instances;
    ChipNets nets;

    int option;			// 
    while ((option = optGetOption(argc, argv)) != EOF) {
	switch (option) {
	case 'A':
	    measures |= MEASURE_AREA;
	    break;
	case 'N':
	    measures |= MEASURE_NET_LENGTH;
	    break;
	case 'I':
	    measures |= MEASURE_INSTANCES;
	    break;
	case 'b':
	    bb = 1;
	    break;
	case 'l':
	    num_longest_nets = atoi(optarg);
	    break;
	}
    }


    octObject facet;

    if (argc == optind+1) {
	ohUnpackDefaults( &facet, "r", ":symbolic:contents" );
	if ( ohUnpackFacetName( &facet, argv[optind] ) != OCT_OK ) optUsage();
    } else {
	optUsage();
    }

    VOV__begin( argc, argv );

    octStatus status = octOpenFacet( &facet );
    if ( status < OCT_OK ) {
        if (status == OCT_NO_PERM) {
	    FatalError("Facet is unreadable" );
        }
        FatalError( "Cannot open the facet");
    }
  
    int pId = VOV__inputFacet( &facet );

    if ( measures ) {
	if ( measures & MEASURE_AREA ) {
	    area.main(&facet);
	    String name = form( "AREA.%s", ohFormatName( &facet ) );
	    VOV__measure( pId, (char*)name, "lambda^2", area.chipArea  );
	    printf( "Area = %d lambda^2\n", area.chipArea );
	}
	if ( measures & MEASURE_NET_LENGTH ) {
	    nets.main(&facet, bb);
	    String name = form( "NETLENGTH.%s", ohFormatName( &facet ) );
	    VOV__measure( pId, (char*)name, "lambda", (int)nets.total_net_length );
	    printf( "Net length = %d lambda\n", (int)nets.total_net_length );
	}
	if ( measures & MEASURE_INSTANCES ) {
	    instances.main( &facet );
	    String name = form( "INSTANCES.%s", ohFormatName( &facet ) );
	    VOV__measure( pId, (char*)name, "instances", instances.num_instances );
	    printf( "Instances = %d\n",  instances.num_instances );
	}
    } else {
	// Behave like old chipstats.
	area.main( &facet );
	instances.main( &facet );
	nets.main( &facet, bb );
	cout << area.format();
	cout << instances.format();
	cout << nets.format( bb,  num_longest_nets );
    }
    octEnd();
    VOV__end( 0 );
}


void FatalError( String& es )
{
    cerr.form("%s: Fatal Error: %s\n",optProgName,(char*)es);
    octError(optProgName);
    exit(-1);
}

void Warning( String& es )			/* error string */
{
    cerr.form("%s: Warning: %s\n", optProgName, (char*)es);
}

