#include "defs.h"

int net_length(octObject *net, int *actual_length, int *bbox_length, int *area )
{
    octStatus status;

    /*
     * Compute the length of all paths attached to the net
     */
    int tot_length = 0;
    int tot_area = 0;
    octGenerator gen;
    octObject path;
    OH_ASSERT(octInitGenContents(net, OCT_PATH_MASK, &gen));
    while ((status = octGenerate(&gen, &path)) == OCT_OK) {
	int32 num_points;
	struct octPoint *points;
	ohGetPoints(&path, &num_points, &points);
	if (num_points != 2) {
	    Warning( form("Net %s has a path with %d points", net->contents.net.name, num_points));
	    return 0; 
	}

	octCoord dx = points[0].x - points[1].x;
	octCoord dy = points[0].y - points[1].y;
	long length = 0;	// Manhattan length of the segment.
	length += dy < 0 ? -dy : dy;
	length += dx < 0 ? -dx : dx;
	tot_length += length;
	tot_area += length * path.contents.path.width;
    }
    OH_ASSERT(status == OCT_GEN_DONE ? OCT_OK : status);
    octFreeGenerator( &gen );

    /*
     *  Compute the bounding box of all terminals on the net
     */
    struct octBox box, bb;
    int first = 1;
    octObject term;
    OH_ASSERT(octInitGenContents(net, OCT_TERM_MASK, &gen));
    while ((status = octGenerate(&gen, &term)) == OCT_OK) {
	if (ohTerminalBB(&term, &bb) != OCT_OK) {
	    Warning( form("%s term %s on %s has no BBox",
			    (term.contents.term.instanceId==0) ? "FORMAL" : "ACTUAL",
			    term.contents.term.name,
			    net->contents.net.name) );
	    continue;		/* Policy violation */
	}
	if (first || bb.lowerLeft.x < box.lowerLeft.x) {
	    box.lowerLeft.x = bb.lowerLeft.x;
	}
	if (first || bb.lowerLeft.y < box.lowerLeft.y) {
	    box.lowerLeft.y = bb.lowerLeft.y;
	}
	if (first || bb.upperRight.x > box.upperRight.x) {
	    box.upperRight.x = bb.upperRight.x;
	}
	if (first || bb.upperRight.y > box.upperRight.y) {
	    box.upperRight.y = bb.upperRight.y;
	}
	first = 0;
    }
    OH_ASSERT(status == OCT_GEN_DONE ? OCT_OK : status);
    octFreeGenerator( &gen );

    *actual_length = tot_length / LAMBDA;
    *bbox_length = ((box.upperRight.x - box.lowerLeft.x) +
		    (box.upperRight.y - box.lowerLeft.y)) / LAMBDA;
    *area = tot_area / (LAMBDA * LAMBDA);
    return 1;
}


net_length_compare( struct net_length_struct *a, struct net_length_struct *b )
{
    if (b->length > a->length) {
	return 1;
    } else if (b->length < a->length) {
	return -1;
    } else {
	return 0;
    }
}

net_bb_compare( struct net_length_struct *a, struct net_length_struct *b )
{
    if (b->bbox_length > a->bbox_length) {
	return 1;
    } else if (b->bbox_length < a->bbox_length) {
	return -1;
    } else {
	return 0;
    }
}



int ChipNets::main( octObject *facet, int bb )
{
    num_nets = all_net_lengths( facet );
    if (num_nets < 0) {
	FatalError("Policy violation while checking nets\n");
    }

    if (bb) {
	qsort(netArray, num_nets, sizeof(struct net_length_struct), net_bb_compare);
    } else {
	qsort(netArray, num_nets, sizeof(struct net_length_struct), net_length_compare);
    }

    if (bb) {
	for( int i = 0; i < num_nets; i++) {
	    total_net_length += netArray[i].bbox_length;
	    total_net_length2 +=  netArray[i].bbox_length * netArray[i].bbox_length;
	    total_net_area += netArray[i].area;
	    total_net_area2 += netArray[i].area *  netArray[i].area;
	}
    } else {
	for( int i = 0; i < num_nets; i++) {
	    total_net_length += netArray[i].length;
	    total_net_length2 +=  netArray[i].length * netArray[i].length;
	    total_net_area += netArray[i].area;
	    total_net_area2 += netArray[i].area *  netArray[i].area;
	}
    }
    if (num_nets > 0) {
	total_net_length_var = (total_net_length2 - (total_net_length * total_net_length)/num_nets) / num_nets;
	total_net_area_var   = (total_net_area2 - (total_net_area * total_net_area)/num_nets) / num_nets;
    }
    return 0;
}


int ChipNets::all_net_lengths( octObject *facet )
{
    octGenerator gen;
    octObject net;
    octStatus status;


    int size = 10;
    netArray = new struct net_length_struct[size];

    num_nets = 0;
    OH_ASSERT(octInitGenContents(facet, OCT_NET_MASK, &gen));
    while ((status = octGenerate(&gen, &net)) == OCT_OK) {
	if (num_nets >= size) {
	    int newsize =  size * 2;
	    struct net_length_struct* tmp = new struct net_length_struct[newsize];
	    for (int i = 0 ; i < size ; i++ ) tmp[i] = netArray[i];
	    delete netArray;
	    netArray = tmp;
	    size = newsize;
	}
	int actual_length, bbox_length, area;
	if (! net_length(&net, &actual_length, &bbox_length, &area)) {
	    return -1;		// Bail out if any errors.
	}
	netArray[num_nets].length = actual_length;
	netArray[num_nets].bbox_length = bbox_length;
	netArray[num_nets].area = area;
	netArray[num_nets].objectId = net.objectId;
	num_nets++;
    }
    OH_ASSERT(status == OCT_GEN_DONE ? OCT_OK : status);
    return num_nets;
}






