#include "defs.h"

#define PRINTF	(void) printf

#define MIN(a,b)	(((a) < (b)) ? (a) : (b))

String ChipInstances::format()
{
    String s;
    s += form("\tNumber of instances    = %d\n", num_instances);
    s += form("\tNumber of connectors   = %d\n" , num_connectors);
    s += form("\tNumber of vias         = %d\n" , num_vias);

    if (num_instances > 0) {
	s += form("\tAverage instance area  = %14.2f\n" , tot_instance_area / num_instances);
    }
    s += form("\tSt.dev. instance area  = %14.2f\n" , sqrt(tot_instance_area_var));

    return s;
}

String ChipArea::format()
{
    String s;
    s += form("\tChip Area W x H        = %6d x%6d\n", chipWidth, chipHeight);
    s += form("\tChip Area              = %14d (100.00%%)\n", chipArea);
    return s;
}

String ChipNets::format( int bb = 0, int num_longest_nets = 6 )
{
    String s;
    s += form("\tTotal   net length     = %14.2f\n" , total_net_length);
    s += form("\tTotal   net area       = %14.2f\n" , total_net_area);
    if (num_nets > 0) {
	s += form("\tAverage net length     = %14.2f\n" , total_net_length / num_nets);
    }
    s += form("\tSt.dev. net length     = %14.2f\n" , sqrt(total_net_length_var));

    s += form("\n***** %d LONGEST NETS\n", MIN(num_nets, num_longest_nets));
    s += form("%-26s %16s %16s\n", "(---) NAME", "LENGTH (lambda)" , "AREA (lambda^2)");
    for( int i = 0; i < MIN(num_nets, num_longest_nets); i++) {
	octObject net;
	OH_ASSERT(ohGetById(&net, netArray[i].objectId));
	String net_name = net.contents.net.name;
	if ( net_name.contains( "/" ) ) {
	    net_name = net_name.after( "/", -1 );
	}
	int length = ( bb) ?  netArray[i].bbox_length : netArray[i].length ;

	s += form("(%3d) %-20s %16d%4s %16d\n", i+1,  (char*)net_name,
		  length, (bb) ? "(bb)" : "",  netArray[i].area);
    }
    return s;
}

