#include "copyright.h"
#include "port.h"
/* LINTLIBRARY */

#include "smm.h"

#ifdef IBM_WATC
#define void int
#endif


/*
 *  Author: Richard Rudell, UC Berkeley, March 1987
 *
 *  These are interface routines to be placed between a program and the
 *  system memory allocator.  
 *
 *  It forces well-defined semantics for several 'borderline' cases:
 *
 *	malloc() of a 0 size object is guaranteed to return something
 *	    which can safely be freed (but not dereferenced, of course)
 *	free() accepts (silently) an 0 pointer
 *	realloc of a 0 pointer is allowed, and is equiv. to malloc()
 *	For the IBM/PC it forces no object > 64K; however the argument
 *	    to malloc/realloc is a 'long' to catch errors.
 *
 *  The function pointer MMoutOfMemory() contains a vector to handle a
 *  'out-of-memory' error (which, by default, points at a simple wrap-up 
 *  and exit routine).
 */

void (*MMoutOfMemory)() = MMout_of_memory;

/* MMout_of_memory -- out of memory for lazy people, flush and exit */
void 
MMout_of_memory(size)
long size;
{
    (void) fflush(stdout);
    (void) fprintf(stderr, "\nout of memory allocating %ld bytes\n", size);
    exit(1);
}


char *
MMalloc(size)
long size;
{
    char *p;

#ifdef IBMPC
    if (size > 65000L) {
	if (MMoutOfMemory != (void (*)()) 0 ) (*MMoutOfMemory)(size);
	return NIL(char);
    }
#endif
    if (size <= 0) size = sizeof(long);
    if ((p = malloc((unsigned) size)) == NIL(char)) {
	if (MMoutOfMemory != (void (*)()) 0 ) (*MMoutOfMemory)(size);
	return NIL(char);
    }
    return p;
}


char *
MMrealloc(obj, size)
char *obj;
long size;
{
    char *p;

#ifdef IBMPC
    if (size > 65000L) {
	if (MMoutOfMemory != (void (*)()) 0 ) (*MMoutOfMemory)(size);
	return NIL(char);
    }
#endif
    if (size <= 0) size = sizeof(long);
    if (obj == NIL(char)) return MMalloc(size);
    if ((p = realloc(obj, (unsigned) size)) == NIL(char)) {
	if (MMoutOfMemory != (void (*)()) 0 ) (*MMoutOfMemory)(size);
	return NIL(char);
    }
    return p;
}


void 
MMfree(p)
char *p;
{
    if (p != NIL(char)) free(p);
}
