/* C++ includes */
#include <stream.h>

/* C includes */

/* Local includes */
#include "bzocket.h"
#include "place.h"
#include "transition.h"
#include "select.h"
#include "annotation.h"
#include "retrace.h"
#include "user.h"
#include "utcc.h"
#include "voverror.h"

/*
* Associate a different port number to each project
* to allow multiple servers to run on the same host.
* Uses a built in hashing function.
*/
int net_portNumber( String& projectname )
{
    const int  VOV_SOCKET_NUMBER =  6200; 	// Base number.
    int hash = hashpjw( (char*)projectname );	// 
    hash &= 0x0FF;				// Reduce in the range 0--255.
    int n = VOV_SOCKET_NUMBER + hash;		// This is the port number.
    return  n;
}


void VovGenerator::pack( Bzocket& z )
{
    z.pack( (int)generator_func );
    z.pack( (int)state );
    z.pack( (int)free_func );
}

void VovGenerator::unpack( Bzocket& z )
{
    generator_func = (octStatus (*)(struct octGenerator*, struct octObject*))z.unpackInt();
    state = (char*)z.unpackInt();
    free_func = (octStatus (*)())z.unpackInt();
}



//
//
//
void VovNode::pack( Bzocket& z )
{
    VovObject::pack( z );
    label.pack( z );
    level.pack( z );
    status.pack( z );
    node_type.pack( z );
    searchId.pack( z );
}

#ifndef VOVCLIENT
void VovNode::packInputs( Bzocket& z )
{
    VovContainersGenerator genInput( this, VOV_PLACE_OCTMASK );
    VovNode node = 0;
    while ( octGenerate( &genInput, &node ) == OCT_OK ) {
	z.pack( 1 );
	node.get();
	if ( node.nodeType() == VOV_PLACE ) {
	    VovPlace place = node.id();
	    place.pack( z );
	} else {
	    VovTransition trans = node.id();
	    trans.pack( z );
	}
    }
    z.pack( 0 );
}
#endif

#ifndef VOVCLIENT
void VovNode::packOutputs( Bzocket& z )
{
    VovContentsGenerator genOutput( this, VOV_PLACE_OCTMASK );
    VovNode node = 0;
    while ( octGenerate( &genOutput, &node ) == OCT_OK ) {
	z.pack( 1 );
	node.get();
	if ( node.nodeType() == VOV_PLACE ) {
	    VovPlace place = node.id();
	    place.pack( z );
	} else {
	    VovTransition trans = node.id();
	    trans.pack( z );
	}
    }
    z.pack( 0 );
}
#endif

void VovNode::unpack( Bzocket& z )
{
    VovObject::unpack( z );
    label.unpack( z );
    level.unpack( z );
    status.unpack( z );
    node_type.unpack( z );
    searchId.unpack( z );
}

void VovNodeSet::pack( Bzocket& zoc ) 
{
    VovObject::pack( zoc );
    owner.pack( zoc );
    created.pack( zoc );
    selectRule.pack( zoc );
}

void VovNodeSet::unpack( Bzocket& zoc ) 
{
    VovObject::unpack( zoc );
    owner.unpack( zoc );
    created.unpack( zoc );
    selectRule.unpack( zoc );
}

//
//
//
void VovPlace::pack(Bzocket& z)
{
    VovNode::pack( z );		
    name.pack( z );
    placeType.pack( z );
    timeStamp.pack( z );
    current.pack( z );
    host.pack( z );
    value.pack( z );
}

void VovPlace::unpack( Bzocket& z)
{
    VovNode::unpack( z );
    name.unpack( z );
    placeType.unpack( z );
    timeStamp.unpack( z );
    current.unpack( z );
    host.unpack( z );
    value.unpack( z );
}

//
//
//
void VovTransition::pack( Bzocket& z )
{
    VovNode::pack( z );
    capsule.pack( z );
    commandLine.pack( z );

    user.pack( z );
    host.pack( z );
    pid.pack( z );
    retracingInfo.pack( z );
    affinity.pack( z );
    cwd.pack( z );

    exitStatus.pack( z );
    okStatus.pack( z );
    started.pack( z );
    ended.pack( z );
}

void VovTransition::unpack( Bzocket& z )
{
    VovNode::unpack( z );
    capsule.unpack( z );
    commandLine.unpack( z );

    user.unpack( z );
    host.unpack( z );
    pid.unpack( z );
    retracingInfo.unpack( z );
    affinity.unpack( z );
    cwd.unpack( z );

    exitStatus.unpack( z );
    okStatus.unpack( z );
    started.unpack( z );
    ended.unpack( z );
}

#ifdef OBSOLETE
void VovGoal::pack( Bzocket& zoc )
{
    VovObject::pack( zoc );
    domain.pack( zoc );
    selectRule.pack( zoc );
    methodName.pack( zoc );
}
void VovGoal::unpack( Bzocket& zoc )
{
    VovObject::unpack( zoc );
    domain.unpack( zoc );
    selectRule.unpack( zoc );
    methodName.unpack( zoc );
}

void VovMeasure::pack( Bzocket& zoc )
{
    VovObject::pack( zoc );
    units.pack( zoc );
    valid.pack( zoc );
}
void VovMeasure:un:pack( Bzocket& zoc )
{
    VovObject::unpack( zoc );
    units.unpack( zoc );
    valid.unpack( zoc );
}
#endif

void VovRetrace::pack( Bzocket& zoc )
{
    VovNodeSet::pack( zoc );
    estimatedDuration.pack( zoc );
    actualDuration.pack( zoc );
    startDate.pack( zoc );
    status.pack( zoc );
}

void VovRetrace::unpack( Bzocket& zoc ) 
{
    VovNodeSet::unpack( zoc );
    estimatedDuration.unpack( zoc );
    actualDuration.unpack( zoc );
    startDate.unpack( zoc );
    status.unpack( zoc );
}

void VovUser::pack( Bzocket& zoc )
{
    VovObject::pack( zoc );
    lastCreation.pack( zoc );
    lastConnection.pack( zoc );
}
void VovUser::unpack( Bzocket& zoc )
{
    VovObject::unpack( zoc );
    lastCreation.unpack( zoc );
    lastConnection.unpack( zoc );
}

void VovAnnotation::pack( Bzocket& zoc )
{
    VovObject::pack( zoc );
    signature.pack( zoc );
    date.pack( zoc );
}
void VovAnnotation::unpack( Bzocket& zoc )
{
    VovObject::unpack( zoc );
    signature.unpack( zoc );
    date.unpack( zoc );
}


void VovSelectRule::pack( Bzocket& zoc )
{
    VovObject::pack( zoc );
    typemask.pack( zoc );
    type2mask.pack( zoc );
    statusmask.pack( zoc );
    rx.pack( zoc );
}

void VovSelectRule::unpack( Bzocket& zoc )
{
    VovObject::unpack( zoc );
    typemask.unpack( zoc );
    type2mask.unpack( zoc );
    statusmask.unpack( zoc );
    rx.unpack( zoc );
}



