#include "String.h"
#include "port.h"
#include "vov.h"
#include "node.h"
#include "base.h"
#include "voverror.h"
extern "C"
{
#include "rpc.h"
#include "oh.h"
}

VovStatus popBag( lsList cmdList, octObject* bag )
{
    if ( lsLength( cmdList ) < 1 )  {
	return VOV_ERROR;
    }
    RPCArg* arg;
    lsFirstItem( cmdList, &(Pointer)arg, 0 );
    if ( arg->argType != VEM_OBJ_ARG ) {
	return VOV_ERROR;
    }
    ohGetById( bag, arg->argData.objArg.theBag);
    return VOV_OK;
}


static vemSelSet selectSet = 0;
static VovBag argBag("VOV RPC ARG BAG");

void vrpc_initSelSet( octObject& facet )
{
   if ( selectSet == 0 ) {
	selectSet = vemNewSelSet( facet.objectId, 60000, 60000, 10000, 1, 1, 4, 4,
				 "0001 0110 0110 1000" );
	if ( selectSet == 0 ) {
	    vovError( "Cannot create selected set" );
	    return ;
	}
	
	octCreate( &facet, &argBag );
	octMarkTemporary( &argBag );
    }
}

void vrpc_addSelSet( VovObject& obj )
{
    if ( selectSet != 0 ) {
	vemAddSelSet( selectSet, obj.id() );
	argBag.add( obj );
    }
}


void vrpc_delSelSet( VovObject& obj )
{
    if ( selectSet != 0 ) {
	vemDelSelSet( selectSet, obj.id() );
	argBag.del( obj );
    }
}


int vrpc_isInSelSet( VovObject& obj )
{
    if ( selectSet != 0 ) {
	return argBag.contains( obj );
    } else { 
	return 0;
    }
}

void vrpc_clearSelSet( )
{
    if ( selectSet != 0 ) {
	vemClearSelSet( selectSet );
	argBag.clear( OCT_ALL_MASK );
    }
}

static octGenerator genArgs;
static int genArgsInit = 0;

void vrpc_initGenSelSet()
{
    if ( genArgsInit ) {
	octFreeGenerator( &genArgs );
	genArgsInit = 0;
    }
    if ( selectSet != 0 ) {
	OH_ASSERT( octInitGenContents( &argBag, OCT_ALL_MASK, &genArgs ) );
	genArgsInit = 1;
    }
}

VovStatus vrpc_genSelSet( VovNode& node )
{
    if ( ! genArgsInit ) {
	vrpc_initGenSelSet();
    }
    if ( octGenerate( &genArgs, &node ) == VOV_OK ) {
	if ( node.get() != VOV_OK ) {
	    vovError("Object is not a node" );
	}
	return VOV_OK;
    } else {
	genArgsInit = 0;
	octFreeGenerator( &genArgs );
	return VOV_ERROR;
    }
}

