#include "String.h"
#include "port.h"
#include "ddsplus.h"			// This should also eliminate including of port.h


#include "vov.h"
#include "trace.h"
#include "transition.h"
#include "place.h"
#include "user.h"
// #include "goal.h"
#include "voverror.h"
#include "version.h"

extern octStatus vovUnpackFacetName( VovObject* facet, char*  name);
extern void vovUnpackDefaults( VovObject* f, char* m, char* s);

extern "C"
{
#include "rpc.h"
#include "oh.h"

extern void  VRPC_initX();

extern int    VRPCexit();
extern int    VRPCviewPlace();
extern int    VRPCchangeStatus();
extern int    VRPCnicePlacement();
extern int    VRPCgetInfo();
extern int    VRPCinput();
extern int    VRPCoutput();
extern int    VRPCdisconnect();
extern int    VRPCforget();
extern int    VRPCselect();
extern int    VRPCfind();
}

extern void vovEditNodeStatus( VovTrace& , VovNode& );



#include "parse.h"


static VovTrace* rpcTraceP = 0;
void      vovSetRpcTrace( VovTrace& trace )
{
    rpcTraceP = &trace;
}
VovTrace& vovRpcTrace()
{
    return *rpcTraceP;
}


RPCFunction CommandArray[] = {
    { VRPCselect, 		"VOV", "getNode", "G", 0},
    { VRPCviewPlace, 		"VOV", "view", "v", 0},
    { VRPCchangeStatus,		"VOV", "status", 0, 0},
    { VRPCnicePlacement, 	"VOV", "beautify", 0, 0},
    { VRPCgetInfo, 		"VOV", "info", 0, 0},
    { VRPCforget, 		"VOV", "forgetNode", "", 0},
    { VRPCfind,			"VOV", "findNode",  "" , 0},

    { VRPCinput, 		"CONNECT", "input", 0, 0},
    { VRPCoutput, 		"CONNECT", "output", 0, 0},
    { VRPCdisconnect, 		"CONNECT", "disconnect", 0, 0},

    { VRPCexit,     		"EXIT", "exit", 0, 0}
};


void VRPCecho( const String& s )
{
    VEMMSG( (char*)s );
}

void VRPCfinish( RPCSpot* spot,const String& s )
{
    VovObject facet(OCT_FACET);
    ohGetById( &facet, spot->facet );
    VovObject box( OCT_BOX );
    if ( octBB( &facet, &box.contents.box ) != OCT_OK ) {
	VEMMSG("cannot get facet bb" );
    }

    // Now RPC should do automatic refreshing.
    // lsList cmdList = lsCreate();
    // vemCommand( "redraw-window", spot, cmdList, 0 );
    vovMessage( form( "vovRpc: %s, done.\n" , (char*)s ) );
    vemPrompt();
}


/*ARGSUSED*/
long UserMain(char* display, RPCSpot* spot,lsList cmdList,long  uow,RPCFunction* array[])
{
    *array = CommandArray;			// Required for RPC.
    vovSetErrorOutputFunction( VRPCecho );	// Send msgs to VEM console.

    vovMessage( "Starting vov_rpc: please wait until I setup ..." );
    util_setenv( "DISPLAY", display );

    VRPC_initX( display );
    vovPartiallyInitializedX();


    VovObject facet(OCT_FACET);
    ohGetById( &facet, spot->facet );
    VovTrace& trace = *(new VovTrace);
    vovSetRpcTrace( trace );
    trace.open( facet );

    VEMMSG( form( "%s : ready.\n", VOV_VERSION ));
    vemPrompt();

    vrpc_initSelSet( facet );
    return(sizeof(CommandArray) / sizeof(RPCFunction));
}

/*ARGSUSED*/
int VRPCexit(RPCSpot* spot, lsList  cmdList, long  uow)
{
    vovMessage( "VOV RPC: done.");
    RPCExit(RPC_OK);
}


int isAttachedToBag( octObject& i , char* bagname )
{
    octObject bag, facet;
    octObject sets;
    octGetFacet( &i, &facet ) ;		
    OH_ASSERT( ohGetByBagName( &facet, &sets, "SETS" ) );
    
    
    if (ohGetByBagName( &facet, &bag, bagname ) != OCT_OK ) {
	if (ohGetByBagName( &sets, &bag, bagname ) != OCT_OK ) {
	    return 0;
	}
    }
    return ( octIsAttached( &bag, &i ) == OCT_OK ) ? 1 : 0 ;
}


int VRPCnicePlacement(RPCSpot* spot, lsList  cmdList, long  uow)
{
    vovMessage( "Improving placement of trace\n\t(expensive operation, please wait)." );
    vovRpcTrace().doPlacement();
    
    VRPCfinish( spot, "Beautify" );
    return RPC_OK;
}

int VRPCgetInfo(RPCSpot* spot, lsList  cmdList, long  uow)
{
    String f = vovRpcTrace().format();

    vovShortMessage( "TRACE INFORMATION", f );
    VRPCfinish( spot, "info" );
    return RPC_OK;
}

int VRPCselect(RPCSpot* spot, lsList  cmdList, long uow)
{
    VovObject inst( OCT_INSTANCE );
    vemStatus s = vuFindSpot( spot, &inst, OCT_INSTANCE_MASK );
    if ( s == VEM_CANTFIND || s == VEM_NOSELECT ) {
	vovError("nothing found under cursor");
    } else {
	if ( vrpc_isInSelSet( inst ) ) {
	    vrpc_delSelSet( inst );
	} else {
	    vrpc_addSelSet( inst );
	}
    }
    return RPC_OK;
}

int VRPCdisconnect(RPCSpot* spot, lsList  cmdList, long uow)
{
    int nodecount = 0;
    
    vrpc_initGenSelSet();
    VovNode node[2];
    if ( vrpc_genSelSet( node[0] ) == VOV_OK ) {
	nodecount = 1;
	if ( vrpc_genSelSet( node[1] ) == VOV_OK ) {
	    nodecount = 2;
	}
    }    
    if ( nodecount == 1 ) {
	vovMessage( "disconnect node" );
	node[0].disconnect(VOV_IOS_MASK);
    } else {
	if ( node[0].isInput( node[1]) == VOV_OK )  {
	    vovMessage("input");
	    node[0].disconnectInput( node[1] );
	} else if ( node[0].isOutput( node[1] ) == VOV_OK ) {
	    vovMessage("output");
	    node[0].disconnectOutput( node[1] );
	} else {
	    vovError( "The two nodes are not connected" );
	}
    }

    VRPCfinish( spot, "disconnect" );
    return RPC_OK;
}

int VRPCio(RPCSpot* spot, lsList  cmdList, int inputFlag)
{
    
    vrpc_initGenSelSet();
    VovNode node[2];
    if ( vrpc_genSelSet( node[0] )!=VOV_OK || vrpc_genSelSet( node[1] ) != VOV_OK ) {
	vovError( "Need two nodes to make a connection" );
	return RPC_OK;
    }

    int t_index = 0;
    int p_index = 1;
    if ( node[0].nodeType() == VOV_TRANSITION && node[1].nodeType() == VOV_PLACE ) {
	// vovMessage( "indices are OK" );
    } else if ( node[1].nodeType() == VOV_TRANSITION && node[0].nodeType() == VOV_PLACE ) {
	// vovMessage( "inverting indices" );
	t_index = 1;
	p_index = 0;
    } else {
	vovError("The two nodes must be of different types" );
	return RPC_OK;
    }

    VovTransition tran = node[t_index];
    VovPlace place = node[p_index];

    if ( inputFlag ) {
	if ( vovRpcTrace().declareInput( tran, place,0 ) != VOV_OK ) {
	    vovError("Cannot connect" );
	} 
    } else {
	String errmsg;
	if ( vovRpcTrace().declareOutput( tran, place, 1, 0, errmsg) != VOV_OK ) {
	    vovError("Cannot connect" + errmsg );
	} 
    }
    VRPCfinish( spot, "connect" );
    return RPC_OK;
}

int VRPCinput(RPCSpot* spot, lsList  cmdList, long  uow)
{
    return VRPCio( spot, cmdList, 1 );
}

int VRPCoutput(RPCSpot* spot, lsList  cmdList, long  uow)
{
    return VRPCio( spot, cmdList, 0 );
}

int VRPCforget(RPCSpot* spot, lsList  cmdList, long  uow)
{
    VovNode node;
    vrpc_initGenSelSet();
    while ( vrpc_genSelSet( node ) == VOV_OK ) {
	if ( vovRpcTrace().destroy( node ) == VOV_ERROR ) {
	    vovError("cannot forget node" );
	}
    }
    VRPCfinish( spot, "forgetNode" );
    return RPC_OK;
}



int VRPCviewPlace(RPCSpot* spot, lsList  cmdList, long  uow)
{
    octObject inst;
    
    vemStatus s = vuFindSpot( spot, &inst, OCT_INSTANCE_MASK );
    if ( s == VEM_CANTFIND || s == VEM_NOSELECT ) {
	VEMMSG("vovRpc:nothing found under cursor\n");
	return RPC_OK;
    }
    
    if ( isAttachedToBag( inst, ";;;TRANSITIONS;;;" ) ) {
	{
	    VovTransition t = inst.objectId;
	    vovRpcTrace().edit( t, 1 );
	}
	return RPC_OK;
    } else if ( isAttachedToBag( inst, "USERS" ) ) {
	{
	    VovUser user = inst.objectId;
	    vovShortMessage( "USER", (char*)user.format() );
	}
	return RPC_OK;
    } else if ( isAttachedToBag( inst, "GOALS" ) ) {
	{
	    // VovGoal goal;
	    // goal.objectId = inst.objectId;
	    // goal.getById( VOV_GOAL_OCTMASK );
	    // goal.get();
	    // vovShortMessage( "USER", (char*)goal.format() );
	}
	return RPC_OK;
    } else if ( isAttachedToBag( inst, ";;;PLACES;;;" ) ) {
	{
	    VovPlace place = inst.objectId;
	    
	    switch ( place.getPlaceType() ) {
	    case VOV_PROGRAM_OPTION:
	    case VOV_EXECUTABLE:
	    case VOV_EXIT_STATUS:
	    case VOV_BOOLEAN:
	    case VOV_UNIX_FILE:
	    case VOV_MEASURE:
		vovRpcTrace().edit( place , 1 );
		break;
	    case VOV_OCT_FACET:
		VEMMSG("vovRpc:opening OCT facet");
		{
		    VovObject facet(OCT_FACET);
		    vovUnpackDefaults( &facet, "r" , ":symbolic:contents" );
		    OH_ASSERT( vovUnpackFacetName( &facet, (char*)place.getName()) );
		    vemOpenWindow( &facet , 0 );
		}
		break;
	    default:
		
		break;
	    }
	}
    } else {
	VEMMSG( "Cannot determine type of object under mouse\n");
    }
    VRPCfinish( spot, "view" );
    return RPC_OK;
}

int VRPCchangeStatus(RPCSpot* spot, lsList  cmdList, long  uow)
{
    octObject inst;
    
    vemStatus s = vuFindSpot( spot, &inst, OCT_INSTANCE_MASK );
    if ( s == VEM_CANTFIND || s == VEM_NOSELECT ) {
	VEMMSG("vovRpc:nothing found under cursor\n");
	return RPC_OK;
    }
    
    if ( ! isAttachedToBag( inst, ";;;PLACES;;;" ) ) {
	vovMessage("vovRpc: not a place" );
	return RPC_OK;
    }
    VovPlace place = inst.objectId;
    
    vovEditNodeStatus( vovRpcTrace(), place );
 
    VRPCfinish( spot, "changeStatus" );
    return RPC_OK;
}


int VRPCfind(RPCSpot* spot, lsList  cmdList, long  uow)
{
    String name;
    if ( lsLength( cmdList ) == 1 ) {
	RPCArg* arg ;
	lsFirstItem( cmdList, (Pointer*)&arg, (lsHandle*)0 );
	if ( arg->argType == VEM_TEXT_ARG ) {
	    name = arg->argData.string;
	}
    }
    if ( name == 0 ) {
	VEMMSG( "ERROR: \"regex\" : findNode\n" );
	vemPrompt();
	return RPC_OK;
    }
    int found = 0;
    for (VovPlaceType t = VOV_FIRST_PLACE_TYPE + 1; !found &&  t != VOV_LAST_PLACE_TYPE; t++ ) {
	VovPlace place( t, name  );
	if ( vovRpcTrace().find( place ) == VOV_OK ) {
	    VEMMSG( (char*)place.format() );
	    vrpc_addSelSet( place );
	    found = 1;
	}
    }
    if ( ! found ) {
	String list;
	int count = vovRpcTrace().list( VOV_PLACE, name , list );
	if ( count == 0 ) {
	    VEMMSG( "Not found" );
	} else {
	    VEMMSG( "Showing list" );
	    vovShortMessage( "PLACES WITH SIMILAR NAME", list );
	}
    }
    VRPCfinish( spot, "findNode" );
    return RPC_OK;
}


