/* C++ includes */
#include <stream.h>

/* C includes */

/* Local includes */
#include "VOVprotocol.h"
#include "clients.h"
#include "slaveset.h"
#include "trace.h"
// #include "goal.h"
#include "place.h"
#include "transition.h"
#include "user.h"
#include "annotation.h"
#include "event.h"
#include "dispatch.h"
#include "voverror.h"
#include "comm.h"
#include "path.h"

extern "C"
{
#include <signal.h>
#include <setjmp.h>
}


extern int VS_getdtablesize();



extern int updateServerDelay( int, VovCode );
extern void VS_dispatch_job( VovTrace& trace, SlaveSet& slaveset, int mode ) ;

void VS_suddenClientDeath( Client& client, SlaveSet& slaveset, SlaveSet& notifyset,
			  VovTrace& trace, VovEventQueue& queue )
{
    if ( client.theTransition != 0) {
	VovTransition trans = client.theTransition;
	if ( trans.id() == client.theTransition ) {
	    if ( trans.getStatus() == VOV_TRACING_NODE || trans.getStatus() == VOV_RETRACING_NODE ) {
		vovMessage( "Sudden client death: terminating \n\t" + trans.getCommandLine() );
		trace.terminate( trans, VOV_ERROR, -1, 0 );
	    }
	} else {
	    vovError( "Bogus transition id in sudden client death" );
	}
	client.theTransition = 0;
    }
    if ( client.clientclass() & CLIENT_CLASS_SLAVE ) {
	slaveset.del( &client );
	queue.enq( VOV_SLAVE_MASK, VE_END, "A slave on " + client.host + " has died.", 0, 0 );
    } 
    if ( client.clientclass() & CLIENT_CLASS_NOTIFY ) {
	notifyset.del( &client );
	// queue.enq( VOV_SLAVE_MASK, VE_END, "A notify client has died" , 0 , 0 );
    }
}

#ifndef VOVCLIENT
static jmp_buf vov_dispatch_loop_env;
SIGNAL_FN handleBrokenPipe( int sig,  int  code, struct sigcontext* scp)
{
    vovError("Caught BROKEN PIPE signal" );
    longjmp( vov_dispatch_loop_env, 1 );
} 

void setDispatchSignalHandlers()
{
    struct sigvec sig, osig;
    sig.sv_handler = handleBrokenPipe;
    sig.sv_mask = 0;
    SIG_FLAGS(sig) = 0;
    sigvec( SIGPIPE, &sig, &osig );
}
#endif

//
// VS_dispatch()
// The main read/dispatch loop.
//
void VS_dispatch(VovTrace& trace)
{
#ifndef VOVCLIENT
    SlaveSet slaveset( VS_getdtablesize() ); 		// Set of all slaves.
    SlaveSet notifyset( VS_getdtablesize() );		// Set of all clients waiting to be notified.
    VovEventQueue latestqueue( trace, 400 );		// Queue of all latest events.
    int delay = updateServerDelay( 0 , 0 );		// Delay: timeout at select.
    
    trace.stopAllTransitions();			// Actually there should be no active transition at this point.
    
    setDispatchSignalHandlers();
    
    extern SlaveSet& serverSlaveSet( SlaveSet* );
    (void)serverSlaveSet( &slaveset ); 		// Register this slaveset.
    VS_InitReceive();				// Must be done before setjmp.
    
    
    if ( setjmp( vov_dispatch_loop_env ) != 0 ) {
	vovMessage( "Back to dispatch after long jump" );
    }
    
    //
    // Get a client that has a request.
    // If no client has a request, slow down activity.
    //
    while (1) {
	// 
	// Notify all clients about the events that have recently happened.
	// Also log all these events in the journal.
	//
	VS_notify( trace, notifyset );
	
	VovCode code = VOV_Ok;			// Some code.
	Client* c = VS_Receive( delay );	// Get a client.
	int timeOut = ( c == 0 );		// Is this a timeout ?
	if ( ! timeOut ) {			// Not a time out, get the code.
	    if ( ! c->active() ) {		// Client is no longer active, declare it dead.
		VS_suddenClientDeath( *c, slaveset, notifyset, trace, trace.getQueue() );
		continue;
	    }
	    if ( c->unpack( (int*)&code) == BUF_ERROR) {  // Make sure a code is coming in.
		c->print();				  // 
		vovError("No CODE from " + c->format() ); // 
		vovMessage( "Closing faulty client" );	  // 
		c->close();				  // 
		continue;				  // 
	    }
	}
	delay = updateServerDelay( timeOut , code );
	if ( timeOut ) {
	    VS_routine_tasks_timeout( trace, slaveset );
	    if ( delay > 32 ) {
		VS_routine_tasks_idle( trace, slaveset );
	    }
	    continue;
	}
	
	Client& client = *c;	// Just a renaming
	
	client.startService();
	octId id = oct_null_id;		// Returned by VovTrace::serverDispatch();
	if ( trace.serverDispatch( client, code, id ) == VOV_OK ) {
	    switch ( code ) {
	    case VOV_EvolveAll:		// In all these cases, start retracing as soon as possible,
	    case VOV_EvolveSet:		// rather than waiting for the next timeout.
	    case VOV_EvolveNode:
		VS_dispatch_job( trace, slaveset, VOV_DISPATCH_MODE_SLOW );
		break;
	    case VOV_StartTransition:
		// vovMessage( form( "setting client.transition id = %d ", id ) );
		client.theTransition = id;
		break;
	    case VOV_EndTransition:
		// If client is slave, make it ready for another job.
		if ( client.clientclass() & CLIENT_CLASS_SLAVE ) {
		    slaveset.slaveEnd( &client ); 
		    client.theTransition = 0;
		    VS_dispatch_job(trace,slaveset,VOV_DISPATCH_MODE_SLOW); // Get the slave immediately busy.
		}
		break;
	    case VOV_EvolveStop:
		// This clears some transitions which had been cashed in VS_dispatch_job.
		void extern VS_stop_retracing();
		VS_stop_retracing();
		break;
	    default:
		break;
	    }
	    client.endService();
	    continue;
	}
	// 
	// Not a service that the trace could provide directly.
	// 
	switch ( code ) {
#ifdef OLDCODE
	case (VovCode)SWAP_CODE:
	    client.setSwapMask( client.unpackInt() ); // 
	    client.pack( SWAP_CODE );		      // Send reply.
	    client.pack( SWAP_TEST_VECTOR );	      // 
	    client.flush();
	    break;
#endif
	case VOV_StartUp:
	    {
		String version = client.unpackString();
		client.user = client.unpackString( );
		client.host = client.unpackString( );
		client.pid  = client.unpackInt();
		client.nesn = vov_nextEventSerialNumber();
		if ( version != VOV_PROTOCOL_VERSION ) {
		    String msg = form( "Wrong VOV protocol version:\n\tclient: %s\n\tserver: %s\n",
				      (char*)version , VOV_PROTOCOL_VERSION );
		    msg += "Please check that you have the latest version of the software.";
		    VS_oops( client, msg );
		} else {
		    client.pack( VOV_StartUp );
		    client.pack( trace.getProjectName() );
		    client.flush();
		}
		VovUser user( client.user );
		if ( trace.find( user ) == VOV_OK ) {
		    user.connected();
		}
	    }
	    break;
	case VOV_SynchClocks:
	    {
		// Two steps.
		int step = client.unpackInt();
		switch (step) {
		case 1:
		    client.pack( VOV_SynchClocks );
		    client.pack( getDate() );
		    client.flush();
		    break;
		case 2:
		    {
			int delta = client.unpackInt( );
			int uncertainty = client.unpackInt( );
			Date repOfClassDate;
			repOfClassDate.updateHostDelta( client.host, delta );
		    }
		    break;
		case 3:		// Force update of a particular host.
		    {
			Date d;
			String h = client.unpackString();
			int delta = client.unpackInt();
			vovMessage("Forcing update of " + h );
			d.updateHostDelta( h, delta );
			vovMessage( d.mapFormat() );
		    }
		    break;
		default:
		    vovError("Protocol error in SynchClocks");
		}
		
	    }
	    break;
	case VOV_Debug:
	    {
		int flag = client.unpackInt();
		vovMessage( form("Turning debugging %s", flag ? "ON":"OFF" ) );
		client.setDebug( flag );
	    }
	    break;
	case VOV_SlaveReady:
	    {
		int power = client.unpackInt(); 		// Estimated power of this slave.
		String affinity = client.unpackString();	// Get type of slave.
		client.clientclass() |= CLIENT_CLASS_SLAVE;	// Mark client as a slave.
		slaveset.add( &client, power, affinity);	// Update power of slave.
	    }
	    break;
	case VOV_CannotExec:	// A slave is reporting on executing a transaction.
	    {
		VovTransition theTransition;
		theTransition.id() = client.theTransition;
		theTransition.get();
		String errmsg = client.unpackString();
		trace.getQueue().enq( VOV_TRANSITION_MASK, VE_ERROR, 
				     form("Slave on %s cannot execute transition:\n%s\n%s\n",
					  (char*)client.host, 
					  (char*)theTransition.getCommandLine(),
					  (char*)errmsg
					  ),  theTransition.id(), 0 );
		client.theTransition =  0; 
		// If client is slave, make it ready for another job.
		slaveset.slaveEnd( &client ); 
		VovTransitionTerminationCode code;
		theTransition.terminate( trace.getQueue(), VOV_ERROR, 0, errmsg, code);
		VOV_ASSERT( trace.stopRetracing( theTransition , errmsg ) );
	    }
	    break;
	case VOV_SynchNotify:
	    { 
		int max = client.unpackInt();
		int unreported = client.unpackInt(); // Only new events, or also old ones?
		
		client.pack( VOV_SynchNotify );
		
		VovEvent event;				// Generic event.
		VovEventQueue& unreportedqueue = trace.getUnreportedQueue();
		if ( unreported ) { 
		    int ul = unreportedqueue.length();
		    if ( ul > max ) ul = max;
		    client.pack( ul );
		    while ( ul-- > 0 ) {
			unreportedqueue.deq( event );
			latestqueue.enq( event );
			event.pack( client );
		    }
		} else {
		    int nl = latestqueue.length();
		    if ( nl > max ) nl = max;
		    client.pack( nl );
		    while ( nl > 0 ) {
			if ( latestqueue.get( event, --nl ) == VOV_ERROR ) {
			    vovError( "This should never happen. I was promised this event." );
			}
			event.pack( client );
		    }
		}
		client.flush();
	    }
	    break;
	case VOV_AsynchNotify:
	    { 
		int begin = client.unpackInt();
		if ( begin ) {
		    client.clientclass() |= CLIENT_CLASS_NOTIFY;
		    notifyset.add( &client, 0, 0 );
		    VovEvent event;
		    VovEventQueue& queue = trace.getUnreportedQueue();
		    int offset = 0;
		    while ( 1 ) {
			if ( queue.get( event, offset ) != VOV_OK ) break;
			if ( event.getSerialNumber() >=  client.nesn  ) {
			    offset++; // Go back another event.
			} else {
			    break;
			}
		    } 
		    if ( offset ) {
			// vovMessage( form( "%d events to be notified immediately", offset ) );
			while ( offset-- > 0 ) {
			    VOV_ASSERT( queue.get( event, offset ) );
			    client.pack( VOV_Notify );
			    event.pack( client );
			    client.flush();
			}
			client.nesn = event.getSerialNumber();
		    }
		} else {
		    client.clientclass() &= ~CLIENT_CLASS_NOTIFY;
		    notifyset.del( &client );
		}
	    }
	    break;
	case VOV_EchoLog:
	    {
		String s = client.unpackString();
		vovMessage( s );
		VS_ok( client );
	    }
	    break;
	case VOV_KillServer:
	    VS_ok( client );
	    String msg = "Kill request by " + client.signature() + "."; // 
	    trace.journalLog( msg );				  // 
	    vovMessage( msg );					  // 
	    trace.checkPlaces( 0 );				  // Check all places before exiting.
	    VS_notify( trace, notifyset );
	    trace.stopAllTransitions();				  // Really, there should be no active transition at this point.
	    trace.killSlave();
	    return;
	    break;
	default:
	    vovError( form( "Unknown code %d", code ) );
	    if ( client.discardPacket() != BUF_OK ) {
		vovFatalError("Cannot discard packet");
	    }
	    break;
	}
	client.endService();
	VS_routine_tasks_always( trace, slaveset );
    }
#endif
}



