#include "trace.h"
#include "VOVprotocol.h"
#include "clients.h"
#include "slaveset.h"
#include "user.h"
#include "voverror.h"

VovStatus vov_dispatchJob( VovTrace& trace, VovTransition& trans, SlaveSet& slaveset )
{
    Client* c;
    VovStatus s = slaveset.bestSlave( &c, trans.getAffinity() );
    if ( s != VOV_OK ) {
	if ( s == VOV_ERROR ) {
	    String msg = trans.getCommandLine()+"\n\tCannot be retraced.\n";
	    msg += form( "\tNo affine slave: transition affinity is `%s'.\n",
			(char*)trans.getAffinity() );
	    trace.getQueue().enq(VOV_TRANSITION_MASK,VE_ERROR,msg, trans.id(), 0 );
	    VOV_ASSERT( trace.stopRetracing( trans, msg ) );
	} else {
	    // All slaves are still busy.
	}
	return s;	     // vovMessage("No best slave");
    }
    
    slaveset.slaveBegin( c, trans.duration() );
    
    Client& slave = *c;
 
    trans.setHost( slave.host );
    trans.setUser( "<vov_server>" );
    VOV_ASSERT( trace.dispatchTransition( slave, trans ) );

    String msg = slave.host + "<-- " + trans.getCommandLine() + "\n";
    trace.getQueue().enq(VOV_TRANSITION_MASK,VE_START, msg, trans.id(), 0 );

    VovUser user( "<vov_server>" );
    VOV_ASSERT( trace.findOrCreate( user ) );
    VOV_ASSERT( trace.activeDesigner( user, trans ) );

    slave.theTransition = trans.id();
    return VOV_OK;
}



