#include "clients.h"
#include "slaveset.h"
#include "VOVprotocol.h"
#include "vov.h"
#include "event.h"
#include "voverror.h"

void VS_notify_clients( SlaveSet& set, VovEvent& event )
{
    for ( int i = set.first() ; i != -1 ;  set.next( i ) ) {
	Client* c = set(i);
	c->pack( VOV_Notify );
	event.pack( *c );
	c->flush();
	c->nesn = event.getSerialNumber();
    }
}

// Used for asyncronous notification.
VovStatus VS_notify( VovTrace& trace, SlaveSet& notifySet )
{
    VovEvent event;
    VovEventQueue&  workQueue = trace.getQueue();
    VovEventQueue&  unreportedQueue = trace.getUnreportedQueue();
    
    while ( workQueue.length() > 0 ) {
	VOV_ASSERT( workQueue.deq( event ) );
	if ( event.getSubjectMask() == VOV_NULL_MASK ) {
	    continue;					// vovError("null event in queue");
	}
	if ( notifySet.length() > 0 ) {			// Send events to the clients if there is any.
	    VS_notify_clients( notifySet, event );
	}
	if ( unreportedQueue.enq( event ) != VOV_OK ) {
	    vovError("Queue overfilled");
	}
	trace.journalLog( event.format() ) ;
    }	
    return VOV_OK;
}

