#include <stream.h>
#include "priority.h"
#include <builtin.h>
#include "utcc.h"

SortedArray::SortedArray( int s, int flag = 1 )
{
    leastFirst = flag;
    size = s;
    if (size > 0) {
	array = new int[size];
	valueArray = new double[size];
    }
    clear();
}

SortedArray::~SortedArray()
{
    if (size > 0) {
	delete array;
	delete valueArray;
    }
}

void SortedArray::clear()
{
    entries = 0;
    index = 0;
}

int SortedArray::next( )
{
    int ret = 0;

    if ( index < entries ) {
	ret = array[index];
    } 
    index++;
    if ( index >= entries ) {
	clear();
    }
    return ret;
}

int SortedArray::length()
{
    return entries ? entries - index : 0; 
}

void SortedArray::undo_next( )
{
    if ( --index < 0 ) index = 0;
}

int SortedArray::get( int i)
{
    if ( i < entries && i >= 0) {
	return array[i];
    }
    return 0;
}

// Shift down existing entries and insert new value.
void SortedArray::insert( int obj, double value , int i)
{
    if ( entries == size ) entries--; 	// lose the last entry if array is full
    int last = entries - 1;
    for ( int j = last ; j >= i ; j-- ) {
	valueArray[j+1] = valueArray[j];
	array[j+1] = array[j];
    }
    array[i] = obj;
    valueArray[i] = value;
    entries++;

    // Double chack that the insertion is correct.
    // This is just for debugging purposes.
    if ( i > 0 ) {
	int s = sign( valueArray[i] - valueArray[i-1] );
	if ( leastFirst && s < 0 || (! leastFirst) && s > 0 ) {
	    utcc_error(form("In priority:1: i = %d, entries = %d, sign(%f-%f)= %d", 
			    i, entries, array[i], array[i-1],s));
	    print();		// debugging
	} 
    }
    if ( i < entries - 1 ) {
	int s = sign( valueArray[i+1] - valueArray[i] );
	if ( leastFirst && s < 0 || ! leastFirst && s > 0 ) {
	    utcc_error("In priority:2");
	    print();		// debugging
	} 
    }
}

void SortedArray::add( int obj, double value )
{
    if (entries == 0) {
	insert( obj, value, 0 );
	return;
    }
    
    // Linear search for the right spot to enter this object.
    int i = 0;
    if ( leastFirst ) {
	while ( i < entries && valueArray[i] <= value ) i++;
    } else {
	while ( i < entries && valueArray[i] >= value ) i++;
    }
    insert( obj, value, i );		// Insert element at index i.
    return;
}

void SortedArray::print()
{
    cerr << "Sorted array contents: Mode is ";
    cerr << ( leastFirst ? "LEAST FIRST\n" : "MOST FIRST\n" );
    if (entries == 0) {
	cerr << "No entries in the array\n";
    } else {
	cerr << "Entries in the array\n";
	for (int i = 0 ; i < entries ; i++ ) {
	    cerr.form("%3d: ptr: %12d: value %7.3f %s\n", 
		      i, array[i], valueArray[i], (index==i)?"***":"");
	}
    }
}
