// server.cc
//
/* C++ includes */
#include "port_internal.h"


/* C includes */
extern "C"
{
#include <sys/time.h>
#include <options.h>
#include <utility.h>
#include <errtrap.h>
#include <setjmp.h>
}
/* Local includes */
#include "trace.h"
// #include "goal.h"
#include "dialog.h"
#include "utcc.h"
#include "voverror.h"
#include "version.h"
#include "ddsplus.h"

extern int  net_portNumber( String& );
extern void net_Initialize( int  );
extern void VS_dispatch(VovTrace&);
extern void net_HangUp();
extern void net_Close();

optionStruct optionList[] = {
    { "C", 0,		"Clear trace, Start from scratch"},
    { "S", 0,		"Start slaves automatically" },
    { "B", 0,		"Run in Batch mode" },
    { OPT_RARG ,        "<project_name>", "" },
    { 0, 0, 0 }
};

void VS_lib_error_handler( const char* k, const char* t )
{
    String liberror = form( "%s:%s", k, t );
    vovError("G++ library error: " + liberror );
    errRaise( "G++lib", -1, "%s:%s", k, t );
}

static int batch_mode = 0;	// 
static jmp_buf  vov_server_jmp_env;

void vov_server_error_handler( char* pkgName, int code, char* message )
{
    static int count_errors = 0;

    vovError( form("Fatal Error detected number %d by %s (code %d)\n\t%s",
		   ++count_errors, pkgName, code,  message ) );

    vovMessage( "Closing all active connections" );
    net_HangUp();

    if ( count_errors > 100 ) {
	vovMessage( "Too many fatal errors. This may be an infinite loop" );
	vovMessage( "User attention is requested" );
	batch_mode = 0;
	count_errors = 0;	// Reset
    }
    if ( !batch_mode && userDecision( "VOV SERVER: A serious error occurred.\n\n", "Abort", "Try to continue" ) == 1  ) {
	net_Close();
	abort() ; 
    } else {
	vovMessage( "Attempting to restart dispatch loop." );
	longjmp( vov_server_jmp_env, 1  );
    }
    
}

//
// Start slaves if possible.
//
void VS_start_slaves( String& host, String& project )
{
    String command = "vov_start_slaves " + host + " " + project + " &";
    // vovMessage("Automatically starting slaves: " + command );
    if ( util_csystem( (char*)command ) == 127 ) {
	vovError("Cannot start slaves") ;
    }
}


#ifdef OBSOLETE
void VS_default_goals( VovTrace& trace ) 
{
    //Design goals: this is still very experimental.
    VovSelectRule rule(VOV_PLACE_MASK, VOV_ALL_PLACE_TYPE_MASK,
		       VOV_ALL_STATUS_MASK, SELECT_ALL );
    VovGoal diskUsageGoal( rule , "DISK_USAGE");
    
    trace.findOrCreate( diskUsageGoal );
}
#endif

main(int argc,  char* argv[])
{
    extern int optind;		// Declared in the option package.
    String mode = "a";		// The trace is opened for "append".
    int option;			// The command line option.
    int startSlaves = 0;	// 
    errCore( 1 );		// Dump core on error.
    
    vovRegisterMainArgs( argc, argv );
    vovMessage( String("VOV server ") + VOV_VERSION );
    while (( option = optGetOption( argc, argv ) ) != EOF) {
	switch (option) {
	case 'S':
	    startSlaves = 1;
	    break;
	case 'B':
	    batch_mode =1;
	    break;
	case 'C':
	    String q = form( "Do you really want to start the trace %s from scratch?",
			    argv[argc-1] );
	    String o1 = "Yes, and forget about everything done so far";
	    String o2 = "Oh, my God, definitely not";
	    if ( ttyUserDecision( q, o1, o2 ) != 1 ) {
		vovMessage( "OK, Try again." );
		exit( -1 );
	    }
	    mode = "w";		// Overwrite the trace. 
	    break;
	default:
	    break;
	}
    }
    
    String host = utcc_gethostname();
    String project = argv[optind];
    if ( project == 0 ) {
	vovError( "You must specify the project name" );
	optUsage();
    }
    
    VovTrace trace( project, host, mode, VOV_TRACE_TYPE_SERVER );
    if ( trace.disabled() ) {
	vovError( "The VOV server cannot be started." );
	exit( 1 );
    }
    trace.journalLog( String("VOV server restarted: ") + VOV_VERSION );
    if ( mode == "w" ) {
	trace.journalLog( "Starting a new design trace" );
	trace.startProject( );
    } else {
	trace.journalLog( "Continuing an existing design trace" ); 
	vovMessage( "Project was started on: " + trace.getProjectStartDate().format() );
    }

    String msg;
    trace.OK( msg );			// Check integrity of trace.

    net_Initialize ( net_portNumber( project ) );
    
    vovMessage( "Server is ready." );
    
    if (startSlaves) VS_start_slaves( host, project );

    set_lib_error_handler( VS_lib_error_handler );

    // VS_default_goals( trace );

    errPushHandler( vov_server_error_handler );
    if ( setjmp( vov_server_jmp_env ) != 0 ) {
	vovMessage( "JUMP!! Restarting dispatch loop." );
	if ( batch_mode || userDecision("Do you want to save the trace now?" , "Yes", "No" ) == 1 ) {
	    trace.saveOntoDisk();		// Save the trace onto disk.
	}

    }
    VS_dispatch ( trace );		// The main server loop.
    
    trace.saveOntoDisk();		// Save the trace onto disk.
    vovMessage( "Server is now exiting." );
    
    net_Close();
    exit( 0 );
}










