#include <stream.h>
#include <String.h>

#include "voverror.h"
#include "utcc.h"
#include "trace.h"
#include "place.h"
#include "retrace.h"
#include "transition.h"
#include "annotation.h"
#include "vovlib.h"
#include "ddsplus.h"
#include "path.h"

extern VovPlaceType guessType( String& name );
extern void vovEditNodeStatus( VovTrace& , VovNode& );
extern VovStatus vovNavigate( VovTrace& trace, VovNode& n, int direction );


//
//  


class StringRing {
    String* array;
    int    length;
    int    index;
public:
    StringRing( int l );
    ~StringRing();
    void add( String& s );
    String getlast();
    String format();
};

StringRing::StringRing( int l )
{
    length = l;
    array = new String[length];
    index = 0;
}

StringRing::~StringRing()
{
    delete array;
}

void StringRing::add( String&  s  )
{
    String ss = s;
    String sub = path_canonicalize( utcc_getwd() ) +"/" ;
    ss.gsub( sub, String("") );
    if ( ss  != array[index] ) {     // Avoid some useless duplicates.
	index++;
	index %= length;
	array[index] = ss;
    }
}

String StringRing::getlast()
{
    return array[index];
}

String StringRing::format()
{
    String s;
    for ( int i = 0; i < length ; i++ ) {
	int j = ( length + index - i ) % length ;
	if ( array[j] != 0 ) {
	    s += array[ j ] + "\n";
	}
    }
    return s;
}



// ----------------------------------------------------
// ----------------------------------------------------
// browser, history, retrace, etc.
// ----------------------------------------------------
// ----------------------------------------------------
// ----------------------------------------------------

typedef struct {
    VovTrace* tracePtr;
    StringRing* historyNamesRingPtr;
    ddsHandle namestarDlgHandle;
    ddsHandle nameDlgHandle;
    ddsHandle recentDlgHandle;
    int hist_level;
    int hist_flags;
    int hist_direction;
} VovLocalBrowseData;

void vov_browse_update( VovLocalBrowseData& data )
{
    VovDialogEditText namestarDlg( data.namestarDlgHandle );
    VovDialogEditText nameDlg( data.nameDlgHandle );
    VovDialogEditText recentDlg( data.recentDlgHandle );
    namestarDlg.modifyText("");
    nameDlg.modifyText( data.historyNamesRingPtr->getlast() );
    recentDlg.modifyText( data.historyNamesRingPtr->format() );
}

void vov_browse_ccb( ddsHandle item )
{
    VovDialogControl button( item );
    VovLocalBrowseData& data = *(VovLocalBrowseData*)button.getData();
    StringRing& historyNamesRing = *data.historyNamesRingPtr;
    
    VovTrace& trace = *data.tracePtr;
    VovDialogEditText namestarDlg (  data.namestarDlgHandle );
    VovDialogEditText nameDlg ( data.nameDlgHandle );

    String n = namestarDlg.getTextItem( );
    if ( n == 0 ) {
	n = nameDlg.getTextItem( );
	if ( n == 0 ) {
	    vovConfirm( "ATTENTION PLEASE" , "Name should not be empty." , "Continue", 0 );
	    return;
	}
    }
    // Only the first word is used as the name of the place.
    if ( n.contains( RXwhite ) ) {
	extern String vov_normalizeString( const String&  );
	n = vov_normalizeString( n );
	n = n.before( RXwhite );
    }

    historyNamesRing.add( n );
    vov_browse_update( data );


    VovPlace thePlace;
    VovTransition theTransition;
    VovNode theNode;
	
    String hist_direction_string = ( data.hist_direction == VOV_FORWARD)?"FORWARD":"BACKWARD" ;


    int found = 0;
    String name =  n;
    for (VovPlaceType t = VOV_FIRST_PLACE_TYPE + 1; !found &&  t != VOV_LAST_PLACE_TYPE; t++ ) {
	VovPlace p( t, name );
	if ( trace.find( p ) == VOV_OK ) {
	    found = 1;
	    thePlace = p;
	}	    
    }
    if ( ! found )  {
	String l;
	String msg = name + ": place not found.";
	int count = trace.list( VOV_PLACE, name, l );
	if ( count ) {
	    msg += "\nPlaces with similar name:\n" + l;
	    vovShortMessage( "ATTENTION" , msg );
	} else {
	    vovConfirm( "ATTENTION PLEASE", msg, "Continue", 0 );
	}
	return ;
    }
	
    String ctrl = button.name();
    if ( ctrl == "History" ) {
	String title = form( "HISTORY OF %s (%d levels %s)", (char*)name, data.hist_level, (char*)hist_direction_string );
	String text;
	trace.history( thePlace, data.hist_flags, data.hist_direction, data.hist_level, text );
	// Improve the look of the text, by shortening the name of the places.
	String wd = path_canonicalize( utcc_getwd() );
	text.gsub( wd , String(".") );
	vovLongMessage( title, text, 20, 70 );
    } else if ( ctrl == "Edit" ) {
	trace.edit( thePlace );
    } else if ( ctrl == "View" ) {
	trace.edit( thePlace, 1 );
    } else if ( ctrl == "In" ||  ctrl == "Out" ) {
	theNode.id() = thePlace.id();
	VOV_ASSERT( trace.getById( theNode ) );
	VovStatus s = vovNavigate( trace, theNode, ctrl == "In" ? VOV_INPUT_MASK : VOV_OUTPUT_MASK );
	if ( s == VOV_CHANGE ) {
	    if ( theNode.nodeType() == VOV_PLACE ) {
		thePlace.id() = theNode.id();
		VOV_ASSERT( trace.getById( thePlace ) );
		historyNamesRing.add( thePlace.getName() );
		trace.edit( thePlace , 1 );
	    } else {
		theTransition.id() = theNode.id();
		VOV_ASSERT( trace.getById( theTransition ) );
		trace.edit( theTransition , 1  );
	    }
	}
    } else if ( ctrl == "Status" ) {
	vovEditNodeStatus( trace, thePlace );
    } else if ( ctrl == "Retrace" ) {
	String msg;
	VovRetrace retrace( hist_direction_string + " " + thePlace.format() );
	VOV_ASSERT( trace.create( retrace ) );
	VovStatus s =  trace.evolve( retrace, thePlace, data.hist_direction, VOV_RETRACE_SLOW , msg );
	vovLongMessage( "RETRACE", msg, 5 ,70 );
    } else {
	vovError( "What button did you press ? " );
    }
}

void vov_browse_level_ccb( ddsHandle item )
{
    VovDialogNumber levelDlg( item );
    VovLocalBrowseData& data = *(VovLocalBrowseData*)levelDlg.getData();
    data.hist_level = (int)levelDlg.getValue();
}

void vov_browse_flags_ccb( ddsHandle item, int offset )
{
    VovDialogList list( item );
    VovLocalBrowseData& data = *(VovLocalBrowseData*)list.getData();
    data.hist_flags = list.reduceOR();
}

void vov_browse_direction_ccb( ddsHandle item, int offset )
{
    VovDialogList list( item );
    VovLocalBrowseData& data = *(VovLocalBrowseData*)list.getData();
    data.hist_direction = (int)list.getWhichOne();
}


void vov_browse_dismiss_ccb( ddsHandle item )
{
    VovDialogControl ctrl( item );
    int& posted = *(int*)ctrl.getData();
    posted = 0;
    dds_destroy( item );
}


void vovBrowser( VovTrace& trace )
{
    // This routine is not reentering. 

    static int browserPosted = 0;
    if ( browserPosted ) {
	vovError( "browser is still posted" );
	return;
    }
    browserPosted = 1;

    static StringRing historyNamesRing(8);			// Name used to find place, transition
    static VovLocalBrowseData data;
    static int initialized = 0;

    if ( ! initialized ) {
	data.hist_level = 2;				// Levels of history search.
	data.hist_direction = VOV_BACKWARD;
	data.hist_flags = VOV_TRANSITION_MASK | VOV_ANNOTATION_MASK ;
	initialized = 1;
    }
    


    VovDialog top_dialog ( "VOV BROWSER", DDS_VERTICAL, 8);
    

    VovDialogEditText namestarDlg( top_dialog, "name*", "" , 1, 50 );
    VovDialogEditText nameDlg( top_dialog, "name", historyNamesRing.getlast() , 1, 50 );
    VovDialogEditText recentDlg( top_dialog, "recent", historyNamesRing.format(), 4, 50 );
    VovDialogEditText cwdDlg( top_dialog, "CWD", path_canonicalize(utcc_getwd()) + "/", 1, 50 ); 

    VovDialogNumber levels( top_dialog, "levels", 1, 40, 1, data.hist_level, "%.0f", vov_browse_level_ccb, (ddsData)&data );
    
    VovDialogList flagDialog( top_dialog, DDS_FALSE );
    flagDialog.addItem( "places", VOV_PLACE_MASK, ( data.hist_flags & VOV_PLACE_MASK ) > 0 );
    flagDialog.addItem( "transitions", VOV_TRANSITION_MASK, ( data.hist_flags & VOV_TRANSITION_MASK ) > 0 );
    flagDialog.addItem( "nodes", VOV_NODE_MASK, ( data.hist_flags & VOV_NODE_MASK ) > 0 );
    flagDialog.addItem( "annotations", VOV_ANNOTATION_MASK, ( data.hist_flags & VOV_ANNOTATION_MASK ) > 0 );
    flagDialog.endOfList( vov_browse_flags_ccb, (ddsData)&data );
    
    VovDialogList dirDialog( top_dialog, DDS_TRUE );
    dirDialog.addItem( "BACKWARD", VOV_BACKWARD, data.hist_direction == VOV_BACKWARD);
    dirDialog.addItem( "FORWARD", VOV_FORWARD,  data.hist_direction == VOV_FORWARD );
    dirDialog.endOfList( vov_browse_direction_ccb, (ddsData)&data );
    
    // top_dialog.setTitle( "VOV: HISTORY DIALOG" );
    top_dialog.addControl( "History", vov_browse_ccb, (ddsData)&data  );
    top_dialog.addControl( "Edit", vov_browse_ccb, (ddsData)&data );
    top_dialog.addControl( "View", vov_browse_ccb, (ddsData)&data );
    top_dialog.addControl( "In", vov_browse_ccb, (ddsData)&data  );
    top_dialog.addControl( "Out", vov_browse_ccb, (ddsData)&data  );
    top_dialog.addControl( "Status", vov_browse_ccb, (ddsData)&data  );
    top_dialog.addControl( "Retrace", vov_browse_ccb, (ddsData)&data );
    top_dialog.addControl( "Dismiss", vov_browse_dismiss_ccb, (ddsData)&browserPosted );
    
    
char* helpString = "\
You can choose a place in the design trace. The choice is done by name.\n\
Use either the `name*' or the `name' field to enter the name of the place\n\
(`name*' overrides `name').\n\n\
The `recent' field is provided for your convenience. It contains the most\n\
recent names used.\n\n\
`History': get a textual description of the history of the place.\n\
\tThe level specifies how far to carry the search.\n\
\n\
Use the following flags to determine the direction of search:\n\
   BACKWARD: look at past (follow inputs of place)\n\
   FORWARD : look at future (follow outputs of place)\n\
\n\
Use the following flags to determine the information appearing\n\
in the history of a place:\n\
   places     : give information on places found along the way\n\
   transitions: same as above, for transitions\n\
   nodes      : same as above, for nodes\n\
   annotations: same as above, for annotations\n\
\n\
`Edit': edit the place\n\
`In' and `Out': navigate through the trace.\n\
`Status': Change status of place.\n\
`Retrace': retrace TO or FROM the place depending on the direction of search.\
";

    top_dialog.addHelp( helpString );


    data.tracePtr = &trace;
    data.historyNamesRingPtr = &historyNamesRing;
    data.namestarDlgHandle = namestarDlg.handle();
    data.nameDlgHandle = nameDlg.handle();
    data.recentDlgHandle = recentDlg.handle();

    top_dialog.postUnmoded( 0 );
    
	
}

