/* Local includes */
#include "trace.h"
#include "transition.h"
#include "place.h"
#include "bzocket.h"
#include "vovlib.h"
#include "utcc.h"
#include "VOVprotocol.h"
#include "ddsplus.h"

#include "voverror.h"
#include "comm.h"

extern VovStatus vov_updateMIPSinfo( VovTrace& , int , String&  );

typedef struct {
    VovTrace* tracePtr;
    ddsHandle powerDlgHandle;
    ddsHandle timeOutDlgHandle;
    ddsHandle resourcesDlgHandle;
    int power;
    int timeOut;
    int stop;
} VovLocalInteractiveSlaveData;


void islave_update_ccb( ddsHandle item )
{
    VovDialogControl ctrl( item );
    VovLocalInteractiveSlaveData& data = *(VovLocalInteractiveSlaveData*)ctrl.getData();
    VovTrace& trace = *data.tracePtr;
    VovDialogEditText resourceDlg( data.resourcesDlgHandle );
    String affinity = resourceDlg.getTextItem();
    VovDialogNumber powerDlg( data.powerDlgHandle );
    vov_updateMIPSinfo( trace, (int)powerDlg.getValue(), affinity);  
}


void islave_timeout_ccb( ddsHandle item )
{
    VovDialogNumber number( item );
    VovLocalInteractiveSlaveData& data = *(VovLocalInteractiveSlaveData*)number.getData();
    data.timeOut = (int)number.getValue();
}

void islave_power_ccb( ddsHandle item )
{
    VovDialogNumber number( item );
    VovLocalInteractiveSlaveData& data = *(VovLocalInteractiveSlaveData*)number.getData();
    data.power = (int)number.getValue();
}

void islave_ready_ccb( ddsHandle item )
{
}


void islave_quit_ccb( ddsHandle item )
{
    VovDialogControl ctrl( item );
    VovLocalInteractiveSlaveData& data = *(VovLocalInteractiveSlaveData*)ctrl.getData();
    data.stop = 1;
    dds_destroy( item );
}

void VC_interactiveSlave( VovTrace& trace )
{
    VovLocalInteractiveSlaveData data;
    data.timeOut = 10000;
    data.stop = 0;

    VovDialog top( "VOV INTERACTIVE SLAVE", DDS_VERTICAL, 3 );
    VovDialogEditText resourcesDlg( top, "RESOURCES", "interactive", 1, 60 );
    VovDialogNumber powerDlg( top, "POWER", 100, 12000, 100, 1000, "%.0f", islave_power_ccb, (ddsData)&data );
    VovDialogNumber timeOutDlg( top, "TIME OUT (us)", 10000, 1000000, 10000, data.timeOut, "%.0f", islave_timeout_ccb, (ddsData)&data );


    //    top.addControl( "Ready for next job" );
    top.addControl( "Update", islave_update_ccb, (ddsData)&data );
    top.addControl( "Quit", islave_quit_ccb, (ddsData)&data );


    data.tracePtr = &trace;
    data.powerDlgHandle = powerDlg.handle();
    data.timeOutDlgHandle = timeOutDlg.handle();
    data.resourcesDlgHandle = resourcesDlg.handle();

    top.postUnmoded();


    Bzocket& zoc = trace.getBzocket();

    while ( ! data.stop ) {
	dds_outstanding( 0 );
	int s =  zoc.waitForPacket( 0, data.timeOut );	// 
	if ( s == BUF_TIMEOUT ) continue;
	if ( s == BUF_ERROR ) {				// 
	    vovError( "Slave lost connection." );	// 
	    break;					// Slave exits.
	}						// 
	if ( VC_get_reply( zoc ) != VOV_OK ) { 		// Get rest of packet.
	    break;					// Lost connection.
	}						// 
	int code = zoc.unpackInt();			// Get packed op-code.
	if ( code == VOV_Exec ) {			// 
	    extern int slave_exec( VovTrace& );
	    slave_exec( trace );			// 
	} else if ( code == VOV_KillSlave ) {		// A bit overloaded code.
	    vovMessage( "Slave instructed to exit." );
	    break;
	} else {					// 
	    vovError( form( "Unexpected code %d.", code ) );
	    break;
	}
    }
    vovMessage( "Interactive slave has been terminated." );
}





