/* Local includes */
#include "trace.h"
#include "transition.h"
#include "place.h"
#include "bzocket.h"
#include "vovlib.h"
#include "utcc.h"
#include "VOVprotocol.h"
#include "voverror.h"
#include "comm.h"
#include "signal.h"


static VovTransition *curTransitionPtr = 0;

SIGNAL_FN killTransition()
{
    if ( curTransitionPtr ) {
	curTransitionPtr->kill( SIGKILL );
    }
    delete curTransitionPtr;
    curTransitionPtr = 0;
}



int slave_exec( VovTrace& trace )       // Returns transition duration.
{					// 
    
    signal( SIGUSR1, killTransition );

    Bzocket& zoc = trace.getBzocket();	// 
					// 
    // int oldumask = umask( 0 );		// Eliminate protections.
					// 
    VovTransition trans;		// The transition to execute.
    trans.unpack( zoc );		// Get transition from socket.
    int fastRetracing = zoc.unpackInt(); // Should it be fast or slow?
    
    String errmsg;			// 
    int transitionStatus;		// Returned by vovTransition::waitCompletion().

    vovMessage( trans.getCommandLine() );
    VovStatus status =  trans.fire(0, fastRetracing, errmsg ); // Redirect output.

    if ( status == VOV_OK ) {
	curTransitionPtr = new VovTransition;
	*curTransitionPtr = trans;

	status = trans.waitCompletion( &transitionStatus, errmsg  );
	if ( status == VOV_ERROR ||  transitionStatus == 127 ) {
	    errmsg += "\nThe transition failed.";
	    status = VOV_ERROR;
	    vovMessage( errmsg );	// Just for debugging.
	}
    }

    int rtn = 0;		// 
    if ( status == VOV_ERROR ) {
	zoc.pack( VOV_CannotExec );
	zoc.pack( errmsg );
	zoc.flush();
    } else {
	if ( trace.terminate( trans, VOV_OK, transitionStatus, errmsg ) != VOV_OK ) {
	    vovError("Cannot properly terminate the transition." );
	}
	rtn = trans.duration();
    }
    if ( curTransitionPtr ) {
	delete curTransitionPtr;
	curTransitionPtr = 0;
    }
    return rtn;
}
