/* Routine to compute load on the machine the slave is running on */
#include <String.h>
#include <stream.h>
#include "voverror.h"
#include "utcc.h"


extern "C"
{
#include <sys/time.h>
#include <sys/resource.h>
}

double vov_loadOnLocalHost()
{
    String sorg = utcc_popen( "uptime" ); 		// 
    String s = sorg;					// 
    Regex l_rx = "[0-9]+\\.[0-9][0-9]";			// Format examples: 0.22 12.35 
    double load[3];					// 
    for ( int i = 0 ; i < 3 ; i++) {			// 
	String number = s.at( l_rx );			// 
	load[i] = atof( (char*)number );		// 
	s = s.after( number );				// 
    }							// 
    // vovMessage(form( "loads = %.2f %.2f %.2f", load[0],load[1],load[2] ));
    if ( load[0] > load[1] ) {
	return load[0];
    } else {
	return load[1];					// Return load.
    }
}    



// Try to fool the optimizer by having a function using
// a bunch of arguments, which are computed by test_routine_for_mips();
static void dummy_function( int i, int ii, double d, double dd )
{
}

int test_routine_for_mips()
{
    int rep = 10000;				// Repeat this many times.
    int ii = 0;					// 
    double d = 0.0;				// 
    double dd = 0.0;				// 
    for (int i = 0; i < rep ; i++) {		// 
	ii += i;				// 
	ii += i * i;				// 
	d  += i + 3.22;				// 
	dd += d + d;				// 
	dd /= d;				// 
	dd = sin( dd );				// Gotta call a function, or the optimizer might turn all this into NOP.
    }						// 
    dummy_function( i, ii, d, dd );		// Use all vars in a function.
    return i * 50;				// 
}

double estimate_mips()
{
    struct rusage start_rusage;			// Used to compute swaps.
    struct rusage end_rusage;			//   "  "    "       " 
    int    maxtry = 10;				// Try measure this many times.
    int    tot_est_op;				// Total estimated operations.
    do {
	if ( ! --maxtry ) {
	    vovError( "Could not estimate mips. Too many swaps." );
	    break;
	}
	getrusage( 0 , &start_rusage );
	tot_est_op = test_routine_for_mips();
	getrusage( 0 , &end_rusage );
    } while (end_rusage.ru_nswap != start_rusage.ru_nswap);

    long sec  = end_rusage.ru_utime.tv_sec - start_rusage.ru_utime.tv_sec;
    long usec = end_rusage.ru_utime.tv_usec- start_rusage.ru_utime.tv_usec;
    double time = sec + usec * 1e-6;
    
    if (time == 0.0) {
	vovError ( "Test time is 0.0!!" );
	return 1.0;
    }
    return (tot_est_op / time);
}




