// This may look like C code, but it is really -*- C++ -*-

#ifndef VOV_PROTO_H
#define VOV_PROTO_H

#define VOV_PROTOCOL_VERSION     "VOV protocol 5.04"

/*
 * Request codes
 */

typedef enum _vovCode {
    VOV_NoCode = -2,		/*  */
    VOV_Error = -1,		/*  */
    VOV_Ok = 0,
    /*
     * Handshake with server.
     */
    VOV_StartUp = 1,		/* Start connection with server. */
    VOV_Close,			/* End connection with server. */
    VOV_KillServer,		/* Tell server to gently exit */
    VOV_KillSlave,		/* Tell slave to exit */

    VOV_Create,
    VOV_GetById,
    VOV_GetIntermediateNodes,
    VOV_List,
    VOV_Find,
    VOV_Modify,
    VOV_Delete,
    VOV_FreeResources,
    VOV_Edit,

    VOV_Lock,
    VOV_Unlock,

    VOV_ChangeStatus,

    VOV_1SetOperation,
    VOV_2SetOperation,
    VOV_SetRecompute,
    
    VOV_InitGenerator,
    VOV_Generate,
    VOV_FreeGenerator,

    VOV_SaveOntoDisk,
    VOV_CheckPlaces,
    VOV_StopAllTransitions,
    /*
     * Miscellaneous codes.
     */
    VOV_SynchClocks,		/* Synchronize clocks. */
    VOV_Debug,			/* Debug the communication through the bzockets. */
    VOV_Sync,			/* Force saving of trace onto disk. */
    VOV_Acknowledge,		/* Just to test the server. */
    VOV_EchoLog,		/* Just another test. */
    VOV_PlaceTrace,		/* Instruct server to nicely place  the trace. */

    /*
     * Basic server functions, for simple interface to existing software.
     */
    VOV_StartTransition,	/* Define the default transition.         	*/
    VOV_EndTransition,		/* Announce transition has ended. 		*/
    VOV_ForgetTransition,	/* Abort and forget the current transition.     */
    VOV_ForgetSet,		/* Forget a set of places.			*/
    VOV_TestConnectivity,	/* Get connectivity info on a node.		*/
    VOV_Disconnect,		/* Disconnect two connected nodes.		*/
    VOV_AddInput,		/* Declare input  of default transition. 	*/
    VOV_AddOutput,		/* Declare output of default transition.	*/
    VOV_AddIOs,			/* Efficient way to declare many IO's at once. 	*/
    VOV_AddAnnotation,		/* 						*/
    VOV_CountAnnotations,       /*						*/
    VOV_GetAnnotations,		/* 						*/
    VOV_ToggleStatus,		/* VOV_TRACE_NODE <-> VOV_PLAN_NODE		*/
    VOV_Get,			/* Get a place or a transition.			*/
    VOV_GetPlaceById,		/* Get a place by id.				*/
    VOV_GetHistory,		/* History of a place.				*/
    VOV_GetInfo,		/* Information about server. 			*/
    
    VOV_Format,
    VOV_UserFormat,

    VOV_MvLib,			/* Change the name of a set of places.		*/

    VOV_ActiveDesigner,
    VOV_EvolveAll,		/* Retrace what you can.			*/
    VOV_EvolveSet,
    VOV_EvolveNode,		/* Retrace to/from a particular node.		*/
    VOV_EvolveStop,		/* Stop retracing. */
    VOV_GetExecutionScript,
    VOV_GetScript,
    /* 
     * Superior functions.
     */
    VOV_SubSelect,		/* Select a set of places based on a Regex.	*/
    VOV_FreezeConfiguration,	/* Declare that 'everything' is precious. 	*/
    VOV_Snapshot,		/* Take a snapshot of the current trace.   	*/
    /*
     * Functions used for retracing.
     * Slaves running on various machines are used.
     */
    VOV_SlaveReady,		/* Slave tells server that it is ready to exec. */
    VOV_Exec,			/* Exec a transition.				*/
    VOV_CannotExec,		/* Error code, slave cannot exec a transition	*/
    VOV_GetHostLoad,

    /*
     *  Codes for active notification.
     */
    VOV_Notify,
    VOV_SynchNotify,
    VOV_AsynchNotify,
    
    /*
     * Codes for the transmission of whole pages from server to clients
     */
    VOV_RequestPage,		/* 						*/
    VOV_StartPage,		/*  						*/
    VOV_EndPage,		/*  						*/

    VOV_Measurement,		/* A client communicating a measurement.	*/

    VOV_LastCode		/* This is a sentinel. 				*/
} VovCode;

#endif /* VOV_PROTO_H */

