#include "String.h"
#include "date.h"
#include "base.h"
#include "voverror.h"


extern int unix_info( char*, int*, int*, Date*);
extern int unix_exists( char* );
extern int unix_timestamp( char* );
extern void unix_chmod( char*, int );
extern int unix_size( char* );
		     
octStatus vovUnpackFacetName( VovObject* facet, char* name)
// Same as ohUnpackFacetName, but without memory leaks.
{
    char *s;

    /* Check for silly input */
    if (name == 0 || strcmp(name, "") == 0) return OCT_ERROR;
    if (strchr(name, ' ') != 0) return OCT_ERROR;
	
    /* Make a copy of the input string (because we will tear it apart) */ 
    s = utcc_strsav_and_track( name );

    if (s[0] != ':') facet->contents.facet.cell = s;
    if ((s = strchr(s, ':')) != 0) {
	*s++ = '\0';
	if (s[0] != ':') facet->contents.facet.view = s;
	if ((s = strchr(s, ':')) != 0) {
	    *s++ = '\0';
	    if (s[0] != ':') facet->contents.facet.facet = s;
	    if ((s = strchr(s, ':')) != 0) {
		*s++ = '\0';
		if (strchr(s, ':') != 0) return OCT_ERROR;
		if (*s == '\0') {
		    facet->contents.facet.version = OCT_CURRENT_VERSION;
		} else {
		    facet->contents.facet.version = s;
		}
	    }
	}
    }

    if (facet->contents.facet.cell == 0 ||
        facet->contents.facet.view == 0 ||
        facet->contents.facet.facet == 0) {
	return OCT_ERROR;
    }        

    return OCT_OK;
}



void vovUnpackDefaults( VovObject* f, char* m, char* s)
{
    f->type=OCT_FACET;
    f->contents.facet.cell=0;
    f->contents.facet.view=0;
    f->contents.facet.facet=0;
    f->contents.facet.version=OCT_CURRENT_VERSION;
    f->contents.facet.mode=m;
    (void) vovUnpackFacetName(f, s );
}

char* vov_filename( VovObject* facet )
{
    static char filename[1024];
    sprintf(filename, "%s/%s/%s;%s",
	    facet->contents.facet.cell,
	    facet->contents.facet.view,
	    facet->contents.facet.facet,
	    facet->contents.facet.version);
    return filename;
}

/*
*  These two functions consider an oct facet as a file
*  and they determine the existence and timestamp of a facet
*  by checking the corresponfing file. They offer an 
*  alternative to octGetFacetInfo
*/
int vov_oct_info( VovObject* facet, int* existflag, int* mode,  Date* timestamp)
{
    return unix_info( vov_filename( facet ), existflag, mode, timestamp );
}

int vov_oct_exists( VovObject* facet )
{
    return unix_exists( vov_filename( facet ) );
}

void vov_oct_chmod( VovObject* facet, int m )
{
    unix_chmod( vov_filename( facet ), m );
}

int vov_oct_timestamp( VovObject* facet )
{
    return unix_timestamp( vov_filename( facet ) );
}


int vov_oct_size( VovObject* facet )
{
    return unix_size( vov_filename( facet ) );
}



//
//
//
int facet_info(char* path, int* existflag, int* mode, Date* timestamp)
{
    VovObject facet( OCT_FACET );
    vovUnpackDefaults( &facet, "r", "::contents" );
    if (  vovUnpackFacetName( &facet, path ) != OCT_OK ) {
	vovError( form( "Cannot unpack facet name %s", path ) );
	return 0;		// File does not exist.
    }

    return vov_oct_info( &facet, existflag, mode, timestamp);
}

int facet_delete(char* path) 
{
    cerr.form("deletion of OCT facet %s, not yet implemented\n" , path);
    return 0;
}

int facet_exists(char* path)
{
    VovObject facet( OCT_FACET );
    vovUnpackDefaults( &facet, "r", "::contents" );
    OH_ASSERT( vovUnpackFacetName( &facet, path) );

    return vov_oct_exists( &facet );
}

int facet_valid(char* path)
{
    return facet_exists(path);
}

int  facet_cp( char* from, char* to )
{
    // utcc_message( form( "facet_cp: %s -> %s", from, to ) );
    VovObject facetold( OCT_FACET );
    VovObject facetnew( OCT_FACET );
    vovUnpackDefaults( &facetold, "r", "::contents" );
    vovUnpackDefaults( &facetnew, "r", "::contents" );
    OH_ASSERT( vovUnpackFacetName( &facetold, from ) );
    OH_ASSERT( vovUnpackFacetName( &facetnew, to ) );
    
    octFacetInfo info;
    if ( octGetFacetInfo( &facetold, &info ) != OCT_OK ) {
	return -1;		// FROM does not exist.
    }
    if ( octGetFacetInfo( &facetnew, &info ) == OCT_OK ) {
	return -1;		// TO exists already
    }
    facetold.contents.facet.facet = "contents";
    facetnew.contents.facet.facet = "contents";
    octStatus s1 =  octCopyFacet( &facetnew, &facetold ) ;

    if ( s1 != OCT_OK ) {
	return -1;
    }

    facetold.contents.facet.facet = "interface";
    facetnew.contents.facet.facet = "interface";
    octStatus s2 =  octCopyFacet( &facetnew, &facetold ) ;

    if ( s2 != OCT_OK ) {
	return -1;
    }
    return 0;
}

void facet_chmod( char* path, int m )
{
    VovObject facet( OCT_FACET );
    vovUnpackDefaults( &facet, "r", "::contents" );
    OH_ASSERT( vovUnpackFacetName( &facet, path) );

    vov_oct_chmod( &facet, m );
}

int facet_timestamp(char* path)
{
    VovObject facet( OCT_FACET );
    vovUnpackDefaults( &facet, "r", "::contents" );
    OH_ASSERT( vovUnpackFacetName( &facet, path) );

    return vov_oct_timestamp( &facet );
}

int facet_size(char* path)
{
    VovObject facet( OCT_FACET );
    vovUnpackDefaults( &facet, "r", "::contents" );
    OH_ASSERT( vovUnpackFacetName( &facet, path) );

    return vov_oct_size( &facet );
}


