// File fire.cc
//
#include "transition.h"
#include "utcc.h"
#include "voverror.h"
#include "buf.h"
#include "utility.h"

static int dbgpipe = 0;
extern String vov_stripWord( String& );

void sprite_modifyCommand( String& command, String& affinity )
    // On SPRITE, migrate the process if possible.
    // Do not migrate if affinity is set.
{
    if (  affinity == "" && utcc_getenv( "SPRITE_OS", 0 ) == "yes" ) {
	if ( common_prefix( command, "(" ) == "(" ) {  // Some commands redirect stderr and stdout
	    command = "( mig -l " + command.after(0);
	} else {
	    command = "mig -l " + command;
	}
    }
}

// Make this global just to test it
static int fildes[2];		// Two file descriptors for the pipe.

void vov_init_redirect()
{
    fildes[0] = fildes[1] = 0;				// Reset file descriptors.
    if ( pipe( fildes ) == -1 ) { 		// Open a pipe to get stderr from process.
	utcc_printerrno( "Opening pipe" );	// Pipe could not be opened (could be serious).
    }
}

void vov_do_redirect()
{
    if ( dup2( fildes[1], fileno(stdout)) == -1  ) {
	utcc_printerrno("dup2 stdout");	
    }
    if ( dup2( fildes[1], fileno(stderr)) == -1 ) {  
	utcc_printerrno("dup2 stderr");	
    }	
}

void vov_get_redirect( Buf& buf )
{
    if ( fildes[0] ) {
	while ( buf.readSocket( fildes[0] ) == BUF_OK ) {
	    // 
	}
    }
}

String vov_end_redirect( Buf& buf )
{
    String pipeOut;
    if ( fildes[0] ) {
	while ( buf.readSocket( fildes[0] ) == BUF_OK ) {
	    if ( dbgpipe ) buf.print();
	} 
	int length = buf.length();
	if ( length ) {
	    buf.put( "\n\0", 2 );	// Make sure the string is null terminated.
	    pipeOut = buf.get( buf.length() );
	    pipeOut.gsub( "\n", "\n||  " );
	    pipeOut = "\n||  " + pipeOut;
	}
	if ( close(fildes[0])==-1 || close(fildes[1])==-1 ) { 
	    utcc_printerrno( "fildes[1]" );			// Never seems to happen.
	}							// 
    }
    fildes[0] = fildes[1] = 0;
    return pipeOut;
}


// Fire transition on local host.
VovStatus VovTransition::fire( 
			      int redirect, 		// 
			      int fastMode,		// 
			      String& errmsg		// Returned message in case of error.
			      )
{
    char* wd = util_file_expand( (char*)cwd.string() );
    if (chdir( wd ) == -1) {				// Change the working directory.
	errmsg =  "Could not cd to: " + cwd.string();	// 
	return VOV_ERROR;				// 
    }							// 
    String command = commandLine.string();		// Command line to execute.
    if ( fastMode && capsule.string() != 0 ) {		// 
	command = capsule.string();			// 
    }							// 
    sprite_modifyCommand( command, affinity.string() );	// On sprite, allow migration.
    
    if ( command.length() == 0 ) {
	vovMessage( "Null command line" );
	return VOV_ERROR;
    }
    
    if ( redirect ) { 		// Open a pipe to get stderr from process.
	vov_init_redirect();
    }

    int useCsh = 0;
    int argc = 0;
    char** argv = 0;
    char* bin = 0;					// The binary to exec.
    if ( command.contains( Regex("[~$*\\?()<>&;\\|]") ) ) {
	// vovMessage( "Csh: " + command );
	useCsh = 1;
    } else {
	char* sep = " ";
	int size = command.freq( sep ) + 2;	// No of args required.
	argv = new char*[size];

	// vovMessage( "Command: " + command );
	while ( command.length() ) {
	    String arg = command.before( sep );
	    if ( arg == 0 ) arg = command;		// Takes care of the last argument.
	    argv[argc++] = utcc_strsav( (char*)arg );
	    command = command.after( sep );
	}
	argv[argc] = 0;					// Terminate arg list.
	bin = util_path_search( argv[0] );		// Find binary for transition.
	if ( bin == NIL(char) ) {
	    vovError( form("Cannot find executable for %s", argv[0] ));
	    return VOV_ERROR;
	}
    }	
	
    pid = vfork();					// Quickly fork (vfork) another process.
    if ( pid.integer() == 0) {				// This is the child process.
	// Do as little work as possible here.
	// Send all stderr/stdout through the pipe.
	if ( redirect ) {
	    vov_do_redirect();
	}						
	int s;						// Status of exec.
	if ( useCsh ) {
	    s = execl("/bin/csh", "csh", "-f", "-c", (char*)command, 0);
	} else {
	    s = execve( bin, argv, environ );
	}
	if ( s == -1 ) {
	    utcc_printerrno( "exec-ing" );
	}
	(void) _exit(127);			
    }
    return VOV_OK;
}

