/*
  These routines are for the most trivial use of 
  VOV. We assume one transaction, one trace.
*/

#include "trace.h"
#include "place.h"
#include "transition.h"
#include "user.h"
#include "event.h"
#include "select.h"

#include "VOVprotocol.h"
#include "voverror.h"
#include "dialog.h"

#include "comm.h"

VovStatus VOVdebugCommunication( VovTrace& trace )
{
    static int debugFlag = 1;
    Bzocket& zoc = trace.getBzocket();

    zoc.setDebug( debugFlag );

    zoc.pack( VOV_Debug );
    zoc.pack( debugFlag );
    zoc.flush();

    debugFlag = ! debugFlag;
}

/*
* Add an input to the "current" transition
* Arguments are type of the input (ASCII, OCT ...)
* and name.
*/
VovStatus VOVaddInput( VovTrace& trace, VovTransition& tran, VovPlace& place )
{
    if ( trace.findOrCreate( place ) != VOV_OK ) {
	vovError( "cannot find or create place" );
	return VOV_ERROR;
    }
    VovStatus s = trace.declareInput( tran, place, 0 );
    if ( s == VOV_INPUT_CONFLICT ) {
	String q = "\n\tINPUT CONFLICT! INPUT CONFLICT! INPUT CONFLICT!\n\n";
	q += place.format() + "\n\n";
	q += tran.getCommandLine();
	q += "\nneeds the above mentioned place which is " + place.getStatusString() + " now.\n" ;
	q += "Do you want to continue this transaction anyway?";
	String o1 = "No, stop, let me take a look, and forget this transition";
	String o2 = "Yes, go ahead";
	if ( userDecision( q, o1, o2 ) == 1 ) {
	    vovMessage( "Transition aborted." );
	    VOV_ASSERT( trace.destroy( tran ) );
	    exit( 255 );
	}
    }
    return VOV_OK;
}
/*
* Add an output to the "current" transition
* Arguments are type of the input (ASCII, OCT ...)
* and name.
*/
VovStatus VOVaddOutput(VovTrace& trace, VovTransition& tran, VovPlace& place )
{
    trace.findOrCreate( place );
    String errmsg;
    if ( trace.declareOutput( tran, place, 0, 0, errmsg ) == VOV_ERROR ) {
	String q = "\n\tOUTPUT CONFLICT! OUTPUT CONFLICT! OUTPUT CONFLICT!\n\n";
	q += "PLACE: " + place.format() + "\n\n";
	q += "The transition currently executing is:\n" + tran.getCommandLine() + "\n\n" + errmsg;
	String option1 = "Force the place to be the output of the current transition";
	String option2 = "Abort this transition";
	if ( userDecision( q, option1, option2 ) != 1 ) {
	    vovMessage( "Transition aborted." );
	    VOV_ASSERT( trace.destroy( tran ) );
	    exit( 255 );
	}
	return  trace.declareOutput( tran, place, 1, 0,  errmsg );
    }
}


VovStatus VOVforget( VovTrace& trace, String& setname )
{
    VovNodeSet set( setname );
    if ( trace.find( VovSuperObject(set) ) == VOV_ERROR ) {
	return VOV_ERROR;
    }
    return trace.destroy( VovSuperObject(set) );
}

VovStatus VOVkillServer( VovTrace& trace )
{
    String h = utcc_getenv("VOV_HOST_NAME");
    String p = utcc_getenv("VOV_PROJECT_NAME");

    String q = ( "Kill VOV server\n\ton HOST: " + h + "\n\tfor PROJECT: " + p );
    String option1 = "Yes, go ahead, make my day";
    String option2 = "No, I changed my mind, I'm sorry";
    if ( userDecision( q, option1, option2 ) == 1 ) {
	Bzocket& zoc = trace.getBzocket();
	zoc.pack( VOV_KillServer );
	zoc.flush();
	return  VC_wait_for_code( zoc, VOV_Ok );
    }
    return VOV_ERROR;
}

/*
 * Have the server echo a message 
 * This function primarily tests the communication with the server
 */
VovStatus VOVecho( VovTrace& trace, String& m)
{
    Bzocket& zoc = trace.getBzocket();

    zoc.pack(VOV_EchoLog);
    zoc.pack(m);
    zoc.flush();

    VovStatus status =  VC_wait_for_ok_reply( zoc );
    return status;
}


VovStatus VOVsync( VovTrace& trace )
{
    trace.checkPlaces( 0 );
    return trace.saveOntoDisk();
}



VovStatus VOVsetClock( VovTrace& trace, char* host, int delta )
{
    Bzocket& zoc = trace.getBzocket();
    zoc.pack( VOV_SynchClocks ); // Overload this code.
    zoc.pack( 3 );		 // A special code. See ../server/dispatch.cc
    zoc.pack( host );		 // 
    zoc.pack( delta );		 // 
    zoc.flush();		 // 
    return VOV_OK;		 // 
}





/*
  These routines are for the most trivial use of 
  VOV. We assume one transaction, one trace.
*/


#ifndef VOVCLIENT
VovStatus VOVselect( VovTrace& trace, String& setname, VovSelectRule& rule )
{
    VovNodeSet set( setname );
    trace.findOrCreate( VovSuperObject(set) );
    return trace.subselect( set, rule );
}
#endif


// Notify client synchronously.
// Send events as long as there are any. Then stop sending.
#ifndef VOVCLIENT
int VOVsynchNotify( VovTrace& trace, VovEventQueue& queue, int max_events, int onlynew = 1)
{
    Bzocket& zoc = trace.getBzocket();
    
    zoc.pack( VOV_SynchNotify );
    zoc.pack( max_events );
    zoc.pack( onlynew );	// 
    zoc.flush();
    
    if ( VC_wait_for_code( zoc, VOV_SynchNotify ) != VOV_OK ) {
	vovFatalError( "Wrong response from server in VOVsynchNotify" );
    }
    
    int eventCount = zoc.unpackInt();
    int count  = eventCount;
    // vovMessage( form( "%d events received.", eventCount ) );
    VovEvent event;
    while ( count-- ) {
	event.unpack( zoc );
	queue.enq( event );
    }
    queue.markObsolete();
    return eventCount;
}
#endif

// 
// Make this routine synchronous, to avoid BROKEN PIPE signals when closing connection.
#ifndef VOVCLIENT
void VOVnotify( VovTrace& trace, int flag )
{
    Bzocket& zoc = trace.getBzocket();
    zoc.pack( VOV_AsynchNotify );
    zoc.pack( flag );		// if flag != 0, begin a notify session
    zoc.flush();
}
#endif

#ifndef VOVCLIENT
VovStatus VOVnextEvent( VovTrace& trace, VovEvent& event )
{
    Bzocket& zoc = trace.getBzocket();
    if ( VC_wait_for_code( zoc, VOV_Notify ) != VOV_OK ) {
	return VOV_ERROR;
    }
    event.unpack( zoc );
    return VOV_OK;
}
#endif

#ifndef VOVCLIENT
VovStatus VOVasynchNotify( VovTrace& trace, VovEventFilterFunction func, VovEventFilter& stopFilter ) 
{
    VOVnotify( trace, 1 );	// Start

    VovEvent event;
    while ( VOVnextEvent( trace, event ) == VOV_OK ) {
	if ( event.getSubjectMask() == VOV_NULL_MASK ) {
	    vovMessage( "No more events" );
	    break;
	}
	if ( (*func)( trace,  event ) != VOV_OK ) {
	    break;
	}
	if ( stopFilter.filter( event ) == VOV_OK ) {
	    break;
	}
    }
    VOVnotify( trace, 0 );	// End
    return VOV_OK;
}
#endif


#ifndef VOVCLIENT
VovStatus VOVedit( VovTrace& trace, VovNode& node )
{
    vovMessage( "Obsolete function VOVedit()" );
    octId lockId;
    int forceLock = 0;
    Date now;
    String desc = utcc_getenv("USER") + " " + now.format();
    String msg;

    if( trace.lock( node, desc, forceLock, lockId, msg ) != VOV_OK ) {
	String q = "\n\n\tLOCK CONFLICT!! LOCK CONFLICT!! LOCK CONFLICT!! LOCK CONFLICT!!\n\n" +
	  node.format() + "\nis already locked.\n\n" + msg;
	String o1 = "Force the lock";
	String o2 = "Abort editing";
	if ( userDecision( q, o1, o2 ) != 1 ) {
	    return VOV_NO_CHANGE;
	}
	forceLock = 1;
	desc += " (FORCING)";
	VOV_ASSERT( trace.lock( node, desc, forceLock, lockId, msg ) );
    }
    
    VovStatus rtn;
    if ( node.nodeType() == VOV_PLACE ) {
	VovPlace place;
	place.id() = node.id();
	VOV_ASSERT( trace.getById( place ) );
	rtn = trace.edit( place );
    } else {
	VovTransition tran;
	tran.id() = node.id();
	VOV_ASSERT( trace.getById( tran ) );
	rtn = trace.edit( tran );
    }
    if ( trace.unlock( node, lockId ) != VOV_OK ) {
	vovError( "Unlocking failed: somebody must have overwritten the lock." );
    }
    return rtn;
}
#endif


