/* -*- C++ -*- */
/*
 * page.h
 *
 * This file describes the data structures to represent the design trace
 */
#ifndef PAGE_H
#define PAGE_H
#pragma once

#include <String.h>
#include "base.h"
#include "date.h"
#include "vovinternal.h"
#include "node.h"
#include "event.h"

extern class VovPlacementMatrix;
/*
 * A page is actually a facet.
 */
class VovPage : public VovObject {
    VovNodeSet places;		// 
    VovNodeSet transitions;	// 
    VovNodeSet nodes;		// 
    VovNodeSet placesToCheck;
    VovNodeSet transitionsToFire; // 
    VovNodeSet transitionsToKill; // Transitions that could be killed, because their inputs have been changed.
    VovNodeSet placesToEvolve;

    VovBag retracings;		// Bag of all retracings.
    VovBag sets;		// 
    VovBag goals;		// 
    VovBag users;		// 

    VovProp goodBackup;		// Flag marking availability of a good backup.
    VovProp lastSave;		// Time stamp of last save operation.
    VovPlacementMatrix* matrix;	// 
    long lastDoubleSave;
 public:
    VovPage();			// 
    VovPage( VovPage& );	// 
    ~VovPage();			// 
    VovStatus open( const String& name, const String& mode );
    VovStatus open( VovObject& facet );
    VovStatus saveOntoDisk(  );
    VovStatus get();
    VovStatus close();

    VovStatus initPlacementMatrix();
    int32 getTimeStamp();	// 

    VovStatus OK();		// Verify consistency of page.
    String format();		// Print info about the page.

    friend class VovTrace;	// 
};

#endif	/* PAGE_H */
