/* -*- C++ -*- */
/*
 * place.h
 */
#ifndef PLACE_H
#define PLACE_H
#pragma once


#include <String.h>
#include "date.h"
#include "node.h" 
#include "chain.h"

extern class Bzocket;
extern class VovChainInfo;
extern class VovEventQueue;
extern class VovTransition;
extern class VovTrace;

#define VOV_PLACE_OCTMASK OCT_INSTANCE_MASK

class VovPlace : public VovNode {
    VovProp      placeType;	// Place type.
    VovProp	 name;		// The official name of the place.
    VovProp	 timeStamp;	// Time the place was changed last.
    VovProp	 value;		// For some places, the value of the place.

    VovProp      host;		// Host on which the place resides.

    // In order to allow tools to run "in place" we need to
    // distinguish between "information" and "chain"
    VovProp	 current; 	// 1 if this is the current place in the chain.
    friend       class VovChainInfo;

    VovBag    goalBag;		// That must be satisfied by the place ( is this really used ? )
    VovBag    measureBag;	// Measures used in conjunction with the goals.

    char*    getCellName( VovPlaceType t );
 public:
    VovPlace();
    VovPlace( octId id );
    VovPlace( VovPlaceType t , const String& nm );
    VovPlace( VovNode& );
    VovPlace( VovPlace&);
    ~VovPlace();
    void operator=( VovPlace& );

    VovStatus create( );
    VovStatus get( );
    VovStatus modify();
    VovStatus del();

    // Duplicate a node which is both input and output of a transition.
    VovStatus duplicate( VovTransition&, VovPlace& , int dir = VOV_FORWARD );	
    VovStatus merge( VovTransition&, VovPlace& , int dir = VOV_FORWARD );
    VovStatus setCurrent( ); 			       
    VovStatus unsetCurrent( );				
    VovStatus getChain( VovChainInfo& );
    VovStatus isChain();
    
    VovStatus correctInputInChain( VovTrace&, VovTransition& );
    VovStatus correctOutputInChain( VovTrace&, VovTransition& );

    VovPlaceType getPlaceType() { return placeType.integer();}

    void   setName( String& n );
    String getName() { return name.string(); }
    String expandedName();
    String getCommonName();

    VovProp& getValue() { return value; }
    VovStatus edit();
    void setupDialog( VovDialog& top_dialog, VovDialogHandles& data , int viewOnly = 0);
    VovStatus readDialog( VovDialogHandles& data );

    String format();

    VovStatus copy( VovPlace& );
    VovStatus mvlib( String& oldpath , String& newpath );
    VovStatus match(VovPlace&);
    VovStatus exists();

    VovStatus changeTransition( VovTransition&, int flag );

    VovStatus info(int*, int*, Date*, int );
    Date getTimeStamp();
    inline void setTimeStamp() { timeStamp = getDate(); }
    inline void setTimeStamp( Date t ) { timeStamp = t.get();}

    VovStatus freeResources(); /* protected version */
    int   getProtection(); 
    void  setProtection( int );

    void  canonicalize();
    void  pack(Bzocket&);
    void  unpack(Bzocket&);

    // VovStatus addGoal( VovGoal& );
    // VovStatus scheduleGoals();
    VovStatus updateMeasure( VovPlace& );
    String formatMeasures();
};

extern VovStatus vov_place_find( VovPlace& );
extern VovStatus vov_place_find_add( VovPlace& );
extern VovStatus vov_place_find_del( VovPlace& );


//int      VovPlace::newRefinement(){ return 0;}
//VovPlace VovPlace::getRefinement(){ return this;}

#endif
