#include "retrace.h"
#include "transition.h"
#include "date.h"
#include "voverror.h"

extern String vov_prettyTime( int s );

VovRetrace::VovRetrace(  )
    :
    ("no name"),
    estimatedDuration( "ESTIMATED_DURATION" , OCT_INTEGER ),
    actualDuration( "ACTUAL_DURATION", OCT_INTEGER ),
    startDate( "START_DATE", OCT_INTEGER ),
    status( "STATUS", OCT_INTEGER )
{
    estimatedDuration = 0;
    actualDuration = 0;
    startDate = 0;
    status = VOV_OK;
}

VovRetrace::VovRetrace(  String& t )
    :
    ((char*)t),
    estimatedDuration( "ESTIMATED_DURATION", OCT_INTEGER ),
    actualDuration( "ACTUAL_DURATION", OCT_INTEGER ),
    startDate( "START_DATE", OCT_INTEGER ),
    status( "STATUS", OCT_INTEGER )
{
    estimatedDuration = 0;
    actualDuration = 0;
    startDate = 0;
    status = VOV_OK;
}

VovRetrace::VovRetrace(  VovRetrace& r )
    :
    (r),
    estimatedDuration( "ESTIMATED_DURATION", OCT_INTEGER ),
    actualDuration( "ACTUAL_DURATION", OCT_INTEGER ),
    startDate( "START_DATE", OCT_INTEGER ),
    status( "STATUS", OCT_INTEGER )
{
    estimatedDuration = r.estimatedDuration;
    actualDuration = r.actualDuration;
    startDate = r.startDate;
    status = r.status;
}

VovRetrace::~VovRetrace()
{
}

#ifndef VOVCLIENT
VovStatus VovRetrace::create()
{
    VovNodeSet::create();
    OH_ASSERT( octCreate( this, &estimatedDuration ) );
    OH_ASSERT( octCreate( this, &actualDuration ) );
    OH_ASSERT( octCreate( this, &startDate ) );
    OH_ASSERT( octCreate( this, &status ) );
    // vovMessage( "Created: " + format() );
    return VOV_OK;
}
#endif

#ifndef VOVCLIENT
VovStatus VovRetrace::get()
{
    getById( OCT_BAG_MASK );
    VovNodeSet::get();
    OH_ASSERT( octGetByName( this, &estimatedDuration ) );
    OH_ASSERT( octGetByName( this, &actualDuration ) );
    OH_ASSERT( octGetByName( this, &startDate ) );
    OH_ASSERT( octGetByName( this, &status ) );
    // vovMessage( "Got: " + format() );
    return VOV_OK;
}
#endif

#ifndef VOVCLIENT
VovStatus VovRetrace::del()
{
    // vovMessage( "Deleting: " + format() );
    OH_ASSERT( octDelete( &estimatedDuration ) );
    OH_ASSERT( octDelete( &actualDuration ) );
    OH_ASSERT( octDelete( &startDate ) );
    OH_ASSERT( octDelete( &status ) );
    VovNodeSet::del();
    return VOV_OK;
}
#endif


#ifndef VOVCLIENT
VovStatus VovRetrace::start( String& msg )
{
    // vovMessage( "Starting: " + format() );
    int totalduration = 0;
    int unknownduration = 0;
    int countTrans = 0;
    VovContentsGenerator genTrans( this, VOV_TRANSITION_OCTMASK );
    VovTransition trans;
    while ( octGenerate( &genTrans, &trans ) == OCT_OK ) {
	VOV_ASSERT( trans.get() );
	countTrans++;
	int d = trans.duration();
	if ( d >= 0 ) {
	    totalduration += d;
	} else {
	    unknownduration++;
	}
    }
    if ( countTrans == 0  ) {
	msg = "\nNothing to retrace.";
	status.modify( VOV_RETRACE_DONE );
	return VOV_ERROR;
    } else {
	msg = form( "%3d transition%s. Estimated retracing time: %s",
		   countTrans, countTrans != 1 ? "s":"", (char*)vov_prettyTime(totalduration) );
	msg += (unknownduration) ? form( " + %d(?).", unknownduration ) : "";
	msg += ".\n";
	startDate.modify( getDate() );
	estimatedDuration.modify( totalduration );
	status.modify( VOV_RETRACE_IN_PROGRESS );
	return VOV_OK;
    }
}
#endif

#ifndef VOVCLIENT
VovStatus VovRetrace::stop( String& msg )
{
    // vovMessage( "Stopping: " + format() );
    int count = length( VOV_TRANSITION_OCTMASK );
    if ( count ) {
	status.modify( VOV_RETRACE_ERROR );
	clear( VOV_TRANSITION_OCTMASK );
	msg = form("Forced stop. %d transition%s not retraced.\n", count, count!=1?"s":"");
	return VOV_OK;
    } else {
	msg = "No transitions were scheduled for this retracing.\n";
    }
    return VOV_OK;
}
#endif

#ifndef VOVCLIENT
VovStatus VovRetrace::terminate( VovTransition& t )
{
    if ( ! contains( t ) ) return VOV_ERROR;
    del( t );
    if ( length( VOV_TRANSITION_OCTMASK ) == 0 ) {
	actualDuration.modify(  getDate() - startDate.integer() );
	if ( status.integer() != VOV_RETRACE_ERROR ) {
	    status.modify( VOV_RETRACE_DONE );
	}
	return VOV_DELETE_NODE;	// A bit overloaded.
    }
    return VOV_OK;
}
#endif

String VovRetrace::format()
{
    String s = form( "RETRACE %s\n\t", (char*)getName() );
    switch ( (VovRetraceStatus)status.integer() ) {
    case VOV_RETRACE_IN_PROGRESS:
	s += "IN PROGRESS: Completing ";
	int r = estimatedDuration.integer() - ( getDate() - startDate.integer() );
	if ( r > 0 ) {
	    s += "in " + vov_prettyTime( r ) + ".";
	} else {
	    s += "anytime soon.";
	}
	break;
    case VOV_RETRACE_DONE:
	s += "DONE:  ";
	s += "Duration: Expected " + vov_prettyTime( estimatedDuration.integer() );
	s += " actual " + vov_prettyTime( actualDuration.integer() );
	s += ".";
	break;
    case VOV_RETRACE_ERROR:
	s += "FAILED.";
	break;
    default:
	vovError( "Illegal retrace status" );
    }
    return s;
}
