#include "select.h"
#include "place.h"
#include "transition.h"

#include "voverror.h"
static int debug = 0;

VovSelectRule::VovSelectRule( )
    : 
    ( OCT_BAG ),
    typemask( "TYPEMASK", OCT_INTEGER ),
    type2mask( "SECONDARYTYPEMASK", OCT_INTEGER ),
    statusmask( "STATUSMASK", OCT_INTEGER ),
    rx( "REGEX", OCT_STRING )
{
    contents.bag.name = "SELECTRULE";
    set( VOV_PLACE_MASK | VOV_TRANSITION_MASK, VOV_ALL_PLACE_TYPE_MASK, VOV_ALL_STATUS_MASK, SELECT_ALL );
    debug = 0;
}

VovSelectRule::VovSelectRule( int tm, int t2m, int sm, String& r )
    :
    ( OCT_BAG ),
    typemask( "TYPEMASK", OCT_INTEGER ),
    type2mask( "SECONDARYTYPEMASK", OCT_INTEGER ),
    statusmask( "STATUSMASK", OCT_INTEGER ),
    rx( "REGEX", OCT_STRING )
{
    contents.bag.name = "SELECTRULE";
    set( tm, t2m, sm, r );
    debug = 0;
}

VovSelectRule::VovSelectRule( VovSelectRule& r )
    :
    ( OCT_BAG ),
    typemask( "TYPEMASK", OCT_INTEGER ),
    type2mask( "SECONDARYTYPEMASK", OCT_INTEGER ),
    statusmask( "STATUSMASK", OCT_INTEGER ),
    rx( "REGEX", OCT_STRING )
{
    contents.bag = r.contents.bag;
    typemask = r.typemask;
    type2mask = r.type2mask;
    statusmask = r.statusmask;
    rx = r.rx;
    debug = 0;
}

VovSelectRule::~VovSelectRule( )
{
}

void VovSelectRule::set( int tm, int t2m, int sm, String& r )
{
    contents.bag.name = "SELECTRULE";
    typemask = tm;
    type2mask = t2m;
    statusmask = sm;
    rx = r;
}

#ifndef VOVCLIENT
void VovSelectRule::create()
{
    octCreate( this, &typemask );
    octCreate( this, &type2mask );
    octCreate( this, &statusmask );
    octCreate( this, &rx );
}
#endif

#ifndef VOVCLIENT
void VovSelectRule::get()
{
    octGetByName( this, &typemask );
    if ( octGetByName( this, &type2mask ) != OCT_OK ) {
	vovError( "Missing type2mask. Creating" );
	octCreate( this, &type2mask );
    }
    octGetByName( this, &statusmask );
    octGetByName( this, &rx );
}
#endif

#ifndef VOVCLIENT
void VovSelectRule::modify()
{
    octModify( &typemask );
    octModify( &type2mask );
    octModify( &statusmask );
    octModify( &rx );
}
#endif

#ifndef VOVCLIENT
void VovSelectRule::del()
{
    octDelete( &typemask );
    octDelete( &type2mask );
    octDelete( &statusmask );
    octDelete( &rx );
}
#endif

// Select a node if it satisfies all conditions determined by
// typemask
// statusmask
// regular expression rx.
#ifndef VOVCLIENT
int VovSelectRule::select( VovNode& n )
{
    if ( (typemask.integer() & VOV_SOURCE_MASK) && ( n.hasInputs() == VOV_OK ) ) {
	if ( debug ) vovMessage("not a source" );
	return 0;
    }
    if ( (typemask.integer() & VOV_SINK_MASK) && (n.hasOutputs() == VOV_OK) ) {
	if ( debug ) vovMessage("not a sink" );
	return 0;
    }
    if ( (typemask.integer() & ( 1 << n.nodeType() )) == 0 ) {
	if ( debug ) vovMessage( "not the right type" );
	return 0;
    } 
    if (statusmask.integer() != VOV_ALL_STATUS_MASK ) {
	if ( ! (statusmask.integer() & ( 1 << n.getStatus() )) ) {
	    if ( debug ) vovMessage( "not the right status" );
	    return 0;
	}
    }
    if ( n.nodeType() == VOV_PLACE ) {
	VovPlace place( n.id() );
	if ( ! (type2mask.integer() & (1<<place.getPlaceType())) ) {
	    if ( debug ) vovMessage( "not the right type2" );
	    return 0;
	}
	if ( rx.string() != SELECT_ALL ) {
	    String name = place.getName();
	    return  name.matches( Regex(rx.string()) );
	}
    } else if ( n.nodeType() == VOV_TRANSITION ) {
	VovTransition tran( n.id() );
	if ( rx.string() != SELECT_ALL ) {
	    String name = tran.getCommandLine();
	    String cwd = tran.getWd();
	    return  (name.matches( Regex(rx.string()) ) || cwd.matches( Regex( rx.string() ) )) ;
	    
	}
    } else {
	vovFatalError( "Illegal node type in select" );
    }
    if ( debug ) vovMessage( n.format() + " is selected" );
    return 1;			// 
}
#endif

String VovSelectRule::format()
{
    String s = form( "typemask: %04x, type2mask: %04x, statusmask: %04x, regex: %s",
		    typemask.integer(), type2mask.integer(), statusmask.integer(), (char*)rx.string() );
    return s;

}

