#include "vovinternal.h"
#include "trace.h"
#include "transition.h"
#include "place.h"
#include "ddsplus.h"
// #include "measure.h"
#include "annotation.h"
#include "voverror.h"
extern "C"
{
#include "errtrap.h"
}

extern void vovEditSet( VovTrace& trace, VovNodeSet& set );
extern VovStatus VOVedit( VovTrace& trace, VovNode& node );

void  test_chains(  VovTrace& trace )
{
    const int LENGTH = 10;
    
    VovPlace  p( VOV_BOOLEAN, "ANYPLACE_FOR_CHAIN" ) ;
    trace.findOrCreate( p );
    VovTransition tarray[20];
    for ( int i = 0 ; i < LENGTH ; i++ ) {
	String cmd = form("t%d",i );
	VovTransition t(cmd , cmd );
	trace.create( t );
	trace.initiate( t );
	trace.declareInput( t, p, 0 );
	String errmsg;
	trace.declareOutput( t, p , 1, 0, errmsg );
	trace.terminate( t, VOV_OK , 0, 0 );
	tarray[i] = t;
    }
    
    
    trace.changeStatus( p , VOV_PLAN_NODE, VOV_CHANGE_STATUS_SELF|VOV_CHANGE_STATUS_RECURSIVE );
    p.setCurrent();
    trace.find( p );
    
    VOV_ASSERT( trace.destroy( tarray[1] ) );
    VOV_ASSERT( trace.destroy( tarray[5] ) );
    VOV_ASSERT( trace.destroy( p ) );
}

void testtrace( VovTrace& project_trace , int testX )
{
    
    VovPlace p1(VOV_UNIX_FILE,"bitslice.bds");
    VOV_ASSERT( project_trace.findOrCreate(p1) );   	 
    
    VovTransition t1("bdsyn   bitslice.bds","bdsyn bitslice.bds");
    VovTransition t2("misII \t bitslice.blif  -s ccc djkd -ddd  ","misII bitslice.blif");
    
    VovPlace p2(VOV_UNIX_FILE,"bitslice.blif");
    
    VovPlace p3(VOV_OCT_FACET,"bitslice:logic");
    
    VovPlace pMSU(VOV_UNIX_FILE,"script.msu");
    VovPlace bool( VOV_BOOLEAN, "a boolean place" );
    
    VovPlace exec( VOV_EXECUTABLE, "pippo" );
    VovPlace opt( VOV_PROGRAM_OPTION, "-f -u _cFamily" );
    
    vovMessage("Initialization done");
    
    
    VOV_ASSERT( project_trace.findOrCreate( exec ) );
    VOV_ASSERT( project_trace.findOrCreate( bool ) );
    VOV_ASSERT( project_trace.findOrCreate( opt ) );
    VOV_ASSERT( project_trace.findOrCreate(p2) );
    
    VOV_ASSERT( project_trace.create(t1) );  		 
    VOV_ASSERT( project_trace.initiate(t1) );  		 
    VOV_ASSERT( project_trace.declareInput(t1,p1, 0) );	
    String errmsg;
    VOV_ASSERT( project_trace.declareOutput(t1,p2,1, 0, errmsg) );
    
    VOV_ASSERT( project_trace.terminate( t1, VOV_OK, 0, 0 ) ); 
    
    VOV_ASSERT( project_trace.create(t2) );
    VOV_ASSERT( project_trace.findOrCreate(p2) );
    VOV_ASSERT( project_trace.findOrCreate(p3) );	 vovMessage( p3.format() );
    VOV_ASSERT( project_trace.findOrCreate(pMSU) );    vovMessage( pMSU.format() );
    VOV_ASSERT( project_trace.declareInput(t2,p2,0) );	 vovMessage("first input");	
    VOV_ASSERT( project_trace.declareInput(t2,pMSU,0) ); vovMessage("second input");
    VOV_ASSERT( project_trace.declareOutput(t2,p3,1, 0, errmsg) );vovMessage("first output"+errmsg );
    
    VOV_ASSERT( project_trace.terminate( t2, VOV_OK, 0, 0 ) );  vovMessage( t2.format() );
    
    test_chains( project_trace );
    
    VovAnnotation note("This is a very nice note");
    VOV_ASSERT( project_trace.addAnnotation( p3, note ) );
    // if (testX )  {
    // VOVedit( project_trace, note );
    //}
    
    vovMessage( "USERS: " + project_trace.userFormat( 1 ) );
    
    VovAnnotation note1("This is another very nice note");
    VovAnnotation note2("This is a third  note");
    VovAnnotation note3("This is the last note");
    VOV_ASSERT( project_trace.addAnnotation( p3, note1 ) );
    VOV_ASSERT( project_trace.addAnnotation( p3, note2 ) );
    VOV_ASSERT( project_trace.addAnnotation( p3, note3 ) );
    VOV_ASSERT( project_trace.addAnnotation( p2, note2 ) );
    VOV_ASSERT( project_trace.addAnnotation( p2, note3 ) );
    
    
    
    if ( testX ) {
	vovShortMessage( "shortmessage" , "Editing: " + note.format() ) ;
	
	project_trace.edit( note1 );
	note1.edit();
	project_trace.edit( p3 );
	project_trace.edit( bool );
	project_trace.edit( opt );
	project_trace.edit( t1 );
    }
    
    int level = 1000;		// A lot of levels.
    String hist;
    int flags = VOV_NODE_MASK | VOV_PLACE_MASK | VOV_TRANSITION_MASK;
    int dir = VOV_BACKWARD;
    if (project_trace.history( p3, flags, dir, level, hist ) != VOV_OK) {
	vovError("Problems getting the history of a place");
    }
    vovMessage( "HISTORY:" + hist );
    
    
    VovSelectRule rule( VOV_PLACE_MASK, VOV_ALL_PLACE_TYPE_MASK, VOV_ALL_STATUS_MASK,  SELECT_ALL );
    // VovGoal diskUsageGoal( rule , "DISK_USAGE");
    
    //    VOV_ASSERT( project_trace.findOrCreate( diskUsageGoal ) );
    
    VOV_ASSERT( project_trace.getById( t1 ) );
    VOV_ASSERT( project_trace.getById( p1 ) );
    //     VOV_ASSERT( project_trace.getById( diskUsageGoal ) );
    
    
    project_trace.doPlacement();
    
    // project_trace.manageGoals();
    
    VovSelectRule selectRule;
    VovSelectRule selectOct( VOV_PLACE_MASK, (1 << VOV_OCT_FACET ), VOV_ALL_STATUS_MASK,  SELECT_ALL );
    VovSelectRule selSource( VOV_SOURCE_MASK, VOV_ALL_PLACE_TYPE_MASK, VOV_ALL_STATUS_MASK,SELECT_ALL );
    VovSelectRule selSink( VOV_SINK_MASK, VOV_ALL_PLACE_TYPE_MASK, VOV_ALL_STATUS_MASK,SELECT_ALL );
    
    VovNodeSet set( "ANY SET", selectRule );
    VovNodeSet setoct( "OCTFACETS", selectOct );
    VovNodeSet sources( "SOURCES", selSource );
    VovNodeSet sinks( "SINKS", selSink );
    
    VOV_ASSERT( project_trace.create( set ) );	
    VOV_ASSERT( project_trace.create( setoct ) );
    VOV_ASSERT( project_trace.create( sources ) );
    VOV_ASSERT( project_trace.create( sinks ) );
    
    VovNodeSet tset( "TRANSITIONS" );
    VOV_ASSERT( project_trace.findOrCreate( tset ) );
    VovSelectRule sel( VOV_TRANSITION_MASK, VOV_ALL_PLACE_TYPE_MASK, VOV_ALL_STATUS_MASK,SELECT_ALL );
    VOV_ASSERT( project_trace.getIntermediateNodes( sources, sinks, sel, tset ) );
    
    if ( testX ) {
	vovEditSet( project_trace, tset );
	vovMessage( tset.format() );
	
	project_trace.edit( p1 );
    }
    
    VOV_ASSERT( project_trace.setOperation( set, setoct, VOV_SET_UNION ) );
    VOV_ASSERT( project_trace.setOperation( set, setoct, VOV_SET_INTERSECTION ) );
    
    VovSelectRule ruleOct( VOV_PLACE_MASK, VOV_ALL_PLACE_TYPE_MASK, VOV_ALL_STATUS_MASK,  ".*contents" );
    VOV_ASSERT( project_trace.subselect( set, ruleOct ) );
    
    vovMessage("Saving the trace ...");
    VOV_ASSERT( project_trace.saveOntoDisk() );
    
    
    VOV_ASSERT( project_trace.destroy( p1 ) );
    VOV_ASSERT( project_trace.destroy( note ) );
    VOV_ASSERT( project_trace.destroy( t1 ) );
    
    if ( testX ) {
	project_trace.getQueue().edit();
    }
    
}


main( int argc, char** argv )
{
    int testX = 1;
    errCore( 1 );		// Dump core on error.
    String msg;
    
    vovRegisterMainArgs( argc, argv );
    
    vovShortMessage( "Starting test" , "test unmoded message" );
    VovTrace project_trace("badge","ic", "w", VOV_TRACE_TYPE_LOCAL);
    project_trace.OK( msg );
    vovMessage( msg );
    
    VovEventQueue queue(project_trace, 200);
    
    testtrace( project_trace ,  testX );
    
    for ( int i = 0 ; i < 50 ; i++ ) {
	project_trace.checkPlaces( 0 );
    } 

    vovMessage( "Exiting test" );
    octEnd();
    utcc_strfreeall();
    exit(0);

}

