#include "trace.h"
#include "travfunc.h"
#include "user.h"
#include "VOVprotocol.h"
#include "place.h"
#include "transition.h"
#include "annotation.h"
// #include "measure.h"		/*  */

#include "voverror.h"
#include "comm.h"


extern VovStatus vov_cycle( VovNode&, VovNode&);
extern void vovEditSet( VovTrace& trace, VovNodeSet& set );

VovStatus VovTrace::setOperation( VovNodeSet& set, VovNode& node, VovSetOperation op )
{
    if ( type == VOV_TRACE_TYPE_DISABLED ) return VOV_OK;
    VovStatus rtn = VOV_OK;
    Bzocket& zoc = getBzocket();
    
    if ( isClient() ) {
	zoc.pack( VOV_1SetOperation );
	zoc.pack( op );
	zoc.pack( set.id() );
	zoc.pack( node.id() );
	zoc.flush();
    } else {
	if ( isServer(VOV_SERVER_ENTRY) ) {
	    op = zoc.unpackInt();
	    set.id() = zoc.unpackInt();
	    node.id() = zoc.unpackInt();
	    rtn = getById( set );
	    if ( rtn == VOV_OK ) {
		rtn = getById( node );
	    }
	}
	if ( rtn == VOV_OK ) {
	    switch ( op )  {
	    case VOV_SET_ADD:
		set.add( node );
		break;
	    case VOV_SET_DEL:
		set.del( node );
		break; 
	    case VOV_SET_CONTAINS:
		if ( ! set.contains( node ) ) rtn = VOV_NOT_OK ;
		break;
	    default:
		vovBug("Illegal set operation" );
	    }
	}
	if ( isServer(VOV_SERVER_EXIT) ) {
	    zoc.pack( rtn );
	    zoc.flush();
	}
    }
    if ( isClient() ) {
	VC_get_reply( zoc );
	rtn = zoc.unpackInt();
    }
    return rtn;
}

VovStatus VovTrace::setOperation( VovNodeSet& acc, VovNodeSet& set, VovSetOperation op )
{
    if ( type == VOV_TRACE_TYPE_DISABLED ) return VOV_OK;
    Bzocket& zoc = getBzocket();
    VovStatus rtn = VOV_OK;
    
    if ( isClient() ) {
	zoc.pack( VOV_2SetOperation );
	zoc.pack( op );
	zoc.pack( acc.id() );
	zoc.pack( set.id() );
	zoc.flush();
    } else {
	if ( isServer(VOV_SERVER_ENTRY) ) {
	    op = zoc.unpackInt();
	    acc.id() = zoc.unpackInt();
	    set.id() = zoc.unpackInt();
	    rtn = getById( acc );
	    if ( rtn == VOV_OK ) {
		rtn = getById( set );
	    }
	}
	
	if ( rtn == VOV_OK ) {
	    switch ( op )  {
	    case VOV_SET_INTERSECTION:
		acc &= set;
		break;
	    case VOV_SET_UNION:
		acc |= set;
		break; 
	    case VOV_SET_COMPLEMENT:
		acc ^= set;
		break;
	    default:
		vovBug("Illegal set operation" );
	    }
	}
	if ( isServer(VOV_SERVER_EXIT) ) {
	    zoc.pack( rtn );
	    zoc.flush();
	}
    }
    if ( isClient() ) {
	VC_get_reply( zoc );
	rtn = zoc.unpackInt();
    }
    return rtn;
}


VovStatus VovTrace::subselect( VovNodeSet& set, VovSelectRule& rule )
{
    if ( type == VOV_TRACE_TYPE_DISABLED ) return VOV_OK;
    Bzocket& zoc = getBzocket();

    if ( isClient() ) {
	zoc.pack( VOV_SubSelect );
	set.pack( zoc );
	rule.pack( zoc );
	zoc.flush( );
    } else {
	if ( isServer(VOV_SERVER_ENTRY) ) {
	    set.unpack( zoc );
	    rule.unpack( zoc );
	}

	int ol = set.length( VOV_NODE_OCTMASK );
	set.subselect( rule  );
	int nl = set.length( VOV_NODE_OCTMASK );
	String msg = form( "Subselection of\n%s\nwith %s\nLength was %d, is %d.",
			  (char*)set.format(), (char*)rule.format(), ol, nl );
	getQueue().enq( VOV_NODE_SET_MASK, VE_CHANGE, msg, set.id(), 0 );

	if ( isServer(VOV_SERVER_EXIT) ) {
	    set.pack( zoc );
	    zoc.flush();
	}
    }
    if ( isClient() ) {
	VC_get_reply( zoc );
	set.unpack( zoc );
    }
    return VOV_OK;
}



