#include "user.h"
#include "oh.h"
#include "date.h"
#include "voverror.h"


extern String vov_prettyTime( int s );

VovUser::VovUser( String& name ) 
    :
    (OCT_INSTANCE),
    lastCreation("LAST", OCT_INTEGER ),
    lastConnection("LASTCONNECTION", OCT_INTEGER ),
    activeSet("ACTIVE")
{
    lastCreation = 0;
    lastConnection = 0;
    
    contents.instance.name = utcc_strsav_and_track( (char*)name );
    contents.instance.master = (char*)getCellName();
    contents.instance.view = "vov";
    contents.instance.facet = "contents";
    contents.instance.version = OCT_CURRENT_VERSION;
    contents.instance.transform.transformType = OCT_NO_TRANSFORM;
    contents.instance.transform.translation.x = 0;
    contents.instance.transform.translation.y = 0;
}

#ifndef VOVCLIENT
VovUser::VovUser( octId id ) 
    :
    (OCT_INSTANCE),
    lastCreation("LAST", OCT_INTEGER ),
    lastConnection("LASTCONNECTION", OCT_INTEGER ),
    activeSet("ACTIVE")
{
    objectId = id;
    if ( objectId ) {
	if ( getById( VOV_USER_OCTMASK ) != OCT_OK || get() != VOV_OK ){
	    vovError( "Not a user" );
	    objectId = 0;
	}
    }
}
#endif

VovUser::~VovUser()
{
}

#ifndef VOVCLIENT
VovStatus VovUser::create()
{
    OH_ASSERT( octCreate( this, &lastCreation ) );
    OH_ASSERT( octCreate( this, &lastConnection ) );
    OH_ASSERT( octCreate( this, &activeSet ) );
    doPlacement();
    return VOV_OK;
}
#endif

#ifndef VOVCLIENT
VovStatus VovUser::get()
{
    OH_ASSERT( octGetByName( this, &lastCreation ) );
    OH_ASSERT( octGetOrCreate( this, &lastConnection ) ); // Backwards compatibility.
    OH_ASSERT( octGetOrCreate( this, &activeSet ) );
    return VOV_OK;
}
#endif


#ifndef VOVCLIENT
VovStatus VovUser::connected()
{
    lastConnection.modify( getDate() );
    return VOV_OK;
}
#endif

#ifndef VOVCLIENT
VovStatus VovUser::active( VovTransition& trans )
{
    activeSet.add( trans );
    lastCreation.modify( getDate() );
    return VOV_OK;
}
#endif

String VovUser::getName()
{
    return contents.instance.name;
}

String VovUser::format( int extended = 0 )
{
    const int IDLE_THRESH = DATE_HOUR;
    const int MAX_TRANSITIONS_REPORTED = 5;
    
    Date now;
    int idle = now.get() - lastConnection.integer();
    int lastExecution = now.get() - lastCreation.integer();
   
    String s = form( "%-12s", (char*)getName() );
    
    if ( idle < IDLE_THRESH ) {
	s += " ACTIVE";
	if ( lastExecution < IDLE_THRESH ) {
	    s += " and growing the trace (" + vov_prettyTime( lastExecution ) + ")";
	}
    } else {
	s += " IDLE: " + vov_prettyTime( idle );
    }
    
    if ( extended ) {
#ifndef VOVCLIENT
	int extras = activeSet.length(VOV_TRANSITION_OCTMASK) - MAX_TRANSITIONS_REPORTED;
	VovContentsGenerator genTran( &activeSet, VOV_TRANSITION_OCTMASK );
	VovTransition transition = 0;
	String transName ;		// Short name for the transition.
	while ( octGenerate( &genTran, &transition ) == OCT_OK ) {
	    VOV_ASSERT( transition.get() );
	    VovNodeStatus st = transition.getStatus();
	    if ( extras-- > 0 && ( st!=VOV_TRACING_NODE && st!=VOV_RETRACING_NODE ) ) {
		activeSet.del( transition );
	    }
	    const int MAX_NAME_LENGTH = 65;
	    String transName = transition.getCommandLine();
	    if ( transName.contains("\n") ) transName = transName.before("\n");
	    if ( transName.length() > MAX_NAME_LENGTH ) {
		transName = transName.at(0,MAX_NAME_LENGTH-4 ) + "...";
	    }
	    s += form( "\n%10s: %s", (char*)transition.getStatusString(),(char*)transName);
	}
#endif
    }
    return s;
}

