/* -*- C++ -*- */
#ifndef VOVINTERNAL_H
#define VOVINTERNAL_H
#pragma once

// Order is important: oct.h before vov.h
//
#include "oct.h"
#include "vov.h"

extern class Bzocket;
extern class String;
extern class VovTrace;
extern class VovObject;
extern class VovProp;
extern class VovNode;
extern class VovNodeSet;
extern class VovPlace;
extern class VovTransition;
extern class VovAnnotation;
extern class VovRetrace;
extern class VovGoal;
extern class VovUser;
extern class VovEvent;
extern class VovEventQueue;
extern class VovEventFilter;
extern class VovDialog;
extern class VovDialogHandles;



/*
**  RETURN STATUS FROM VOV FUNCTIONS:
*/
typedef enum _vovStatus {
    VOV_OUTPUT_CONFLICT = -5,   /*  */
    VOV_INPUT_CONFLICT = -4,	/*  */
    VOV_CYCLE = -3,		/*  */
    VOV_NOT_FOUND = -2,		/*  */
    VOV_ERROR = -1 ,		/*  */
    VOV_NOT_OK = 0,		/*  */
    VOV_OK,			/*  */
    VOV_STOP,			/*  */
    VOV_CHANGE,			/*  */
    VOV_NO_CHANGE,		/*  */
    VOV_TIMEOUT,		/*  */
    VOV_DELETE_NODE		/*  */
  } VovStatus;
  

/*
**  STATUS OF NODES IN TRACE.
*/
typedef enum _nodeStatus {
    VOV_FIRST_STATUS_NODE = -2,	/*  Used only as a sentinel. */
    VOV_UNSPECIFIED_NODE = -1,	/*   */
    VOV_TRACE_NODE = 0,		/*  Valid transition or place. */
    VOV_PLAN_NODE,		/*  Invalid transition or place. */
    VOV_TRACING_NODE,		/*  Transition executing, and its outputs. */
    VOV_RETRACING_NODE,		/*  Transition retracing and its outputs. */
    VOV_DEAD_NODE,		/*  Inactive part of the trace. */
    VOV_NONEXISTING_NODE,	/*   */
    VOV_LAST_STATUS_NODE	/*  Used only as a sentinel. */
  } VovNodeStatus ;



#define VOV_INFO_TRACE  0x01
#define VOV_INFO_SLAVE  0x02
#define VOV_INFO_SERVER 0x04
#define VOV_INFO_DATE   0x08
#define VOV_INFO_USER   0x10
#define VOV_INFO_CWD    0x20
#define VOV_INFO_STATS  0x40
#define VOV_INFO_ALL    0x7F

#define VOV_HISTORY_PLACE_MASK 0x01
#define VOV_HISTORY_TRANS_MASK 0x02
#define VOV_HISTORY_NODE_MASK  0x04

#define VOV_ALL_PLACE_TYPE_MASK 0xffff

typedef long VovTime;

//
// Possible direction of search.
//
#define VOV_NO_SEARCH 	    -1
#define VOV_FORWARD   0
#define VOV_BACKWARD  1


#define VOV_ALL_STATUS_MASK 0xffff

// Flags to be used to change status of nodes.
#define VOV_CHANGE_STATUS_SELF         0x01
#define VOV_CHANGE_STATUS_RECURSIVE    0x02
#define VOV_CHANGE_STATUS_SELF_REC     0x03

// Retracing can be either SLOW (normal) or FAST (risky).
#define VOV_RETRACE_SLOW 0
#define VOV_RETRACE_FAST 1

typedef VovStatus (*VovStatusFunc)(VovNode&,void*);

typedef enum _vovType {
    VOV_UNDEFINED = -1,
    VOV_NODE = 0,		// These should not be changed.
    VOV_PLACE = 1,		// These should not be changed.
    VOV_TRANSITION = 2,		// These should not be changed.
    VOV_ANNOTATION,
    VOV_GOAL,	
    VOV_NODE_SET, 
    VOV_USER,
    VOV_TRACE,
    VOV_RETRACE,
    VOV_SLAVE,
    VOV_DETAIL
} VovType;


#define VOV_NULL_MASK         0
#define VOV_NODE_MASK         ( 1 << VOV_NODE )
#define VOV_PLACE_MASK        ( 1 << VOV_PLACE )
#define VOV_TRANSITION_MASK   ( 1 << VOV_TRANSITION )
#define VOV_GOAL_MASK         ( 1 << VOV_GOAL )
#define VOV_NODE_SET_MASK     ( 1 << VOV_NODE_SET )
#define VOV_ANNOTATION_MASK   ( 1 << VOV_ANNOTATION )
#define VOV_USER_MASK         ( 1 << VOV_USER )
#define VOV_TRACE_MASK        ( 1 << VOV_TRACE )
#define VOV_RETRACE_MASK      ( 1 << VOV_RETRACE )
#define VOV_SLAVE_MASK        ( 1 << VOV_SLAVE )
#define VOV_DETAIL_MASK	      ( 1 << VOV_DETAIL )
class VovSuperObject {
    int     allocated;
public:
    VovType type;
    union {
	VovObject*	object;
	VovNode*	node;
	VovPlace* 	place;
	VovTransition*	transition;
	// VovGoal*  	goal;
	VovNodeSet* 	set;
	VovUser*	user;
	VovAnnotation*  note;
	VovRetrace*     retrace;
    } contents;

    VovSuperObject();
    VovSuperObject( VovSuperObject& );

    VovSuperObject( VovPlace& x);
    VovSuperObject( VovTransition& x);
    VovSuperObject( VovNode& x);
    VovSuperObject( VovNodeSet& x);
    VovSuperObject( VovUser& x );
    VovSuperObject( VovAnnotation& x );
    VovSuperObject( VovRetrace& x );
    ~VovSuperObject();

    octId  getId();
    String format();
    void  setupDialog( VovDialog&, VovDialogHandles& data, int viewonly = 0 );
    VovStatus readDialog( VovDialogHandles& data );
    VovStatus get();
    void pack( Bzocket& );
    void unpack( Bzocket& );
};


extern String vov_getStatusString( VovNodeStatus status );
extern String vov_getPlaceTypeString( VovPlaceType type );

typedef VovStatus (*VovEventFilterFunction)( VovTrace&, VovEvent& ) ;


#define VOV_ASSERT( f ) {if((f)!=VOV_OK)vovFatalError(form("Assertion failed: %s,%d",__FILE__,__LINE__));}


#endif /* VOV_H */
