#include "port_internal.h"

#include "date.h"
#include "dateVHMap.h"
#include "utcc.h"

// static StringintVHMap* date_global_map = 0;	// Map used to manage dates on other hosts.

#include <time.h>


StringintVHMap date_global_map( 0, 30 );



TIME_T getDate()
{
    return time( (long)0 );
}

void  getDate( struct timeval& tp )
{
    // struct timezone tzp;
    // tzp.tz_minuteswest = 0;	// This is really ignored.
    // tzp.tz_dsttime = 0;		// This is too.
    if ( gettimeofday( &tp, 0 ) == -1 ) {
	utcc_printerrno( "gettimeofday" );
    }
}

double getDateInterval( struct timeval& tp )
{
    struct timeval t;
    getDate( t );
    return ( t.tv_sec - tp.tv_sec ) + ( t.tv_usec - tp.tv_usec ) / 1.0E6;
}

Date::Date()
{
    clock = getDate();
    host = 0;
}

Date::Date( Date& d )
{
    clock = d.clock;
    host = d.host ?  utcc_strsav( d.host ) : (char*)0;
}

Date::Date(TIME_T c)
{
    if (c < 0) {
	cerr.form("Time (%ld) cannot be negative\n",c);
	clock = getDate();
    } else {
	clock = c;
    }
    host = 0;
}

Date::~Date()
{
    if ( host ) delete host;
}

#ifdef FIXED_FUNNY_BUG
int Date::operator<( Date& d )
{
    Date adj= d.adjusted( host );
    return ( clock <  adj.clock ); 
}

int Date::operator>( Date& d )
{
    Date adj= d.adjusted( host );
    return ( clock >  adj.clock );
}

int Date::operator>=( Date& d )
{ 
    Date adj = d.adjusted( host );
    return ( clock >= adj.clock ); 
}

int Date::operator<=( Date& d )
{ 
    Date adj= d.adjusted( host );
    return ( clock <= adj.clock );
}

int Date::operator==( Date& d )
{ 
    Date adj= d.adjusted( host );
    return ( clock == adj.clock );
}

int Date::operator!=( Date& d )
{ 
    Date adj= d.adjusted( host );
    return ( clock != adj.clock );
}
#else
int Date::operator<( Date& d )
{
    return ( clock <  d.clock ); 
}

int Date::operator>( Date& d )
{
    return ( clock >  d.clock );
}

int Date::operator>=( Date& d )
{ 
    return ( clock >= d.clock ); 
}

int Date::operator<=( Date& d )
{ 
    return ( clock <= d.clock );
}

int Date::operator==( Date& d )
{ 
    return ( clock == d.clock );
}

int Date::operator!=( Date& d )
{ 
    return ( clock != d.clock );
}
#endif


void Date::operator=( Date& d)
{
    clock = d.clock;
    set( d.host );
}


String Date::format()
{
    if ( clock < DATE_REMOTE_PAST_THRESHOLD ) {
	return "------";	// Not a legal date.
    }
    String s = (char*)ctime( (TIME_T*)&clock );
    s = s.before( "\n" );

    if ( host ) {
	s += form("(%s)", host );
    }
    return s;
}

void Date::now()
{
    clock = getDate();
}

void Date::set(TIME_T c)
{
    if (c < 0) {
	cerr.form("Time (%ld) cannot be negative\n",c);
    } else {
	clock = c;
    }
}

void Date::set( const char* h )
{
    if (host) delete host;
    host = h ? utcc_strsav( (char*)h ) : 0;
}

// Convert the date to a date relative to the new host.
Date Date::adjusted( char* newhost )
{
    Date newdate;
    newdate.set( newhost );

    if ( String(host) == newhost ) {
	newdate.clock = clock;
    } else {
	int delta = date_global_map[newdate.getHost()] - date_global_map[getHost()];
	newdate.clock = clock + delta;
    }
    return newdate;
}


void Date::updateHostDelta( const char* h, int delta )
{
    String hostToUpdate = h ? h : DATE_DEFAULT_HOST;
    if ( date_global_map.contains( hostToUpdate ) ) {
	int prev_delta = date_global_map[hostToUpdate];
	if ( abs( prev_delta - delta ) > 5 ) {
	    cerr << ("Date on " + hostToUpdate + " is drifting very rapidly\n" );
	}
    }
    // cerr << ( "Date addHost: " + hostToUpdate + " " + dec( delta ) + "\n");
    date_global_map[hostToUpdate] = delta;
}

String Date::mapFormat()
{
    String s;
    
    s = "DATES: ";
    
    for (Pix i = date_global_map.first(); i != 0; date_global_map.next( i ) ) {
	String& onehost = date_global_map.key( i );
	if ( onehost == 0 ) onehost = DATE_DEFAULT_HOST;
	int delta = date_global_map.contents( i );
	s += ( onehost + "(" + dec( delta ) + ") " );
    }
    return s;
}

