#include "port_internal.h"
#include "utcc.h"
#include "voverror.h"
#include "ddsplus.h"

extern "C"
{
#include "errtrap.h"
}

static int debug = 0;		// Just a debug flag. Should be 0.


static int ddsinitialized = 0;

static int vovArgc = 0;
static char** vovArgv;
void vovRegisterMainArgs( int argc,  char** argv )
    // The arguments must be copied, because you never know how they 
    // have been generated  and if they will last very long.
{
    vovArgc = argc;
    vovArgv = new char*[argc+1];
    for (int i = 0 ; i < argc; i++ ) {
	argv;
	vovArgv[i] = utcc_strsav( argv[i] );
    }
    vovArgv[argc] = 0;		// Always terminate with 0.
}

int vovInitX()
{
    if ( ! ddsinitialized ) {
	if ( getenv( "DISPLAY" ) == 0 ) {
	    errRaise( "VOV", 1, "Enviroment variable DISPLAY variable is not set." );
	}
	if ( !vovArgc ) {
	    vovArgc = 1;
	    vovArgv[0] = "vov";
	}
	dds_initialize( &vovArgc, vovArgv );
	ddsinitialized = 1;

	extern int ddsReallyDestroyFlag; // This is temporary!!
	ddsReallyDestroyFlag = 1;	 // Will go away when dds/X11 are fixed.

    }
    return 1;
}

void vovPartiallyInitializedX()
{
    ddsinitialized = 1;
}

int vov_Xhandler( XEvent* evt, ddsData data )
{
    // vovMessage( "Extraneous event" );
    return  0;
}


// Main callback functions.

void vov_ctrl_ccb( ddsHandle item )
{
    VovDialogControl button( item );
    String& ctrl = *(String*)button.getData();
    ctrl = button.name();
    dds_destroy( item );
}

// Unpost, destroy a dialog.
// 
void vov_destroy_ccb( ddsHandle item )
{
    VovDialogControl ctrl( item );
    char* helpString = (char*)ctrl.getData();
    if ( helpString ) {
	delete helpString;
    }
    dds_destroy( item );
}

// Pop up an help dialog.
void vov_help_ccb( ddsHandle item )
{
    VovDialogControl ctrl( item );
    
    char* text = (char*)ctrl.getData();
    ddsComposite topDesc;
    topDesc.title = "HELP";
    topDesc.just = DDS_VERTICAL;
    ddsHandle top = dds_component( 0, DDS_TOP, (ddsData)&topDesc );
    
    ddsShowText textDesc;
    textDesc.where = DDS_FROM_STRING;
    textDesc.text = text;
    textDesc.rows = 20;
    textDesc.cols = 80;
    textDesc.word_wrap_p = DDS_FALSE;
    textDesc.compute_size_p = DDS_TRUE;
    dds_component( top, DDS_SHOW_TEXT, (ddsData)&textDesc );

    ddsControl ctrlDesc;
    ctrlDesc.name = "Dismiss";
    ctrlDesc.accelerator = 0;
    ctrlDesc.callback = vov_destroy_ccb;
    ctrlDesc.user_data = 0;
    dds_component( top, DDS_CONTROL, (ddsData)&ctrlDesc );

    ddsPosition pos;
    pos.pos = DDS_WINDOW;
    pos.win = dds_window( item );
    pos.vjust = DDS_MIDDLE;
    pos.hjust = DDS_RIGHT;
    pos.x = 0;
    pos.y = 0;
    dds_post( top, &pos , 0 );	// Post help dialog unmoded.
		
}


VovDialog::VovDialog()
{
    vovInitX();
    dialog = 0;
}


VovDialog::VovDialog( const String& title, ddsCompJust just, int items  )
{
    vovInitX();
    ddsComposite desc;
    desc.title = (char*)title;
    desc.just = just;
    desc.max_items = items;

    help = "";
    dialog = dds_component( 0, DDS_TOP, (ddsData)&desc );
}

VovDialog::VovDialog( VovDialog& top, const String& title, ddsCompJust just, int items  )
{
    vovInitX();
    ddsComposite desc;
    desc.title = (char*)title;
    desc.just = just;
    desc.max_items = items;

    help = "";
    dialog = dds_component( top.dialog, DDS_COMPOSITE, (ddsData)&desc );
}



VovDialog::~VovDialog()
{
    // Do not delete anything. The dialogs will be deleted
    // in the ccb function.
}


VovDialog::VovDialog( VovDialog& d)
{
    vovInitX();
    vovMessage("VovDialog(VovDialog&)");
    dialog = d.dialog;
}
void VovDialog::operator=( VovDialog& d)
{
    vovMessage("VovDialog::operator=");
    dialog = d.dialog;
}


//
void VovDialog::addControl( const String& label, VovCCB cb,  ddsData data , char* acc = 0 )
{
    VovDialogControl control( *this, label, acc, cb, data) ;
}

// By default, a control button calls a call back function which 
// registers the control button label.
void VovDialog::addControl( const String& label, char* acc = 0 )
{
    addControl( label, vov_ctrl_ccb, (ddsData)&ctrl, acc );
}


String VovDialog::getControl()
{
    return ctrl;
}

void VovDialog::addHelp( const String& text )
{
    help += text;
}



void VovDialog::post( int mode, int cancel )
{
    ddsEvtHandler dflt_handler;
    dflt_handler.other_handler = 0;
    dflt_handler.user_data = 0;

    char* helpString = 0;
    
    if ( help != 0 ) {
	helpString = utcc_strsav( (char*)help );
    }


    if ( mode == VOV_POST_MODED ) {
	dflt_handler.other_handler =  vov_Xhandler ;
    }
    
    if ( cancel ) addControl( mode == VOV_POST_MODED ? "Cancel" : "Dismiss",
			     vov_destroy_ccb, (ddsData)helpString, "Meta<Key>Delete" );
    
    if ( help != 0 ) {
	ddsControl desc;
	desc.name = "Help";
	desc.accelerator = "<Ctrl>h|<Key>Help";
	desc.callback = vov_help_ccb;
	desc.user_data = (ddsData)helpString;
	dds_component( dialog, DDS_CONTROL, (ddsData)&desc );
    } 
    
    
    ddsPosition pos;
    pos.pos = DDS_MOUSE;
    pos.hjust = DDS_CENTER;
    pos.vjust = DDS_MIDDLE;
    dds_post( dialog, &pos, &dflt_handler );
}

void VovDialog::postModed( int addCancel_p = 1 )
{
    post( VOV_POST_MODED, addCancel_p );
}

void VovDialog::postUnmoded( int addCancel_p = 1 )
{
    post( VOV_POST_UNMODED, addCancel_p );
}



// Special class for list of items.

VovDialogList::VovDialogList( VovDialog& t, ddsBoolean exclusive_p  )
{
    size = 0;
    count = 0;
    array = 0;
    suspended = 0;

    top = t.dialog;
    desc.exclusive_p = exclusive_p;
}

VovDialogList::~VovDialogList()
{
    clear();
    if ( size ) {
	delete array;
    }
}

int VovDialogList::length()
{
    return count;
}

void VovDialogList::addItem( const String& label, int data, int setFlag = 0 )
{
    if ( count >= size ) {
	int newsize = size * 2;
	if ( newsize < 8 ) newsize = 8; // We need a minimum size;
	resizeItems( newsize );
    }

    int i = count;
    array[i].label = utcc_strsav( (char*)label );
    array[i].selected_p = setFlag ? DDS_TRUE : DDS_FALSE ;
    array[i].user_data = (ddsData)data;
    if ( desc.exclusive_p == DDS_TRUE && setFlag ) {
	desc.selected = i;
    }
    count++;
}


void VovDialogList::clear()
{
    for ( int i = 0 ; i < count ; i++ ) {
	delete array[i].label;
	array[i].label = 0;
    }
    count = 0;
}

void VovDialogList::endOfList( VovListCCB cb, ddsData data, int max_items = 20, int scrollFlag = 1 )
{
    desc.num_items = count;
    desc.items = array;
    desc.scroll_p = ( scrollFlag && count > 20 ) ?  DDS_TRUE : DDS_FALSE;
    desc.max_items = max_items;
    desc.callback = cb;
    desc.user_data = data;

    if ( dialog == 0 ) {
	dialog = dds_component( top, DDS_ITEM_LIST,  (ddsData)&desc );
    } else {
	dds_set( dialog, (ddsData)&desc );
    }
}

void VovDialogList::get()
{
    dds_get( dialog, (ddsData)&desc ); 			// Get current status of dialog.
    for ( int i = 0 ; i < count ; i++ ) {
	array[i].selected_p = desc.items[i].selected_p;
    }
}


void VovDialogList::set()
{
    desc.num_items = count;
    desc.items = array;

    dds_set( dialog, (ddsData)&desc );		// Set new status of dialog.
}

void VovDialogList::modifyItem( String& newLabel, ddsData newData, int index )
{
    if ( ! suspended ) {
	get();
    }
    if ( index < 0 || index >= desc.num_items ) {	// Check bounds.
	vovError( "Index out of bound" );		// 
	return;						// 
    }							// 
    if ( array[index].label ) delete array[index].label; // Delete old label.
    array[index].label = utcc_strsav( (char*)newLabel ); // Change label.
    array[index].user_data = newData;			// 
    
    if ( ! suspended ) {				// 
	set();
    }
}

void VovDialogList::suspendRefresh()
{
    if ( ! suspended ) {
	suspended = 1;
	dds_get( dialog, (ddsData)&desc );
    }
}

void VovDialogList::refresh()
{
    if ( suspended ) {
	suspended = 0;
	set();
    }
}

void VovDialogList::doAll( ddsBoolean select = DDS_TRUE )
{
    get();
    for ( int i = 0 ; i < desc.num_items; i++ ) {
	array[i].selected_p = select ;
    }
    set();
}

ddsData VovDialogList::getData()
{
    get();
    return desc.user_data;
}


void VovDialogList::resizeItems( int newsize )
{
    if ( newsize <= size ) return; 			// Plenty of space.
    if ( size == 0 ) {					// If first allocation
	array = new ddsFlagItem[newsize];		// 
    } else {						// 
	if ( newsize < 2 * size ) {		// If resizing, resize by at least doubling.
	    newsize = 2 * size ;
	}
	ddsFlagItem* tmp = new ddsFlagItem[newsize];
	for ( int  i = 0 ; i < count; i++ ) {
	    tmp[i] = array[i];
	}
	delete array;
	array = tmp;
    }
    size = newsize;
}


int VovDialogList::selectedIndex()
{
    ddsItemList d;
    dds_get( dialog, (ddsData)&d );
    return d.selected;
}

ddsData VovDialogList::getWhichOne()
{
    if ( dds_type( dialog ) == DDS_ITEM_LIST ) {
	ddsItemList tmpDesc;
	dds_get( dialog, (ddsData)&tmpDesc );
	int i = tmpDesc.selected;
	// vovMessage( form( "Chosen is %d",i ) );
	return tmpDesc.items[i].user_data;
    } else {
	vovError( "getWhichOne: wrong type");
	return (ddsData)0;
    }
}

int VovDialogList::isSelected( int idx )
{
    ddsItemList tmpDesc;
    dds_get( dialog, (ddsData)&tmpDesc );

    if ( idx < 0 || idx >= tmpDesc.num_items ) {
	return 0;
    }
    return ( tmpDesc.items[idx].selected_p == DDS_TRUE );
}

ddsData VovDialogList::getItemData( int idx = -1 )
{
    ddsItemList tmpDesc;
    dds_get( dialog, (ddsData)&tmpDesc );
    if ( idx >=  0 && idx < tmpDesc.num_items ) {
	return tmpDesc.items[idx].user_data;
    } else {
	return tmpDesc.items[ tmpDesc.selected ].user_data;
    }
}


int VovDialogList::reduceOR()
{
    int mask = 0;
    ddsItemList tmpDesc;
    dds_get( dialog, (ddsData)&tmpDesc );
    for ( int i = 0 ; i < tmpDesc.num_items; i++ ) {
	if ( tmpDesc.items[i].selected_p == DDS_TRUE ) {
	    mask |= (int)tmpDesc.items[i].user_data;
	}
    }
    return mask;
}







VovDialogEditText::VovDialogEditText( VovDialog& top, const String& label, char* text, int rows, int cols)
{
    ddsEditText desc;
    desc.label = (char*)label;
    desc.text = text;
    desc.rows = rows;
    desc.cols = cols;
    dialog = dds_component( top.dialog, DDS_EDIT_TEXT, (ddsData)&desc );
}

VovDialogEditText::VovDialogEditText( VovDialog& top, const String& label, String& text, int rows, int cols)
{
    ddsEditText desc;
    desc.label = (char*)label;
    desc.text = (char*)text;
    desc.rows = rows;
    desc.cols = cols;
    dialog = dds_component( top.dialog, DDS_EDIT_TEXT, (ddsData)&desc );
}

char* VovDialogEditText::getTextItem()
{
    ddsEditText desc;
    dds_get( dialog, (ddsData)&desc );
    return desc.text;
}

void  VovDialogEditText::modifyText( String& text ) 
{
    modifyText( (char*)text );
}

void  VovDialogEditText::modifyText( char* text ) 
{
    ddsEditText desc;
    dds_get( dialog, (ddsData)&desc );
    desc.text = text;
    dds_set( dialog, (ddsData)&desc );
}



// Enumerate dialog.
VovDialogEnumerate::VovDialogEnumerate( VovDialog& t, String& l )
{
    size = 0;
    items = 0;
    count = 0;
    
    label = l;
    top = t.dialog;
}

VovDialogEnumerate::~VovDialogEnumerate()
{
    if ( count ) {
	for ( int i = 0; i < count ; i++ ) {
	    delete items[i];
	}
    }
    delete items;
}

void VovDialogEnumerate::resize( int newsize )
{
    if ( newsize <= size ) return; 			// Plenty of space.
    if ( size == 0 ) {					// If first allocation
	items = new char*[newsize];			// 
    } else {						// 
	if ( newsize < 2 * size ) {			// If resizing, resize by at least doubling.
	    newsize = 2 * size ;
	}
	char** tmp = new char*[newsize];
	for ( int  i = 0 ; i < count; i++ ) {
	    tmp[i] = items[i];
	}
	delete items;
	items = tmp;
    }
    size = newsize;
}

void VovDialogEnumerate::addItem( String item )
{
    resize( count + 1 );	// make sure we have space.
    items[count] = utcc_strsav( (char*)item );
    count++;
}

void VovDialogEnumerate::endOfList( VovListCCB cb, ddsData data )
{
    ddsEnumerate desc;
    desc.label = (char*)label;
    desc.num_items = count;
    desc.items = items;
    desc.selected = 0;
    desc.callback = cb;
    desc.user_data = data;
    dialog = dds_component( top, DDS_ENUMERATE, (ddsData)&desc );
}

ddsData VovDialogEnumerate::getData()
{
    ddsEnumerate desc;
    dds_get( dialog, (ddsData)&desc );
    return desc.user_data;
}

String VovDialogEnumerate::getItem( int i = -1 )
{
    ddsEnumerate desc;
    dds_get( dialog, (ddsData)&desc );
    
    if ( i >= 0 && i < desc.num_items ) {
	return desc.items[ i ];
    } else {
	return desc.items[ desc.selected ];
    } 
}


// Show Text dialog.

VovDialogShowText::VovDialogShowText( VovDialog& top, const String& label, const String& text, int rows, int cols )
{
    // Show text does not have a label.
    // Get around this by concatenating label and text.
    String longText = text;
    if ( rows == 1 ) {
	longText = label + ": " + text;
    }
    ddsShowText desc;
    desc.where = DDS_FROM_STRING;
    desc.rows = rows;
    desc.cols = cols;
    desc.text = (char*)longText;
    desc.word_wrap_p = DDS_FALSE;
    desc.compute_size_p = DDS_FALSE;

    dialog = dds_component( top.dialog, DDS_SHOW_TEXT, (ddsData)&desc );
}

VovDialogShowText::VovDialogShowText( VovDialog& top, const String& label, const String& text )
{
    // Show text does not have a label.
    // Get around this by concatenating label and text.
    ddsShowText desc;
    desc.where = DDS_FROM_STRING;
    desc.rows = 24;
    desc.cols = 80;
    desc.text = (char*)text;
    desc.word_wrap_p = DDS_FALSE;
    desc.compute_size_p = DDS_TRUE;

    dialog = dds_component( top.dialog, DDS_SHOW_TEXT, (ddsData)&desc );
}

VovDialogShowText::VovDialogShowText( VovDialog& top, const String& filename )
{
    ddsShowText desc;
    desc.where = DDS_FROM_FILE;
    desc.rows = 20;
    desc.cols = 80;
    desc.text = (char*)filename;
    desc.word_wrap_p = DDS_FALSE;
    desc.compute_size_p = DDS_FALSE;

    dialog = dds_component( top.dialog, DDS_SHOW_TEXT, (ddsData)&desc );
}

void  VovDialogShowText::modifyText( const String& text ) 
{
    ddsShowText desc;
    dds_get( dialog, (ddsData)&desc );
    desc.text = (char*)text;
    dds_set( dialog, (ddsData)&desc );

    // Debugging attempt.
    ddsShowText d;
    dds_get( dialog, (ddsData)&d );
    String s = d.text;
    if ( s != text ) {
	vovMessage( "Usual minor bug in DDS modifyText detected again." );
    }
}

VovDialogNumber::VovDialogNumber(VovDialog& top, String& label, double min, double max, double inc, double value, char* format, VovCCB cb, ddsData data )
{
    ddsNumber desc;
    desc.label = (char*)label;
    desc.minimum = min;
    desc.maximum = max;
    desc.increment = inc;
    desc.value = value;
    desc.format = format;
    desc.callback = cb;
    desc.user_data = data;
    dialog = dds_component( top.dialog, DDS_NUMBER, (ddsData)&desc );
}

double VovDialogNumber::getValue()
{
    ddsNumber desc;
    dds_get( dialog, (ddsData)&desc );
    return desc.value;
}
 
ddsData VovDialogNumber::getData()
{
    ddsNumber desc;
    dds_get( dialog, (ddsData)&desc );
    return desc.user_data;
}

void VovDialogNumber::modifyValue( double newvalue )
{
    ddsNumber olddesc, newdesc;
    dds_get( dialog, (ddsData)&olddesc );
    newdesc = olddesc;
    newdesc.value = newvalue;
    dds_set( dialog, (ddsData)&newdesc );
}

void VovDialogNumber::modifyBounds( double  min, double max )
{
    if ( min >= max ) {
	vovError( "min should be less than max" );
	return;
    }
    ddsNumber olddesc, newdesc;
    dds_get( dialog, (ddsData)&olddesc );
    newdesc = olddesc;
    newdesc.minimum = min;
    newdesc.maximum = max;
    if ( newdesc.value < min ) {
	newdesc.value = min;
    }
    if ( newdesc.value > max ) {
	newdesc.value = max;
    }
    dds_set( dialog, (ddsData)&newdesc );
}


// Control dialog.

VovDialogControl::VovDialogControl( VovDialog& top, const String& label, char* acc, VovCCB cb, ddsData data )
{
    ddsControl desc;
    desc.name = (char*)label;
    desc.accelerator = acc;
    desc.callback = cb;
    desc.user_data = data;
    dialog = dds_component( top.dialog, DDS_CONTROL, (ddsData)&desc );
}

VovDialogControl::VovDialogControl( ddsHandle item )
{
    dialog = item;
}

ddsData VovDialogControl::getData()
{
    ddsControl desc;
    dds_get( dialog, (ddsData)&desc );
    return desc.user_data;
}

String VovDialogControl::name()
{
    ddsControl desc;
    dds_get( dialog, (ddsData)&desc );
    return desc.name;
}

void VovDialogControl::modifyName( const String& label )
{
    ddsControl desc;
    dds_get( dialog, (ddsData)&desc );
    desc.name = (char*)label;
    dds_set( dialog, (ddsData)&desc );
}


// Utilities:

String vovConfirm( String& title, String& text, String& o1, String& o2 )
{
    VovDialog top( title, DDS_VERTICAL, 3 );
    VovDialogShowText textDlg( top, "", text, 10, 60 );
    top.addControl( o1 );
    if ( o2 != 0 ) top.addControl( o2 );
    top.postModed( 0 );
    return top.getControl();
}

void vovLongMessage( String& title, String& msg, int rows, int cols )
{
    VovDialog top_dialog( title, DDS_VERTICAL, 1 );
    VovDialogShowText dialog( top_dialog, "", msg, rows, cols );
    top_dialog.postUnmoded();
}

void vovShortMessage( String& title, String& msg )
{
    VovDialog top_dialog( title, DDS_VERTICAL, 1 );
    VovDialogShowText dialog( top_dialog, "", msg );
    top_dialog.postUnmoded();
}



VovDialogHandles::VovDialogHandles( int newsize )
{
    size = newsize;
    array = new ddsHandle[size];
    count = 0;
    index = 0;
}

VovDialogHandles::~VovDialogHandles()
{
    delete array;
}

void VovDialogHandles::add( ddsHandle handle )
{
    if ( count < size ) {
	array[count++] = handle;
    } else {
	vovError( "VovDialogHandles overflow" );
    }
}

ddsHandle VovDialogHandles::next()
{
    if ( count && index < count ) {
	return array[index++];
    } else {
	vovError( "VovDialogHandles::next()" );
	return 0;
    }
}
