// This may look like C code, but it is really -*- C++ -*-
#ifndef DDSPLUS_H
#define DDSPLUS_H
#pragma once

extern "C" 
{
#ifndef PORT_H
#define PORT_H 1		// Avoid loading port.h
#endif
#include "dds.h"
}

#define VOV_POST_MODED   0
#define VOV_POST_UNMODED 1


extern String vovConfirm( String& title, String& text, String& o1, String& o2 );
extern void vovLongMessage( String& title, String& msg, int rows, int cols );
extern void vovShortMessage( String& title, String& msg );

extern int useXinterface( int status = -1 );

extern void vovRegisterMainArgs( int argc, char** argv );
extern int vovInitX();
extern void vovPartiallyInitializedX();

// Two types of callback functions:
// One for control buttons, one for lists.
typedef void (*VovCCB)( ddsHandle );
typedef void (*VovListCCB)( ddsHandle, int );


class VovPostedDialog {
    String help;
    String ctrl;
public:
    VovPostedDialog( String& h ) { help = h; }
    
    
};

// This is the composite dialog.
class VovDialog {
    ddsHandle dialog;		// Pointer to the actual dds dialog.
    String help;		// The help message.
    String ctrl;		// The control button label.
    void post( int moded , int addCancel_p );
 public:
    VovDialog();
    VovDialog( ddsHandle handle ) { vovInitX(); dialog = handle; }
    VovDialog( const String& title, ddsCompJust just, int items ); 	// TOP level dialog.
    VovDialog( VovDialog& dialog, const String& title, ddsCompJust just, int items ); // COMPOSITE dialog.
    VovDialog( VovDialog& );
    void operator=( VovDialog& );
    ~VovDialog();

    void get( ddsHandle handle ) { dialog = handle; }
    ddsHandle handle() { return dialog; }

    void addControl( const String& label, char* acc = 0 );
    void addControl( const String& label, VovCCB cb, ddsData data, char* acc = 0 );
    void addHelp( const String& text );

    void postModed( int addCancel_p = 1 );
    void postUnmoded( int addCancel_p = 1 );
    int  posted();		// Is the dialog still posted ?

    String getControl();

    friend class VovDialogList;
    friend class VovDialogEditText;
    friend class VovDialogEnumerate;
    friend class VovDialogShowText;
    friend class VovDialogNumber;
    friend class VovDialogControl;
};

class VovDialogList : public VovDialog {
    ddsItemList desc;
    ddsFlagItem* array;
    int count;
    int size;
    int suspended;
    ddsHandle top;
    ddsBoolean exclusive;
    void resizeItems( int newsize );
public:
    VovDialogList(VovDialog& top, ddsBoolean exclusive  );
    VovDialogList( ddsHandle handle ) : ( handle ) { size = 0; array = 0; count = 0;};
    ~VovDialogList();
    int  length();

    void get();
    void set();
    void   clear();
    void   addItem( const  String& label, int data, int setFlag = 0 );
    void   endOfList( VovListCCB, ddsData, int max_items = 20, int scrollFlag = 1 );

    void modifyItem( String& label, ddsData value, int index );

    void doAll( ddsBoolean select = DDS_TRUE );

    int    selectedIndex();
    int    isSelected( int index  = -1 );
    ddsData    getItemData( int index = -1 );
    ddsData    getWhichOne();
    String getTextItem( const String& label);

    ddsData getData();

    int reduceOR();

    void suspendRefresh();	// So we can do many modify items at once and quickly.
    void refresh();
};

class VovDialogEditText : public VovDialog {
public:
    VovDialogEditText( VovDialog& top, const String& label, char* text, int rows, int cols );
    VovDialogEditText( VovDialog& top, const String& label, String& text, int rows, int cols );
    VovDialogEditText( ddsHandle handle ) : ( handle ) {};
    void modifyText( char* text );
    void modifyText( String& text );
    char* getTextItem();

};

class VovDialogEnumerate: public VovDialog {
    int size;
    char** items;
    int count;
    ddsHandle top;
    String label;
    void resize( int newsize );
public:
    VovDialogEnumerate( VovDialog& top, String& label );
    VovDialogEnumerate( ddsHandle handle ) : (handle) { size = 0; count = 0; items = 0; }
    ~VovDialogEnumerate();

    void addItem( String item );
    void endOfList( VovListCCB cb, ddsData data );

    ddsData getData();
    String  getItem( int i = -1 );
};

class VovDialogShowText : public VovDialog {
public:
    VovDialogShowText( VovDialog& top, const String& label, const String& text, int rows, int cols );
    VovDialogShowText( VovDialog& top, const String& label, const String& text );
    VovDialogShowText( VovDialog& top, const String& filename );
    VovDialogShowText( ddsHandle handle ) : ( handle ) {};
    void modifyText( const String& text );
};

class VovDialogNumber : public VovDialog {
public:
    VovDialogNumber( VovDialog& top, String& label, 
		    double min, double max, double inc, double value,
		    char* format, VovCCB cb, ddsData data );
    VovDialogNumber( ddsHandle handle ) : ( handle ) {};
    ddsData getData();
    double getValue();
    void modifyValue( double newvalue );
    void modifyBounds( double min, double max );
};

class VovDialogControl : public VovDialog {
public:
    VovDialogControl( VovDialog& top, const String& label, char* acc, VovCCB cb, ddsData data );
    VovDialogControl( ddsHandle item );
    String name();
    void modifyName( const String& label );
    ddsData getData();
};



class VovDialogHandles {
    ddsHandle* array;
    int size;
    int count;
    int index;
 public:
    VovDialogHandles( int newsize );
    ~VovDialogHandles();
    void add( VovDialog& dialog) { add( dialog.handle() ); }
    void add( ddsHandle handle );
    void reset() { index = 0; }
    ddsHandle next();
};
#endif
