#include "stream.h"
#include "date.h"
#include "String.h"
#include "path.h"
#include "utcc.h"
#include "dialog.h"
#include "ddsplus.h"
#include "voverror.h"

void testAllPathFunctions( String& path )
{
    cout << "\nINITIAL PATH:" << path ;
    cout << "\nIS DIRECTORY:" << path_isDirectory( path );
    cout << "\nIS SYMBOLIC LINK:" << path_isSymbolicLink( path );
    cout << "\nREMOVEDOTS:" << path_removedotdot( path );
    cout << "\nLOCAL FULL:" << path_localFull( path );
    cout << "\nREMOVE SYMLINKS:" << path_removeSymbolicLinks( path );
    cout << "\nLAST COMPONENT:" << path_lastComponent( path );
    // cout << "\nHOST:" << path_host( path );
    cout << "\nEXPAND:" << path_expand( path );
    cout << "\nFINAL:" << path_canonicalize( path );
    cout << "\n";
    fflush( stdout );
}

void testBigDialog( ddsCompJust just)
{
    VovDialog top( "BIG DIALOG", just, 5 );

    VovDialogShowText file( top, "/etc/motd" );


    VovDialogEnumerate e( top, "ENUM DIALOG" );
    e.addItem( "AAA" );
    e.addItem( "BBB" );
    e.addItem( "CCC" );
    e.addItem( "DDD" );
    e.endOfList( 0, 0 );
    


    VovDialogList list( top, DDS_TRUE )   ;
    list.addItem( "DDDD", 1, 1 );
    list.addItem( "FFFF", 2, 1 );
    list.addItem( "HHHHH", 2, 1 );
    list.endOfList( 0 ,  0 );



    top.addHelp( "silly help" );
    top.addControl( "Little control" );
    top.postModed();
}

main( int argc, char** argv)
{
    int N = 1;
    if ( argc >= 2 ) {
	 N = atoi( argv[1] );
     }


    utcc_cat( "test.cc" );

    vovMessage( utcc_which( "cc" ) );
    vovMessage( utcc_which( "util-test" ) );
    vovMessage( utcc_which( "g++" ) );
    vovMessage( utcc_which( "nosuchprogram" ) );


    int testX = 1;
    if ( testX ){
	vovRegisterMainArgs( argc, argv );
	if ( vovConfirm( "USER DECISION", "Test X ? " , "Yes", "No" ) == "No" ) {
	    testX = 0;
	} else {
	    vovShortMessage("Message", "Testing X and dds" );
	    testBigDialog( DDS_VERTICAL );
	    testBigDialog( DDS_HORIZONTAL );
	}

    }


    int testEnv = 1;
    if ( testEnv ) {
	for ( int i = 0 ; i < N ; i++ ) {
	    // Test manipulation of environment variables.
	    util_printenv( stderr );
	    util_setenv( "FOOO", "fooo" );
	    util_setenv( "PIPPO", "pippo" );
	    util_printenv(stderr);
	    util_unsetenv( "PIPPO" );
	    util_setenv( "FOOO", "bar" );
	    util_printenv(stderr);
	}
	
    }    

    

    if ( testX ) {
	userDecision( "test the dialog" , "option1", "option2" );
	String t = "this is some text" ;
	cout << "Result:" << t << "\n";
	if (userDecision( "Do you want to continue the test" , "Yes", "No" ) == 2) {
	    exit( -1 );
	}
    }
    // Test dates.
    
    int testDate = 1;
    if ( testDate ) {
	for ( int j = 0; j < N ; j++ ) {
	    Date d;
	    Date d1((long)304);
	    Date bb;
	    
	    bb.set((long)0);
	    d1.set("orca");
	    
	    cout << bb.format() << "\n";
	    cout << d.format() << "\n";
	    cout << d1.format() << "\n";
	    d1.now();
	    cout << d1.format() << "\n";
	    
	    cout << d1.operator<( bb ) << " ";
	    cout << (int)(d1 < bb) << " ";
	    cout << (int)(d1 > bb) << " ";
	    cout << (d1 == bb) << " ";
	    cout << (d1 != bb) << " ";
	    cout << (d1 >= bb) << " ";
	    cout << (d1 <= bb) << " ";
	    cout << "\n";
	    
	    d1.set( "canova" );
	    d1.updateHostDelta( "canova", 55 );
	    bb.updateHostDelta( "orca",   33 );
	    
	    cout << d1.adjusted( "orca" ).format() << "\n";
	    cout << d1.adjusted( "dummyhost" ).format() << "\n";
	    
	    cout << d1.mapFormat() << "\n";
	}
    }    
    int testpath = 1;
    if ( testpath ) {
	for ( int j = 0 ; j < N ; j++ ) {
	    testAllPathFunctions( "/net/xxx/../aaa/./bbb/./ttt/../../eee" );
	    testAllPathFunctions( "/./././sss/ddd/../../www" );
	    testAllPathFunctions( "/usr/local/mac" );
	    testAllPathFunctions( "~/src/Vov" );
	    testAllPathFunctions( "/users/casotto/projects/vov/develop" );
	    testAllPathFunctions( "/users/casotto/projects/vov/develop/util/test.cc" );
	    testAllPathFunctions( "../vovinclude/vov.h" );
	}
    }
    utcc_strfreeall();
    exit( 0 );
}

