/* -*- C++ -*- */
#ifndef BASE_H
#define BASE_H
#pragma once

#include <String.h>
#include "utcc.h" 
#include "oct.h"
#include "vovinternal.h"

extern "C"
{
#include "oh.h"
}

extern class Bzocket;		// 
extern class VovDialog;		// 
extern class VovAnnotation;	// 

/*
 * BASE CLASS: VovObject
 */
class VovObject : public  octObject {
  public:
    VovObject();
    VovObject( int t );				// 	
    VovObject( VovObject& o );			// 
    ~VovObject();				// 
    void operator = ( VovObject& o ) ;		// 
    octId& id() { return objectId; }		// 
    octStatus getById( int typemask = OCT_ALL_MASK ); // Get with type checking.

    String format();

    void      addNote( VovAnnotation& note );
    int	      countNotes();
    int       getNotes( VovAnnotation* noteArray, int length );
    void      copyNotes( VovObject& );
    void      deleteNotes();
    String    formatNotes();

    String    getName();	// 

    VovStatus lock( String& desc, int force, octId& lockId, String&  msg );
    VovStatus unlock( octId lockId );

    void pack( Bzocket& );	// Pack object for socket transmission.
    void unpack( Bzocket& );	// 
};

class VovBag : public VovObject {
 public:
    VovBag( char* name ) ;
    VovBag( VovBag& );
    ~VovBag();
    void operator |= ( VovBag& ); // UNION
    void operator &= ( VovBag& ); // INTERSECTION
    void operator ^= ( VovBag& ); // COMPLEMENT of 'this' in set
    int length( int );
    int contains( VovObject& );
    void add( VovObject& );
    void del( VovObject& );
    void clear( int );

    String format();
};

class VovProp : public VovObject {
 public:
    VovProp( char* name , octPropType t ) : (OCT_PROP)
      { 
	  contents.prop.name = name; 
	  contents.prop.type = t;
	  contents.prop.value.string = "<no value>";
      }
    VovProp( VovProp& );
    ~VovProp();

    void operator=( VovProp& p )
      {
	  VovObject::operator=( p );
      }

    void operator=( String& s ) 
      {
	  contents.prop.type = OCT_STRING;
	  contents.prop.value.string = utcc_strsav_and_track( (char*)s );
      }
    void operator=( int i ) 
      {
	  contents.prop.type = OCT_INTEGER;
	  contents.prop.value.integer = i;
      }
    void operator=( double r ) 
      { 
	  contents.prop.type = OCT_REAL;
	  contents.prop.value.real = r;
      }

    void modify( String& s );
    void modify( int i );
    void modify( double r );

    String format();
    String string();
    int    integer();
    double real();
};

class VovGenerator : public octGenerator {
public:
    void pack( Bzocket& );
    void unpack( Bzocket& );
};

class VovContentsGenerator : public VovGenerator {
 public:
 VovContentsGenerator( VovObject* o, octObjectMask mask )
   { if ( octInitGenContents(o,mask,this)!=OCT_OK) utcc_fatalerror("init gen contents") ; }
 ~VovContentsGenerator()
   { if(octFreeGenerator(this)!=OCT_OK) utcc_fatalerror("free gen contents"); }
};

class VovContainersGenerator : public octGenerator {
 public:
 VovContainersGenerator( VovObject* o, octObjectMask mask )
   { if(octInitGenContainers(o,mask,this)!=OCT_OK) utcc_fatalerror("init gen container"); }
 ~VovContainersGenerator()
   { if (octFreeGenerator(this)!=OCT_OK) utcc_fatalerror("free gen container"); }
};
#endif
