/* -*- C++ -*- */
//
// Buffer: used in bzockets for bidirectional communication.
//
#ifndef BUF_H
#define BUF_H
#pragma once


#define DEFBUFSIZE 4096		/* Default buffer size */
#define MAXBUFSIZE (1<<20)	/* Maximum buffer size (1Mbyte)*/

/* Return codes */
#define BUF_TIMEOUT 1
#define BUF_OK      0
#define BUF_ERROR  -1

/* To determine byte swapping */
// #define SWAP_TEST_VECTOR 0x00010203  /* A swap variant integer */
// #define SWAP_CODE        0xEEEEEEEE  /* A swap invariant word */

class Buf {
    char* buf;			/* The buffer */
    int   size;			/* Size of the buffer */
    char* rptr;			/* Read  pointer */
    char* wptr;			/* Write pointer */
    int   cnt;			/* Counts unread bytes available in the buffer */
    char* pptr;			/* Beginning of next packet */
    int   debug;		/* If 1, then print what is happening on stderr. */
    // int   swapMask;		/*  */
 public:
    Buf();			/* Constructor. */
    Buf(Buf&);			/* Constructor. */
    ~Buf();			/* Destructor. */
    // int  setSwapMask(int);	/* To do correct byte swapping */
    // int  swapBytes(int);	/*  */
    int  resize(int);		/* Make sure the buffer is large enough */
    void freeMemory();		/* Delete buffer */
    void clear();		/* Clear buffer */
    void cleanup();		/* Discard characters already read */
    int  put(int);	        /* Add an integer at the end of the buffer */
    int  put(char*, int);	/* Add a string of chars at the end of the buffer */
    char* get(int);		/* Get some bytes from the buffer */
    int   getInt();		/*  Fragile. */
    int   length();		/* Get number of chars in the buffer. */
    char* getPacket( int* );	/*  */
    int  unget(int);		/* Undo get for some number of characters */
    int  sendSocket(int);	/* Send buffer onto the socket */
    int  readSocket(int);	/* Get bytes from socket  */
    void pad( int );		/* Add padding to allow 4 byte alignement of ints */
    int  packetReady();		/* Is there a ready packet in the buffer ? */
    void print();		/*  */
    void setDebug( int d )	/*  */
      { debug = d; }		/*  */
};


#endif /* BUF_H */


