// This may look like C code, but it is really -*- C++ -*-

/*
* Bidirectional communication using sockets.
*/
#ifndef BZOCKET_H
#define BZOCKET_H
#pragma once

#include <String.h>
#include "buf.h"


class Bzocket {
    int   on;			/* 1 active 0 inactive */
    int   fd;			/* file descriptor */
    int   server;		/* 1 is the server, 0 is the client */
    Buf   inBuf;		/* Input buffer, to hold data from socket */
    Buf   outBuf;		/* Output  buffer, to hold packet to be sent out */
    int   debug;		/* If 1, then print log of what is happening */
    String errorString;		/* Description of last error,if any. */
 public:
    Bzocket( int srv = 1 );
    Bzocket( Bzocket& );
    ~Bzocket();
    // Start, end and test connection.
    void       start(int);	/* Start the connection */
    // int        setSwapMask(int);/*  */
    virtual void  close();	/* Close connection socket */
    int	       check();		/* check that connection is still good */
    int        active() { return on; }
    int	       connectionNumber() { return fd; }
    //  Functions to receive.
    int	       waitForPacket( int sec = 0 , int usec = 0 );
    int        read();		/* Read as much as you can from socket */
    int        ready();		/* Is there a full message ready ? */

    int        unpack(int*);	/* Robust. */
    int	       unpack(char**);	/* Robust. */
    int	       unpackInt();	/* Fragile */
    char*      unpackStr();	/* Fragile */
    String     unpackString();	/* Fragile */
    int	       discardPacket();	/*  */

    // Functions to send.
    void       clear();		/* clear current packet */
    void       pack(int);	/* pack an integer */
    void       pack(char*);	/* pack a null terminated string */
    void       pack( String& );	/* Pack a string  */
    void       flush();		/* flush current packet */

    // Debugging.
    void       print();
    void       setDebug( int d ) { debug = d; }
};


#endif /* BZOCKET_H */


