/* -*- C++ -*- */
#ifndef DATE_H
#define DATE_H
#pragma once


#include "String.h"
#include "time.h"

/* Portability mess. The type of time changes 
** depending on the machine. 
*/
#if defined(sun4) || defined(sun3)
#define TIME_T long
#else 
#define TIME_T time_t
#endif


#define DATE_MINUTE	60
#define DATE_HOUR	3600
#define DATE_DAY	(24*3600)
#define DATE_WEEK	(7*24*3600)

#define DATE_REMOTE_PAST_THRESHOLD (DATE_WEEK * 52)

#define DATE_DEFAULT_HOST "<local>"

extern class StringintVHMap;

class Date {
    TIME_T   clock;		// Seconds since Jan 1 1970.
    char*  host;		// Date is relative to each host.
 public:
    Date();			// Default constructor.
    Date(TIME_T);			// 
    Date(Date&);		// 
    ~Date();
    void set(TIME_T);		// Set the clock.
    void set( const char*);	// Set the host.
    void now();			// On the local host.

    inline TIME_T get()		// 
      { return clock;}		// 
    inline String getHost()	// 
      { return host ? host : DATE_DEFAULT_HOST; }		// 

    String format();		// 

    int operator<( Date& d );
    int operator>( Date& d );
    int operator<=( Date& d );
    int operator>=( Date& d );
    int operator==( Date& d );
    int operator!=( Date& d );
    void operator=( Date& d);

    Date adjusted( char* h ); // Time adjusted for host 'h' given in argument.
    void updateHostDelta( const char*, int );
    String mapFormat();

};

extern TIME_T getDate();
extern void getDate( struct timeval& );
extern double getDateInterval( struct timeval& );
#endif /* DATE_H */
