/* -*- C++ -*- */
#ifndef EVENT_H
#define EVENT_H
#pragma once


#include "String.h"
#include "vovinternal.h"
#include "bzocket.h"
#include "date.h"

extern class VovTrace;

extern int vov_nextEventSerialNumber();


// Possible Event
#define VE_NULL_EVENT   0x0000
#define VE_START	0x0001
#define VE_END		0x0002
#define VE_CONNECT	0x0004
#define VE_DISCONNECT	0x0008
#define VE_CHANGE	0x0010
#define VE_FORGET       0x0020
#define VE_ERROR	0x1000

class VovEvent {
    int 	serialNumber;		// Number of the event.
    Date 	date;			// Date in which event happened.
    int		subjectmask;		// What we are talking about.
    int 	eventmask;		// 
    int 	actionMask;		// Expected actions for this event.  (-- not used ).
    String 	description;		// Textual description of the event.
    int 	id;			// id of the object involved in the event.
    int		obsolete;		// Flag. Another event happened later for the same subject.
 public:
    VovEvent();
    VovEvent( int sm, int em, String& desc, int id, int am );
    VovEvent( VovEvent& );
    ~VovEvent();
    void operator = ( VovEvent& );
    void set( int sm, int em, String& desc, int id, int am);

    int getSerialNumber(){ return serialNumber; }
    int& getId(){ return id; }
    int& getSubjectMask() { return subjectmask; }
    int& getEventMask() { return eventmask; }
    String& getDescription() { return description; }

    String format();

    void pack( Bzocket& );
    void unpack( Bzocket& );
    
    friend class VovEventFilter;
    friend class VovEventQueue;
};

class VovEventQueue {
    VovTrace*  eqtrace;		// Trace the events belong to.
    VovEvent** array;		// Array of VovEvent*.
    int size;			// Size of the array.
    int count;			// Actual elements in array.
    int writeIndex;		// Next element to write.
    int readIndex;		// Next element to be read.
public:
    VovEventQueue( );
    VovEventQueue( VovTrace& t );
    VovEventQueue( VovTrace& t, int size );
    VovEventQueue( VovEventQueue& );
    ~VovEventQueue();
    void init( VovTrace& t, int  size );
    int length();
    VovStatus enq( int sm, int em, String& desc, int id, int am );
    VovStatus enq( VovEvent& );
    VovStatus deq( VovEvent& );
    VovStatus get( VovEvent&, int offset ); // Get last 'offset' event.

    VovEvent& get( int index );
    VovTrace& getTrace() { return *eqtrace; }

    int newIndexForEventQueueEdit( int cur, int sm, int em, int dir );

    void markObsolete();
    VovStatus edit();
    String format();
    
};

class VovEventFilter {
    int subjectmask;
    int eventmask;
    int id;			// 
    String rx;			// Regular expression;
public:
    VovEventFilter( int subjectmask, int eventmask, String& rx );
    VovEventFilter( int subjectmask, int eventmask, int match_id );
    ~VovEventFilter();
    VovStatus filter( VovEvent& );
};
#endif /* EVENT_H */




