/* -*- C++ -*- */
#ifndef TRANSITION_H
#define TRANSITION_H
#pragma once


#include <String.h>
#include "node.h"

#define VOV_TRANSITION_OCTMASK OCT_INSTANCE_MASK
/* 
 * Masks used to determine the type of a transition
 */
#define DETERMINISTIC_TT   0x001
#define INTERACTIVE_TT     0x002

extern class Bzocket;
extern class VovTrace;		
extern class VovEventQueue;
extern class VovPlace;

typedef enum _VovTransitionTerminationCode {
    VOV_TRANSITION_TERMINATION_OK,
    VOV_TRANSITION_TERMINATION_WRONG_STATUS,
    VOV_TRANSITION_TERMINATION_FAILED,
    VOV_TRANSITION_TERMINATION_INVALID_INPUT,
    VOV_TRANSITION_TERMINATION_NO_OUTPUTS,
    VOV_TRANSITION_TERMINATION_DEFECTIVE_OUTPUT,
    VOV_TRANSITION_TERMINATION_INCONSISTENT
} VovTransitionTerminationCode;


class VovTransition : public VovNode {

    VovProp commandLine;	// Example: vov_bdsyn xx.bds
    VovProp capsule;		// Example: bdsyn xx.bds > xx.blif
    
    /* Historical information on last run */

    VovProp user;		// Who did it.
    VovProp host;		// On which machine.
    VovProp cwd;		// Current working directory.
    VovProp pid;		// Process id.
    VovProp started;		// Start date.
    VovProp ended;		// Finish date.
    VovProp exitStatus;		// Exit status of the transition.
    VovProp okStatus;		// List of the good status.

    VovProp affinity;		// String explaining where the transition can run.
    VovProp retracingInfo;	// Used by the retracing mechanism.


    void checkAllInputs( VovEventQueue&, VovTransitionTerminationCode& , String&  );
    void checkAllOutputs( VovEventQueue&, VovTransitionTerminationCode& , String&  );

 public:
    VovTransition();
    VovTransition( const String& actual, const String& capsule );
    VovTransition( VovTransition& );
    VovTransition( VovNode& );
    VovTransition( octId );
    ~VovTransition();

    void operator = (VovTransition& );
    VovStatus create();
    VovStatus get();
    VovStatus modify();
    VovStatus del();
    VovStatus retracingInvariants( VovTrace& trace, VovTransition& rt );

    char* getCellName();
    long  duration();		// In seconds. -1 if not known.

    String getWd() { return cwd.string(); }
    String getCommandLine();
    void   setCommandLine( String& cmd );
    void   setPid( int id ) { pid.modify( id ); }

    VovStatus edit();
    void setupDialog( VovDialog& top_dialog, VovDialogHandles& data, int viewonly = 0 );
    VovStatus readDialog( VovDialogHandles& data );

    String format();

    VovStatus mvlib( String& oldpath , String& newpath );
    VovStatus match( VovTransition& );

    void setHost( const String& h ) { host = h; OH_ASSERT( octModify( &host ) ); }
    void setUser( const String& u ) { user = u; OH_ASSERT( octModify( &user ) ); }
    void setExitStatus( const String& s ) { okStatus = s; }
    void setAffinity( const String& s );
    String getAffinity() { return affinity.string(); }
    String getUser() { return user.string(); }
    String getHost() { return host.string(); }
    VovStatus initiate( VovNodeStatus );
    VovStatus terminate( 
			VovEventQueue& queue, 			// IN: The event queue to log events to.
			VovStatus termStatus,			// IN: Termination status.
			int exit_status,			// IN: Exit status of trans.
			String& pipedMsg,			// IN: Piped message from stderr.
			VovTransitionTerminationCode& code	// OUT: Termination code.
			);
    VovStatus setRetracingInfo( VovTransition& );
    VovStatus manageRetracing( VovEventQueue& );

    VovStatus fire( int redirect, int fastMode, String& errmsg );
    VovStatus waitCompletion( int* ret, String& errmsg );
    VovStatus kill( int );

    VovStatus isReadyToFire( String& msg );
    VovStatus isUserInitiated();
    VovStatus isSuccessful();

    void  pack( Bzocket& );
    void  unpack( Bzocket& );
};

#endif	/* TRANSITION_H */


